%%%-------------------------------------------------------------------
%%% File    : agent.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : 
%%%
%%% Created :  6 May 2004 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id$
%%
%%**
%% 
%% @doc
%% @end
%%*
-module(lgi.agent).

-behaviour(mdisp).

-export([aid/4, send/3, cast/4, new/3, new/4, delete/1, delete/2, delete/3]).
-export([start_link/0, stop/0]).

%% Mdisp exports
-export([init/2, awake/2, terminate/3]).

%% Internal exports
-export([law_do/1]).

-import(erlang).
-import(lists).
-import(rpc).
-import(mdisp).
-import(wslib.log).
-import(wslib.props).

-define(agent(Law, Name), {Law, Name}).
-define(idle_timeout, 5000).

-record(agent_state, {pid,
		      context,
		      law_module
		     }).
-record(aid, {law,
	      pid,
	      name,
	      node
	     }).

aid(Law, Pid, Name, Node) ->
  #aid{law=Law, pid=Pid, name=Name, node=Node}.

cast(From, To, Msg, Timeout) -> 
  Label = erlang:make_ref(),
  send_agent(From, To, {send, {Label, self()}, From, To, Msg}),
  recv_agent_ack(Label, Timeout).

send(From, To, Msg) -> 
  send_agent(From, To, {send, undefined, From, To, Msg}).

send_agent(From=#aid{law=L, name=Nm, node=Node}, To=#aid{law=L}, Msg) ->
  mdisp:send({lgi_agent, Node}, ?agent(Nm, L), Msg).

new(Law, Name, Props) ->
  new(Law, Name, self(), Props).

new(Law, Name, Pid, Props) ->
  Agent = ?agent(Name, Law),
  ok = mdisp:new(lgi_agent, ?MODULE, Agent, {Pid, Props}),
  {ok, aid(Law, undefined, Name, node())}.

delete(#aid{law=Law, name=Name, node=Node}) ->
  delete(Law, Name, Node).

delete(Law, Name) ->
  delete(Law, Name, node()).

delete(Law, Name, Node) ->
  mdisp:send({lgi_agent, Node}, ?agent(Name, Law), {stop, self()}).
%% mdisp:delete({lgi_agent, Node}, ?agent(Name, Law)).

start_link() ->
  mdisp:start_link(lgi_agent).

stop() ->
  erlang:exit(erlang:whereis(lgi_agent), normal).

%%
%% Law operations
%%

%% aid - get agent own id
%% {context_get, K}
%% {context_set, K, V}
%% {context_delete, K}
%% {forward, To, Msg}
%% {forward, To}
%% forward
%% {deliver, To, Msg}
%% {deliver, To}
%% deliver

law_do(Op) when element(1, Op) == context_get;
		Op == aid ->
  worker_send(Op),
  Answer = worker_recv(),
  Answer;
law_do(Op) ->
  worker_send(Op),
  ok.

%%
%% Mdisp callbacks
%% 
init(Key=?agent(Name, Law), {Pid, Props}) ->
  Mod = props:gv(law_module, Props, Law),
  CS = agent_context:set(Mod:initial_context(props:delete(law_module, Props)),
			 agent_context:new()),
  start_agent(Key, #agent_state{pid=Pid, context=CS, law_module=Mod}).

awake(Key, State) ->
  start_agent(Key, State).

terminate(Reason, _Key, _State) ->
  exit(Reason).

%%
%% Agent
%%
start_agent(Key=?agent(Nm, L), AS=#agent_state{pid=AP}) ->
  receive
    {send,
     Proc,
     From=#aid{law=L, name=Nm},
     To=#aid{law=L, name=ToNm},
     Msg
    } ->
      Req={sent, ToNm, Msg},
      OpParams = {L, From, To, Msg}, 
      {Success, NxtCS} = apply_law(Key, Req, OpParams, AS),
      proc_answer(Proc, Success),
      start_agent(Key, AS#agent_state{context=NxtCS});
    {forward,
     From=#aid{law=L, name=FromNm},
     To=#aid{law=L, pid=P, name=Nm},
     Msg
    } ->
      Req = {arrived, FromNm, Msg},
      To1 = To#aid{pid=choose_defined(P, AP)},
      OpParams = {L, To1, From, Msg},
      apply_law(Key, Req, OpParams, AS);
    {stop, FromPid} ->
      stop_agent(Key, FromPid, AS);
    Other ->
      start_agent(Key, AS)
  after
    ?idle_timeout ->
      mdisp:sleep(lgi_agent, Key, AS)
  end.

stop_agent(Key, Pid, AS=#agent_state{pid=P}) when P == Pid;
						  P == undefined ->
  mdisp:die(lgi_agent, Key);
stop_agent(Key, _Pid, AS) ->
  start_agent(Key, AS).

choose_defined(undefined, B) -> B;
choose_defined(A, _) -> A.

apply_law(Key=?agent(Name, Law), Req, OpParams,
	  AS=#agent_state{context=CS, law_module=LM}) ->
  process_flag(trap_exit, true),

  {LawSuccess, LawOperations, LawCS} = do_law(Key, Law, LM, Req, CS),

  {Success, OpCS} =
    case {LawSuccess, do_operations(OpParams, LawOperations)} of
      {true, true} ->
	{true, LawCS};
      _ ->
	{false, CS}
    end,
  %% post condition rule can only change context
  {_, _, NxtCS} = do_law(Key, Law, LM, {post_condition, Req, Success}, OpCS),
  process_flag(trap_exit, false),
  {Success, NxtCS}.

do_law(Key, Law, LawMod, Req, CS) ->
  Self = self(),
  WorkerPid = spawn_link(fun () -> start_worker(Self, LawMod) end),
  WorkerPid ! {Self, apply_law, Req},
  Ret =
    case do_law_loop(WorkerPid, LawMod, {Key, CS, []}) of
      {error, Reason} ->
	log:write_error("\n\tlaw: ~s"
			"\n\tcrash reason: ~p", [Law, Reason]),
	{false, [], CS};
      false ->
	{false, [], CS};
      Other={true, _, _} ->
	Other
    end,
  %% WorkerPid ! {Self, stop},
  Ret.

do_law_loop(WorkerPid, LawMod, State={Key, CS, Operations}) ->
  receive
    %% for simplicity we allow user to do
    %% A = B
    %% and don't treat it as error
    {'EXIT', WorkerPid, {{badmatch, _}, [{LawMod, _, _}|_]}} ->
      false;
    %% acceptable nonfatal error
    %% !!FixME add error logging here
    {'EXIT', WorkerPid, {lgi_error, _}} ->
      false;
    {'EXIT', WorkerPid, normal} ->
      {true, lists:reverse(Operations), CS};
    {'EXIT', _, Reason} ->
      {error, Reason};
    {WorkerPid, WorkerReq} ->
      NxtState = do_law_worker_req(WorkerPid, WorkerReq, State),
      do_law_loop(WorkerPid, LawMod, NxtState)
  end.

%% post-law operations
do_law_worker_req(WorkerPid, Op={OpName, _, _}, State={Key, CS, Ops})
  when OpName == forward;
       OpName == deliver ->
  {Key, CS, [Op|Ops]};
do_law_worker_req(WorkerPid, Op={OpName, _}, State={Key, CS, Ops})
  when OpName == forward;
       OpName == deliver ->
  {Key, CS, [Op|Ops]};
do_law_worker_req(WorkerPid, Op=OpName, State={Key, CS, Ops})
  when OpName == forward;
       OpName == deliver ->
  {Key, CS, [Op|Ops]};
%% context operations
do_law_worker_req(WorkerPid, aid, State={Key=?agent(Name, Law), CS, Ops}) ->
  WorkerPid ! {self(), aid(Law, undefined, Name, node())},
  State;
do_law_worker_req(WorkerPid, {context_get, K}, State={_Key, CS, Ops}) ->
  WorkerPid ! {self(), agent_context:get(K, CS)},
  State;
do_law_worker_req(WorkerPid, {context_set, K, V}, State={Key, CS, Ops}) ->
  {Key, agent_context:set(K, V, CS), Ops};
do_law_worker_req(WorkerPid, {context_delete, K}, State={Key, CS, Ops}) ->
  {Key, agent_context:delete(K, CS), Ops}.

do_operations(Props, []) ->
  true;
do_operations(Props, [Op|Rest]) ->
  case do_operation(Op, Props) of
    true ->
      do_operations(Props, Rest);
    Other=false ->
      Other
  end.

%%Params = {Law, From, To, Msg}
do_operation(Op=deliver, Params={_, Self, _, Msg}) ->
  do_operation({Op, Self, Msg}, Params);
do_operation(Op=forward, Params={_, _, To, Msg}) ->
  do_operation({Op, To, Msg}, Params);
do_operation({Op, self}, Params={_, Self, _, Msg}) ->
  do_operation({Op, Self, Msg}, Params);
do_operation({Op, ToName}, Params={_, _, To=#aid{name=ToName}, Msg}) ->
  do_operation({Op, To, Msg}, Params);
do_operation({Op, self, Msg}, Params={_, Self, _, _}) ->
  do_operation({Op, Self, Msg}, Params);
do_operation({Op, ToName, Msg}, Params={_, _, To=#aid{name=ToName}, _}) ->
  do_operation({Op, To, Msg}, Params);
do_operation({deliver, To, Msg}, _) ->
  deliver(To, Msg);
do_operation({forward, To, Msg}, Params={Law, Self, _, _}) ->
  forward(Law, Self, To, Msg).

deliver(To=#aid{pid=Pid, node=Node}, Msg) ->
  deliver({Pid, Node}, Msg);
deliver(undefined, Msg) ->
  true;
deliver(Pid, Msg) when is_pid(Pid); is_atom(Pid) ->
  Pid ! Msg,
  true;
deliver({undefined, Node}, Msg) ->
  true;
deliver({To, Node}, Msg) when Node == node() ->
  To ! Msg,
  true;
deliver(To={Pid, Node}, Msg) when is_atom(Pid) ->
  To ! Msg;
deliver({To, Node}, Msg) ->
  rpc:cast(Node, erlang, send, [To, Msg]),
  true.

forward(_, From, To=#aid{law=Law, name=ToName, node=ToNode}, Msg) ->
  mdisp:send({lgi_agent, ToNode}, ?agent(ToName, Law),
	     {forward, From, To, Msg}),
  true;
forward(Law, From, ToName, Msg) when is_atom(ToName) ->
  forward(Law, From, aid(Law, undefined, ToName, node()), Msg);
forward(Law, From, {ToName, ToNode}, Msg) ->
  forward(Law, From, aid(Law, undefined, ToName, ToNode), Msg);
forward(Law, From, {ToPid, ToName, ToNode}, Msg) ->
  forward(Law, From, aid(Law, ToPid, ToName, ToNode), Msg).

%%
%% Worker
%%
start_worker(Master, LawMod) ->
  put(lgi_master, Master),
  receive
    {Master, apply_law, Req} ->
      worker_apply_law(Req, LawMod);
    {Master, stop} ->
      exit(normal)
  end.

worker_apply_law({sent, To, Msg}, Law) ->
  Law:sent(To, Msg);
worker_apply_law({arrived, From, Msg}, LawMod) ->
  LawMod:arrived(From, Msg);
worker_apply_law({post_condition, {Op, Agent, Msg}, Success}, LawMod) ->
  LawMod:post_condition(Op, Agent, Msg, Success).

worker_send(Req) ->
  Master = get(lgi_master),
  Master ! {self(), Req}.

worker_recv() ->
  Master = get(lgi_master),
  receive
    {Master, Answer} ->
      Answer
  end.

recv_agent_ack(Label, Timeout) ->
  proc_recv_answer(Label, Timeout).

proc_recv_answer(Label, Timeout) ->
  receive
    {Label, answer, Answer} ->
      Answer;
    {Label, exit, Error} ->
      exit(Error)
  after Timeout ->
      exit(timeout)
  end.

proc_answer(undefined, Answer) ->
  ok;
proc_answer({Label, Pid}, Answer) ->
  Pid ! {Label, answer, Answer}.
