%%%-------------------------------------------------------------------
%%% File    : connection_pool.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : Connection pool
%%%
%%% Created :  9 Dec 2002 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id: connection_pool.erl,v 1.1.1.1 2003/05/05 20:20:58 vsekissov Exp $
%%
%% $Log: connection_pool.erl,v $
%% Revision 1.1.1.1  2003/05/05 20:20:58  vsekissov
%% Initial import
%%
%%
%%**
%% 
%% @doc It is a connection pool
%% <p>Now it supports two modes
%% <dl>
%%    <dt><code>shared</code></dt>
%%       <dd>client requests spread between connection servers. On request
%%           server with minimal number of users selected.
%%       </dd>
%%    <dt><code>exclusive</code></dt>
%%       <dd>only one client per connection server. Servers are selected in
%%           round-robin manner.
%%       </dd>
%% </dl>
%% </p>
%%  <p> Client can return server explicitly with
%%      <code>disconnect/1</code> or will be returned implicitly after its
%%      exit.
%%  </p>
%% @end
%%*
-module(connection_pool).

-behaviour(gen_server).
%%--------------------------------------------------------------------
%% Include files
%%--------------------------------------------------------------------

%%--------------------------------------------------------------------
%% External exports
-export([start_link/2, stop/1, stop_insure/1, info/1, connect/1, connect/2, disconnect/1, call/2, call/3, cast/2, conn/2]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2, code_change/3]).

-export([worker_wrapper/3]).

-import(wslib.getopt).
-import(wslib.misc).

-record(state, {pool,% {Workers, Users} Users = [{User, Worker}]
		pool_type = exclusive, % exclusive | shared
		user_queue = queue:new(), %waiting for worker
		pool_sup,% pool superviser
		pool_start_len=0,
		pool_max_len=1,
		worker_spec,
		num_started=0 %started workers but not yet initialized
	       }).

-record(pool, {workers,
	       users=users_new(),
	       num_workers=0
	      }).

-record(conn_id, {pool_pid,
		  worker_pid,
		  ref
		 }).

connect(Pool) ->
  connect(Pool, infinity).

%%**
%% @spec connect( Pool, Timeout::integer() ) -> connection() | error()
%%
%%       Pool    = name() | pid()
%%       
%% @doc Request connection from connection pool
%%
%% @end
%%*
connect(Pool, Timeout) ->
  {ok, ConnRef} = gen_server:call(Pool, {connect, self()}),
  Ref = erlang:monitor(process, Pool),
  receive
    {'DOWN', Ref, _, _, Info} ->
      exit({server_down, Info});
    {ConnRef, Answer} ->
      erlang:demonitor(Ref),
      Answer
  after Timeout ->
      erlang:demonitor(Ref),
      gen_server:cast(Pool, {connect_timeout, self(), ConnRef}),
      exit(timeout)
  end.

%%**
%% @spec disconnect( Connection::connection() ) -> ok
%%
%% @doc Return server to connection pool
%%
%% @end
%%*
disconnect(Cid=#conn_id{pool_pid=Pool, worker_pid=W, ref=Ref}) ->
  gen_server:cast(Pool, {disconnect, W, Ref}).

%%**
%% @spec conn( worker, Connection::connection() ) -> pid()
%%
%% @doc Get server pid from connection returned by <code>connect</code>
%%
%% @end
%%*
conn(worker, C=#conn_id{worker_pid=W}) ->
  W.

info(Pool) ->
  gen_server:call(Pool, {info, state}).

%% gen_server bridge
call(Cid=#conn_id{worker_pid=Worker}, Request) ->
  gen_server:call(Worker, Request).

call(Cid=#conn_id{worker_pid=Worker}, Request, Timeout) ->
  gen_server:call(Worker, Request, Timeout).

cast(Cid=#conn_id{worker_pid=Worker}, Request) ->
  gen_server:cast(Worker, Request).

%%====================================================================
%% External functions
%%====================================================================


%%**
%% @spec start_link( Name, Opts ) -> {ok, Pid} | error()
%%
%%       Name = {local, atom()} | {global, atom()}
%%       Opts = [{{worker_spec, WorkerSpec}| OptionalOpts]
%%       OptionalOpts = {supervisor_spec, SuperSPec}
%%                       | {pool_type, PoolType}
%%                       | {pool_max_len, integer()}
%%                       | {pool_start_len, integer()}
%%       
%% @doc Start connection pool.
%%      <h4>Defaults</h4>
%%      <dl>
%%      <dt><code>supervisor_spec</code></dt>
%%         <dd><code>{simple_one_for_one, 1,  1000}</code></dd>
%%      <dt><code>pool_type</code></dt>
%%         <dd>exclusive</dd>
%%      <dt><code>pool_start_len</code></dt>
%%         <dd>0</dd>
%%      <dt><code>pool_max_len</code></dt>
%%         <dd>1</dd>
%%      </dl>
%% @end
%%*

start_link(Name, Opts) ->
  gen_server:start_link(Name, ?MODULE, Opts, []).

stop_insure(Srv) ->
  Ref = erlang:monitor(process, Srv),
  stop(Srv),
  receive
    {'DOWN', Ref, _, _, _} ->
      ok
  end.

stop(Srv) ->
  Srv ! {self(), stop},
  ok.

%%====================================================================
%% Server functions
%%====================================================================

%%--------------------------------------------------------------------
%% Function: init/1
%% Description: Initiates the server
%% Returns: {ok, State}          |
%%          {ok, State, Timeout} |
%%          ignore               |
%%          {stop, Reason}
%%--------------------------------------------------------------------
init(UserOpts) ->
  process_flag(trap_exit, true),
  case catch
    begin
      Opts = options(UserOpts),
      [SupSpec, WorkerSpec, Max, Start, PoolType] =
	[getopt:value(Key, Opts) 
	 || Key <- [supervisor_spec, worker_spec, pool_max_len, pool_start_len, pool_type]],
      {ok, PoolSup} = start_pool_sup(SupSpec, WorkerSpec),
      Pool = pool_new(PoolType),
      init_pool(PoolSup, Start),
      {ok, #state{pool=Pool, pool_type=PoolType, pool_sup=PoolSup,
		  pool_max_len=Max, pool_start_len=Start,
		  worker_spec=WorkerSpec, num_started=Start}}
    end of
    Result = {ok, _} ->
      Result;
    Error ->
      {stop, misc:better_error(Error)}
  end.

%%--------------------------------------------------------------------
%% Function: handle_call/3
%% Description: Handling call messages
%% Returns: {reply, Reply, State}          |
%%          {reply, Reply, State, Timeout} |
%%          {noreply, State}               |
%%          {noreply, State, Timeout}      |
%%          {stop, Reason, Reply, State}   | (terminate/2 is called)
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_call({connect, User}, From,
	    S=#state{user_queue=UQ}) ->
  Ref = erlang:monitor(process, User),
  gen_server:reply(From, {ok, Ref}),
  NxtS = assign_user(S#state{user_queue=user_queue_in(User, Ref, UQ)}),
  {noreply, NxtS};
handle_call({info, state}, _From,
	    S=#state{pool=#pool{workers=Ws, users=Us, num_workers=NW},
		     user_queue=UQ, num_started=NumS}) ->
  {reply, [{workers, workers_to_list(Ws)},
	   {num_workers, NW},
	   {users, users_to_list(Us)},
	   {user_queue, user_queue_to_list(UQ)},
	   {num_started, NumS}
	  ], S};
handle_call(Request, From, State) ->
  Reply={invalid_request, Request},
  log(error, "unexpected call ~w from ~w", [Reply, From]),
  {reply, Reply, State}.
%%--------------------------------------------------------------------
%% Function: handle_cast/2
%% Description: Handling cast messages
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------

handle_cast({new_worker, NW},
	    S=#state{pool=P, num_started=NumS}) ->
  NxtP = pool_add_worker(NW, P),
  NxtNumS = if (NumS > 0) -> (NumS - 1);
	     true -> NumS
	  end,
  NxtS = assign_user(S#state{pool=NxtP, num_started=NxtNumS}),
  {noreply, NxtS};
handle_cast({disconnect, _Worker, Ref}, S=#state{pool=P}) ->
  NxtS = assign_user(S#state{pool=pool_delete_user(Ref, P)}),
  {noreply, NxtS};
handle_cast({connect_timeout, Ref}, S=#state{pool=P, user_queue=UQ}) ->
  erlang:demonitor(Ref),
  NxtS = assign_user(S#state{pool=pool_delete_user(Ref, P),
			     user_queue=user_queue_delete(Ref, UQ)
			    }),
  {noreply, NxtS};
handle_cast(Request, State) ->
  Info={invalid_request, Request},
  log(error, "unexpected cast ~w from ~w", [Info]),
  {noreply, State}.

%%--------------------------------------------------------------------
%% Function: handle_info/2
%% Description: Handling all non call/cast messages
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_info({'EXIT', PS, Reason}, S=#state{pool_sup=PS}) ->
  {stop, Reason, S};
handle_info({'EXIT', Pid, Reason},
	    S=#state{pool=P,
		     pool_sup=PS,
		     pool_max_len=Max,
		     user_queue=UQ,
		     num_started=NumS
		    }) ->
  NxtP =
    case pool_delete_worker(Pid, P) of
      {false, NP} ->
	NP;
      {WorkerUsers, NP} ->
	lists:foldl(fun (Ref, Pool) ->
			erlang:demonitor(Ref),
			pool_delete_user(Ref, Pool)
		    end, NP, WorkerUsers)
    end,
  {_, NxtS} = start_worker(S#state{pool=NxtP}),
  {noreply, NxtS};
handle_info({Pid, stop}, State) when is_pid(Pid) ->
  {stop, normal, State};
handle_info({'DOWN', Ref, process, _, _}, S=#state{pool=P, user_queue=UQ}) ->
  erlang:demonitor(Ref),
  {NxtP, NxtUQ} =
    case is_pool_user(Ref, P) of
      true ->
	{pool_delete_user(Ref, P), UQ};
      false ->
	{P, user_queue_delete(Ref, UQ)}
    end,
  NxtS = assign_user(S#state{pool=NxtP, user_queue=NxtUQ}),
  {noreply, NxtS};
handle_info(Info, State) ->
  {noreply, State}.



%%--------------------------------------------------------------------
%% Function: terminate/2
%% Description: Shutdown the server
%% Returns: any (ignored by gen_server)
%%--------------------------------------------------------------------
terminate(Reason, S) ->
  ok.

%%--------------------------------------------------------------------
%% Func: code_change/3
%% Purpose: Convert process state when code is changed
%% Returns: {ok, NewState}
%%--------------------------------------------------------------------
code_change(OldVsn, State, Extra) ->
  {ok, State}.

%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------

assign_user(S=#state{pool=Pool, user_queue=UQ}) ->
  assign_user(S, user_queue_empty(UQ)).

assign_user(S=#state{pool=Pool, user_queue=UQ}, false) ->
  case pool_has_free_workers(Pool) of
    true ->
      {W, Pool1} = pool_select_worker(Pool),
      {{User, Ref}, UQ1} = user_queue_out(UQ),
      Pool2 = pool_add_user(W, User, Ref, Pool1),
      User ! {Ref, {ok, conn_id(W, self(), Ref)}},
      assign_user(S#state{pool=Pool2, user_queue=UQ1});
    false ->
      case start_worker(S) of
	{true, NxtS} ->
	  NxtS;
	false ->
	  case pool_select_worker(Pool) of
	    {W, Pool1} ->
	      {{User, Ref}, UQ1} = user_queue_out(UQ),
	      Pool2 = pool_add_user(W, User, Ref, Pool1),
	      User ! {Ref, {ok, conn_id(W, self(), Ref)}},
	      assign_user(S#state{pool=Pool2, user_queue=UQ1});
	    false ->
	      S
	  end
      end
  end;
assign_user(S, _) ->
  S.

start_pool_sup(SupSpec, WorkerSpec) ->
  connection_pool_sup:start_link({SupSpec, WorkerSpec}).
    
init_pool(PoolSup, 0) ->
  ok;
init_pool(PoolSup, Num) ->
  start_worker(PoolSup),
  init_pool(PoolSup, Num-1).

start_worker(S=#state{pool=P, pool_sup=PS,
		      pool_max_len=Max, num_started=NumS}) ->
  case ((pool_num_workers(P) + NumS) =< Max) of
    true ->
      start_worker(PS),
      {true, S#state{num_started=NumS+1}};
    false ->
      false
  end;
start_worker(PoolSup) ->
  case connection_pool_sup:start_child(PoolSup) of
    {error, Error} -> % crash if server can't start
      exit(Error);
    Other ->
      Other
  end.

wrap_worker({Fun, Args}) when is_function(Fun) ->
  {?MODULE, worker_wrapper, [self(), Fun, Args]};
wrap_worker({Module, Fun, Args}) ->
  WrapFun = fun (As) ->
		apply(Module, Fun, Args)
	    end,
  {?MODULE, worker_wrapper, [self(), WrapFun, [Args]]}.
  
worker_wrapper(PoolMgr, Fun, Args) ->
  case apply(Fun, Args) of
    {ok, Pid} ->
      link(Pid),
      new_pool_worker(PoolMgr, Pid),
      {ok, Pid};
    Error ->
      Error
  end.

check_worker_spec({FW, Restart, Shutdown, Type, Modules}) ->
  WS = [{make_ref(), wrap_worker(FW), Restart, Shutdown, Type, Modules}],
  case supervisor:check_childspecs(WS) of
    ok ->
      {ok, WS};
    Error ->
      Error
  end;
check_worker_spec(_) ->
  {error, bad_worker_spec}.

new_pool_worker(PoolMgr, Pid) ->
  gen_server:cast(PoolMgr, {new_worker, Pid}).

%% Pool
pool_new(exclusive) ->
  #pool{workers={fun excl_pool/2, excl_pool(new)}};
pool_new(shared) ->
  #pool{workers={fun shared_pool/2, shared_pool(new)}}.

%
pool_num_workers(P=#pool{num_workers=Num}) ->
  Num.

pool_has_free_workers(P=#pool{workers=Ws}) ->
  workers_have_free(Ws).

pool_select_worker(P=#pool{workers=Ws}) ->
  case workers_select(Ws) of
    {false, _} ->
      false;
    {Worker, NewWs} ->
      {Worker, P#pool{workers=NewWs}}
  end.

pool_add_worker(Worker, P=#pool{workers=Ws, users=Us, num_workers=WN}) ->
  link(Worker),
  P#pool{workers=workers_add(Ws, Worker), num_workers=WN+1}.

pool_delete_worker(Worker, P=#pool{workers=Ws, users=Us, num_workers=WN}) ->
  case workers_delete(Ws, Worker) of
    {false, _} ->
      false;
    {WorkerUsers, NxtWs} ->
      NxtUs = lists:foldl(fun (U, Users) -> users_delete(Users, U) end,
			  Us, WorkerUsers),
      NxtWN = WN - 1,
      {WorkerUsers, P#pool{workers=NxtWs, users=NxtUs, num_workers=NxtWN}}
  end.

pool_add_user(Worker, User, Ref, P=#pool{workers=Ws, users=Us}) ->
  NxtWs = workers_add_user(Ws, Worker, Ref),
  NxtUs = users_add(Us, User, Ref, Worker),
  P#pool{workers=NxtWs, users=NxtUs}.

pool_delete_user(Ref, P=#pool{workers=Ws, users=Us}) when is_reference(Ref) ->
  {W, NxtUs} = users_delete(Us, Ref),
  NxtWs = workers_delete_user(Ws, W, Ref),
  P#pool{workers=NxtWs, users=NxtUs}.

is_pool_user(Ref, P=#pool{users=Us}) ->
  users_is_user(Us, Ref).

%%% Users
users_new() ->
  gb_trees:empty().

users_add(Us, User, Ref, Worker) ->
  gb_trees:insert(Ref, {User, Worker}, Us).

users_delete(Us, Ref) ->
  {_, W} = gb_trees:get(Ref, Us),
  {W, gb_trees:delete(Ref, Us)}.

users_is_user(Us, Ref) ->
  gb_trees:is_defined(Ref, Us).

users_to_list(Us) ->
  gb_trees:to_list(Us).

%%% Workers

workers_to_list({Fun, Pool}) ->
  Fun(to_list, Pool).

workers_have_free({Fun, Pool}) ->
  Fun(has_free, Pool).

workers_select({Fun, Pool}) ->
  {Ret, NewPool} = Fun(select, Pool),
  {Ret, {Fun, NewPool}}.

workers_add({Fun, Pool}, Worker) ->
  {Fun, Fun(add, {Pool, Worker})}.

workers_delete({Fun, Pool}, Worker) ->
  {Deleted, NewPool} = Fun(delete, {Pool, Worker}),
  {Deleted, {Fun, NewPool}}.

workers_add_user({Fun, Pool}, Worker, User) ->
  {Fun, Fun(add_user, {Pool, Worker, User})}.

workers_delete_user({Fun, Pool}, Worker, User) ->
  {Fun, Fun(delete_user, {Pool, Worker, User})}.

shared_pool(new) ->
  orddict:new().

shared_pool(add, {Pool, Worker}) ->
  orddict:store(Worker, [], Pool);

shared_pool(select, Pool) ->
  {Worker, _} =
    orddict:fold(fun (Worker, Users, Acc={_, PrevCount}) ->
		     Count = length(Users),
		     if Count < PrevCount -> {Worker, Count};
			true -> Acc
		     end;
		     (Worker, Users, _) ->
		     {Worker, length(Users)}
		 end, {false, false}, Pool),
  {Worker, Pool};

shared_pool(to_list, Pool) ->
  orddict:to_list(Pool);

shared_pool(has_free, Pool) ->
  (catch
   orddict:fold(fun (Worker, [], Res) -> 
		    throw(true);
		    (_, _, Res) ->
		    Res
		end, false, Pool));

shared_pool(delete, {Pool, Worker}) ->
  case orddict:find(Worker, Pool) of
    {ok, Users} ->
      {Users, orddict:erase(Worker, Pool)};
    false ->
      {false, Pool}
  end;

shared_pool(add_user, {Pool, Worker, User}) ->
  orddict:append(Worker, User, Pool);

shared_pool(delete_user, {Pool, Worker, User}) ->
  {ok, Users} = orddict:find(Worker, Pool), 
  orddict:store(Worker, lists:delete(User, Users), Pool).

%%%
excl_pool(new) ->
  {queue:new(), shared_pool(new)}.

excl_pool(select, Pool={Free, Busy}) ->
  case queue:out(Free) of
    {empty, _} ->
      {false, Pool};
    {{value, Worker}, NewFree} ->
      {Worker, {NewFree, Busy}}
  end;

excl_pool(has_free, Pool={Free,_}) ->
  case queue:out(Free) of
    {empty, _} ->
      false;
    _ ->
      true
  end;

excl_pool(to_list, Pool={Free, Busy}) ->
  shared_pool(to_list, Busy);

excl_pool(add, {Pool={Free, Busy}, Worker}) ->
  {queue:in(Worker, Free), shared_pool(add, {Busy, Worker})};

excl_pool(delete, {Pool={Free, Busy}, Worker}) ->
  NewFree = queue:from_list(lists:delete(Worker, Free)),
  {Ret, NewBusy} = shared_pool(delete, {Busy, Worker}),
  {Ret, {NewFree, NewBusy}};

excl_pool(add_user, {Pool={Free, Busy}, Worker, User}) ->
  {Free, shared_pool(add_user, {Busy, Worker, User})};

excl_pool(delete_user, {Pool={Free, Busy}, Worker, User}) ->
  {queue:in(Worker, Free), shared_pool(delete_user, {Busy, Worker, User})}.

%%%
user_queue_empty(Q) ->
  case queue:out(Q) of
    {empty, _} ->
      true;
    _ ->
      false
  end.

user_queue_in(User, Ref, Q) ->
  queue:in({User, Ref}, Q).

user_queue_out(Q) ->
  case queue:out(Q) of
    {{value, V}, Q1} ->
      {V, Q1};
    Result -> %{empty, Q1}
      Result
  end.

user_queue_delete(Ref, Q) ->
  queue:from_list(lists:keydelete(Ref, 2, queue:to_list(Q))).

user_queue_to_list(Q) ->
  queue:to_list(Q).

conn_id(Worker, Mgr, Ref) ->
  #conn_id{pool_pid=Mgr, worker_pid=Worker, ref=Ref}.

log(error, Fmt, Args) ->  
  error_logger:format(Fmt, Args).

options(Opts) ->
  MkChkLen = fun (Min) ->
		 fun (V) when integer(V), V >= Min ->
		     {ok, V};
		     (_) ->
		     {error, bad_uinteger}
		 end
	     end,
  ChkType = fun (Type) when Type == shared; Type == exclusive ->
		{ok, Type};
		(Type) ->
		{error, {bad_pool_type, Type}}
	    end,
  getopt:options(Opts,
		 [{worker_spec, required, fun check_worker_spec/1},
		  {supervisor_spec, {simple_one_for_one, 1,  1000}, term},
		  {pool_type, exclusive, ChkType},
		  {pool_max_len, 1, MkChkLen(1)},
		  {pool_start_len, 0, MkChkLen(0)}],
		 defined_strict).
