%%%-------------------------------------------------------------------
%%% File    : odbcdb.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : 
%%%
%%% Created :  4 Dec 2002 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id: odbcdb.erl,v 1.6 2003/10/31 18:02:49 vsekissov Exp $
%%
%%**
%% 
%% @doc
%% @end
%%*
-module(odbcdb).

-export([start/0, stop/0, install/1, connect/0, disconnect/1, is_connection/1]).
-export([select_count/3, select_count/5, next/1, next/3, do/3, do/5]).
-export([fold/5, fold/6]).
-export([transaction/2, transaction/4, commit/2, commit/3]).
-export([mk_query/2, mk_select_query/2, group_by/2]).
-export([select_rows/3,select_rows/4, delete_row/3, insert_row/3, read_row/3]).
-export([update_row/3, fetch_row/3, current_rowid/2, next_rowid/2]).
-export([add_query_src/1, add_query_phrase/4, update_query_dictionary/0]).
%% Internal exports
-export([odbc_connect/4, start_odbc/2, generate_schema_phrases/3, generate_schema_phrase/4, mk_phrase_src/1, info/1]).

-import(wslib.listn).
-import(wslib.str).
-import(wslib.props).
-import(wslib.hof).

-include("dbg.hrl").

-record(db, {conn}).
-record(srv_state, {result_as=raw,
		    result_fmt=fun def_result_fmt/2,
		    check_fatal
		   }).
 
-define(dict, odbcdb_phrase_dict).

install(Nodes) ->
  mnesia:start(),
  {atomic,ok} = odbcdb_schema:install(Nodes),
  {atomic,ok} = phrase.mnesia_dict:install(?dict, Nodes),
  ok.
  
start() ->
  application:start(?MODULE).

stop() ->
  application:stop(?MODULE).

connect() ->
  {ok, Conn} = connection_pool:connect(?MODULE),
  {ok, #db{conn=Conn}}.

disconnect(_Db=#db{conn=C}) ->
  connection_pool:disconnect(C).

is_connection(Db) when is_record(Db, db) ->
  true;
is_connection(_) ->
  false.

transaction(Db, Fun) ->
  transaction(Db, Fun, [], infinity).

transaction(Db, Fun, FunParams, Timeout) ->
  case catch apply(Fun, FunParams) of
    {EK, Error} when EK == error;
		     EK == 'EXIT' ->
      commit(Db, rollback, Timeout),
      exit(Error);
    Res ->
      case commit(Db, commit, Timeout) of
	{error,not_an_explicit_commit_connection} ->
	  Res;
	ok ->
	  Res;
	 Error ->
	  exit(Error)
      end
  end.

current_rowid(Db, Table) ->
  {ok, [{Key, Id}]} = fetch_row(Db, {Table, current_rowid}, []),
  rowid_to_pkey_type(Table, Key, Id).

next_rowid(Db, Table) ->
  {ok, [{Key, Id}]} = fetch_row(Db, {Table, next_rowid}, []),
  rowid_to_pkey_type(Table, Key, Id).

rowid_to_pkey_type(Table, Key, Rowid) ->
  case odbcdb_schema:attribute_property(Table, Key, type) of
    undefined ->
      Rowid;
    Type ->
      odbcdb.attr_type:coerce_type(in, Type, Rowid)
  end.

select_rows(Db, Table, WhereParams) ->
  select_rows(Db, Table, WhereParams, []).

%% @spec select_rows(Db::dbh(), Table::atom(), WhereParams, GroupBy)
%%       -> {selected, Rows} | Error
%% WhereParams = [where_param]
%% GroupBy = group_by()
%% 
select_rows(Db, Table, WhereParams, GroupBy) ->
  {Where, Params} =
    lists:foldl(
      fun ({G=in, Name, []}, {W, P}) ->
	  %% PostgreSQL doesn't like empty lists
	  {[{'=', Name}|W], [{Name, undefined}|P]};
	  ({G=in, Name, Values}, {W, P}) ->
	  ValStr =
	    "("++str:join([format_query_param(V) || V <- Values], $,) ++")",  
	  {[{G, Name}|W], [{Name, {noquote, ValStr}}|P]};
	  ({Guard, Name, Value}, {W, P}) ->
	  {[{Guard, Name}|W], [{Name, Value}|P]};
	  ({Name, Value}, {W, P}) ->
	  {[{'=', Name}|W], [{Name, Value}|P]}
      end, {[], []}, WhereParams),
  Query1 = mk_select_query(Table, Where),
  Query = if GroupBy == [] ->
	      Query1;
	     true ->
	      group_by(Query1, GroupBy)
	  end,
  do(Db, Query, Params, tuple, infinity).

read_row(Db, Table, Row) ->
  fetch_row(Db, mk_read_query(Table), Row).

insert_row(Db, Table, Row) ->
  CoerceType = odbcdb_schema:table_property(Table, coerce_attr_type),
  Row1 = [{K, CoerceType(out, K, V)} || {K, V} <- Row],
  do(Db, {Table, insert_row}, Row1, tuple, infinity).

delete_row(Db, Table, Row) ->
  do(Db, {Table, delete_row}, Row, tuple, infinity).

update_row(Db, Table, Row) ->
  CoerceType = odbcdb_schema:table_property(Table, coerce_attr_type),
  Row1 = [{K, CoerceType(out, K, V)} || {K, V} <- Row],
  do(Db, {Table, update_row}, Row1, tuple, infinity).

fetch_row(Db, Query, QueryParams) ->
  Data = do(Db, Query, QueryParams, tuple, infinity),
  case Data of
    {selected, [Row]} ->
      {ok, Row};
    {selected, []} ->
      {error, no_data_found};
    {selected, _} ->
      {error, too_many_rows}
  end.

select_count(Db, Query, Params) ->
  select_count(Db, Query, Params, tuple, infinity).

select_count(Db, Query, Params, ResAs, Timeout) when ResAs == raw;
						     ResAs == tuple;
						     ResAs == row ->
  Query1 = mk_query(Query),
  Query1(Params,
	 fun (QueryStr, TypeFmt) ->
	     send_recv(Db, {select_count, QueryStr, ResAs, TypeFmt, Timeout})
	 end).

next(Db) ->
  next(Db, 1, infinity).

next(Db, N, Timeout) ->
  send_recv(Db, {next, N, Timeout}).

%% @spec fold(Db, Query, QueryParams, Fun, InitAcc, Params) -> Result
%% Fun = fun (Db, Row, Acc) -> NxtAcc
%% 
fold(Db, Query, QueryParams, Fun, InitAcc) ->
  fold(Db, Query, QueryParams, Fun, InitAcc, []).

fold(Db, Query, QueryParams, Fun, InitAcc, Params) ->
  [OrderBy, InitOffset, Limit, ChunkSize, Timeout] =
    props:gv_map([{order_by, []},
		  {offset, 0},
		  {limit, 10000},
		  {chunk_size, 1000},
		  {timeout, infinity}
		 ], Params),
  RowF =
    fun (Acc) ->
	{selected, Rows} = odbcdb:next(Db, ChunkSize, Timeout),
	if Rows == [] -> {false, Acc};
	   true -> {true, lists:foldl(Fun, Acc, Rows)}
	end
    end,
  QueryF =
    fun ({Offset, Acc}) ->
	Q = odbcdb:group_by(Query,
			    {OrderBy, [{limit, Limit}, {offset, Offset}]}),
	{ok, NumRows} =
	  odbcdb:select_count(Db, Q, QueryParams, tuple, Timeout),
	NxtAcc = hof:while(RowF, Acc),
	if NumRows < Limit -> {false, Acc};
	   true -> {true, {Offset+Limit, NxtAcc}}
	end
    end,

  hof:while(QueryF, {0, InitAcc}).

do(Db, Query, Params) ->
  do(Db, Query, Params, tuple, infinity).

do(Db, Query, Params, ResAs, Timeout) when ResAs == raw;
					   ResAs == tuple;
					   ResAs == row ->
  Query1 = mk_query(Query),
  Query1(Params,
	 fun (QueryStr, TypeFmt) ->
	     send_recv(Db, {do, QueryStr, ResAs, TypeFmt, Timeout})
	 end).

commit(Db, Mode) ->
  commit(Db, Mode, infinity).
commit(Db, Mode, Timeout) when Mode == commit;
			       Mode == rollback ->
  send_recv(Db, {commit, Mode, Timeout}).

send_recv(Db, Query) ->
  SPid = dbh(Db),
  Ref = erlang:monitor(process, SPid),
  SPid ! {self(), Query},
  Return =
    receive
      {'DOWN', Ref, _, _, Info} ->
	exit(Info);
      {SPid, {fatal, Reson}} ->
	exit(Reson);
      {SPid, Result} ->
	Result
    end,
  erlang:demonitor(Ref),
  Return.

dbh(Db=#db{conn=C}) ->
  connection_pool:conn(worker, C).

add_query_src(Fun) ->
  phrase:add_src(?dict, mk_phrase_src(Fun)).

add_query_phrase(Key, Format, ParamNames, ResultTypes) ->
  odbcdb_schema:add_properties({{db_query, Key, result_types}, ResultTypes}),
  phrase:add_phrase(?dict, Key, Format, ParamNames).

update_query_dictionary() ->  
  phrase:update_dict(?dict).

%% server stuff

odbc_connect(Parent, ConnStr, ConnOpts, CheckFatal) ->
  {ok, Conn} = odbc:connect(ConnStr, ConnOpts),
  proc_lib:init_ack(Parent, {ok, self()}),
  odbc_srv_loop(Conn, #srv_state{check_fatal=CheckFatal}).

start_odbc(ConnStr, Opts) ->
  [ConnOpts, ErrorFatalFun] =
    [proplists:get_value(K, Opts) || K <- [odbc_opts, odbc_fatal_fun]],
  proc_lib:start_link(?MODULE, odbc_connect,
		      [self(), ConnStr, ConnOpts, ErrorFatalFun]).

odbc_srv_loop(Conn, S=#srv_state{check_fatal=CF}) ->
  {User, Result, NxtS} =
    receive
      {Pid, {select_count, Query, ResAs, ResFmt, Timeout}} ->
	?dbg("query ~s", [Query]),
	{Pid, odbc:select_count(Conn, Query, Timeout),
	 S#srv_state{result_as=ResAs, result_fmt=ResFmt}};
      {Pid, {next, N, Timeout}} ->
	{Pid, odbc:select(Conn, next, N, Timeout), S};
       {Pid, {do, Query, ResAs, ResFmt, Timeout}} ->
	?dbg("query ~s", [Query]),
	{Pid, odbc:sql_query(Conn, Query, Timeout),
	 S#srv_state{result_as=ResAs, result_fmt=ResFmt}};
      {Pid, {commit, Mode, Timeout}} ->
	{Pid, odbc:commit(Conn, Mode), S};
      _Msg ->
	{undefined, ignore, S}
    end,
  case Result of
    Error={error, ErrorMsg} ->
      case CF(Result) of
	true ->
	  User ! {self(), {fatal, Error}},
	  exit(Error);
	false ->
	  User ! {self(), Error}
      end;
    Good ->
      #srv_state{result_as=RA, result_fmt=RF} = NxtS,
      Ret =
	case catch format_query_result(RA, RF, Good) of
	  {'EXIT', Error} ->
	    {error, {format_result, Error}};
	  Other ->
	    Other
	end,
      User ! {self(), Ret};
    ignore ->
      ok
  end,
  odbc_srv_loop(Conn, NxtS).

generate_schema_phrases(Table, PrimaryKeys, Attrs) ->
  lists:foreach(
    fun (Phrase) ->
	ok = generate_schema_phrase(Phrase, Table, PrimaryKeys, Attrs)
    end, [delete_row, insert_row, read_row, update_row, select]),
  ok.

generate_schema_phrase(K=keygenerator, Table,
		       {sequence, SeqName, AttrName}, _) ->
  generate_schema_phrase(K=keygenerator, Table, {sequence, SeqName}, [AttrName]);
generate_schema_phrase(K=keygenerator, Table,
		       {sequence, SeqName}, [AttrName]) ->
  NxtQuery = io_lib:format("select nextval('~s') as ~s", [SeqName, AttrName]),
  CurQuery = io_lib:format("select currval('~s') as ~s", [SeqName, AttrName]),
  phrase:add_phrase(?dict, {Table, next_rowid}, NxtQuery, []),
  phrase:add_phrase(?dict, {Table, current_rowid}, CurQuery, []);
generate_schema_phrase(K=insert_row, Table, _PrimaryKeys, Attrs) ->
  AttrStr = str:join([atom_to_list(A) || A <- Attrs], $,),
  AttrValStr = str:join(lists:duplicate(length(Attrs), "~s"), $,),
  Query = io_lib:format("insert into ~s(~s) values(~s)",
			[Table, AttrStr, AttrValStr]),
  phrase:add_phrase(?dict, {Table, K}, Query, Attrs);

generate_schema_phrase(K=delete_row, Table, PrimaryKeys, _Attrs) ->
  PkStr = str:join([atom_to_list(P)++"=~s" || P <- PrimaryKeys], " AND "),
  Query = io_lib:format("delete from ~s where ~s", [Table, PkStr]),
  phrase:add_phrase(?dict, {Table, K}, Query, PrimaryKeys);

generate_schema_phrase(K=update_row, Table, PrimaryKeys, Attrs) ->
  Attrs1 = Attrs--PrimaryKeys,
  PkStr = str:join([atom_to_list(P)++"=~s" || P <- PrimaryKeys], " AND "),
  AttrStr = str:join([atom_to_list(A)++"=~s" || A <- Attrs1], $,),
  Query = io_lib:format("update ~s set ~s where ~s", [Table, AttrStr, PkStr]),
  phrase:add_phrase(?dict, {Table, K}, Query, Attrs1++PrimaryKeys);

generate_schema_phrase(K=read_row, Table, PrimaryKeys, Attrs) ->
  PkStr = str:join([atom_to_list(P)++"=~s" || P <- PrimaryKeys], " AND "),
  AttrStr = str:join([atom_to_list(A) || A <- Attrs], $,),
  Query = io_lib:format("select ~s from ~s where ~s", [AttrStr, Table, PkStr]),
  phrase:add_phrase(?dict, {Table, K}, Query, PrimaryKeys);

generate_schema_phrase(K=select, Table, _PrimaryKeys, Attrs) ->
  AttrStr = str:join([atom_to_list(A) || A <- Attrs], $,),
  Query = io_lib:format("select ~s from ~s where ~~s", [AttrStr, Table]),
  phrase:add_phrase(?dict, {Table, K}, Query, [sql_where]).
%%
%% Where = [{Match, PlHolderName}]
%% 

mk_read_query(Table) ->
  mk_query1({Table, read_row}, get_table_result_fmt(Table)).

%% mk_select_query(Table) ->
%%   mk_query1({Table, select}, get_table_result_fmt(Table)).

%% GroupBy = OrderBy | {OrderBy, Limits}
%% OrderBy = [OrderByField]
%% OrderByField = field_name() | {field_name(), desc} | {field_name(), asc}
%% Limits = [Limit]
%% Limit = {limit, integer()} | {offset, integer()}
%% 
group_by(Query, []) ->
  Query;
group_by(Query, Spec) ->
  GroupByStr = format_group_by(Spec),
  fun(Params, Cont) ->
      Query(Params, mk_group_by_cont(GroupByStr, Cont))
  end.

mk_group_by_cont(GroupByStr, Cont) ->
  fun (QueryStr, TypeFmt) ->
      Cont(QueryStr++GroupByStr, TypeFmt)
  end.

mk_select_query(Table, Where) ->
  mk_query1({Table, select}, get_table_result_fmt(Table), Where).

mk_query(Query) when is_function(Query) ->
  Query;
mk_query(QueryName) ->
  mk_query1(QueryName, get_query_result_fmt(QueryName)).

mk_query(QueryName, Where) ->
  mk_query1(QueryName, get_query_result_fmt(QueryName), Where).

mk_query1(QueryName, TypeFmt) ->
  Query = phrase:msg(?dict, QueryName),
  fun (Params, Cont) ->
      QueryStr = format_query_str(Query, Params),
      Cont(QueryStr, TypeFmt)
  end.

mk_query1(QueryName, TypeFmt, Where) ->
  Query = phrase:msg(?dict, QueryName),
  SetQueryParams = mk_query_params(Where),
  fun (Params, Cont) ->
      Params1 = SetQueryParams(Params),
      QueryStr = format_query_str(Query, Params1),
      Cont(QueryStr, TypeFmt)
  end.

mk_query_params(Where) ->
  Query = mk_query_where(Where),
  fun (Params) ->
      WhereStr = Query(Params),
      [{sql_where, {noquote, WhereStr}}|Params]
  end.

mk_query_where([]) ->
  fun (_) -> "1=1" end;
mk_query_where(Where) ->
  {PlHolderNames, MatchFormat} = format_match(wslib.listn:mk_list(Where)),
  fun (Params) ->
      PlHolderVals = [format_query_param(proplists:get_value(PN, Params))
		      || PN <- PlHolderNames],
      WhereStr = io_lib:format(MatchFormat, PlHolderVals),
      WhereStr
  end.

format_group_by([]) -> "";
format_group_by(GroupBy) when is_list(GroupBy) ->
  format_order_by(GroupBy);
format_group_by({GroupBy, Limits}) ->
  format_order_by(GroupBy) ++ format_query_limits(Limits).

format_order_by([]) -> "";
format_order_by(GroupBy) ->
  " order by "++str:join([format_order_by_column(F) || F <- GroupBy], $,).

format_order_by_column({Name, desc}) ->
  atom_to_list(Name) ++ " desc";
format_order_by_column({Name, asc}) ->
  atom_to_list(Name);
format_order_by_column(Name) ->
  atom_to_list(Name).

format_query_limits("") ->
  "";
format_query_limits(Limits) ->
  [Limit, Offset] = props:gv_map([limit, offset], Limits),
  LimitStr = if Limit == undefined -> "";
		true -> io_lib:format(" limit ~w", [Limit])
	     end,
  OffsetStr = if Offset == undefined -> "";
		true -> io_lib:format(" offset ~w", [Offset])
	      end,
  LimitStr++OffsetStr.

format_query_str(Query, Params) ->
  Query([{K, format_query_param(V)} || {K, V} <- Params]).

format_query_param({noquote, S}) ->
  S;
format_query_param(S) when is_list(S) ->
  [$',str:substitute(S, $', "''"), $'];
format_query_param(undefined) ->
  null;
format_query_param(A) when atom(A) ->
  [$',str:substitute(atom_to_list(A), $', "''"), $'];
format_query_param(V) when is_float(V) ->
  float_to_list(V);
format_query_param(V) when is_integer(V) ->
  integer_to_list(V);
format_query_param(V) ->
  "'"++ odbcdb.attr_type:coerce_type(out, undefined, V)++"'".

format_query_result(FmtAs, undefined, Res) ->
  format_query_result(FmtAs, fun def_result_fmt/2, Res);
format_query_result(FmtAs, VF,
		    {selected, ColNames, Rows}) when FmtAs == row;
						     FmtAs == tuple ->
  ColNames1 = [list_to_atom(CN) || CN <- ColNames],
  Rows1 = [listn:applyn(fun (K, V) ->
			    {K, VF(K, null_to_undef(V))}
			end,
			[ColNames1, tuple_to_list(Row)])
	   || Row <- Rows],
  {selected, Rows1};
format_query_result(_FmtAs, VF, {selected, ColNames, Rows}) ->
  ColNames1 = [list_to_atom(CN) || CN <- ColNames],
  Rows1 = [list_to_tuple(
	     listn:applyn(fun (K, V) -> VF(K, null_to_undef(V)) end,
			  [ColNames1, tuple_to_list(Row)]))
	   || Row <- Rows],
  {selected, ColNames1, Rows1};
format_query_result(_, _, Result) ->
  Result.

format_match(Where) ->
  {PlHolderNames, PlHolderStrs} =
    lists:foldl(fun (ColumnSpec, {PNs, PSs}) ->
		    case format_match_placeholder(ColumnSpec) of
		      {PN, PS} ->
			{[PN|PNs], [PS|PSs]};
		      PS ->
			{PNs, [PS|PSs]}
		    end
		end,
		{[], []}, Where),
  MatchFormat = lists:flatten(wslib.str:join(PlHolderStrs, " AND ")),
  {PlHolderNames, MatchFormat}.

format_match_placeholder(Name) when is_atom(Name) ->
  format_match_placeholder({'=', Name});
format_match_placeholder({Match, Name}) ->
  {Name, io_lib:format("~s ~s ~~s", [Name, Match])};
format_match_placeholder({Match, Name, Value}) ->
  io_lib:format("~s ~s ~s", [Name, Match, format_query_param(Value)]).

mk_phrase_src({M, F}) ->
  mk_phrase_src(fun () -> M:F() end);
mk_phrase_src(F) when is_function(F) ->
  fun () ->
      case F() of
	{Chunk, Cont} ->
	  {lists:map(fun prepare_query_phrase/1, Chunk), mk_phrase_src(Cont)};
	Other ->
	  Other
      end
  end.

prepare_query_phrase({Key, FormatStr, Params}) ->
  prepare_query_phrase({Key, FormatStr, Params, []});
prepare_query_phrase({Key, FormatStr, Params, ResultTypes}) ->
  odbcdb_schema:add_properties({{db_query, Key, result_types}, ResultTypes}),
  {Key, FormatStr, Params}.

get_table_result_fmt(Table) ->
  case
    odbcdb_schema:table_property(Table, coerce_attr_type) of
    undefined ->
      undefined;
    F ->
      fun (K, V) -> F(in, K, V) end
  end.

get_query_result_fmt(Query) ->
  case
    odbcdb_schema:query_property(Query, coerce_result_type) of
    undefined ->
      undefined;
    F ->
      fun (K, V) -> F(in, K, V) end
  end.

def_result_fmt(_, V) ->
  V.

null_to_undef(null) -> undefined;
null_to_undef(V) -> V.

%% undef_to_null(undefined) -> null;
%% undef_to_null(V) -> V.V.

info(mnesia_tables) ->
  [?dict].
