%%%-------------------------------------------------------------------
%%% File    : phrase.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : phrase dictionary server
%%%
%%% Created :  4 Dec 2002 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id: phrase.erl,v 1.5 2003/10/31 18:02:49 vsekissov Exp $

%%**
%% 
%% @doc Phrase dictionary server
%%
%% <p> It keeps phrases in dictionary with parameters specifications
%% and on user request returns function which accepts parameters as
%% <code>[{Key,Value}]</code> and produces result message with call
%% to <code>io_lib:format</code> or to user provided format function
%% if phrase value is different from <code>string()</code>
%% </p>
%% <h3>Phrases</h3>
%%   <h4>Loading phrases</h4>
%% <p>
%% Phrase's sources are loaded as
%% <code>[{Key, {PhraseString, [ParamName]}}]</code>.
%% To get them server calls source functions on startup or by calling
%% <code>phrase:add_src</code>
%% </p>
%% @end
%%
%% @type phrase_src() = {Key, {phrase_str(), [ParamName]}}
%%       Key = term()
%%       phrase_str() = format_string() | term()
%%       ParamName = atom().
%%      
%% A phrase source format.
%%
%% @type source_fun().
%%      function(Cont) -> [phrase_src()].
%%
%%      Cont = start | term() | NoMore
%%      NoMore = []
%% 
%% @type phrase_fun().
%%      function([{ParamName, ParamValue}]) -> string() |
%%      function(CustomFormat, [{ParamName, ParamValue}]) -> term()
%%      
%% @end
%%      
%%*
-module(phrase).

-behaviour(gen_server).

%%--------------------------------------------------------------------
%% Include files
%%--------------------------------------------------------------------

%%--------------------------------------------------------------------
%% External exports
-export([start_link/4,stop/1,msg/2,find_msg/2,add_src/2,add_phrase/4,update_dict/1,clean_dict/1,info/2]).
-export([behaviour_info/0]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2, code_change/3]).

-import(wslib.props).

-record(state, {handler,   % dictinary
		hdata,
		default_parameter_value,
		src=[] % functions which feed us with sources
	       }).

%%====================================================================
%% External functions
%%====================================================================

behaviour_info() ->
  [{lookup_phrase, 2}, {write_phrase, 4}, {clear_dict, 1},
   {terminate, 1}, {init, 1}].

%%**
%% @spec msg(Srv::server_ref(), Key) -> PhraseFun | exit
%%
%%      Key = term()

%%      PhraseFun = phrase_fun()
%%      ParamName = atom()
%%      ParamValue = term()
%%       
%% @doc Return function for phrase formatting or <code>exit()</code>
%%      if <code>Key</code> was not found.
%% @end
%%*
msg(Srv, Key) ->
  msg_get(Srv, Key).

find_msg(Srv, Key) ->
  msg_find(Srv, Key).

%%**
%% @spec add_src( source_fun() ) -> ok
%%
%%       
%% @doc Add new source to phrase dictionary
%%
%% @end
%%*
add_src(Srv, Src) ->
  gen_server:call(Srv, {add_src, mk_src_fun(Src)}, infinity).

add_phrase(Srv, Key, Format, ParamNames) ->
  gen_server:call(Srv, {add_phrase, {Key, Format, ParamNames}}, infinity).

%%**
%% @spec update_dict(  ) -> ok
%%
%%       
%% @doc Update dictionary from source functions
%%
%% @end
%%*
update_dict(Srv) ->
  gen_server:call(Srv, update, infinity).


%%**
%% @spec clean_dict(  ) -> ok
%%
%%       
%% @doc Clean phrase dictionary
%%
%% @end
%%*
clean_dict(Srv) ->
  gen_server:call(Srv, clean, infinity).


%%**
%% @spec start_link( Opts ) -> {ok, pid()} | error()
%%
%%      Opts = [{src, source_fun()}] | []
%%      
%% @doc Start server
%%
%% @end
%%*
start_link(Name, Handler, HandlerArg, Opts) ->
  case gen_server:start_link(Name, ?MODULE, {Handler, HandlerArg, Opts}, []) of
    Good={ok,Pid} ->
      update_dict(Pid), % well, we now ready to work
      Good;
    Error ->
      Error
  end.

info(Srv, state) ->
  gen_server:call(Srv, {info, state}, infinity).

stop(Srv) ->
  gen_server:cast(Srv, stop).

%%====================================================================
%% Server functions
%%====================================================================

%%--------------------------------------------------------------------
%% Function: init/1
%% Description: Initiates the server
%% Returns: {ok, State}          |
%%          {ok, State, Timeout} |
%%          ignore               |
%%          {stop, Reason}
%%--------------------------------------------------------------------
init({Handler, HandlerArg, UserOpts}) ->
  {ok, HD} = Handler:init(HandlerArg),
  Opts = options(UserOpts),
  Src = [mk_src_fun(S) || S <- proplists:append_values(src, Opts)],
  DefParamVal = proplists:get_value(default_parameter_value, Opts),
  {ok, #state{handler=Handler, hdata=HD, src=Src,
	      default_parameter_value=DefParamVal}}.

%%--------------------------------------------------------------------
%% Function: handle_call/3
%% Description: Handling call messages
%% Returns: {reply, Reply, State}          |
%%          {reply, Reply, State, Timeout} |
%%          {noreply, State}               |
%%          {noreply, State, Timeout}      |
%%          {stop, Reason, Reply, State}   | (terminate/2 is called)
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_call({get, Key}, _From, S=#state{handler=H, hdata=HD}) ->
  case H:lookup_phrase(Key, HD) of
    {undefined, NxtHD} ->
      {reply, undefined, S#state{hdata=NxtHD}};
    {ok, Msg, _, NxtHD} ->
      {reply, Msg, S#state{hdata=NxtHD}}
  end;
handle_call({info, state}, _From, State) ->
  {reply, State, State};
handle_call({add_phrase, {Key, Format, ParamNs}}, _From,
	    S=#state{handler=H, hdata=HD, src=Src,
		     default_parameter_value=D}) ->
  {ok, NxtHD} = H:write_phrase(Key, mk_msg(Format, ParamNs, D), ParamNs, HD),
  {reply, ok, S#state{hdata=NxtHD}};

handle_call({add_src, Fun}, _From, S=#state{handler=H, hdata=HD, src=Src,
					    default_parameter_value=D}) ->
  NxtHD = load_phrases(H, HD, D, [Fun]),
  {reply, ok, S#state{hdata=NxtHD, src=[Fun|Src]}};

handle_call(update, _From, S=#state{handler=H, hdata=HD, src=Src,
				    default_parameter_value=D}) ->
  NxtHD = load_phrases(H, HD, D, Src),
  {reply, ok, S#state{hdata=NxtHD}};
handle_call(clean, _From, S=#state{handler=H, hdata=HD}) ->
  NxtHD = H:clear_dict(HD),
  {reply, ok, S#state{hdata=NxtHD}}.

%%--------------------------------------------------------------------
%% Function: handle_cast/2
%% Description: Handling cast messages
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_cast(stop, State) ->
  {stop, normal, State}.

%%--------------------------------------------------------------------
%% Function: handle_info/2
%% Description: Handling all non call/cast messages
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_info(Info, State) ->
  {noreply, State}.

%%--------------------------------------------------------------------
%% Function: terminate/2
%% Description: Shutdown the server
%% Returns: any (ignored by gen_server)
%%--------------------------------------------------------------------
terminate(Reason, S=#state{handler=H, hdata=HD}) ->
  (catch H:terminate(HD)),
  ok.

%%--------------------------------------------------------------------
%% Func: code_change/3
%% Purpose: Convert process state when code is changed
%% Returns: {ok, NewState}
%%--------------------------------------------------------------------
code_change(OldVsn, State, Extra) ->
  {ok, State}.

%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------

msg_get(Srv, Key) ->
  case msg_find(Srv, Key) of
    undefined ->
      exit({undefined, Key});
    Msg ->
      Msg
  end.

msg_find(Srv, Key) ->
  gen_server:call(Srv, {get, Key}).

load_phrases(H, HD, D, Src) ->
  lists:foldl(fun (Fun, HD1) ->
		  load_phrases0(H, HD1, D, Fun())
		end, HD, Src).

load_phrases0(_H, HD, _D, end_of_input) ->
  HD;
load_phrases0(H, HD, D, {Msgs, Cont}) ->
  NxtHD =
    lists:foldl(
      fun ({Key, Msg, ParamNs}, HD1) ->
	  {ok, HD2} = H:write_phrase(Key, mk_msg(Msg, ParamNs, D), ParamNs, HD1),
	  HD2
      end, HD, Msgs),
  load_phrases0(H, NxtHD, D, Cont()).

options(Opts) ->
  {ok, Props} = props:check_required([{src, []},
				      {default_parameter_value, undefined}
				     ],
				     Opts),
  Props.

mk_msg(MsgStr, ParamNames, DefParamVal) when is_list(MsgStr) ->
  FlatMsgStr = lists:flatten(MsgStr),
  fun (KVs) ->
      io_lib:format(FlatMsgStr, [kv(K, KVs, DefParamVal) || K <- ParamNames])
  end;
mk_msg(MsgFun, ParamNames, DefParamVal) when is_function(MsgFun) ->
  fun (KVs) ->
      MsgFun([kv(K, KVs, DefParamVal) || K <- ParamNames])
  end;
mk_msg(MsgTerm, ParamNames, DefParamVal) ->
  fun (KVs) ->
      {MsgTerm, ParamNames, [kv(K, KVs, DefParamVal) || K <- ParamNames]}
  end.

kv(K, L, D) ->
  case lists:keysearch(K, 1, L) of
    {value, {_, V}} ->
      V;
    _ ->
      D
  end.

mk_src_fun(F) when is_function(F) ->
  F;
mk_src_fun({M, F}) ->
  fun () -> M:F() end.
