%%%----------------------------------------------------------------------
%%% File    : property_file.erl
%%% Author  :  <svg@surnet.ru>
%%% Purpose : 
%%% Created :  2 Oct 2002 by  <svg@surnet.ru>
%%%----------------------------------------------------------------------

%%
%% $Id: property_file.erl,v 1.1.1.1 2002/10/04 00:00:00 svg Exp $
%%

%% @doc property_file is a configuration file parser.
%% 
%% <p>It supports two popular configuration formats - Apache and C-like,
%% property substitution and file inclusion. It works with Unix, DOS and
%% Mac files. Both formats are line-oriented, declaration may be
%% continued at next line with <em>'\'</em> before new line.</p>
%% 
%% <p>Property substitution may be used in quoted and unquoted string
%% values. If property doesn't exists or its value is not a string, empty
%% string is silently substituted. Syntax - <code>${property_name}</code>.
%% </p>
%% 
%% <h3>C-like Grammar</h3>
%% <pre>
%% config := decls
%%  
%% decls := decl*
%%  
%% decl := bool_decl
%%      |  simple_decl
%%      |  struct_decl
%%      |  include_decl
%% 
%%  
%% bool_decl := property_name : {property_name, true}
%%           |  property_name = bool : {property_name, bool} 
%% 			       
%% bool := 'true' | 'false' | 'yes' | 'no'
%%  
%% simple_decl := property_name '=' simple_value+
%%             : {property_name, [simple_value+]}
%% 
%% struct_decl := property_name property_attr* '{' decl* '}'
%%             : {property_name, [property_attr], [decl]}
%% 
%% property_attr := simple_value
%% 
%% include_decl := '$INCLUDE' string
%% 
%% simple_value := string
%%              |  integer
%%              |  float
%%              |  ipv4_address
%%              |  ipv4_netmask
%% 
%% property_name := squoted_string
%%               | atom_string
%% 
%% string := unquoted_string
%%        |  "double quoted string"
%%        |  'single quoted string'
%% </pre>
%% 
%% <h3>apache-like Grammar</h3>
%% <pre>
%% config := decls
%%  
%% decls := decl*
%%  
%% decl := bool_decl
%%      |  simple_decl
%%      |  struct_decl
%%      |  include_decl
%% 
%%  
%% bool_decl := property_name : {property_name, true}
%%           |  property_name = bool : {property_name, bool} 
%%           |  property_name bool : {property_name, bool} 
%% 			       
%% bool := 'true' | 'false' | 'yes' | 'no'
%%  
%% simple_decl := property_name '=' simple_value+
%%             |  property_name simple_value+
%%             : {property_name, [simple_value+]}
%% 
%% struct_decl := '&lt;'property_name property_attr* '&gt;' decl* '&lt;'/property_name'&gt;'
%%             : {property_name, [property_attr], [decl]}
%% 
%% property_attr := simple_value
%% 
%% include_decl := 'Include' string
%% 
%% simple_value := string
%%              |  integer
%%              |  float
%%              |  ipv4_address
%%              |  ipv4_netmask
%% 
%% property_name := squoted_string
%%               | atom_string
%% 
%% string := unquoted_string
%%        |  "double quoted string"
%%        |  'single quoted string'
%%
%% ipv4_netmask := ipv4 '/' ipv4        : {ipv4, bitmask}
%%              | ipv4 '/' integer_mask : {ipv4, bitmask}
%% 
%% integer_mask =&lt; 32
%%  
%% </pre>
%%
%% @end

-module(property_file).
-author('svg@surnet.ru').

-export([behaviour_info/1]).
-export([parse/2, parse/3, format_error/1]).
-export([parse_config/3, store_config/4, remove_config/2,
	 postload_config/3, walk_config/3]).

-import(wslib.listn).
-import(wslib.str).
-import(wslib.log).

behaviour_info(callbacks) ->
  [{parse_config, 4}, {store_config, 2}, {remove_config, 1},
   {post_load, 2} % used by property_server
		  % post_load(ConfigServer::pid(), Arg::term()) -> void()
  ].

%%@spec parse(FileName::string(), Mode::mode()) -> Result | Error
%%
%%@doc The same as <code>parse/3</code> but with empty initial environment
%%
%%@see parse/3

parse(FileName, Mode) ->
  parse(FileName, Mode, []).

%%@spec parse(FileName::string(), Mode::mode(), Env::env_list()) ->
%% Result | {error, Error}
%%
%% Result = list()
%% mode() = apache | clike
%% env_list() = [{Key, Value}]
%% Key = atom()
%% Value = string()
%% Error = term()
%%
%% @doc Parse file and return result or error. Error can be formatted
%% with <code>format_error/1</code>. Env is a list of parameters which
%% can be substituted in properties values.

parse(FileName, Mode=apache, Env) ->
  apache_property_parse:parse(FileName, property_mlex:grammar(Mode), Env);
parse(FileName, Mode=clike, Env) ->
  clike_property_parse:parse(FileName, property_mlex:grammar(Mode), Env).

%% @spec format_error(Error::term()) -> string()
%% 
%% @doc Format error returned by <code>parse/3</code>

format_error({Line, Module, Error}) ->
  format_error(str:fmt("Line ~p: ~s", [Line, Module:format_error(Error)]));
format_error(Message) ->
  case io_lib:deep_char_list(Message) of
    true ->
      lists:flatten(Message);
    _ ->
      io_lib:write(Message)
  end.

%% @spec parse_config(ParsedConfig, Modules, VarBindings) -> Result | exit()
%% Result = [term()]
%% Modules = [module_name()]
%% VarBindings = [{Key, Value}]
%% Key = atom()
%% Value = term()
%%
%% @doc Apply <code>Module:parse_config(ReversePath, Attrs, Content, Seed)</code>
%% for every <code>Module</code> from <code>Modules</code>.
%% @end

parse_config(ParsedConfig, Modules, VarBindings) when is_list(Modules) ->
  ok = verify_modules(Modules),
  walk_config(ParsedConfig, parse_config, Modules, VarBindings).

%% @spec store_config(ConfEntries, Modules, StoreFun, DB)
%% -> {ok, NewDB} | exit()
%% ConfEntries = [ConfEntry]
%% ConfEntry = [term()] | term()
%% Modules = [module_name()]
%% StoreFun = store_fun(StoreEntry::term(), DB) -> NewDB
%% DB = term()
%% NewDB = term()
%% 
%% @doc Apply
%% <code>Module:store_config(ConfEntry, ConfEntries) -> StoreEntries|exit()</code>
%% for every <code>Module</code> from <code>Modules</code> and
%% call <code>StoreFun(StoreEntry) for every item in <code>StoreEntries</code>.
%% If noone of modules returns result <code>StoreFun(StoreEntry) is called.
%% @end

store_config(ConfEntries, Modules, StoreFun, DB) ->
  NxtDB =
    lists:foldl(
      fun (ConfEntry, DB1) ->
	  lists:foldl(
	    StoreFun,
	    DB1,
	    case map_config_property(store_config,
				     [ConfEntry, ConfEntries],
				     Modules) of
	      {ok, NewConfEntries} ->
		NewConfEntries;
	      false ->
		[ConfEntry]
	    end
	   )
      end, DB, ConfEntries),
  {ok, NxtDB}.

%% @spec remove_config(ConfDB, Modules) -> ok | exit()
%% ConfDB = term()
%% Modules = [module_name()]
%% DB = term()
%% NewDB = term()
%% 
%% @doc Apply <code>Module:remove_config(ConfDB) -> void()</code>
%% for every <code>Module</code> from <code>Modules</code>.
%% @end

remove_config(ConfigDB, Modules) ->
  {FunName, FunArity, FunArgs} = {remove_config, 1, [ConfigDB]},
  lists:foreach(
    fun (M) ->
	case is_exported(M, {FunName, FunArity}) of
	  true ->
	    case catch apply(M, FunName, FunArgs) of
	      {'EXIT', Reason} ->
		log:write_info("remove config error"
			       "\n\tmodule:~s"
			       "\n\terror:~s",
			       [M, str:to_str(Reason)]),
		ok;
	      _ ->
		ok
	    end;
	  false ->
	    ok
	end
    end, Modules),
  ok.

%% @spec postload_config(ConfigDB, Arg) -> void() | exit()
%% ConfDB = term()
%% Arg = term()
%% @doc Could be used to start services, for example to add children to
%% supervisor tree.
%% @end
postload_config(ConfigDB, Arg, Modules) ->
  lists:foreach(fun (M) ->
		    case is_exported(M, {postload_config, 2}) of
		      true ->
			M:postload_config(ConfigDB, Arg);
		      false ->
			ok
		    end
		end, Modules),
  ok.

%% @spec >Module:parse_config(ReversePath, Attrs, Content, Seed) -> NewContent|exit()
%% ReversePath = [node_name()]
%% Attrs = [term()]
%% Content = [term()]
%% Seed = term()
%% NewContent = [term()] | term()
%% @doc Transform configuration entry.
%% <ul>
%% <li>ReversePath - path to current configuration entry
%%     including entry name
%%  </li>
%%  <li>Attrs - entry attributes</li>
%%  <li>Content - entry content</li>
%%  <li>Seed - wsxmlt seed</li>
%% </ul>
%% It is not an error if function is absent in module. It must match only
%% entries it is interested in.
%% <p>
%% <code>wsxmlt</code> transformer is used to walk configuration list so
%% all <code>wsxmlt</code> information and variable manupulation functions
%% are available.
%% <p>
%% <code>ReversePath = [eof]</code> signaled about end of config and the whole
%% config is in <code>Content</code>. It is a right place to validate it.
%% @see walk_config/4
%% @end
%%
%% @spec Module:store_config(ConfEntry, ConfEntries) -> StoreEntry|exit()
%% ConfEntries = [ConfEntry]
%% ConfEntry = term()
%% StoreEntry = term() | [term()]
%% @see Module:parse_config/4
%% @end
%%
%% @spec Module:remove_config(ConfDB) -> void()
%% ConfDB = term()
%% @see Module:parse_config/4
%% @end
%% @spec Module:postload_config(ConfDB, Arg) -> void()
%% ConfDB = term()
%% Arg = term()
%% @doc Called when configuration is parsed and loaded. _Arg_ is setted by
%% configuration store handler and its content depends from its
%% implementation only. _property_supervisor_ sets supervisor _pid()_ in it.
%% @see Module:parse_config/4
%% @end

walk_config(ParsedConfig, FunName, Modules, VarBindings) ->
  walk_config(ParsedConfig, mk_config_walker(FunName, Modules), VarBindings).

walk_config(ParsedConfig, WalkFun, VarBindings) ->
  wsxmlt:transform(walk_config_rules(WalkFun),
		   VarBindings,
		   [{eof, [], ParsedConfig}]
		  ).
		   
walk_config_rules(WalkF) ->
  [wsxmlt:default(
     {mk_config_walker(WalkF),
      fun (_, _) -> walk_config_rules(WalkF) end
     })
  ].

mk_config_walker(FunName, Modules) ->
  WF =
    fun (ReversePath, Attrs, Content, Seed) ->
	FunArgs = [ReversePath, Attrs, Content, Seed],
	case map_config_property(FunName, FunArgs, Modules) of
	  false ->
	    if (ReversePath =/= [eof]) ->
		Reason = str:fmt("bad ~s configuration directive", [FunName]),
		exit(Reason);
	       true ->
		Content
	    end;
	  {ok, NewContent} ->
	    NewContent
	end
    end,
  WF.

mk_config_walker(WalkerF) ->
  fun ({Node, Attrs}, Content, Seed) ->
      RP = [Node|rpath(Seed)],
      case catch WalkerF(RP, Attrs, Content, Seed) of
	{'EXIT', ExitReason} ->
	  if (RP == [eof])
	     and (element(1, ExitReason) == function_clause) ->
	      Content;
	     true ->
	      NodePath = case lists:reverse(RP) of
			   [P] ->
			     P;
			   P ->
			     list_to_tuple(P)
			 end,
	      Reason = str:fmt("config error at ~p:\n\t~s",
			       [{NodePath, Attrs}, str:to_str(ExitReason)]),
	      exit(Reason)
	  end;
	Other->
	  Other
      end
  end.

map_config_property(FunName, FunArgs, Modules) ->
  map_config_property(FunName, FunArgs, length(FunArgs), Modules, [], false).

map_config_property(FunName, FunArgs, FunArity, [], Content, false) ->
  false;
map_config_property(FunName, FunArgs, FunArity, [], Content, true) ->
  {ok, Content};
map_config_property(FunName, FunArgs, FunArity, [M|Ms], Content, Status) ->
  case is_exported(M, {FunName, FunArity}) of
    true ->
      case catch apply(M, FunName, FunArgs) of
	{'EXIT', {function_clause, [{M, FunName, _}|_]}} ->
	  map_config_property(FunName, FunArgs, FunArity, Ms, Content, Status);
	{'EXIT', Reason} ->
	  exit(str:fmt("module:~s~n\treason:~s~n", [M, str:to_str(Reason)])); 
	ModContent ->
	  NxtContent = Content ++ listn:mk_list(ModContent),
	  map_config_property(FunName, FunArgs, FunArity, Ms, NxtContent, true)
      end;
    false ->
      map_config_property(FunName, FunArgs, FunArity, Ms, Content, Status)
  end.

rpath(Seed) ->
  rpath(del_eof(lists:reverse(wsxmlt:ancestors(Seed))), []).

rpath([], Ret) ->
  Ret;
rpath([{Node, _}|Rest], Acc) ->
  rpath(Rest, [Node|Acc]).

is_exported(Module, ToFind) ->
  Exports = Module:module_info(exports),
  lists:member(ToFind, Exports).

verify_modules([]) ->
  ok;
verify_modules([Mod|Rest]) ->
  case code:which(Mod) of
    non_existing ->
      exit(str:fmt("module ~s does not exist", [Mod]));
    Path ->
      verify_modules(Rest)
  end.

del_eof([N]) -> % eof
  [];
del_eof([_,_|Path]) ->
  Path.
