%%%-------------------------------------------------------------------
%%% File    : property_server.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : 
%%%
%%% Created : 27 May 2004 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id$
%%
%%**
%% 
%% @doc
%% @end
%%*
-module(property_server).

-behaviour(gen_server).
%%--------------------------------------------------------------------
%% Include files
%%--------------------------------------------------------------------

%%--------------------------------------------------------------------
%% External exports
-export([start_link/4, stop/1]).
-export([behaviour_info/1,check/2,info/1,list/1,load/2, handler_op/2,parse/2,remove/1]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2, code_change/3]).

-import(wslib.props).
-import(wslib.log).

%% Handler interface
%% H:init(Arg) -> {ok, HS}
%% H:stop(HS) -> void()
%% H:config_file(HS) -> {ok, FileName}
%% H:bootstrap(Config, HS) -> {ok, Modules, ParseEnv}
%% H:init_postload(HS) -> {ok, Arg, NewHS}
%% H:store_entry(Entry, HS) - > {ok, NewHS}
%% H:new_config(HS) -> {ok, NewHS}
%% H:delete_config(HS) -> {ok, NewHS}
%% H:handle_op(Operation, HS) -> {ok, Reply, NxtHS}
%% H:config_to_list(HS) -> {ok, Entries}
%% 

-define(timeout, 180*1000). %three minutes should be enough

-record(state, {hlr,
		hlr_state,
		config_status=noconfig, % noconfig|ready
		file_format=clike,
		parse_env=[],
		worker,
		modules=[]}).

behaviour_info(callbacks) ->
  [{init, 1},
   {config_file, 1},
   {bootstrap, 2},
   {store_entry, 2},
   {new_config, 1},
   {delete_config, 1},
   {handle_op, 2},
   {config_to_list, 1},
   {init_postload, 1}
  ].

%% Args = [{file_format, clike}, {parse_env, []}]
start_link(Name, HlrMod, HlrArg, Args) ->
  gen_server:start_link(Name, ?MODULE, {HlrMod, HlrArg, Args}, []).

parse(Srv, ConfigFile) ->
  CF = case ConfigFile of
	 {file, F} when is_list(F) -> ConfigFile;
	 F when is_list(F) -> {file, F}
       end,
  gen_server:call(Srv, {parse, CF}, ?timeout).

check(Srv, ConfigFile) ->
  parse(Srv, ConfigFile).

%% ConfigSource = empty | default
%%               | {list, ParsedConfig}
%%               | {file, FileName}
%%               | {config, Modules, ConfigEntries}
load(Srv, ConfigSource) when ConfigSource == empty;
			     ConfigSource == default;
			     element(1, ConfigSource) == list;
			     element(1, ConfigSource) == file ->
  gen_server:call(Srv, {load, ConfigSource}, ?timeout).

remove(Srv) ->
  gen_server:call(Srv, remove, ?timeout).

list(Srv) ->
  gen_server:call(Srv, list, ?timeout).

info(Srv) ->
  gen_server:call(Srv, info).

%% 
handler_op(S=#state{hlr=H, hlr_state=HS}, Operation) ->
  {ok, Reply, _} = H:handle_op(Operation, HS),
  Reply;
handler_op(Srv, Operation) ->
  case gen_server:call(Srv, {handler_op, Operation}) of
    {ok, Result} ->
      Result;
    {error, Error} ->
      exit(Error)
  end.

stop(Srv) ->
  gen_server:cast(Srv, stop).

%%====================================================================
%% Server functions
%%====================================================================

%%--------------------------------------------------------------------
%% Function: init/1
%% Description: Initiates the server
%% Returns: {ok, State}          |
%%          {ok, State, Timeout} |
%%          ignore               |
%%          {stop, Reason}
%%--------------------------------------------------------------------
init({HlrMod, HlrArg, Args}) ->
  [FileFormat, ParseEnv] =
    props:gv_map([file_format, parse_env], check_props(Args)),
  case HlrMod:init(HlrArg) of
    {ok, HlrState} ->
      process_flag(trap_exit, true),
      {ok, #state{hlr=HlrMod, hlr_state=HlrState,
		  file_format=FileFormat, parse_env=ParseEnv}};
    {error, Reason} ->
      {stop, Reason}
  end.

%%--------------------------------------------------------------------
%% Function: handle_call/3
%% Description: Handling call messages
%% Returns: {reply, Reply, State}          |
%%          {reply, Reply, State, Timeout} |
%%          {noreply, State}               |
%%          {noreply, State, Timeout}      |
%%          {stop, Reason, Reply, State}   | (terminate/2 is called)
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------

handle_call({Op, ConfigSource}, From, State) when Op == parse;
						  Op == check ->
  spawn_worker(From, fun do_parse_config/2, [ConfigSource, State], false),
  {noreply, State};
%%
handle_call({load, ConfigSource}, From,
	    S=#state{hlr=H, config_status=noconfig}) ->
  case do_load_config(ConfigSource, S) of
    {ok, Modules, HS} ->
      Self = self(),
      {ok, ModArg, NxtHS} = H:init_postload(HS),
      NxtS = S#state{hlr_state=NxtHS,
		     modules=Modules,
		     config_status=loading
		    },
      Pid = spawn_worker(Self,
			 fun (Srv, State) ->
			     {do_postload(Srv, ModArg, State), From}
			 end,
			 [Self, NxtS],
			 true),
      {noreply, NxtS#state{worker=Pid}};
    Error ->
      {reply, Error, S}
  end;
handle_call({load, _C}, From, S=#state{config_status=CS}) ->
  {reply, {error, CS}, S};
%%
handle_call(remove, From, S=#state{config_status=ready}) ->
  Self = self(),
  Pid = spawn_worker(Self,
		     fun (Srv, State) ->
			 {do_remove_config(Srv, State), From}
		     end,
		     [Self, S],
		     true),
  {noreply, S#state{worker=Pid, config_status=removing}};
handle_call(remove, From, S=#state{config_status=CS}) ->
  {reply, {error, CS}, S};
%%
handle_call(list, From, S=#state{config_status=ready}) ->
  spawn_worker(From, fun do_config_to_list/1, [S], false),
  {noreply, S};
handle_call(list, From, S=#state{config_status=CS}) ->
  {reply, {error, CS}, S};
handle_call(info, From, S=#state{config_status=CS}) ->
  {reply, {ok, CS}, S};
handle_call({handler_op, _O}, _F, S=#state{config_status=noconfig}) ->
  {reply, {error, S#state.config_status}, S};
handle_call({handler_op, Operation}, From, S=#state{hlr=H, hlr_state=HS}) ->
  {ok, Reply, NxtHS} = H:handle_op(Operation, HS),
  {reply, {ok, Reply}, S#state{hlr_state=NxtHS}}.

%%--------------------------------------------------------------------
%% Function: handle_cast/2
%% Description: Handling cast messages
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_cast(stop, State) ->
  {stop, normal, State};
handle_cast({handler_op, Operation}, S=#state{hlr=H, hlr_state=HS}) ->
  {ok, _Reply, NxtHS} = H:handle_op(Operation, HS),
  {noreply, S#state{hlr_state=NxtHS}};
handle_cast({config_loaded, Receiver},
	    S=#state{config_status=loading}) ->
  gen_server:reply(Receiver, ok),
  {noreply, S#state{worker=undefined, config_status=ready}};
handle_cast({config_removed, Receiver},
	    S=#state{hlr=H, hlr_state=HS, config_status=removing}) ->
  {ok, NxtHS} = H:delete_config(HS),
  gen_server:reply(Receiver, ok),
  NxtS = S#state{worker=undefined,
		 hlr_state=NxtHS,
		 modules=[],
		 config_status=noconfig
		},
  {noreply, NxtS}.

%%--------------------------------------------------------------------
%% Function: handle_info/2
%% Description: Handling all non call/cast messages
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_info({'EXIT', W, Reason}, S=#state{worker=W}) ->
  {stop, Reason, S};
handle_info(Info, State) ->
  {noreply, State}.

%%--------------------------------------------------------------------
%% Function: terminate/2
%% Description: Shutdown the server
%% Returns: any (ignored by gen_server)
%%--------------------------------------------------------------------
terminate(Reason, S=#state{hlr=H, hlr_state=HS, config_status=ready}) ->
  do_remove_config(S, S),
  {ok, HS1} = H:delete_config(HS),
  ok;
terminate(Reason, S=#state{worker=W}) ->
  if is_pid(W) ->
      exit(W, kill),
      ok;
     true ->
      ok
  end.

%%--------------------------------------------------------------------
%% Func: code_change/3
%% Purpose: Convert process state when code is changed
%% Returns: {ok, NewState}
%%--------------------------------------------------------------------
code_change(OldVsn, State, Extra) ->
  {ok, State}.

%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------
do_load_config(empty, State) ->
  do_load_config({list, []}, State);
do_load_config(default, S=#state{hlr=H, hlr_state=HS}) ->
  case H:config_file(HS) of
    {ok, ConfigFile} ->
      do_load_config({file, ConfigFile}, S);
    Error ->
      Error
  end;
do_load_config(ConfigSource, State) ->
  case spawn_worker(fun do_parse_config/2, [ConfigSource, State], false) of
    {ok, ConfigEntries, Modules} ->
      do_store_config(ConfigEntries, Modules, State);
    Error ->
      Error
  end.

do_store_config(ConfigEntries, Modules, S=#state{hlr=H, hlr_state=HS}) ->
  StoreF = fun (Entry, HlrState) ->
	       {ok, NxtHlrState} = H:store_entry(Entry, HlrState),
	       NxtHlrState
	   end,
  {ok, HS1} = H:new_config(HS),
  {ok, NxtHS} =
    property_file:store_config(ConfigEntries, Modules, StoreF, HS1),
  {ok, Modules, NxtHS}.

do_parse_config({file, ConfigFile}, S=#state{file_format=FF, parse_env=PE}) ->
  case property_file:parse(ConfigFile, FF, PE) of
    {error, Error} ->
      exit(property_file:format_error(Error));
    Config ->
      do_parse_config({list, Config}, S)
  end;
do_parse_config({list, Config}, S=#state{hlr=H, hlr_state=HS}) ->
  {ok, Modules, ParseEnv} = H:bootstrap(Config, HS),
  ConfigEntries = property_file:parse_config(Config, Modules, ParseEnv),
  {ok, ConfigEntries, Modules};
do_parse_config({config, Modules, ConfigEntries}, _) ->
  {ok, ConfigEntries, Modules}.

do_remove_config(Srv, S=#state{modules=Ms}) ->
  ok = property_file:remove_config(Srv, lists:reverse(Ms)),
  config_removed.

do_postload(Srv, Arg, S=#state{modules=Ms}) ->
  ok = property_file:postload_config(Srv, Arg, Ms),
  config_loaded.

do_config_to_list(S=#state{hlr=H, hlr_state=HS}) ->
  H:config_to_list(HS).

%%
spawn_worker(FunName, FunArgs, AllowCrash) ->
  run_worker(self(), FunName, FunArgs, AllowCrash).

spawn_worker(From, Fun, FunArgs, AllowCrash) ->
  spawn_link(fun () -> run_worker(From, Fun, FunArgs, AllowCrash) end).

run_worker(Receiver, Fun, FunArgs, AllowCrash) ->
  Self = self(),
  if AllowCrash ->
      ok;
     true ->
      process_flag(trap_exit, true)
  end,
  WorkerPid = spawn_link(fun () -> start_worker(Self, Fun, FunArgs) end),
  Result =
    receive
      {WorkerPid, Msg} ->
	Msg;
      {'EXIT', WorkerPid, Reason} ->
	{error, Reason};
      {'EXIT', _, Reason} ->
	exit(Reason)
    end,
  if Self == Receiver ->
      Result;
     is_pid(Receiver) ->
      gen_server:cast(Receiver, Result);
     true ->
      gen_server:reply(Receiver, Result)
  end.

start_worker(Master, Fun, FunArgs) ->
  Master ! {self(), apply(Fun, FunArgs)}.

check_props(Props) ->
  {ok, Props1} = props:check_required([{file_format, clike},
				       {parse_env, []}
				      ],
				      Props),
  validate_props(Props1).

validate_props(Props) ->
  [{K, validate_prop(K, V)} || {K, V} <- Props].

validate_prop(file_format, V) when V == clike;
				   V == apachelike ->
  V;
validate_prop(parse_env, V) when is_list(V) ->
  V;
validate_prop(K, V) ->
  exit({bad_property, {K, V}}).
