%%%-------------------------------------------------------------------
%%% File    : mt19937.erl
%%% Author  :  <svg@surnet.ru>
%%% Description :  Mersenne Twister RND
%%%
%%% Created : 22 Jul 2004 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id$
%%
%%**
%% 
%% @doc  Mersenne Twister RND (Cokus's C code translated to Erlang)
%% Functions of this module use <code>mtrnd_seed</code> dictionary key.
%% @end
%%*
-module(mt19937).

-export([seed/0, seed/1, create_seed/1]).
-export([uniform/0, uniform/1, uniform_s/1, uniform_s/2]).

-define(N, 624).
-define(M, 397).
-define(MATRIX_A, 16#9908b0df).
-define(UMASK, 16#80000000).
-define(LMASK, 16#7fffffff).
-define(DEFAULT_SEED, 5489).
-define(ul32(_I), ((_I) band 16#ffffffff)).
-define(mixbits(_U,_V), (((_U) band ?UMASK) bor (((_V) band ?LMASK)))).
-define(twist(_U,_V), ((?mixbits((_U),(_V)) bsr 1)
		       bxor
		       (if ((_V) band 1) == 1 -> ?MATRIX_A; true -> 0 end))).

-record(mtrnd_seed, {state, next=?N+1}).

seed() ->
  seed(?DEFAULT_SEED).

seed(Seed) ->
  put(mtrnd_seed, create_seed(Seed)).

create_seed(Seed) when is_list(Seed) ->
  seed_by_array(Seed);
create_seed(Seed) ->
  State = list_to_tuple(seed([?ul32(Seed)], ?N-1)),
  MySeed = #mtrnd_seed{state=State}.

uniform() ->
  real().

uniform(N) when is_integer(N), N >= 1 ->
  F = uniform(N),
  trunc(F * N) + 1.

uniform_s(Seed) ->
  real(Seed).

uniform_s(N, Seed) when is_integer(N), N >= 1 ->
  {F, NxtSeed} = uniform_s(Seed),
  {trunc(F * N) + 1, NxtSeed}.

int32() ->
  {Ret, NxtSeed} = int32(get_seed()),
  put(mtrnd_seed, NxtSeed),
  Ret.

int32(Seed=#mtrnd_seed{next=?N+1}) ->
  int32(next_state(Seed));
int32(Seed=#mtrnd_seed{state=S, next=N}) ->
  Y = element(N, S),
  Y1 = Y bxor (Y bsr 11),
  Y2 = Y1 bxor ((Y1 bsl 7) band 16#9d2c5680),
  Y3 = Y2 bxor ((Y2 bsl 15) band 16#efc60000),
  Y4 = Y3 bxor ((Y3 bsr 18)),
  {Y4, Seed#mtrnd_seed{next=N+1}}.

real() ->
  Int = int32(),
  Int/4294967296.0.

real(Seed) ->
  {Int, NxtSeed} = int32(Seed),
  {Int/4294967296.0, NxtSeed}.

seed(Xs, 0) ->
  lists:reverse(Xs);
seed(Xs=[X|_], N) ->
  NX = ?ul32(1812433253 * (X bxor (X bsr 30)) + ?N - N),
  seed([NX|Xs], N-1).

seed_by_array([]) ->
  seed(19650218);
seed_by_array(Xs) ->
  seed(19650218),
  Seed = get(mtrnd_seed),
  [S0|Ss] = tuple_to_list(Seed#mtrnd_seed.state),
  LenXs = length(Xs),
  N = max(?N,LenXs),
  [_|NxtSs] = seed_by_array1(Xs, Ss, 0, N, S0, Xs, [S0]),
  Seed#mtrnd_seed{state=list_to_tuple([16#80000000|NxtSs])}.

seed_by_array1(_, Ss, _, 0, S0, _, Acc) ->
  seed_by_array2(Ss, length(Acc), ?N-1, S0, Acc);
seed_by_array1([], Ss, _, N, S0, As, Acc) ->
  seed_by_array1(As, Ss, 0, N, S0, As, Acc);
seed_by_array1(Xs, [], XI, N, _, As, Acc=[S0|_]) ->
  Ss = tl(lists:reverse(Acc)),
  seed_by_array1(Xs, Ss, XI, N, S0, As, [S0]);
seed_by_array1([X|Xs], [S1|Ss], XI, N, S0, As, Acc) ->
  NA = ?ul32((S1 bxor ((S0 bxor (S0 bsr 30)) * 1664525)) + X + XI),
  seed_by_array1(Xs, Ss, XI+1, N-1, NA, As, [NA|Acc]).

seed_by_array2(Ss, _, 0, _, Acc) ->
  lists:reverse(Acc)++Ss;
seed_by_array2([], I, N, _, Acc=[S0|_]) ->
  Ss = tl(lists:reverse(Acc)),
  seed_by_array2(Ss, 1, N, S0, [S0]);
seed_by_array2([S1|Ss], I, N, S0, Acc) ->
  NA = ?ul32((S1 bxor ((S0 bxor (S0 bsr 30)) * 1566083941)) - I),
  seed_by_array2(Ss, I+1, N-1, NA, [NA|Acc]).

next_state(Seed=#mtrnd_seed{state=State}) ->
  {State1, P1} = next_state(State, 1, ?M, ?N - ?M),
  {State2, P2} = next_state(State1, P1, ?M-?N, ?M-1),
  S0 = element(P2, State2),
  S1 = element(1, State2),
  S2 = element(P2+?M-?N, State2),
  NxtState = setelement(P2, State2, S2 bxor ?twist(S0, S1)),
  #mtrnd_seed{next=1, state=NxtState}.

next_state(State, P0, L, 0) ->
  {State, P0};
next_state(State, P0, L, N) ->
  P1 = P0+1,
  S0 = element(P0, State),
  S1 = element(P1, State),
  SL =  element(P0+L, State),
  NxtState = setelement(P0, State, SL bxor ?twist(S0, S1)), 
  next_state(NxtState, P1, L, N-1).

max(A, B) when A > B -> A;
max(_, B) -> B.

get_seed() ->
  case get(mtrnd_seed) of
    undefined ->
      seed(),
      get_seed();
    Seed ->
      Seed
  end.
