%%%-------------------------------------------------------------------
%%% File    : netrandom.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : Get random numbers from free network servers
%%%
%%% Created : 26 Jul 2004 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id$
%%
%%**
%% 
%% @doc Get random numbers from free network servers.
%% 
%% **Sources**
%% 
%% _random.org_ :: http://www.random.org
%% _hd.org_ :: http://random.hd.org
%% _hotbits_ :: http://www.fourmilab.ch
%%
%% Default source - _random.org_
%%
%% Please note that default _http_timeout_ here is *infinity*
%% @end
%%*
-module(netrandom).

-export([integers/1, integers/2, bytes/1, bytes/2, sequence/2, sequence/3]).
-export([source_quality/1, sources/0]).

-import(wslib.props).
-import(wslib.str).

-define(DEFAULT_SRC, 'random.org').
-define(PARAMS, [num, interval, src]).

integers(Num) ->
  integers(Num, []).

%% @spec integers(Num, Params) -> {ok, [integer()]} | {error, Error}
%% Num = integer()
%% Params = [Param]
%% Param = {src, Source} | {interval, {Min, Max}} | HttpParam
%% 
%% @doc Get a list of _Num_ random integers using _Source_ from
%% interval _[Min,Max]_
%% @end
integers(Num, Params) ->
  do_request(integers, [{num, Num}|props:delete(num, Params)]).

bytes(Num) ->
  bytes(Num, []).

%% @spec bytes(Num, Params) -> {ok, [integer()]} | {error, Error}
%% Num = integer()
%% Params = [Param]
%% Param = {src, Source} | HttpParam
%% @doc Get a list of _Num_ random bytes using _Source_
%% @end
bytes(Num, Params) ->
  do_request(bytes, [{num, Num}|props:delete(num, Params)]). 

sequence(Min, Max) ->
  sequence(Min, Max, []).

%% @spec sequence(Min, Max, Params) -> {ok, [integer()]} | {error, Error}
%% Num = integer()
%% Params = [Param]
%% Param = {src, Source} | HttpParam
%% @doc Get a list of _min(10000, Min-Max+1)_ random integers using _Source_
%% from interval _[Min,Max]_. Numbers are not duplicated.
%% @end
sequence(Min, Max, Params) ->
  do_request(sequence, [{interval, {Min, Max}}|props:delete(interval, Params)]).

%% @spec source_quality(Source::atom()) -> good | bad | undefined
%% @doc Determine quality of _Source_ if possible
%% @end
source_quality(Src='random.org') ->
  case do_request(source_quality, [{src, Src}]) of
    {ok, Int} when Int > 20 ->
      good;
    _ ->
      bad
  end;
source_quality(_) ->
  undefined.
  
sources() -> [{source_quality, ['random.org']},
	      {integers, ['random.org']},
	      {sequence, ['random.org']},
	      {bytes, ['random.org',
		       'hd.org',
		       'hotbits'
		      ]}
	     ].

do_request(FunName, Opts) ->
  {Opts1, HttpOpts1} = proplists:split(Opts, ?PARAMS),
  {ok, HttpOpts} = props:check_required([{http_timeout, infinity}], HttpOpts1),
  {Src, Uri, Params} = check_params(FunName, lists:flatten(Opts1)),
  Req = format_request(FunName, Src, Params),
  case get_response(Uri, Req, HttpOpts) of
    {ok, Resp} ->
      parse_response(FunName, Src, Resp, Params);
    Error ->
      Error
  end.

get_response(Uri, Req, HttpOpts) ->
  case http:request_sync(get, {lists:flatten(Uri++"?"++Req), []}, HttpOpts) of
    Error={error, _} ->
      Error;
    Result ->
      {ok, Result}
  end.

format_request(source_quality, 'random.org', _) ->
  "";
format_request(integers, 'random.org', Params) ->
  [Num, {Min, Max}] = props:gv_map([num, interval], Params),
  io_lib:fwrite("num=~w&min=~w&max=~w&col=1", [Num, Min, Max]);
format_request(bytes, 'random.org', Params) ->
  io_lib:fwrite("nbytes=~w&format=f", [props:gv(num, Params)]);
format_request(sequence, 'random.org', Params) ->
  {Min, Max} = props:gv(interval, Params),
  io_lib:fwrite("min=~w&max=~w", [Min, Max]);
format_request(bytes, 'hd.org', Params) ->
  io_lib:fwrite("numBytes=~w&type=bin", [props:gv(num, Params)]);
format_request(bytes, 'hotbits', Params) ->
  io_lib:fwrite("nbytes=~w&fmt=bin", [props:gv(num, Params)]).

parse_response(FunName, Src, {Status=200, Headers, Body}, _) ->
  parse_result(FunName, Src, props:gv('Content-Type', Headers), Body);
parse_response(_, _, {Status, _, _}, _) ->
  {error, {http_status, httpd_util:reason_phrase(Status)}}.
  
parse_result(source_quality, 'random.org', "text/plain", Body) ->
  {ok, list_to_integer(str:trim(Body, " \t\n\r%"))};
parse_result(FN, 'random.org', "text/plain", Body)
  when ((FN == integers) or (FN == sequence)) -> 
  {ok, [list_to_integer(S) || S <- string:tokens(Body, " \t\r\n")]};
parse_result(bytes, _, "application/octet-stream", Body) ->
  {ok, Body};
parse_result(FN, Src, ContentType, Body) ->
  {error, {parse, {FN, Src, ContentType}}}.

check_params(FunName, Opts) ->
  Src = props:gv(src, Opts, ?DEFAULT_SRC),
  Uri = check_source(FunName, Src),
  Params = check_params(FunName, Src, props:delete(src, Opts)),
  {Src, Uri, Params}.

check_source(source_quality, 'random.org') ->
  "http://www.random.org/cgi-bin/checkbuf";
check_source(integers, 'random.org') ->
  "http://www.random.org/cgi-bin/randnum";
check_source(bytes, 'random.org') ->
  "http://www.random.org/cgi-bin/randbyte";
check_source(sequence, 'random.org') ->
  "http://www.random.org/cgi-bin/randseq";
check_source(bytes, 'hd.org') ->
  "http://random.hd.org/getBits.jsp";
check_source(bytes, 'hotbits') ->
  "http://www.fourmilab.ch/cgi-bin/uncgi/Hotbits";
check_source(FunName, Src) ->
  exit({badsrc, {FunName, Src}}).

check_params(source_quality, _, Opts) ->
  Opts;
check_params(FunName, Src, Opts) ->
  {ok, Opts1} = props:check_required(defaults(FunName, Src), Opts),
  [check_param(FunName, Src, P) || P <- Opts1].

defaults(integers, 'random.org') ->
  [{interval, {1, 1000000000}},
   {num, 100}
  ];
defaults(sequence, 'random.org') ->
  [{interval, {1, 1000000000}}];
defaults(bytes, _) ->
  [{num, 256}].

check_param(FN, 'random.org', P={interval, {Min, Max}})
  when ((FN == integers) or (FN == sequence))
       and (Min =< 1000000000) and (Max =< 1000000000)
       and (Min >= -1000000000) and (Max >= -1000000000)
       and (Min < Max) ->
  P;
check_param(integers, 'random.org', P={num, N}) when N >= 0,
						     N =< 10000 ->
  P;
check_param(bytes, 'random.org', P={num, N}) when N >= 0,
						  N =< 16384 ->
  P;
check_param(bytes, 'hd.org', P={num, N}) when N >= 0,
					      N =< 256 ->
  P;
check_param(bytes, 'hotbits', P={num, N}) when N >= 0,
					       N =< 2048 ->
  P;
check_param(_, _, P) ->
  exit({badopt, P}).
