%%%-------------------------------------------------------------------
%%% File    : randomx.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : Random's server
%%%
%%% Created : 27 Jul 2004 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id$
%%
%%**
%% 
%% @doc Random's server.
%% It tries to get data for _seed_ from the network servers and uses
%% *Mersenne Twister RND* to generate randoms.
%% @end
%%*
-module(randomx).

-export([uniform/0, uniform/1, alnum_letter/0, alnum_string/1, bool/0]).
-export([start/0, start_link/0, stop/0]).

-import(wslib.log).

-include("plain_fsm.hrl").

-define(server, ?MODULE).
-define(netrandom_props, [{http_timeout, 60*1000}]).
-define(seed_length, 256).

start() ->
  ensure_started(spawn).

start_link() ->
  ensure_started(spawn_link).

stop() ->
  ?server ! stop,
  ok.

%% @spec uniform() -> float()
%% @doc Returns the _float()_ from the (0, 1]
%% 
uniform() ->
  Pid = ensure_started(),
  Self = self(),
  Pid ! {Self, uniform},
  rec_answer(?server).

%% @spec uniform(Int::integer()) -> integer()
%% @doc Returns the _integer()_ from the (0, Int]. _Int_ must be >= 1.
%% 
uniform(Int) when is_integer(Int),
		  Int >= 1 ->
  Pid = ensure_started(),
  Self = self(),
  Pid ! {Self, uniform, Int},
  rec_answer(?server).

%% @spec bool() -> boolean()
%% 
bool() ->
  case (uniform(2) - 1) of
    1 ->
      true;
    0 ->
      false
  end.

%% @spec alnum_string(Num::integer()) -> [char()].
%% @doc Returns alpha-numeric string of length _Num_.
%% 
alnum_string(Num) when Num >= 0 ->
  alnum_string([], Num).

alnum_string(Str, 0) ->
  lists:reverse(Str);
alnum_string(Str, Num) ->
  Letter = alnum_letter(),
  alnum_string([Letter|Str], Num-1).

%% @spec alnum_letter() -> char()
%% 
alnum_letter() ->
  RetLetter =
    case uniform(75) + $/ of
      A when A < $0 ->
        $0 + uniform(9);
      A when A =< $9 ->
        A;
      A when A < $A ->
	case bool() of
	  true ->
	    $A + uniform(25);
	  false ->
	    $0 + uniform(9)
	end;
      A when A =< $Z ->
        A;
      A when A < $a ->
	case bool() of
	  true ->
	    $a + uniform(25);
	  false ->
	    $A + uniform(25)
	end;
      Ret ->
        Ret
    end.

rec_answer(Server) ->
  receive
    {Server, Answer} ->
      Answer
  after 5000 ->
      exit(timeout)
  end.

init() ->
  Self = self(),
  process_flag(trap_exit, true),
  %% we don't want to wait the seed from network, so start child for
  %% and use locally available source before
  SeedFeeder = spawn_link(fun () -> seed_feeder(Self) end),
  SeedFeeder ! {Self, get_seed},
  Seed = new_seed(initial_seed()),
  loop({Seed, SeedFeeder}).

loop({Seed, SeedFeeder}) ->
  Self = self(),
  plain_fsm:extended_receive(
    receive
      {SeedFeeder, Msg} ->
	Data =
	  case Msg of
	    {new_seed, D} ->
	      D;
	    error ->
	      log:write_error("~s:no available random sources, using defaults",
			      [?server]),
	      initial_seed()
	  end,
	SeedFeeder ! {Self, stop},
	loop({new_seed(Data), undefined});
      {From, uniform} ->
	{Ret, NxtSeed} = uniform_s(Seed),
	From ! {?server, Ret},
	loop({NxtSeed, SeedFeeder});
      {From, uniform, Int} ->
	{Ret, NxtSeed} = uniform_s(Int, Seed),
	From ! {?server, Ret},
	loop({NxtSeed, SeedFeeder});
      stop ->
	exit(normal);
      _ ->
	loop({Seed, SeedFeeder})
    end).

uniform_s(S) ->
  mt19937:uniform_s(S).

uniform_s(I, S) ->
  mt19937:uniform_s(I, S).

new_seed(Data) ->
  mt19937:create_seed(Data).

initial_seed() ->
  case (catch list_to_binary(
		os:cmd("dd if=/dev/urandom ibs=12 count=1 2>/dev/null"))) of
    <<X:32, Y:32, Z:32>> ->
      [X, Y, Z];
    _ ->
      tuple_to_list(now())
  end.

seed_feeder(Master) ->
  Self = self(),
  receive
    {Master, get_seed} ->
      case seed_feeder_try(proplists:get_value(bytes, netrandom:sources())) of
	{ok, Bytes} ->
	  Master ! {Self, {new_seed, Bytes}};
	_ ->
	  Master ! {Self, error}
      end,
      seed_feeder(Master);
    {Master, stop} ->
      exit(normal)
  end.
      
seed_feeder_try([]) ->
  error;
seed_feeder_try([Src|Rest]) ->
  SrcQ = netrandom:source_quality(Src),
  if (SrcQ == good) or (SrcQ == undefined) ->
      case netrandom:bytes(?seed_length, [{src, Src}|?netrandom_props]) of
	Good={ok,_} ->
	  log:write_info("using random source ~w", [Src]),
	  Good;
	Error ->
	  seed_feeder_try(Rest)
      end;
     true ->
      seed_feeder_try(Rest)
  end.

ensure_started() ->
  ensure_started(spawn).

ensure_started(SpawnFun) ->
  case whereis(?server) of
    undefined ->
      do_start(SpawnFun);
    Pid ->
      Pid
  end.

do_start(SpawnFun) ->
  case global:trans({?server, self()},
		    fun () ->
			Pid = plain_fsm:SpawnFun(?MODULE, fun init/0),
			register(?server, Pid),
			Pid
		    end,
		    [node()], 1) of
    aborted ->
      ensure_started(SpawnFun);
    Ret ->
      Ret
  end.
