%%%================================================================
%%% File    : stl.erl
%%% Author  :  <svg@surnet.ru>
%%% Purpose : Templates processing
%%% Created : 16 Sep 2001 by  <svg@surnet.ru>
%%% 
%%% ----------------------------------------------------------------
%%% NOTES:
%%% 
%%% STL as "Simple Template Language" is a clone of Bruce R. Lewis's BRL
%%% (http://brl.sourceforge.net) implemented in Erlang. It deals with
%%% template processing and has most capabilities which user expects from
%%% web template engines.
%%% 
%%% 1. Templates loading and processing
%%% 
%%% Templates are transformed into ordinary Erlang modules and
%%% 
%%% - are compiled and loaded dynamically on demand;
%%% - are recompiled and reloaded if template code was changed on disk.
%%% 
%%% It is possible to use only preloaded templates code (for embedded systems).
%%% 
%%% 2. Templates including
%%% 
%%% Template can execute and include output of another template.
%%% 
%%% 3. Context support
%%% 
%%% Template can get values from context and set values into context. Last
%%% thing would be useful for organizing transformation chains.
%%% 
%%% Context parameters could be declared in template '-input' declaration
%%% as template bindings and then used as ordinary Erlang variables.
%%% 
%%% 4. Language.
%%% 
%%% 4.1 Tags
%%% 
%%% STL has only two special tags '(*' and '*)'. From HTML point of view
%%% all text enclosed by this tags is Erlang code. From Erlang point of
%%% view all text into '*)' '(*\' is ordinary string.
%%% 
%%% 4.2 Processing
%%% 
%%% Return values from all constructs which are syntactically direct
%%% subtrees of main form and can produce reasonable output (constants,
%%% variables, lists, function calls but not matches, guards, etc) are
%%% flushed to output stream.
%%% 
%%% 5. Example
%%% 
%%% 5.1 Template code
%%% 
%%% --> index.stl
%%% <html>
%%% <body>
%%% (*
%%% -input([HttpEnv, Input1, Input2, Method, LColor, TxtColor]),
%%% %input declaration must be the first line into Erlang code
%%% 
%%%  stl:include(StlServ, "top"), %process and include output from another template
%%% *)
%%% 
%%% <table cellpadding="0" cellspacing="0" border="0">
%%% 
%%% (*, [[ *)
%%% <tr><td>(*, Key, *)</td><td>(*, Value, *)</td></tr>
%%% (* ] || {Key, Value} <- HttpEnv], *)
%%% 
%%% </table>
%%% (*, stl:include(StlServ, "bottom"), *)
%%% </body>
%%% </html>
%%% <-- index.stl
%%% 
%%% 5.2 Generated Erlang code
%%% 
%%% --> stl_index.erl
%%% 
%%% -module(stl_index).
%%% 
%%% -export([process/7, input/0, modtime/0]).
%%% 
%%% %context variable StlServ is always added to this function at first position
%%% process(StlServ, HttpEnv, Input1, Input2,
%%%         Method, LColor, TxtColor) ->
%%%     [],
%%%     stl:write(StlServ, stl:include(StlServ, "top")),
%%%     stl:write(StlServ,
%%%               "\n\n<table cellpadding=\"0\" cellspacing=\"0\" border=\"0\">\n"),
%%%     stl:write(StlServ,
%%%               [["\n<tr><td>", Key, "</td><td>", Value,
%%%                 "</td></tr>\n"]
%%%                  || {Key, Value} <- HttpEnv]),
%%%     stl:write(StlServ, "\n\n</table>\n"),
%%%     stl:write(StlServ,
%%%               stl:include(StlServ, "bottom")),
%%%     stl:write(StlServ, "\n").
%%% 
%%% input() ->
%%%     ['StlServ', 'HttpEnv', 'Input1', 'Input2',
%%%      'Method', 'LColor', 'TxtColor'].
%%% 
%%% modtime() -> 0.
%%% 
%%% <-- stl_index.erl
%%%================================================================

-module(stl).
-vsn('$Id: stl.erl,v 1.3 2003/10/31 18:04:59 vsekissov Exp $.').
-author('svg@surnet.ru').

-behaviour(gen_server).

%% External exports
-export([start_link/1, start_link/2, start/1, start/2, stop/0, stop/1]).
-export([template/1, template/2, info/0, info/1, info/2]).
-export([is_action/1, is_action/2]).
-export([source/2, process/2, process/3, process/4, process/6]).
-export([include/2, include/3, write/2]).
-export([context_new/0, context_value/2, context_value/3, context_set/2, context_set/3, context_delete/2]).
-export([print_error/1, format_error/1, compile/1, int/1, to_list/1]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2, code_change/3]).

-import(wslib.getopt).
-import(wslib.str).

-include("stl.hrl").
-include_lib("kernel/include/file.hrl").

-record(state, {dir, tmpl_suffix=".stl", mod_prefix="stl", default_tmpl="index",
                actions=[], reload=true, opts=[], cache=ordsets:new()}).

%%%----------------------------------------------------------------------
%%% User API
%%%----------------------------------------------------------------------
%%================================================================
%% 
%% Function   : process/2
%% Purpose    : Process template with context
%% Arguments  : TemplateName
%%              Context
%% Returns    : {ok, {NewContext, Out}} | Error
%% Types      : Context, NewContext = context_new()
%%              TemplateName = string()
%%              Out = string()
%% Description:
%%================================================================

process(Tmplt, Context) ->
  process(stl, Tmplt, [], Context, fun process_format/1, ?PROCESS_TIMEOUT).

process(Tmplt, Actions, Context) ->
  process(stl, Tmplt, Actions, Context, fun process_format/1,
	  ?PROCESS_TIMEOUT).

process(Srv, Tmplt, Actions, Context) ->
  process(Srv, Tmplt, Actions, Context, fun process_format/1,
	  ?PROCESS_TIMEOUT).

process(Srv, Tmplt, Actions, Ctxt, FormatFun, Timeout) ->
  case process_actions(Srv, Actions, {ok, {Tmplt, Ctxt, []}}) of
    {ok, {Tmplt1, Ctxt1, Out1}} ->
      case process_template(Srv, Tmplt1, Ctxt1, FormatFun, Timeout) of
	{ok, {Ctxt2, Out2}} ->
	  {ok, {Ctxt2, lists:reverse(Out1) ++ Out2}};
	Error ->
	  return_error(Error)
      end;
    {return, Ret} ->
      {ok, Ret};
    Error ->
      return_error(Error)
  end.

process_template(StlSrv, Tmplt, Context, FormatFun, Timeout) ->
  case template(StlSrv, Tmplt) of
    {ok, Module} ->
      Srv = self(),
      CtxtFun = mk_context(StlSrv, Srv, Context, FormatFun),
      %%strip Ctxt decl from template bindings and fill the rest
      %%as default unknown values would be setted to []
      Inputs = [context_value(B, Context) || B <- tl(Module:input())],
      process_flag(trap_exit, true),
      Pid = proc_lib:spawn_link(Module,process,[CtxtFun|Inputs]),
      process_loop(Pid, Context, [], Timeout);
    Error ->
      Error
  end.

mk_context(StlSrv, Srv, Context, FormatFun) ->
  fun (get, context) -> Context;
      (get, ?TEMPLATE_SERVER_KEY) -> StlSrv;
      (make_context, NewContext) ->
      mk_context(StlSrv, Srv, NewContext, FormatFun);
      (value, {Key, Default}) -> context_value(Key, Default, Context);
      (value, Key) -> context_value(Key, Context);
      (set, {Key, Val}) ->
      Srv ! {self(), {set_context, Key, Val}}, "";
      (delete, Key) ->
      Srv ! {self(), {delete_context, Key}}, "";
      (write, Val) ->
      Srv ! {self(), {write, FormatFun(Val)}}, "";
      (Msg, Val) ->
      Srv ! {self(), {Msg, Val}}, ""
  end.

process_actions(_, [], Result) ->
  Result;
process_actions(Srv, [{Module, Fun}|RA], {ok, {Tmpl, Ctxt, Out}}) ->
  NxtCont =
    case is_action(Srv, Module) of
      true ->
	Module:Fun(Tmpl, Ctxt, Out);
      _ ->
	error(author, {author, Module})
    end,
  process_actions(Srv, RA, NxtCont);
process_actions(Srv, [Fun|RA], {ok, {Tmpl, Ctxt, Out}}) when function(Fun) ->
  NxtCont = Fun(Tmpl, Ctxt, Out),
  process_actions(Srv, RA, NxtCont);
process_actions(_, _, {ok, Resp={_Ctxt, _Out}}) ->
  {return, Resp};
process_actions(_, _, Error={error, _}) ->
  Error.

%%================================================================
%% Function   : source/2
%% Purpose    : create Erlang source for template and/or dump it to file
%% Arguments  : Template
%%              Options
%% Returns    : {ok, FileName}
%%              {ok, Source, TemplFullPath}
%%              Error
%% Types      : Template = string()
%%              Options = [Option]
%%              FileName = string()
%%              TemplFullPath = string()
%%              Option = {Key, Value}
%%              Key = atom()
%%              Value = term()
%% Description: Options can be selected as follows:
%%              {outdir, OutDir}
%%                   directory where generated source must be placed, default current
%%              file
%%                   if setted output file will be generated
%%              other options are the same as for start/1 and start_link/1 functions
%%================================================================

source(Name, UserOpts) ->
  {ok, Opts} = check_opts(UserOpts,
			  [{outdir, ".", mk_dir_exists(write)},
			   {file, false, atom}] ++ default_opts()),
  Save = getopt:value(file, Opts),
  case get_source(Name, UserOpts) of
    {ok, Source, Module, FName} when Save =/= false ->
      SName = filename:join(getopt:value(outdir, Opts),
			    atom_to_list(Module) ++ ".erl"),
      d("write to file: ~p", [SName]),
      case file:open(SName, [write]) of
	{ok, FH} ->
	  io:put_chars(FH, Source),
	  file:close(FH),
	  {ok, SName};
	Error ->
	  Error
      end;
    {ok, Source, FName} ->
      {ok, Source, FName};
    Error ->
      Error
  end.

compile([Dir, OutDir|Templates]) ->
  lists:map(fun(T) ->
		catch source(T, [{dir, Dir},
				 {outdir, OutDir},
				 file])
	    end, Templates).

%%================================================================
%% Function   : include/2
%% Purpose    : Process template and send its output to parent
%% Arguments  : CtxtFun
%%              TemplateName
%% Returns    : ""
%% Types      : CtxtFun = function()
%% Description: 
%%================================================================
include(CtxtFun, Name) when function(CtxtFun) ->
  include(CtxtFun, Name, []).

include(CtxtFun, Name, AddToContext) when function(CtxtFun) ->
  d("process include ~p", [Name]),
  process_flag(trap_exit, true),
  TmplSrv = CtxtFun(get, ?TEMPLATE_SERVER_KEY),
  case template(TmplSrv, Name) of
    {ok, Module} ->
      Context = context_set(AddToContext, CtxtFun(get, context)),
      Inputs = [context_value(B, Context) ||
		 B <- tl(apply(Module, input, []))],
      IncCtxtFun = CtxtFun(make_context, Context),
      Pid = proc_lib:spawn_link(Module,process,
			    [IncCtxtFun|Inputs]),
      include_wait(Pid, CtxtFun);
    Error ->
      CtxtFun(write, print_error(Error))
  end,
  "".

include_wait(Pid, CtxtFun) ->
  receive
    {'EXIT', Pid, normal} ->
      ok;
    {'EXIT', Pid, Reson} ->
      CtxtFun(write, print_error(Reson));
    _ ->
      include_wait(Pid, CtxtFun)
  end.

process_loop(Pid, Context, Out, Timeout) ->
  receive
    {CPid, {write, Val}}  ->
      d("receive write:~p", [Val]),
      process_loop(Pid, Context, [Val|Out], Timeout);
    {CPid, {set_context, Key, Val}} ->
      d("receive set:~p", [{Key, Val}]),
      process_loop(Pid, context_set(Key, Val, Context), Out, Timeout);
    {CPid, {delete_context, Key}} ->
      d("receive delete:~p", [Key]),
      process_loop(Pid, context_delete(Key, Context), Out, Timeout);
    {CPid, Msg} when pid(CPid) ->
      d("receive unknown message, ignore:~p", [Msg]),
      process_loop(Pid, Context, Out, Timeout);
    {'EXIT', Pid, normal} ->
      {ok, {Context, lists:flatten(lists:reverse(Out))}};
    stop ->
      error(stopped, {process, Context,
		      lists:flatten(lists:reverse(Out)),
		      stopped});
    {'EXIT', Pid, Reson} ->
      error(Reson, {process, Context, lists:flatten(lists:reverse(Out)), Reson});
    {'EXIT', _, normal} ->
      process_loop(Pid, Context, Out, Timeout);
    {'EXIT', Unknown, Reson} ->
      d("exit from unknown pid ~p, terminating: ~p", [Unknown, Reson]),
      exit(Reson);
    _ ->
      process_loop(Pid, Context, Out,  Timeout)
  after Timeout ->
      {error, timeout}
  end.

process_format(V) ->
  to_list(V).

%%================================================================
%% Function   : write/2
%% Purpose    : Send value to parent for including into template output
%% Arguments  : CtxtFun
%%              Val
%% Returns    : ""
%% Types      : CtxtFun = function()
%%              Value   = term()
%% Description: 
%%================================================================

write(Ctxt, Val) ->
  Ctxt(write, Val).

%%================================================================
%% Function   : int/1
%% Purpose    : Format parameter to its string representation if it is integer
%% Arguments  : Val
%% Returns    : term()
%% Types      : Val = term()
%% Description: Single integers in template output stream are treated as chars,
%%              so we need some checks
%%================================================================

int(Val) when integer(Val) ->
  io_lib:format("~w", [Val]);
int(Val) ->
  Val.

%%================================================================
%% Context functions. Self explanatory
%%================================================================

context_new() ->
  orddict:new().

context_value(Keys=[Key|_], Ctxt) when list(Key);
				       atom(Key) ->
  [context_value(K, Ctxt) || K <- Keys];
context_value(Key, Ctxt) ->
  context_value(Key, [], Ctxt).

context_value(Key, Default, Ctxt) when function(Ctxt) ->
  Ctxt(value, {Key, Default});
context_value(Key, Default, Ctxt) ->
  case orddict:find(Key, Ctxt) of
    {ok, Val} -> Val;
    _ -> Default
  end.

context_set(KVs, Ctxt) ->
  lists:foldl(fun ({K, V}, C) ->
		  context_set(K, V, C)
	      end, Ctxt, KVs).

context_set(Key, Val, Ctxt) when function(Ctxt) ->
  Ctxt(set, {Key, Val});
context_set(Key, Val, Ctxt) ->
  orddict:store(Key, Val, Ctxt).

context_delete(Key, Ctxt) when function(Ctxt) ->
  Ctxt(delete, Key);
context_delete(Key, Ctxt) ->
  orddict:erase(Key, Ctxt).

%%%----------------------------------------------------------------------
%%% Server API
%%%----------------------------------------------------------------------

%%================================================================
%% Function   : start/1, start_link/1
%% Purpose    : Start STL server
%% Arguments  : Options
%% Returns    : Pid | Error
%% Types      : Options = [Option]
%%              Option = {Key, Value}
%%              Key = atom()
%%              Value = term()
%% Description: Accepted options:
%%              {dir, Dir}
%%                   root directory where templates are located
%%              {default_tmpl, Name}
%%                   template for directories, default - "index"
%%              {reload, true|false|static}
%%                   true - reload template if it was changed on disk
%%                   false - load template only once,
%%                   static - use only preloaded template module
%%              {debug, true|false}
%%                   true - write debug messages to output
%%                   false - don't write
%%              {mod_prefix, Prefix}
%%                   prefix for generated template modules, default - "stl"
%%              {tmpl_suffix, Suffix}
%%                   suffix which would be appended to template name if it
%%                   hasn't any, default - ".stl"
%%              {actions, [Module]}
%%                   if specified only modules from this list are allowed to
%%                   to execute actions
%%================================================================

start_link(Options) ->
  start_link({local, stl}, Options).

start_link(Name, Options) ->
  gen_server:start_link(Name, stl, Options, []).

start(Options) ->
  start({local, stl}, Options).

start(Name, Options) ->
  gen_server:start(Name, stl, Options, []).

stop() ->
  stop(stl).

stop(Srv) ->
  Srv ! stop.

template(Name) ->
  template(stl, Name).

template(Srv, Name) ->
  gen_server:call(Srv, {template, Name}, infinity).

is_action(Module) ->
  is_action(stl, Module).

is_action(Srv, Module) ->
  gen_server:call(Srv, {is_action, Module}, infinity).

info() ->
  info(stl).

info(Query=default_template) ->
  info(stl, Query=default_template);
info(Srv) ->
  gen_server:call(Srv, info, infinity).

info(Srv, Query=default_template) ->
  gen_server:call(Srv, {info, Query}, infinity).

%%%----------------------------------------------------------------------
%%% Callback functions from gen_server
%%%----------------------------------------------------------------------

%%----------------------------------------------------------------------
%% Func: init/1
%% Returns: {ok, State}          |
%%          {ok, State, Timeout} |
%%          ignore               |
%%          {stop, Reason}
%%----------------------------------------------------------------------
init(UserOpts) ->
  case catch
    begin
      {ok, Opts} =
	check_opts(UserOpts ++
		   [{tmp_dir, filename:join(["/tmp", ?MODULE_STRING])}]),

      TmpDir = getopt:value(tmp_dir, Opts),
      SelfId = lists:nth(2, string:tokens(pid_to_list(self()), ".")),
      PrivTmpDir = str:fmt("~s/~s", [TmpDir, random_string(8)]),

      put(debug, getopt:value(debug, Opts)),

      case file:write_file_info(TmpDir,  #file_info{mode=8#1777}) of
	ok ->
	  ok;
	{error, Error1} ->
	  error_logger:error_msg(
	    "STL:can't change permissions for ~s, reason ~s",
	    [TmpDir, Error1]),
	  ok
      end,
      case file:make_dir(PrivTmpDir) of
	Good when Good == {error, eexist};
		  Good == ok ->
	  ok;
	{error, Error2} ->
	  throw({error,
		 str:fmt("can't create ~s, reason ~s", [PrivTmpDir, Error2])})
      end,
      case file:write_file_info(PrivTmpDir,  #file_info{mode=8#0700}) of
	ok ->
	  ok;
	{error, Error3} ->
	  error_logger:error_msg(
	    "STL:can't change permissions for ~s, reason ~s~n",
	    [PrivTmpDir, Error3]),
	  ok
      end,
      {ok, #state{dir=getopt:value(dir, Opts),
		  tmpl_suffix=getopt:value(tmpl_suffix, Opts),
		  mod_prefix=getopt:value(mod_prefix, Opts),
		  default_tmpl=getopt:value(default_tmpl, Opts),
		  reload=getopt:value(reload, Opts),
		  actions=getopt:value(actions, Opts),
		  opts=getopt:set(tmp_dir, PrivTmpDir, Opts)}}
    end of
    Ret={ok, _} ->
      Ret;
    Error ->
      {stop, Error}
  end.

%%----------------------------------------------------------------------
%% Func: handle_call/3
%% Returns: {reply, Reply, State}          |
%%          {reply, Reply, State, Timeout} |
%%          {noreply, State}               |
%%          {noreply, State, Timeout}      |
%%          {stop, Reason, Reply, State}   | (terminate/2 is called)
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_call({template, Name}, From, State) ->
  case catch get_template(Name, State) of
    {ok, Module} ->
      {reply, {ok, Module},
       State#state{cache=ordsets:add_element(Module,
					     State#state.cache)}};
    Error ->
      {reply, Error, State}
  end;

%%action modules are not specified, all allowed
handle_call({is_action, Name}, From, State=#state{actions=[]}) ->
  {reply, true, State};
handle_call({is_action, Name}, From, State=#state{actions=Actions}) ->
  {reply, lists:member(Name, Actions), State};

handle_call({info, default_template}, From, State) ->
  {reply, State#state.default_tmpl, State};
handle_call(info, From, State) ->
  {reply, State, State};
handle_call(Request, From, State) ->
  {reply, {error, {unknown_request, Request}}, State}.

%%----------------------------------------------------------------------
%% Func: handle_cast/2
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_cast(Msg, State) ->
  {noreply, State}.

%%----------------------------------------------------------------------
%% Func: handle_info/2
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_info(stop, State) ->
  {stop, normal, State};
handle_info(Info, State) ->
  {noreply, State}.

%%----------------------------------------------------------------------
%% Func: terminate/2
%% Purpose: Shutdown the server
%% Returns: any (ignored by gen_server)
%%----------------------------------------------------------------------
terminate(Reason, S=#state{opts=O}) ->
  file:del_dir(getopt:value(tmp_dir, O)),
  ok.

%%----------------------------------------------------------------------
%% Func: code_change/3
%% Purpose: Convert process state when code is changed
%% Returns: {ok, NewState}
%%----------------------------------------------------------------------
code_change(OldVsn, State, Extra) ->
  {ok, State}.

%%%----------------------------------------------------------------------
%%% Internal functions
%%%----------------------------------------------------------------------
get_source(Tmpl, UserOpts) ->
  case init(UserOpts) of
    {ok, State} ->
      get_template(Tmpl, State#state{reload=source});
    {stop, Error} ->
      Error
  end.

get_template([], State=#state{default_tmpl=DefTmpl}) ->
  get_template(DefTmpl, State);
get_template(Path,
	     State=#state{dir=Dir, mod_prefix=Pref, tmpl_suffix=Suf,
			  reload=Reload, opts=Opts, default_tmpl=DefTmpl}) ->
  SearchPaths = search_paths(short_path(string:strip(Path, both, $/)),
			     Dir, Pref, Suf, DefTmpl),
  case find_template(SearchPaths, Reload) of
    false ->
      error(undefined, {search, Path, SearchPaths});
    {ModuleName, FileName} ->
      load_template(ModuleName, FileName, Opts, Reload)
  end.

load_template(ModuleName, FileName, Opts, static) ->
  case code:is_loaded(ModuleName) of
    true -> {ok, ModuleName};
    _ -> error(load_template, {load_template, undefined, ModuleName})
  end;

load_template(Module, File, Opts, source) ->
  case stl_trans:form_list(read_file(File),
			   getopt:set(module, Module, Opts)) of
    {ok, {Forms, _}} ->
      {ok, erl_prettypr:format(erl_syntax:form_list(Forms)), Module, File};
    {error, Error} ->
      error(load_template, {load_template, Error, File})
  end;

load_template(ModuleName, FileName, Opts, LoadOpt) ->
  {ok, FI} = file:read_file_info(FileName),
  FileMTime = calendar:datetime_to_gregorian_seconds(FI#file_info.mtime),
  MustLoad =
    case code:is_loaded(ModuleName) of
      false ->
	true;
      {file, _} when LoadOpt == false ->
	false;
      {file, _} when LoadOpt == true ->
	ModMTime = apply({ModuleName, modtime}, []),
	(FileMTime > ModMTime) and LoadOpt
    end,
  if MustLoad == true ->
      d("load/reload template ~p, ~p", [ModuleName, FileName]),
      load_template(ModuleName, FileName, getopt:set(mtime, FileMTime, Opts));
     true ->
      {ok, ModuleName}
  end.

load_template(Module, File, Opts) ->
  case stl_trans:form_list(read_file(File),
			   getopt:set(file, File,
				      getopt:set(module, Module, Opts))) of
    {ok, {Forms, _}} ->
      case compile:forms(Forms) of
	{ok,Module,Binary} ->
	  code:purge(Module),
	  code:delete(Module),
	  case code:load_binary(Module, File, Binary) of
	    {module, Module} ->
	      {ok, Module};
	    Error ->
	      error(load_template, {load_template, Error, File})
	  end;
	Error ->
	  error(load_template, {load_template, Error, File})
      end;
    {error, Error} ->
      error(load_template, {load_template, Error, File})
  end.

find_template([], _) ->
  false;
find_template(Path, static) ->
  FindFun = fun ({M, FN}) -> code:is_loaded(M) end,
  find_template(Path, FindFun, static);
find_template(Path, LoadOpt) ->
  FindFun =
    fun ({M, FN}) -> 
	case catch file:read_file_info(FN) of
	  {ok, FI=#file_info{access=Access, type=regular}}
	  when Access =/= none ->
	    true;
	  _ ->
	    false
	end
    end,
  find_template(Path, FindFun, LoadOpt).

find_template([], _, _) ->
  false;
find_template([TI|Rest], TestFun, LoadOpt) ->
  case TestFun(TI) of
    true -> TI;
    _ -> find_template(Rest, TestFun, LoadOpt)
  end.

search_paths(Path, Dir, Pref, Suf, DefTmpl) ->
  [{module_name(Pref, P), filename:join(Dir, PN)} ||
    {P, PN} <- search_paths(Path, Suf, DefTmpl)].

search_paths([], Suf, DefTmpl) ->
  [{DefTmpl, DefTmpl ++ Suf}];
search_paths(Path, Suf, DefTmpl) ->
  case filename:extension(Path) of
    [] ->
      TP=filename:join(Path, DefTmpl),
      [{Path, Path ++ Suf},
       {TP, TP ++ Suf}];
    _ ->
      [{filename:rootname(Path), Path}]
  end.

module_name(Pref, Path) ->
  SPath0 = lists:flatmap(fun(E) -> [$_|E] end, filename:split(Path)),
  SPath1 = lists:flatmap(fun(E) -> [$_|E] end, string:tokens(SPath0, ".")),
  list_to_atom(Pref ++ SPath1).

read_file(FileName) ->
  case file:read_file(FileName) of
    {ok, Bin} ->
      binary_to_list(Bin);
    {error, Error} ->
      throw(error(Error, io_lib:format("couldn't read template file ~p: ~p",
				       [FileName, Error])))
  end.

short_path("") ->
  "";
short_path(Path) -> 
  short_path(filename:split(Path), []).

short_path([], Path) ->
  filename:join(lists:reverse(Path));
short_path([".."|Path], []) ->
  short_path(Path, []);
short_path([".."|Path0], [_|Path1]) ->
  short_path(Path0, Path1);
short_path([Dir|Path0], Path1) ->
  short_path(Path0, [Dir|Path1]).

check_opts(UserOpts) ->
  check_opts(UserOpts, default_opts()).

check_opts(UserOpts, OptDef) ->
  {ok, getopt:options(UserOpts, OptDef)}.

default_opts() ->
  [{dir, required, mk_dir_exists(read)},
   {tmp_dir, required, mk_dir_exists(read_write)},
   {include_dir, [], list},
   {include, [], list},
   {default_tmpl, "index", list},
   {reload, true, atom},
   {debug, false, atom},
   {mod_prefix, "stl", list},
   {tmpl_suffix, ".stl", list},
   {actions, [], list}].

mk_dir_exists(AccType) ->
  fun(F) ->
      Dir = filename:absname(F),
      case file:read_file_info(Dir) of
	{ok, #file_info{type=Type, access=Access}} ->
	  if (Type =/= directory) and (Type =/= other) ->
	      {error,str:fmt("~p is not a directory or symlink", [F])};
	     (Access =/= AccType) and (Access =/= read_write) ->
	      {error,
	       str:fmt("no permissions for ~s on ~p ", [AccType, F])};
	     true ->
	      {ok, Dir}
	  end;
	_ ->
	  case file:make_dir(Dir) of
	    ok ->
	      {ok, Dir};
	    {error, Error} ->
	      {error,
	       str:fmt("can't create ~p, reason ~s", [F, Error])}
	  end
      end
  end.

d(F,A) -> d(get(debug),F,A).

d(true,F,A) ->
  io:format("STL_DBG:" ++ F ++ "~n",A);
d(_,_F,_A) -> 
  ok.

error(Info, Desc) ->
  {error, {Info, ?MODULE, Desc}}.

format_error(ErrorDesc) ->
  lists:flatten(error_format(ErrorDesc)).

print_error({_, Desc}) ->
  print_error(Desc);
print_error(ErrDesc={_, Module, Desc}) ->
  case catch apply(Module, format_error, [Desc]) of
    {'EXIT', Reson} ->
      io_lib:write(Desc);
    PrintError ->
      PrintError
  end;
print_error(Error) ->
  io_lib:write(Error).

error_format({process, Context, Out, Reson}) ->
  str:fmt("terminated with reason ~p~n;context ~p~n;output:~p",
		[Reson, Context, Out]);
error_format({search, Path, SearchPaths}) ->
  str:fmt("couldn't find source for template ~p in ~p",
		[Path, SearchPaths]);
error_format({load, ModuleName}) ->
  str:fmt("~p is not loaded", [ModuleName]);
error_format({author, ModuleName}) ->
  str:fmt("~p is not authorized", [ModuleName]);
error_format({load_template, Error, File}) ->
  case io_lib:deep_char_list(Error) of
    true ->
      str:fmt("load_template:~s ~s", [File, Error]);
    false ->
      str:fmt("load_template:~s ~p", [File, Error])
  end;
error_format(Info) ->
  str:fmt("~p", [Info]).

to_list({'(*', V}) ->
  V;
to_list(V) when is_integer(V), V =< 256 ->
  V;
to_list(Vs) when is_list(Vs) ->
  to_list(Vs, [], fun (L) -> lists:reverse(L) end);
to_list(V) ->
  io_lib:format("~w", [V]).

to_list([], Cs, Cont) ->
  Cont(Cs);
to_list([L|Rest], Cs, Cont) when is_list(L) ->
  to_list(L, Cs, fun (L1) -> to_list(Rest, L1, Cont) end);
to_list([C|Rest], Cs, Cont) ->
  to_list(Rest, [to_list(C)|Cs], Cont).

return_error(Error) ->
  Msg = return_error1(Error),
  error_logger:error_msg("STL:~s~n", [Msg]),
  {error, Msg}.

return_error1({error, Error}) ->
  return_error1(Error);
return_error1({_, Module, ErrorDescriptor}) ->
  apply(Module, format_error, [ErrorDescriptor]);
return_error1(Error) ->
  case io_lib:deep_char_list(Error) of
    true ->
      Error;
    _ ->
      io_lib:write(Error)
  end.

random_string(Len) ->
  randomx:alnum_string(Len).
