%%%----------------------------------------------------------------------
%%% File    : stl_scan.erl
%%% Author  :  <svg@surnet.ru>
%%% Purpose : Transform STL template to Erlang syntax form
%%% Created : 15 Sep 2001 by  <svg@surnet.ru>
%%%----------------------------------------------------------------------

-module(stl_trans).
-vsn('$Id: stl_trans.erl,v 1.3 2003/10/31 18:04:59 vsekissov Exp $.').
-author('svg@surnet.ru').

%%-compile(export_all).
-export([form_list/2, check_opts/1, format_error/1]).

-import(wslib.getopt).
-import(wslib.str).

-import(erl_syntax, [make_tree/2,
                     subtrees/1,
                     copy_attrs/2,
                     application/3,
                     revert/1,
                     prefix_expr_argument/1,
                     application_arguments/1,
                     application_operator/1,
                     atom_value/1,
                     get_pos/1,
                     is_proper_list/1,
                     list_elements/1,
                     nil/0,
                     operator_name/1,
                     prefix_expr_operator/1,
                     variable_name/1,
                     attribute/2,
                     clause/3,
                     variable/1,
                     eof_marker/0,
                     function/2,
                     type/1,
                     integer/1,
                     string/1,
                     tuple/1,
                     atom/1,
                     list/1,
                     flatten_form_list/1,
                     form_list_elements/1]).

-include("stl.hrl").

form_list(Str, UserOpts) ->
  Opts = check_opts(UserOpts),
  WriteFun = {getopt:value(writefun, Opts), getopt:value(writeargs, Opts)},
  IncludeDir = getopt:value(include_dir, Opts),
  Include = getopt:value(include, Opts),
  TmpDir = getopt:value(tmp_dir, Opts),
  case catch
    begin
      {ok, Data} = scan_string(Str++"\n"),
      {ok, PrepForms, Forms} = parse_string(Data, IncludeDir, Include, TmpDir),
      %%{ok, Tokens, _} =  erl_scan:string(Data),
      %%{ok, Forms} =  erl_parse:parse_exprs(Tokens),
      {ok, {Tree, Ctxt}} =
	transform(erl_syntax:form_list(Forms), WriteFun),
      module_make(erl_syntax:form_list(PrepForms), Tree, Ctxt, Opts)
    end of
    Ret={ok, _} ->
      Ret;
    Error ->
      return_error(wslib.misc:better_error(Error))
  end.

transform(Forms, WriteFun) ->
  NewTree =
    case subtrees(Forms) of
      [] ->
	{[], Forms};
      [G|Groups] ->
	Res = [topprocess(G, WriteFun, fun attribute/4)|
	       [topprocess(G, WriteFun, fun stlize_node/4)
		|| Group <- Groups]],
	{copy_attrs(Forms,
		    make_tree(erl_syntax:type(Forms),
			      [element(1, E) || E <- Res]
			     )),
	 lists:append([element(2, E) || E <- Res])
	}
    end,
  {ok, NewTree}.

attribute_make(file, [Name]) ->
  attribute_make(file, [Name, 1]);

attribute_make(Name, Vals) ->
  attribute(atom(Name), attribute_make_args(Vals)).

attribute_make_args(Vals) ->
  TypeFun = fun (V) when atom(V) ->
		atom(V);
		(V) when list(V) ->
		string(V);
		(V) when integer(V) ->
		integer(V)
	    end,
  [TypeFun(V) || V <- Vals].

module_make(PreForms, Tree, Ctxt, Opts) ->
  ProcName = getopt:value(procfun, Opts),
  InputName = getopt:value(inputfun, Opts),
  ProcParams0 = getopt:value(procargs, Opts, []),
  InputParams0 = getopt:value(input, Ctxt, []),
  ProcParams =  case lists:prefix(ProcParams0, InputParams0) of
		  true -> InputParams0;
		  _ -> ProcParams0 ++ InputParams0
		end,
  MTimeName = getopt:value(modtfun, Opts),
  MTime = getopt:value(mtime, Opts, 0),
  Attrs = [attribute_make(Name, [Vals]) ||
	    {Name, Vals} <- [{Name, getopt:value(Name, Opts)} ||
			      Name <- [file, module, author, vsn]],
	    Vals =/= undefined],
  ResTree = make_tree(type(Tree),
		      [Attrs ++
		       form_list_elements(flatten_form_list(PreForms)) ++
		       [
			{attribute, 4, export,[{ProcName, length(ProcParams)},
					       {InputName, 0},
					       {MTimeName, 0}]},
			function_make(ProcName, ProcParams,
				      form_list_elements(flatten_form_list(Tree))),
			function_make(InputName, [], [list([atom(A) || A <- ProcParams])]),
			function_make(MTimeName, [], [integer(MTime)]),
			eof_marker()
		       ]]),
  {ok, {erl_syntax:data(revert(ResTree)), ProcParams}}.

topprocess(Tree, WriteFun, Fun) ->
  {NewTree, {_, Ctxt}} =
    lists:mapfoldl(fun(Node, {TrunsFun, Ctxt}) ->
		       TrunsFun(Node, WriteFun, Ctxt, erl_syntax:type(Node))
		   end, {Fun, []}, Tree),
  {NewTree, Ctxt}.

function_make(Name, Params, Body) ->
  function(atom(Name), [clause([variable(V) || V <- Params], none, Body)]).

attribute(Node, WriteFun, Ctxt, prefix_expr) ->
  case is_attribute(Node, prefix_expr, input) of
    true ->
      {nil(), {fun stlize_node/4, Ctxt ++ [{input, attribute_args(Node)}]}};
    _ ->
      {stlize_node(Node, WriteFun, prefix_expr), {fun stlize_node/4, Ctxt}}
  end;
attribute(Node, WriteFun, Ctxt, string) ->
  {stlize_node(Node, WriteFun, string), {fun attribute/4, Ctxt}};
attribute(Node, WriteFun, Ctxt, NodeType) ->
  {stlize_node(Node, WriteFun, NodeType), {fun stlize_node/4, Ctxt}}.

attribute_args(Node) ->
  attribute_args(Node, application_arguments(prefix_expr_argument(Node))).

attribute_args(Node, []) ->
  [];
attribute_args(Node, [ArgList]) ->
  case is_proper_list(ArgList) of
    true ->
      [binding_name(B) || B <- list_elements(ArgList)];
    _ ->
      throw(error(get_pos(Node), "Bad directive format"))
  end;

attribute_args(Node, ArgList) ->
  throw(error(get_pos(Node), "Bad directive format")).

binding_name(Node) ->
  binding_name(Node, erl_syntax:type(Node)).
binding_name(Node, variable) ->
  variable_name(Node);
binding_name(Node, _) ->
  throw(error(get_pos(Node), "Not a proper binding name")).

is_attribute(Node, prefix_expr, AttrName) ->
  case catch is_attribute(Node, AttrName) of
    true ->
      true;
    _ ->
      false
  end;

is_attribute(_, _, _) -> false.

is_attribute(Node, AttrName) ->
  AppNode = prefix_expr_argument(Node),
  AppOp = application_operator(AppNode),
  (operator_name(prefix_expr_operator(Node)) == '-')
    and (erl_syntax:type(AppNode) == application)
    and (erl_syntax:type(AppOp) == atom)
    and (atom_value(AppOp) == AttrName).

stlize_node(Node, WriteFun, Ctxt, NodeType) ->
  {stlize_node(Node, WriteFun, NodeType), {fun stlize_node/4, Ctxt}}.

stlize_node(Node, {{Module, Fun}, FunParams}, NodeType) ->
  stlize_node(Node, Module, Fun, FunParams, NodeType);
stlize_node(Node, {Fun, FunParams}, NodeType) ->
  stlize_node(Node, none, Fun, FunParams, NodeType);
stlize_node(Node, Fun, NodeType) ->
  stlize_node(Node, none, Fun, [], NodeType).

stlize_node(Node, Module, Fun, FunParams, NodeType)  when NodeType == variable;
                                                          NodeType == application;
                                                          NodeType == case_expr;
                                                          NodeType == if_expr;
                                                          NodeType == block_expr;
                                                          NodeType == infix_expr;
                                                          NodeType == prefix_expr;
                                                          NodeType == catch_expr;
                                                          NodeType == receive_expr;
                                                          NodeType == atom;
                                                          NodeType == binary;
                                                          NodeType == float;
                                                          NodeType == integer;
                                                          NodeType == char;
                                                          NodeType == string;
                                                          NodeType == list;
                                                          NodeType == record;
                                                          NodeType == tuple;
                                                          NodeType == list_comp ->
  application(atom(Module), atom(Fun), [variable(V) || V <- FunParams] ++ [Node]);

stlize_node(Node, _, _, _, _) ->
  Node.

scan_string(Chars) ->
  case expand_text(Chars, [], 0, 0, []) of
    {ok, List} ->
      {ok, lists:reverse(List) ++ ". "};
    Error ->
      throw(Error)
  end.

parse_string(Str, Opts, Include, TmpDir) ->
  FileName =
    str:fmt("~s/~s~s.erl", [TmpDir, ?MODULE, random_string(8)]),
  IncludeStr =
    lists:flatten([io_lib:format("-include(\"~s\").\n", [I]) || I <- Include]),
  Ret =
    (catch
      begin
	{ok, Out} = file:open(FileName, [write]),
	ok = file:write(Out, IncludeStr ++ "'__never_use_it__' () -> " ++ Str),
	file:close(Out),
	case epp:parse_file(FileName, Opts, []) of
	  {ok, Forms} ->
	    split_parser_output(Forms);
	  Error ->
	    Error
	end
      end),
  file:delete(FileName),
  Ret.

split_parser_output(Forms) ->
  {Prep, Rest} =
    lists:splitwith(fun ({function, _, '__never_use_it__', 0, _}) ->
			false;
			(_) ->
			true
		    end, Forms),
  case Rest of
    [MyExpr, {eof, _}] ->
      case MyExpr of
	{function, _, _, _,[{clause, _,[],[],Exprs}]} ->
	  {ok, Prep, Exprs};
	Error ->
	  Error
      end;
    _ ->
      case lists:keysearch(error, 1, Forms) of
	{value, {error, {Line, Mod, Desc}}} ->
	  ErrorStr = str:fmt("line:~w, ~s",[Line, Mod:format_error(Desc)]),
	  {error, {parse_error, ErrorStr}};
	_ ->
	  {error, parse_error}
      end
  end.

skip_blanks([Ch|Chars], Blanks, C, L, Res) when Ch == $\n ->
  skip_blanks(Chars, [Ch|Blanks], C, L+1, Res);
skip_blanks([Ch|Chars], Blanks, C, L, Res) when Ch == $\t;
                                                Ch == $\r;
                                                Ch == $\040 ->
  skip_blanks(Chars, [Ch|Blanks], C+1, L, Res);
skip_blanks(Chars, _, C, L, Res) ->
  {Chars, C, L}.

expand_code("\n"++Rest, Token, C, L, Res) ->
  expand_code(Rest, [$\n|Token], C, L+1, Res);

expand_code("*)"++Rest, Token, C, L, Res) ->
  expand_text(Rest, [], C+2, L, Token ++ Res);
expand_code("\\*\\)" ++ Rest, Str, C, L, Res) ->
  expand_code(Rest, ")*"++Str, C+4, L, Res);
expand_code([$\\,Ch|Rest], Token, C, L, Res) ->
  expand_code(Rest, [Ch, $\\|Token], C+2, L, Res);
expand_code([Ch|Rest], Token, C, L, Res) ->
  expand_code(Rest, [Ch|Token], C+1, L, Res);
expand_code([], Token, C, L, Res) ->
  error(L, lists:reverse(Token)).

expand_text("(*"++Rest, Str, C, L, Res) ->
  {Code, NC, NL} = skip_blanks(Rest, [], C+2, L, []),
  expand_code(Code, [], NC, NL, quote(Str) ++ Res);
expand_text("\n"++Rest, Str, C, L, Res) ->
  expand_text(Rest, [$\n|Str], C, L+1, Res);
expand_text("\\(\\*" ++ Rest, Str, C, L, Res) ->
  expand_text(Rest, "*("++Str, C+4, L, Res);
expand_text([$\\|Rest], Str, C, L, Res) ->
  expand_text(Rest, "\\\\"++Str, C+1, L, Res);
expand_text([$"|Rest], Str, C, L, Res) ->
  expand_text(Rest, "\"\\"++Str, C+1, L, Res);
expand_text([Ch|Rest], Str, C, L, Res) ->
  expand_text(Rest, [Ch|Str], C+1, L, Res);
expand_text([], Str, C, L, []) ->
  {ok, quote(Str)};
expand_text([], Str, C, L, Res) ->
  DotPos = string:span(Res, " \t\n\r") + 1,
  SkipPos = case catch lists:nth(DotPos, Res) of
	      Ch when Ch == $.; Ch == $, ->
		DotPos+1;
	      Ch when integer(Ch) ->
		DotPos;
	      {'EXIT', _} ->
		DotPos
	    end,
  {ok, quote(Str) ++ "," ++ string:sub_string(Res, SkipPos)}.

quote([]) ->
  [];
quote(Str) ->
  "}\"" ++ Str ++ "\",'*('{".

check_opts(Opts) ->
  getopt:options(Opts, [{include_dir, [], list},
			{include, [], list},
			{tmp_dir, "/tmp", list},
			{file,      undefined, list},
			{module,    required,  atom},
			{author,    undefined, atom},
			{vsn,       undefined, atom},
			{writefun,  {stl, write}, any},
			{writeargs, [?CONTEXT_BINDING], list},
			{procfun,   process, atom},
			{inputfun,  input,   atom},
			{procargs,  [?CONTEXT_BINDING], list},
			{mtime,   0, integer},
			{modtfun,   modtime, atom}]).

format_error(Message) ->
    case io_lib:deep_char_list(Message) of
	true ->
	    Message;
	_ ->
	    io_lib:write(Message)
    end.

error(Pos, Text) ->
  {error, {Pos, ?MODULE, Text}}.

return_error({error, {ErrorLine, Module, ErrorDescriptor}, EndLine}) ->
  {error,
   io_lib:format("line: ~w:~w ", [ErrorLine, EndLine]) ++
   apply(Module, format_error, [ErrorDescriptor])};
return_error({error, {ErrorLine, Module, ErrorDescriptor}}) ->
  {error,
   io_lib:format("line ~w ", [ErrorLine]) ++
   apply(Module, format_error, [ErrorDescriptor])};
return_error(Error) ->
  Error.

random_string(Len) ->
  randomx:alnum_string(Len).
