%%%-------------------------------------------------------------------
%%% File    : stlfgen.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : STL file generation script
%%%
%%% Created : 11 Mar 2003 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id: stlfgen.erl,v 1.1.1.1 2003/05/05 20:20:58 vsekissov Exp $
%%
%%**
%% 
%% @doc STL file generation script
%% @end
%%*
-module(stlfgen).

-export([main/1, start/1, do/0]).

-import(wslib.misc).

start(Opts) ->
  process(Opts).

main(Opts) ->
  process(Opts).

do() ->
  P = erlang:open_port({fd, 0, 1}, [eof]),
  Str = get_line(P, ""),
  {ok, Opts} = regexp:split(string:strip(Str), "[ \t\n\r]*--[ \t\n\r]*"),
  process(Opts).

get_line(Port, Acc) ->
  receive
    {Port, {data, Data}} ->
      get_line(Port, Acc++Data);
    {Port, eof} ->
      Acc
  end.

display_help() ->

  Help = "
Usage: stlfgen --template=TEMPLATE [switch...]
switches: --template-suffix=SUFFIX
                  template suffix (default is suffix of TEMPLATE)
          --template-dir=DIR
                  root directory of templates location
          --var-file=FILE
                  file to read list of template variables from,
                  must contain list of tuples {NAME, VALUE} in
                  Erlang format. NAME must be the atom
          --out-file=FILE
                  Output file (default is standard output)
          --VARNAME=VALUE
          --help
                  display this message.
",
  stderr(Help).

process(Opts) ->
  case catch
    begin
      {Props, Vars} = parse_opts(Opts),
      Ctxt = create_context(Props, Vars),
      [Tmpl, Suffix, Dir, File] =
	[proplists:get_value(K, Props) ||
	  K <- [template, 'template-suffix', 'template-dir', 'out-file']],
      StlOpts = [{dir, Dir}, {tmpl_suffix, Suffix}],
      {ok, _} = stl:start_link(StlOpts),
      {ok, {_, Res}} = stl:process(Tmpl, [], Ctxt),
      case File of
	undefined ->
	  ok = io:put_chars(Res);
	_ ->
	  ok = file:write_file(File, list_to_binary(Res))
      end,
      ok
    end of
    ok ->
      "";
    {bad_option, Opt} ->
      stderr(io_lib:format("Bad option format ~s", [Opt]));
    show_help ->
      display_help();
    Error ->
      stderr(io_lib:format("~p", [misc:better_error(Error)]))
  end.

parse_opts([[]]) ->
  throw(show_help);
parse_opts(Opts) ->
  {Props, Vars} = sort_opts(lists:flatmap(fun parse_opt/1, Opts)),
  ValidProps = [validate_prop(P, Props) || P <- Props],
  {ValidProps, Vars}.
  
sort_opts(Opts) ->
  {Props, Vars} =
    lists:foldl(fun (O={K,_}, {Ps,Vs}) ->
		    case lists:member(K, props()) of
		      true ->
			{[O|Ps], Vs};
		      false ->
			{Ps, [O|Vs]}
		    end
		end, {[], []}, Opts),
  SortedProps = [{K, proplists:get_value(K, Props)} || K <- props()],
  {SortedProps, Vars}.

validate_prop({Key, Val}, Props) ->
  {Key, validate_prop_val(Key, Val, Props)}.

validate_prop_val('help', true, Ps) ->
  throw(show_help);
validate_prop_val(template, undefined, Ps) ->
  throw(show_help);
validate_prop_val('template-suffix', undefined, Ps) ->
  default_suffix(proplists:get_value(template, Ps));
validate_prop_val('template-dir', undefined, _) ->
  ".";
validate_prop_val(_K, V, _Ps) ->
  V.

default_suffix(Str) ->
  {RevSuf, RevBase} =
    lists:splitwith(fun ($.) -> false;
			(_) -> true
		    end, lists:reverse(Str)),
  case RevBase of
    [] ->
      "";
    _ ->
      [$.|lists:reverse(RevSuf)]
  end.

parse_opt([]) ->
  [];
parse_opt(O) ->
  P = {Name, Val} = parse_var_str(O),
  if Name == '' ->
      throw({bad_option, "--" ++ O});
     true ->
      [P]
  end;
parse_opt(O) ->
  throw({bad_option, O}).

create_context(Props, Vars) ->
  FileVars =
    case proplists:get_value('var-file', Props) of
      undefined ->
	[];
      FileName ->
	{ok, FV} = file:consult(FileName),
	FV
    end,
  EnvVars = [parse_var_str(Str) || Str <- os:getenv()],
  Ctxt = stl:context_set(EnvVars ++ FileVars ++ Vars, stl:context_new()),
  Ctxt.

parse_var_str(Str) -> 
  {Name, Value} =
    case
      lists:splitwith(fun ($=) -> false;
			  (_) -> true
		      end, Str)
      of
      {N, [$=|V]} ->
	{N, V};
      {N, []} ->
	{N, true}
    end,
  {list_to_atom(Name), Value}.

stderr(Str) ->
  P = erlang:open_port({fd, 0, 2}, []),
  erlang:port_command(P, Str),
  erlang:port_command(P, "\n"),
  erlang:port_close(P).

props() ->
  [help, template, 'template-suffix', 'template-dir', 'out-file', 'var-file'].

