%%%-------------------------------------------------------------------
%%% File    : db.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : 
%%%
%%% Created :  9 Mar 2004 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id$
%%
%%**
%% 
%% @doc
%% @end
%%*

-module(aaa.db).

-include("aaa_schema.hrl").

-export([lookup_service/2, lookup_command/2, lookup/1, lookup/2]).
-export([write/1]).

-import(lists).
-import(mnesia).
-import(regexp).
-import(wslib.mnesiadb).
-import(wslib.props).
-import(wslib.listn).
-import(wslib.tuple).
-import(wslib.misc).

-define(default_user, "DEFAULT").
-define(default_authorization, default_author).
-define(record_tuple(_N), list_to_tuple([_N|record_info(fields, _N)])).

%%
%% @spec write({ObjType, Row}) -> ok | exit()
%% ObjType = user | enable | group | service | service_group
%%          | service_constraint | service_group_constraint | nas | global
%% Row = [{Key, Value}]
%% Key = atom()
%% Value = term()
%% 
write({ObjType, Row}) ->
  Table = objtype_to_mnesia_table(ObjType),
  EmptyRow = mnesiadb:make_object(Table),
  mnesiadb:write_row(Table, mnesiadb:make_object(Row, EmptyRow)).

%%
%% @spec lookup_command(User, Cmd) -> permit | deny | undefined | PermRegexps
%%
lookup_command(UserId, Cmd) ->
  case lookup({service, UserId, {UserId, "shell", Cmd}}) of
    {service, Service} ->
      props:gv(attrs, Service);
    Other ->
      Other
  end.

lookup_service(UserId, ServiceType) ->
  lookup({service, UserId, {UserId, ServiceType}}).

lookup(Key, Default) ->
  case lookup(Key) of
    undefined ->
      Default;
    Value ->
      Value
  end.
%%
%% @spec lookup({global, Term}) -> term() | undefined
%%
%% @spec lookup({user, UserId::string()}) -> {user, KVs} | undefined
%%
%% @spec lookup({enable, Id::integer()}) -> {enable, KVs} | undefined
%%
%% @spec lookup({nas, NasId::string()}) -> {nas, KVs} | undefined
%% 
%% @spec lookup({group, GroupId::string()}) -> {group, KVs} | undefined
%% 
%% @spec lookup({service, UserId::string(), ServiceId::term()})
%% -> {service, KVs} | permit | deny | undefined
%%              
lookup(Key) ->
  ObjType = if is_tuple(Key) -> element(1, Key);
	       true -> global
	    end,
  return_object(ObjType, do_lookup(Key)).

return_object(ObjType, Rec) when ((ObjType == user)
				  or (ObjType == enable)
				  or (ObjType == group)
				 ) and is_record(Rec, aaa_user) ->
  {ObjType, mnesiadb:make_object(aaa_user, Rec)};
return_object(OT=service, Service) when is_list(Service) ->
  {OT, Service};
return_object(OT=nas, Rec) when is_record(Rec, aaa_nas) ->
  {OT, mnesiadb:make_object(aaa_nas, Rec)};
return_object(global, Rec) when is_record(Rec, aaa_global) ->
  Rec#aaa_global.value;
return_object(_, Other) ->
  Other.
  
do_lookup(_Key={_Type=user, Id}) ->
  fill_member_group(
    case fetch(aaa_user, Id) of
      undefined ->
	fetch(aaa_user, ?default_user);
      User ->
	User
    end
   );
do_lookup(Key={_Type=enable, _Id}) ->
  fill_member_group(fetch(aaa_user, Key));
do_lookup({_Type=group, Id}) ->
  fill_member_group(fetch(aaa_group, Id));
do_lookup({_Type=nas, Id}) ->
  fetch(aaa_nas, Id);
do_lookup({service, UserId, ServiceId}) ->
  case do_lookup({user, UserId}) of
    undefined ->
      undefined;
    _U=#aaa_user{id=UserId,
		 default_service=DS,
		 group_member_id=GMIs
		} ->
      ServiceIds =
	[{service, ServiceId}|[{group, setelement(1, ServiceId, GI)} || GI <- GMIs]],
      case find_service(ServiceIds) of
	undefined ->
	  if DS == undefined ->
	      lookup(?default_authorization);
	     true -> DS
	  end;
	Service ->
	  Service
      end
  end;
do_lookup({global, Global}) ->
  fetch(aaa_global, Global);
do_lookup(Global) when is_atom(Global) ->
  fetch(aaa_global, Global).

%%
fill_member_group(undefined) ->
  undefined;
fill_member_group(User) ->
  Groups = fetch_user_groups(User),
  GroupIds = [G#aaa_user.id || G <- Groups],
  merge_users([User#aaa_user{group_member_id=GroupIds}|Groups]).

fetch_user_groups(_U=#aaa_user{group_member_id=GrId}) ->
  fetch_user_groups(GrId, []).

fetch_user_groups(undefined, Groups) ->
  lists:reverse(Groups);
fetch_user_groups(GrId, Groups) ->
  case fetch(aaa_group, GrId) of
    undefined ->
      lists:reverse(Groups);
    G=#aaa_user{group_member_id=NxtGrId} ->
      fetch_user_groups(NxtGrId, [G|Groups])
  end.

find_service(Ids) ->
  {Services, Cs} =
    lists:unzip([S || S <- [fetch_service_with_constraints(Id) || Id <- Ids],
		       S =/= undefined]),
  Constraints = [C || C <- Cs, C =/= undefined],
  Service =
    if Services == [] ->
	undefined;
       true ->
	Constraint =
	  if Constraints == [] ->
	      [];
	     true ->
	      ConstrObj =
		mnesiadb:make_object(aaa_service_constraint,
				     merge_service_constraints(Constraints)),
	      props:delete(service_id, ConstrObj)
	  end,
	mnesiadb:make_object(aaa_service, merge_services(Services))
	  ++ Constraint
    end,
  Service.

fetch_service_with_constraints(Id={Tbl, _}) ->
  case fetch_service(Id) of
    undefined ->
      undefined;
    Service ->
      Constraints =
	fetch_service_constraints({Tbl, Service#aaa_service.service_id}),
      {Service, Constraints}
  end.

fetch_service({service, Id}) ->
  fetch(aaa_service, Id);
fetch_service({group, Id}) ->
  fetch(aaa_service_group, Id).

fetch_service_constraints({_, undefined}) ->
  undefined;
fetch_service_constraints({service, Id}) ->
  fetch(aaa_service_constraint, Id);
fetch_service_constraints({group, Id}) ->
  fetch(aaa_service_group_constraint, Id).

%%
fetch(Tab, Key) ->
  case find(Tab, Key) of
    {ok, Value} ->
      Value;
    Other ->
      Other
  end.

find(Tab, Key) ->
  case mnesia:dirty_read(Tab, Key) of
    [Rec] ->
      {ok, Rec};
    [] ->
      undefined
  end.

objtype_to_mnesia_table(ObjType) when ObjType == user;
				      ObjType == enable ->
  aaa_user;
objtype_to_mnesia_table(group) ->
  aaa_group;
objtype_to_mnesia_table(service) ->
  aaa_service;
objtype_to_mnesia_table(service_constraint) ->
  aaa_service_constraint;
objtype_to_mnesia_table(service_group) ->
  aaa_service_group;
objtype_to_mnesia_table(service_group_constraint) ->
  aaa_service_group_constraint;
objtype_to_mnesia_table(global) ->
  aaa_global;
objtype_to_mnesia_table(nas) ->
  aaa_nas;
objtype_to_mnesia_table(account) ->
  aaa_account.

merge_services(Services) ->
  RecFields = ?record_tuple(aaa_service),
  tuple:merge([RecFields|Services], fun merge_records/1).

merge_service_constraints(Constraints) ->
  RecFields = ?record_tuple(aaa_service_constraint),
  tuple:merge([RecFields|Constraints], fun merge_records/1).

merge_users(Users) ->
  RecFields = ?record_tuple(aaa_user),
  tuple:merge([RecFields|Users], fun merge_records/1).

merge_records([N|Vs]) ->
  case lists:member(N, [attrs]) of
    false ->
      case listn:first(fun (E) -> E =/= undefined end, Vs) of
	{ok, V} -> V;
	false -> undefined
      end;
    true ->
      lists:append([case V of undefined -> []; _ -> V end || V <- Vs])
  end.

	
