%%%-------------------------------------------------------------------
%%% File    : odbcsync.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : 
%%%
%%% Created : 12 Oct 2004 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%%**
%% 
%% @doc
%% @end
%%*
-module(aaa.odbcsync).
-author(svg@surnet.ru).

-behaviour(plain_fsm).

-export([import/0, pull_changes/0, mnesia_subscribe/0, translate_by_service/2]).
-export([update/3]).

%% internal exports
-export([data_vsn/0, code_change/3]).

-import(erlang).
-import(odbcdb).
-import(odbcdb_schema).
-import(error_logger).
-import(proplists).
-import(lists).
-import(error_logger).
-import(string).
-import(mnesia).
-import(plain_fsm).
-import(wslib.mnesiadb).
-import(wslib.props).
-import(wslib.cal).
-import(wslib.passwd).

-include("plain_fsm.hrl").
-include("aaa_schema.hrl").

import() ->
  {ok, Db} = odbcdb:connect(),
  import(Db),
  odbcdb:disconnect(Db).

import(Db) ->
  ServiceQuery = odbcdb:mk_query('AAA_SERVICE', []),
  {ok, RowDb} = odbcdb:connect(),
  Ret =
    odbcdb:fold(Db, ServiceQuery, [],
		fun (Row, {GoodCnt, BadCnt}) ->
		    case catch translate_by_service(RowDb, Row) of
		      ok ->
			{GoodCnt+1, BadCnt};
		      Error ->
			error_logger:error_report([?MODULE, {service, Row}, Error]),
			{GoodCnt, BadCnt+1}
		    end
		end,
		{0, 0}
	       ),
  odbcdb:disconnect(RowDb),
  Ret.

mnesia_subscribe() ->
  Pid = plain_fsm:spawn_link(
	  ?MODULE,
	  fun() ->
	      mnesia:subscribe({table, aaa_ar, simple}),
	      mnesia_subscribe_loop([])
	  end),
  {ok, Pid}.

mnesia_subscribe_loop(_) ->
  plain_fsm:extended_receive(
    receive
      {mnesia_table_event, {write, AR, _}} when is_record(AR, aaa_ar) ->
	{ok, Db} = odbcdb:connect(),
	TrF =
	  fun () ->
	      {updated, 1} =
		odbcdb:insert_row(
		  Db, account_transaction,
		  [{id, odbcdb:next_rowid(Db, account_transaction)},
		   {post_date, erlang:universaltime()}
		   |translate(AR, Db)
		  ])
	  end,
	odbcdb:transaction(Db, TrF),
	odbcdb:disconnect(Db),
	mnesia_subscribe_loop([]);
      _ ->
	mnesia_subscribe_loop([])
    end
   ).

pull_changes() ->
  {ok, Db} = odbcdb:connect(),
  Ret = pull_changes(Db),
  odbcdb:disconnect(Db),
  case Ret of
    {'EXIT', Error} ->
      exit(Error);
    Other ->
      Other
  end.

pull_changes(Db) ->
  SeqKeyN = {odbcsync, audit_trail_last},
  LastSeq = db:lookup({global, SeqKeyN}, 0),
  Query = odbcdb:mk_select_query(audit_trail_log,
				 [{'>', audit_trail_seq}]
				),
  {ok, RowDb} = odbcdb:connect(),
  Ret =
    case catch
      odbcdb:fold(Db, Query, [{audit_trail_seq, LastSeq}],
		  fun (Row, Acc) ->
		      case catch sync_by_audit(RowDb, Row) of
			ok ->
			  props:gv(audit_trail_seq, Row);
			Error ->
			  error_logger:error_report([?MODULE, {pull_changes, Row}, Error]),
			  throw({Acc, error})
		      end
		  end,
		  LastSeq,
		  [{order_by, [audit_trail_seq]}]
		 ) of
      I when is_integer(I) ->
	db:write({global, [{id, SeqKeyN}, {value, I}]}),
	ok;
      {I, error} when is_integer(I) ->
	db:write({global, [{id, SeqKeyN}, {value, I}]}),
	error;
      Other ->
	Other
    end,
  odbcdb:disconnect(RowDb),
  Ret.

sync_by_audit(Db, Row) ->
  [Action, Table, AuditId] =
    props:gv_map([action, table_name, audit_trail_seq], Row),
  sync_by_audit(Action, Table, AuditId, Db).

%% FixMe - we now using audit_trail_seq instead of unique table _id_
%% update, insert
sync_by_audit(Op, party_relationship_service, Id, Db) when Op == 'UPDATE';
							   Op == 'INSERT' ->
  ServiceRow = get_service(odbc, Db, Id),
  {_, WriteServiceF} = party_relationship_service(Db, ServiceRow),
  {UserRow, WriteUserF} = service_user(Db, ServiceRow),
  {_, WritePartiesF} =
    parties(Db, lists:usort([props:gv(party_id, UserRow),
			    props:gv(party_id_to, ServiceRow),
			    props:gv(party_id_from, ServiceRow)
			   ])),
 
  TrF =
    fun () ->
	mk_delete_service(Id),
	WriteServiceF(),
	WritePartiesF(),
	WriteUserF()
    end,
  {atomic, _} = mnesia:sync_transaction(TrF),
  ok;
sync_by_audit(Op, user_login, RowId, Db) when Op == 'UPDATE';
					      Op == 'INSERT' ->
  update(user, Db, RowId);
sync_by_audit('INSERT', T=account_transaction, RowId, Db) ->
  {ok, TrRow} = odbcdb:read_row(Db, T, [{id, RowId}]),
  [TrType, SourceType, AccountId] =
    props:gv_map([transaction_type_id, source_type_id, account_id], TrRow),
  [AccountRow] = 
    odbcdb:select_rows(Db, party_relationship_account, [{id, AccountId}]),
  {ok, PartyRel} = odbcdb:read_row(Db, party_relationship, AccountRow),
  [CurrencyTypeId, Amount] =
    props:gv_map([currency_type_id, amount], AccountRow),
  PartyRelId = props:gv(id, PartyRel),
  TrF =
    fun () ->
	[Account=#aaa_account{debit=D, credit=C}] =
	  mnesia:read(aaa_account, {PartyRelId, CurrencyTypeId}),
	NewAccount =
	  case TrType of
	    charge ->
	      Account#aaa_account{debit=D+Amount};
	    payment ->
	      Account#aaa_account{credit=C+Amount}
	  end,
	mnesia:write(NewAccount)
    end,
  {atomic, _} = mnesia:transaction(TrF),
  ok;
sync_by_audit('INSERT', T=party_relationship_account, RowId, Db) ->
  {ok, AccountRow} = odbcdb:read_row(Db, T, [{id, RowId}]),
  {ok, PartyRel} = odbcdb:read_row(Db, party_relationship, AccountRow),
  AccountObj = translate_account(odbc, AccountRow, PartyRel),
  mnesiadb:write_row(aaa_account, AccountObj);
sync_by_audit('UPDATE', T=party_relationship_account, RowId, Db) ->
  {ok, AccountRow} = odbcdb:read_row(Db, T, [{id, RowId}]),
  {ok, PartyRel} = odbcdb:read_row(Db, party_relationship, AccountRow),
  NewAccountObj = translate_account(odbc, AccountRow, PartyRel),
  TrF =
    fun () ->
	[OldAccountObj] = mnesiadb:read(aaa_account, NewAccountObj),
	AccountObj =
	  [proplists:lookup(credit_limit, NewAccountObj)|OldAccountObj],
	mnesiadb:write(aaa_account, AccountObj)
    end,
  {atomic, _} = mnesia:transaction(TrF),
  ok;
%% delete
sync_by_audit('DELETE', party_relationship_service, RowId, Db) ->
  {atomic, _} = mnesia:transaction(mk_delete_service(RowId)),
  ok;
sync_by_audit('DELETE', user_login, RowId, Db) ->
  mnesia:dirty_delete(aaa_user, RowId);
sync_by_audit('DELETE', party, RowId, Db) ->
  mnesia:dirty_delete(aaa_party, RowId);
sync_by_audit('DELETE', T=party_relationship_account, RowId, Db) ->
  {ok, AccountRow} = odbcdb:read_row(Db, T, [{id, RowId}]),
  {ok, PartyRel} = odbcdb:read_row(Db, party_relationship, AccountRow),
  AccountObj = translate_account(odbc, AccountRow, PartyRel),
  mnesiadb:delete_row(aaa_account, AccountObj);
sync_by_audit(_, _, _, _) ->
  ok.

translate_by_service(Db, ServiceId) when is_integer(ServiceId) ->
  ServiceRow = get_service(odbc, Db, ServiceId),
  translate_by_service(Db, ServiceRow);
translate_by_service(Db, ServiceRow) ->
  {_, WriteServiceF} = party_relationship_service(Db, ServiceRow),
  {UserRow, WriteUserF} = service_user(Db, ServiceRow),
  {_, WriteAccountsF} = party_relationship_accounts(Db, ServiceRow),
  {_, WritePartiesF} =
    parties(Db, lists:usort([props:gv(party_id, UserRow),
			     props:gv(party_id_to, ServiceRow),
			     props:gv(party_id_from, ServiceRow)
			    ])),
  TrF =
    fun () ->
	WriteServiceF(),
	WritePartiesF(),
	WriteAccountsF(),
	WriteUserF()
    end,
  {atomic, _} = mnesia:sync_transaction(TrF),
  ok.

party_relationship_service(_Db, ServiceRow) ->
  {Services, ServiceConstraints} =
    translate_service(odbc, ServiceRow),
  {ServiceRow,
   fun () ->
       lists:foreach(fun (S) -> mnesiadb:write(aaa_service, S) end,
		     Services
		    ),
       lists:foreach(fun (S) ->  mnesiadb:write(aaa_service_constraint, S) end,
		     ServiceConstraints
		    )
   end
  }.

%% Update
update(user, Db, RowId) ->
  {ok, UserRow} = odbcdb:read_row(Db, user_login, [{login_id, RowId}]),
  {_, WriteUserPartyF} = user_party(Db, UserRow),
  TrF =
    fun () ->
	NewUser = translate_user(odbc, UserRow, []),
	AaaUser =
	  case mnesiadb:read(aaa_user, [{id, RowId}]) of
	    [OldUser] ->
	      %% SQL-use hasn't yet group field, preserve it
	      mnesiadb:make_object(props:delete(group_member_id, NewUser),
				   OldUser);
	    [] ->
	      NewUser
	  end,
	mnesiadb:write(aaa_user, AaaUser),
	WriteUserPartyF()
    end,
  {atomic, _} = mnesia:transaction(TrF),
  ok.

%% party_relationship_parties(Db,  PartyRel) ->
%%   IntrestingParties = [props:gv(party_id_to, PartyRel),
%% 		       props:gv(party_id_from, PartyRel)
%% 		      ],
%%   parties(Db, IntrestingParties).

parties(Db, Ids) ->
  {selected, PartyRows} = odbcdb:select_rows(Db, party, [{in, id, Ids}]),
  {PartyRows,
   fun () ->
       lists:foreach(fun (P) ->
			 PartyObj = translate_party(odbc, P),
			   mnesiadb:write(aaa_party, PartyObj)
		     end, PartyRows)
   end
  }.
   
service_user(Db, ServiceRow) ->
  UserId = [proplists:lookup(K, ServiceRow)
	    || K <- odbcdb_schema:primary_key(user_login)],
  {ok, UserRow} = odbcdb:read_row(Db, user_login, UserId),
  AaaUser = translate_user(odbc, UserRow, ServiceRow),
  {UserRow,
   fun () -> mnesiadb:write(aaa_user, AaaUser) end
  }.

user_party(Db, UserRow) ->
  {ok, PartyRow} =
    odbcdb:read_row(Db, party, [{id, props:gv(party_id, UserRow)}]),
  {PartyRow,
   fun () -> mnesiadb:write(aaa_party, translate_party(odbc, PartyRow)) end
  }.

party_relationship_accounts(Db, PartyRel) ->
  {selected, AccRows} =
    odbcdb:select_rows(
      Db, party_relationship_account,
      props:lookup_map(odbcdb_schema:primary_key(party_relationship),
		       PartyRel)),
  {AccRows,
   fun () ->
       lists:foreach(fun (AR) ->
			 AccObj = translate_account(odbc, AR, PartyRel),
			 mnesiadb:write(aaa_account, AccObj)
		     end,
		     AccRows)
   end
  }.

convert_row(From, Spec) ->
  ValueF = fun ({_, {FromAttr, Default}}) ->
	       props:gv(FromAttr, From, Default);
	       ({_, FromAttr}) ->
	       props:gv(FromAttr, From);
	       ({_, FromAttrs, ConvF}) ->
	       apply(ConvF, props:gv_map(FromAttrs, From));
	       (FromAttr) when is_atom(FromAttr) ->
	       props:gv(FromAttr, From)
	   end,
  Attrs =
    lists:map(fun (ToAttr) when is_atom(ToAttr) ->
		  {ToAttr, ValueF(ToAttr)};
		  (R) ->
		  ToAttr = element(1,R),
		  {ToAttr, ValueF(R)}
	      end, Spec),
  Attrs.
		
translate_service(odbc, From) ->
  ServiceSpec =
    [
     {id,
      [login_id, service_type_id],
      fun (LoginId, ServiceTypeId) ->
	  case ServiceTypeId of
	    "shell" ->
	      {LoginId, ServiceTypeId};
	    "dialup" ->
	      {LoginId, "ppp"};
	    _ ->
	      {LoginId, ServiceTypeId}
	  end
      end
     },
     party_relationship_id,
     {service_id, id},
     {info, comments},
     status,
     tariff_id
    ],
  ConstSpec =
    [{service_id, id},
     {access_time,
      [restday_mask,
       restday_disabled_time_to, restday_disabled_time_from,
       workday_disabled_time_to, workday_disabled_time_from
      ],
      fun (RestMask, RestFrom, RestTo, WrkFrom, WrkTo) ->
	  #access_time{restday_mask=RestMask,
		       restday_disabled=time_interval(RestFrom, RestTo),
		       workday_disabled=time_interval(WrkFrom, WrkTo)
		      }
      end
     },
     {caller_phone,
      [caller_phone],
      fun (undefined) -> any;
	  (V) -> string:tokens(V, ",")
      end
     },
     {time_limit,
      [day_time_limit, month_time_limit, year_time_limit, total_time_limit],
      fun (D, M, Y, T) ->
	  ConvF = fun (0) -> unlimited;
		      (V) -> V
		  end,
	  #time_limit{day=ConvF(D),
		      month=ConvF(M),
		      year=ConvF(Y),
		      total=ConvF(T)
		     }
      end
     },
     max_sessions,
     nas
    ],
  ServiceAttrs = convert_row(From, ServiceSpec),
  ConstAttrs = convert_row(From, ConstSpec),
  Service1 = [{default_attr, deny}, {attrs, []}|ServiceAttrs],
  Services =
    case props:gv(id, Service1) of
      {UserId, ServiceType} when ServiceType == "ppp";
				 ServiceType == "slip" ->
	[Service1, lists:keyreplace(id, 1, Service1, {id, {UserId, "exec"}})];
      _ ->
	[Service1]
    end,
  ServiceConstraints = [ConstAttrs],
  {Services,ServiceConstraints}.

translate_user(odbc, From, Service) ->
  From1 =
    case props:gv(service_type_id, Service) of
      "dialup" ->
	[{group_member_id, "ppp"}|From];
      _ ->
	From
    end,
  Spec =
    [{id, login_id},
     party_id,
     {password,
      [{password, ""}],
      fun (Str) -> passwd:from_string(Str, des) end
     },
     enabled,
     disabled_datetime,
     group_member_id,
     {attrs, {attrs, []}}
    ],
  convert_row(From1, Spec).

translate_party(odbc, From) ->
  Spec =
    [id,
     party_type_id,
     name,
     {attrs, {attrs, []}}
    ],
  convert_row(From, Spec).

translate_account(odbc, From, PartyRel) ->
  Spec =
    [{id,
      [party_relationship_id, currency_type_id],
      fun (RelId, CurId) -> {RelId, CurId} end
     },
     debit,
     credit,
     credit_limit
    ],
  From1 = [{party_relationship_id, props:gv(id, PartyRel)}|From],
  convert_row(From1, Spec).

translate(
  AR=#aaa_ar{account_id={PartyRelId, CurrencyType},
	     entry_time=ET,
	     ss=SsId,
	     ce=#ce{quantity=Q,
		    amount=A,
		    discount=D,
		    tariff_id=TI
		   }
	    },
  Db) ->
  {selected, [PartyRel]} =
    odbcdb:select_rows(Db, party_relationship, [{id, PartyRelId}]),
  PartyRelPk = [proplists:lookup(K, PartyRel)
		|| K <- odbcdb_schema:primary_key(party_relationship)],
  AccountPk = [{currency_type_id, CurrencyType}|PartyRelPk],
  {ok, Account} = odbcdb:read_row(Db, party_relationship_account, AccountPk),
  AccountId = props:gv(id, Account),
  [#ss{se=#se{id=SI}}] = mnesia:dirty_read(aaa_ar_session, SsId),
  [#aaa_service{service_id=ServiceId}] = mnesia:dirty_read(aaa_service, SI),
  Rec = [{account_id, AccountId},
	 {transaction_type_id, charge},
	 {entry_date, ET},
	 {source_type_id, service},
	 {source_id, ServiceId},
	 {tariff_id, TI},
	 {quantity, Q},
	 {amount, A},
	 {discount, D}
	].
  
get_service(odbc, Db, ServiceId) ->
  Query = odbcdb:mk_query('AAA_SERVICE', [party_relationship_service.id]),
  {ok, ServiceRow} =
    odbcdb:fetch_row(Db, Query, [{party_relationship_service.id, ServiceId}]),
  ServiceRow.

mk_delete_service(Id) ->
  fun () ->
      mnesia:index_delete(aaa_service, Id, #aaa_service.service_id),
      mnesia:index_delete(aaa_service_constraint, Id, #aaa_service.service_id)
  end.

time_interval(From, To) when From == To ->
  undefined;
time_interval(From, To) ->
  cal:create_time_interval(From, To).

%% plain_fsm exports
data_vsn() ->
  1.

code_change(_, S, _) ->
  {ok, S}.
