%%%-------------------------------------------------------------------
%%% File    : session.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : Session logging, etc
%%%
%%% Created : 25 Aug 2004 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id$
%%
%%**
%% 
%% @doc
%% @end
%%*
-module(aaa.session).

-behaviour(gen_event).

-export([add_handler/0, delete_handler/0]).

%% gen_event callbacks
-export([init/1, handle_event/2, handle_call/2, handle_info/2, terminate/2, code_change/3]).

-import(erlang).
-import(calendar).
-import(lists).
-import(mnesia).
-import(tac_event).
-import(wslib.props).
-import(wslib.cal).
-import(wslib.str).
-import(wslib.misc).

-include("aaa_schema.hrl").
-include("tac_event.hrl").

-define(ss_params, [{"task_id", string},
		    {"service", string},
		    {"start_time",
		     integer,
		     {apply,
		      fun () ->
			  cal:datetime_to_epoch(erlang:localtime())
		      end
		     }
		    },
		    {"pre-session-time", integer, {value, 0}},
		    {"elapsed_time", integer, {value, 0}},
		    {"disc-cause", integer},
		    {"protocol", string}
		   ]).

-define(ue_params, [{"elapsed_time", integer, {value, 0}},
		    {"bytes_in", integer, {value, 0}},
		    {"bytes_out", integer, {value, 0}},
		    {"nas-tx-speed", string},
		    {"nas-rx-speed", string}
		   ]).

-record(state, {}).

add_handler() ->
  tac_event:add_handler(?MODULE, []).

delete_handler() ->
  tac_event:delete_handler(?MODULE, []).

init(_) ->
  {ok, #state{}}.

handle_event(#tac_event{reason=account,
			userid=#userid{user=User, peer=UserPeer},
			serverid=#serverid{peer=NasPeer},
			node=N,
			event_args=EAs,
			msg_args=MAs
		       }, S) ->
  RecordType = case props:gv(phase, EAs) of
		 Interim when Interim == continue;
			      Interim == duplicate ->
		   interim;
		 StartStop ->
		   StartStop
	       end,
  MsgKVs = [{K, V} || #msg_arg{name=K, value=V} <- MAs],
  [TaskId, Service, StartTimeEpoch, PreSessionTime, ElTime, DiscCause, Proto] =
    select_params(?ss_params, MsgKVs),
  SessionId = make_session_id(N, TaskId, StartTimeEpoch),
  EndTimeEpoch = StartTimeEpoch+ElTime+PreSessionTime,
  Session1 = if (RecordType == start) ->
		 undefined;
		true ->
		 read_session(SessionId)
	     end,

  Session2 =
    case Session1 of
      undefined ->
	StartTime = cal:epoch_to_datetime(StartTimeEpoch),
	{NasIp, _} = NasPeer,
	{_, UserPort} = UserPeer,
	{user, UserKVs} = db:lookup({user, User}, {user, []}),
	{nas, NasKVs} = db:lookup({nas, NasIp}, {nas, []}),
	ServiceKVs =
	  case db:lookup_service(User, Service) of
	    {service, SKVs} -> SKVs;
	    _ -> []
	  end,
	SE = make_se(NasPeer, NasKVs, ServiceKVs),
	SC = make_sc(User, UserPeer, UserKVs),
	#ss{id=SessionId,
	    service=Service,
	    start_time=StartTime,
	    sc=SC,
	    se=SE
	   };
      _ ->
	Session1
    end,
  {EndTime, TerminateCause} =
    if RecordType == stop ->
	{cal:epoch_to_datetime(EndTimeEpoch), DiscCause};
       true ->
	{undefined, undefined}
    end,
  UE = make_ue(Service, Proto, MsgKVs),
  Session3 = Session2#ss{record_type=RecordType,
			 end_time=EndTime,
			 terminate_cause=TerminateCause,
			 ue=UE
			},
  ChargeRet = charge:rate_session_usage(Session3),
  case {RecordType, ChargeRet} of
    {stop, {ok, _}} ->
      delete_session(Session3#ss.id);
    Other ->
      write_session(Session3)
  end,
  {ok, S};
handle_event(_, State) ->
  {ok, State}.

handle_call(_, S) ->
  {ok, unknown_call, S}.

handle_info(_, S) ->
  {ok, S}.

terminate(_, S) ->
  ok.

code_change(_OldVsn, State, _Extra) ->
  {ok, State}.

make_session_id(Node, TaskId, StartTimeEpoch) ->
  str:fmt("~s:~s:~w", [Node, TaskId, StartTimeEpoch]).

make_se(NasPeer, NasKVs, ServiceKVs) ->
  ProviderId = props:gv(provider_id, NasKVs),
  [ServiceId, TariffId] =
    props:gv_map([id, tariff_id], ServiceKVs),
  #se{id=ServiceId,
      tariff_id=TariffId,
      provider_id=ProviderId,
      provider_host=NasPeer
     }.

make_sc(User, UserPeer, UserKVs) ->
  SubscriberId = props:gv(party_id, UserKVs), 
  #sc{user_id=User,
      subscriber_id=SubscriberId,
      context_id=undefined,
      access_point=UserPeer
     }.

make_ue(Service, Proto, MsgKVs) when Service == "ppp";
				     Service == "slip" ->
  [ElapsedTime, BytesIn, BytesOut, NasTxSpeed, NasRxSpeed] =
    select_params(?ue_params, MsgKVs),
  #ue_ia{duration=ElapsedTime,
	 input_octets=BytesIn,
	 output_octets=BytesOut,
	 input_bandwidth=NasRxSpeed,
	 output_bandwidth=NasTxSpeed,
	 transport_proto=Proto,
	 nas_port_type=undefined
	};
make_ue(Service, Proto, MsgKVs) ->
  undefined.
  
select_params(ParamDescs, KVs) ->
  [param_value(ParamDesc, KVs) || ParamDesc <- ParamDescs].

param_value({Key, ValueType}, KVs) ->
  coerce_param_type(props:gv(Key, KVs), ValueType);
param_value({Key, ValueType, Default}, KVs) ->
  case coerce_param_type(props:gv(Key, KVs), ValueType) of
    undefined ->
      param_default(Default);
    V ->
      V
  end.

param_default({value, V}) ->
  V;
param_default({apply, F}) ->
  apply(F, []).

coerce_param_type(V=undefined, _) ->
  undefined;
coerce_param_type(I, integer) when is_integer(I) ->
  I;
coerce_param_type(Str, integer) when is_list(Str) ->
  list_to_integer(Str);
coerce_param_type(Str, string) ->
  Str.
  
%% Mnesia stuff
read_session(Id) ->
  case mnesia:dirty_read(aaa_session, Id) of
    [] ->
      undefined;
    [Ss] ->
      Ss
  end.

write_session(Session) ->
  mnesia:dirty_write(aaa_session, Session).

delete_session(Id) ->
  mnesia:dirty_delete(aaa_session, Id).
