%%%-------------------------------------------------------------------
%%% File    : tariff.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : 
%%%
%%% Created : 16 Sep 2004 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id$
%%
%%**
%% 
%% @doc
%% @end
%%*

-module(aaa.tariff).

-export([apply_tariff/3]).
-export([delete/1, delete_ces/2, delete_ce/3]).
-export([find_ce/3, info/1]).
-export([get_rec/1, find_rec/1]).
-export([write_obj/2, read_obj/1, write_ce_obj/4]).

-import(erlang).
-import(lists).
-import(mnesia).
-import(ets).
-import(wslib.cal).
-import(wslib.mnesiadb).

-include("aaa_schema.hrl").
-include_lib("stdlib/include/ms_transform.hrl").

%% @spec apply_tariff(TariffId::term(),
%%                    ChargeTime::datetime(),
%%                    ServiceSession::ss()
%%                    ) -> [ChargeEvent::ce()] | exit()
%%                    
apply_tariff(TariffId, ChargeTime, ServiceSession) ->
  InitCe = #ce{charge_time=ChargeTime,
	       tariff_id=TariffId
	      },
  case find_ces(TariffId, ChargeTime) of
    undefined ->
      exit(tariff_not_found);
    TariffCes ->
      apply_tariff_ces(TariffCes, ServiceSession, InitCe)
  end.

%% @spec find_ces(TariffId::term(), ChargeTime::datetime()) ->
%% [TariffChargeEvent::tariff_ce()] | undefined
%% 
find_ces(TariffId, ChargeTime) ->
  Tab = aaa_tariff_ce,
  QueryId =
    ets:fun2ms(fun (#aaa_tariff_ce{id={TI, FT}})
		   when TI == TariffId, FT =< ChargeTime ->
		   FT
	       end),
  case mnesia:dirty_select(Tab, QueryId) of
    [] ->
      undefined;
    ChangeTimes ->
      FromTime = lists:max(ChangeTimes),
      mnesia:dirty_read(Tab, {TariffId, FromTime})
  end.

get_rec(TariffId) ->
  {ok, Obj} = find_rec(TariffId),
  Obj.

find_rec(TariffId) ->
  mnesiadb:fetch(mnesia:dirty_read(aaa_tariff, TariffId)).

write_obj(TariffId, Attrs) ->
  mnesiadb:write_row(aaa_tariff, [{id, TariffId}|Attrs]).

read_obj(TariffId) ->
  mnesiadb:read_row(aaa_tariff, [{id, TariffId}]).
    
delete(TariffId) ->
  CeTab = aaa_tariff_ce,
  Tab = aaa_tariff,
  QueryId =
    ets:fun2ms(fun (#aaa_tariff_ce{id={TI, _}}) when TI == TariffId ->
		   TI
	       end),
  DelF = fun () ->
	     lists:foreach(fun (Key) ->
			       mnesia:delete(CeTab, Key)
			   end,
			   lists:usort(mnesia:select(CeTab, QueryId))
			  ),
	     mnesia:delete(Tab, TariffId)
	 end,
  do_transaction(DelF).

delete_ces(TariffId, FromTime) ->
  CeTab = aaa_tariff_ce,
  DelF = fun () ->
	     mnesia:delete(CeTab, {TariffId, FromTime})
	 end,
  do_transaction(DelF).

%% @spec info(TariffId::term()) -> TariffInfo | undefined
%% TariffInfo = [{id, TariffId::term()},
%%               {info, Info::string()},
%%               {charge_type, ChargeType::atom()},
%%               {currency_type, CurrencyType::atom()},
%%               {from_time, [FromTime::datetime()]}
%%              ]
%%              
info(TariffId) ->
  CeTab = aaa_tariff_ce,
  Tab = aaa_tariff,
  QueryId =
    ets:fun2ms(fun (#aaa_tariff_ce{id={TI, FT}}) when TI == TariffId ->
		   FT
	       end),
  SelF = fun () ->
	     case mnesiadb:read(Tab, TariffId) of
	       [] ->
		 undefined;
	       [Tariff] ->
		 ChangeDates = lists:usort(mnesia:select(CeTab, QueryId)),
		 Tariff ++[{from_time, ChangeDates}]
	     end
	 end,
  do_transaction(SelF).

write_ce_obj(TariffId, FromDate={_, _, _}, SeqNum, Ce) ->
  write_ce_obj(TariffId, {FromDate, {0, 0, 0}}, SeqNum, Ce);
write_ce_obj(TariffId, FromTime={Date, Time}, SeqNum, CeKVs) ->
  cal:valid_datetime(FromTime) orelse exit({bad_datetime, FromTime}),
  F = fun () ->
	  delete_ce(TariffId, FromTime, SeqNum),
	  mnesiadb:write(aaa_tariff_ce,
			 [{id, {TariffId, FromTime}},
			  {seq_num, SeqNum}
			  |CeKVs
			 ])
      end,
  do_transaction(F).

delete_ce(TariffId, FromTime, TariffSeqNum) ->
  F =
    fun () ->
	case find_ce(TariffId, FromTime, TariffSeqNum) of
	  undefined ->
	    ok;
	  TariffCe ->
	    mnesia:delete_object(TariffCe)
	end
    end,
  do_transaction(F).

find_ce(TariffId, FromTime, TariffSeqNum) ->
  CeTab = aaa_tariff_ce,
  Query =
    ets:fun2ms(fun (Obj=#aaa_tariff_ce{id={TI, FT}, seq_num=SN})
		   when TI == TariffId,
			FT == FromTime,
			SN == TariffSeqNum ->
		   Obj
	       end),
  F = fun () ->
	  mnesiadb:fetch(mnesia:select(CeTab, Query), undefined)
      end,
  do_transaction(F).

%%
apply_tariff_ces(TariffCes, ServiceSession, InitCe) ->
  [apply_tariff_ce(TCe, ServiceSession, InitCe) || TCe <- TariffCes].

apply_tariff_ce(TCe=#aaa_tariff_ce{id={Id, _FromDate},
				   chargeable_unit=CU,
				   measured_unit=MU,
				   customer_relationship_fun=CRF,
				   quantity_fun=QF,
				   amount_fun=AF,
				   discount_fun=DF
				  },
		SS,
		Ce0) ->
  
  Ce1 = Ce0#ce{chargeable_unit=CU, measured_unit=MU, tariff_id=Id},
  CustomerRelId = apply(CRF, [SS, Ce1]),
  Ce2 = Ce1#ce{customer_relationship_id=CustomerRelId},
  Quantity = apply(QF, [SS, Ce2]),
  Ce3 = Ce2#ce{quantity=Quantity},
  {Amount, CurrencyType} = apply(AF, [SS, Ce3]),
  Ce4 = Ce3#ce{amount=Amount, currency_type=CurrencyType},
  Discount = apply(DF, [SS, Ce4]),
  Ce = Ce4#ce{discount=Discount}.

do_transaction(F) ->
  case get(mnesia_activity_state) of
    TrId={_, _, _} ->
      F();
    _ ->
      {atomic, Res} = mnesia:transaction(F),
      Res
  end.
