%%%----------------------------------------------------------------------
%%% File    : tac_body.erl
%%% Author  :  <svg@surnet.ru>
%%% Purpose : 
%%% Created : 20 Nov 1999 by  <svg@surnet.ru>
%%% $Id: tac_body.erl,v 1.8 2002/09/04 01:05:52 svg Exp $
%%%----------------------------------------------------------------------

-module(tac_body).
-author('svg@surnet.ru').

-include("tacacs.hrl").

-export([parse/2, pack/2, authen_type/1, format_error/1]).

%% internal exports
-export([authen_service/1, authen_type/1, authen_method/1]).
-import(wslib.listn).
-import(wslib.str).

-import(tac_lib, [bytes_to_int/4, int_to_bytes/1]).

-define(dbg(_F, _A), tac_lib:dbg(get(dbg), _F, _A)).

%%You must catch functions parse/2 and pack/1, they throw when logical error occures 
%%Return packed/parsed packet or {error, packet_error}
%%
parse(Type, Data) ->
  Packet = parse0(Type, Data),
  ?dbg({"packet ", fun fmt_packet/1}, Packet),
  Packet.

pack(Sess, Body) when is_record(Body, error_reply) ->
  ?dbg({"packet", fun fmt_packet/1}, Body),
  {Sess#session{unencrypt=1}, []};
pack(Sess, Body) ->
  ?dbg({"packet ", fun fmt_packet/1}, Body),
  {Type, Data} = pack0(Body),
  {Sess#session{type=Type}, Data}.

parse0(_Type, []) -> %zero length packet
  #error_reply{};
%%Sended by client
parse0(PT, Data) when PT == authen_start;
		      PT == author_request;
		      PT == acct_request ->
  {Head, Data0} =
    case PT of
      acct_request ->
	listn:split_on(2, Data);
      _ ->
	listn:split_on(1, Data)
    end,
  [PrivLvl, AT, AS, UserLen, PortLen, RemAddrLen, DataLen=ArgCnt|Data1]
    = Data0,
  check_priv_lvl(PrivLvl),
  [User, Port, RemAddr|Data2] =
    case PT of
      authen_start ->
	split_on_positions(Data1, [UserLen, PortLen, RemAddrLen, DataLen]);
      _ ->
	{ArgLens, Data3} = split_on(ArgCnt, Data1),
	split_on_positions(Data3, [UserLen, PortLen, RemAddrLen] ++ ArgLens)
    end,
  Authen = #authen{priv_lvl=PrivLvl,
		   type=authen_type(AT),
		   service=authen_service(AS)
		  },
  UserId = #userid{user=User, peer={RemAddr, Port}},
  case PT of
    authen_start ->
      [Action] = Head,
      [UserData] = Data2,
      #authen_start{action=authen_action(Action),
		    authen=Authen,
		    user=UserId,
		    data=UserData
		   };
    author_request ->
      [AM] = Head,
      #author_request{authen=Authen#authen{method=authen_method(AM)},
		      user=UserId,
		      args=parse_args(Data2)
		     };
    acct_request ->
      [Flags, AM] = Head,
      #acct_request{flags=acct_type(Flags),
		    authen=Authen#authen{method=authen_method(AM)},
		    user=UserId,
		    args=parse_args(Data2)
		   }
  end;
parse0(authen_continue, [U1, U2, D1, D2, Flags|Body]) ->
  Lens = [bytes_to_int(0, 0, U2, U1), bytes_to_int(0, 0, D2, D1)],
  [UsrMsg, Data] = split_on_positions(Body, Lens),
  #authen_continue{flags=Flags, user_msg=UsrMsg, data=Data};
%%Sended by server
parse0(authen_reply, [Status, Flags, S2, S1, D2, D1|Body]) ->
  Lens = [bytes_to_int(0, 0, S2, S1), bytes_to_int(0, 0, D2, D1)],
  [ServerMsg, Data] = split_on_positions(Body, Lens),
  #authen_reply{status=authen_status(Status),
		flags=Flags,
		server_msg=ServerMsg,
		data=Data
	       };
parse0(author_response, [Status, ArgCnt, S2, S1, D2, D1| Tail]) ->
  {ArgLens, Body} = split_on(ArgCnt, Tail),
  Lens = [bytes_to_int(0, 0, S2, S1), bytes_to_int(0, 0, D2, D1)] ++ ArgLens,
  [ServerMsg, Data| Args] = split_on_positions(Body, Lens),
  #author_response{status=author_status(Status),
                   server_msg=ServerMsg, data=Data, args=parse_args(Args)};
parse0(acct_reply, [S2, S1, D2, D1, Status| Body]) ->
  Lens = [bytes_to_int(0, 0, S2, S1), bytes_to_int(0, 0, D2, D1)],
  [ServerMsg, Data] = split_on_positions(Body, Lens),
  #acct_reply{status=acct_status(Status), server_msg=ServerMsg, data=Data};

parse0(Type, _) ->
  throw_error(undef_type, Type).

%%Sended by server
pack0(#authen_start{action=Action,  
		    authen=#authen{priv_lvl=PrivLvl, 
                                   type=AT,
                                   service=AS
                                  }, 
		    user=#userid{user=User, peer={RemAddr, Port}}, 
		    data=Data}) ->
  Body =
    [authen_action(Action), PrivLvl,
     authen_type(AT), authen_service(AS),
     length(User), length(Port), length(RemAddr), length(Data)] ++
    User ++ Port ++ RemAddr ++ Data,
  {?TAC_PLUS_AUTHEN, Body};

pack0(#authen_continue{flags=Flags, user_msg=UsrMsg, data=Data}) ->
  [_, _, U2, U1] = int_to_bytes(length(UsrMsg)),
  [_, _, D2, D1] = int_to_bytes(length(Data)),
  Body = [U2, U1, D2, D1, Flags] ++ UsrMsg ++ Data,
  {?TAC_PLUS_AUTHEN, Body};

pack0(#author_request{authen=#authen{method=AuthenMethod,
                                     priv_lvl=PrivLvl,
                                     type=AT,
                                     service=AS
                                    },
		      user=#userid{user=User, peer={RemAddr, Port}},
		      args=Args}) ->
  {ArgLens, PArgs} = pack_args(Args),
  [_, _, _, L1] = int_to_bytes(length(ArgLens)),
  Body = [authen_method(AuthenMethod), PrivLvl,
	  authen_type(AT), authen_service(AS),
	  length(User), length(Port), length(RemAddr), L1]
    ++ ArgLens ++ User ++ Port ++ RemAddr ++ PArgs,
  {?TAC_PLUS_AUTHOR, Body};

pack0(#acct_request{flags=Flags,
		    authen=#authen{method=AuthenMethod,
                                   priv_lvl=PrivLvl,
				   type=AT,
				   service=AS
                                  },
		    user=#userid{user=User, peer={RemAddr, Port}},
		    args=Args}) ->
  {ArgLens, PArgs} = pack_args(Args),
  [_, _, _, L1] = int_to_bytes(length(ArgLens)),
  Body = [acct_type(Flags), authen_method(AuthenMethod), PrivLvl,
	  authen_type(AT), authen_service(AS),
	  length(User), length(Port), length(RemAddr), L1] ++
    ArgLens ++ User ++ Port ++ RemAddr ++ PArgs,
  {?TAC_PLUS_ACCT, Body};

%%Sended by client
pack0(#authen_reply{status=Status,
		    flags=Flags,
		    server_msg=ServerMsg,
		    data=Data
		   }) ->
  [_, _, S2, S1] = int_to_bytes(length(ServerMsg)),
  [_, _, D2, D1] = int_to_bytes(length(Data)),
  Body = [authen_status(Status), Flags, S2, S1, D2, D1] ++ ServerMsg ++ Data,
  {?TAC_PLUS_AUTHEN, Body};
pack0(_AR=#author_response{status=Status, server_msg=ServerMsg,
			   data=Data, args=Args
			  }) ->
  {ArgLens, PArgs} = pack_args(Args),
  [_, _, S2, S1] = int_to_bytes(length(ServerMsg)),
  [_, _, D2, D1] = int_to_bytes(length(Data)),
  [_, _, _, L1] = int_to_bytes(length(ArgLens)),
  Body = [author_status(Status), L1, S2, S1, D2, D1] ++ ArgLens ++
    ServerMsg ++ Data ++ PArgs,
  {?TAC_PLUS_AUTHOR, Body};
pack0(#acct_reply{status=Status, server_msg=ServerMsg, data=Data}) ->
  [_, _, S2, S1] = int_to_bytes(length(ServerMsg)),
  [_, _, D2, D1] = int_to_bytes(length(Data)),
  Body = [S2, S1, D2, D1, acct_status(Status)] ++ ServerMsg ++ Data,
  {?TAC_PLUS_ACCT, Body};
pack0(Data) ->
  throw_error(bad_packet_type, Data).

parse_args([]) ->
  [];
parse_args(Args) ->
  parse_args(Args, []).

parse_args([], PA) ->
  PA;
parse_args([A| Tail], PAs) ->
  case parse_arg(A) of
    {error, invalid_argument} ->
      ?dbg("invalid argument ~p", [A]),
      parse_args(Tail, PAs);
    PA ->
      parse_args(Tail, PAs ++ [PA])
  end.

parse_arg(Arg) ->
  case string:cspan(Arg, "*=") of
    0 ->
      {error, invalid_argument};
    N ->
      ArgName = string:substr(Arg, 1, N),
      ArgValue = string:substr(Arg, N + 2),
      Mandatory =
        case lists:nth(N + 1, Arg) of
          $= ->
            true;
          $* ->
            false
        end,
      #msg_arg{name=ArgName, value=ArgValue, mandatory=Mandatory}
  end.

pack_args(Args) ->
  lists:foldl(fun(A, {ALens, As}) ->
                  PA = pack_arg(A),
                  {ALens ++ [length(PA)], As ++ PA}
              end, {[], []}, Args).

pack_arg(#msg_arg{name=Name, value=Value, mandatory=Mandatory}) ->
  if Mandatory == true ->
      Name ++ "=" ++ Value;
     (Mandatory == false) or (Mandatory == undefined) ->
      Name ++ "*" ++ Value
  end.

authen_status(Status) ->
  case Status of
    ?TAC_PLUS_AUTHEN_STATUS_PASS ->
      ok;
    ok ->
      ?TAC_PLUS_AUTHEN_STATUS_PASS;
    ?TAC_PLUS_AUTHEN_STATUS_FAIL ->
      false;
    false  ->  
      ?TAC_PLUS_AUTHEN_STATUS_FAIL;
    ?TAC_PLUS_AUTHEN_STATUS_GETDATA ->
      getdata;
    getdata ->
      ?TAC_PLUS_AUTHEN_STATUS_GETDATA;
    ?TAC_PLUS_AUTHEN_STATUS_GETUSER ->
      getuser;
    getuser ->
      ?TAC_PLUS_AUTHEN_STATUS_GETUSER;
    ?TAC_PLUS_AUTHEN_STATUS_GETPASS ->
      getpass;
    getpass ->
      ?TAC_PLUS_AUTHEN_STATUS_GETPASS;
    ?TAC_PLUS_AUTHEN_STATUS_RESTART ->
      restart;
    restart ->
      ?TAC_PLUS_AUTHEN_STATUS_RESTART;
    ?TAC_PLUS_AUTHEN_STATUS_ERROR   ->
      error;
    error ->
      ?TAC_PLUS_AUTHEN_STATUS_ERROR;
    ?TAC_PLUS_AUTHEN_STATUS_FOLLOW    ->
      follow;
    follow ->
      ?TAC_PLUS_AUTHEN_STATUS_FOLLOW;
    _ ->
      throw_error(bad_authen_status, Status)
  end.

authen_service(Serv) ->
  case Serv of
    login ->
      ?TAC_PLUS_AUTHEN_SVC_LOGIN;
    ?TAC_PLUS_AUTHEN_SVC_LOGIN ->
      login;
    enable ->
      ?TAC_PLUS_AUTHEN_SVC_ENABLE;
    ?TAC_PLUS_AUTHEN_SVC_ENABLE ->
      enable;
    ppp ->
      ?TAC_PLUS_AUTHEN_SVC_PPP;
    ?TAC_PLUS_AUTHEN_SVC_PPP ->
      ppp;
    arap ->
      ?TAC_PLUS_AUTHEN_SVC_ARAP;
    ?TAC_PLUS_AUTHEN_SVC_ARAP ->
      arap;
    pt ->
      ?TAC_PLUS_AUTHEN_SVC_PT;
    ?TAC_PLUS_AUTHEN_SVC_PT ->
      pt;
    rcmd ->
      ?TAC_PLUS_AUTHEN_SVC_RCMD;
    ?TAC_PLUS_AUTHEN_SVC_RCMD ->
      rcmd;
    x25 ->
      ?TAC_PLUS_AUTHEN_SVC_X25;
    ?TAC_PLUS_AUTHEN_SVC_X25 ->
      x25;
    nasi ->
      ?TAC_PLUS_AUTHEN_SVC_NASI;
    ?TAC_PLUS_AUTHEN_SVC_NASI ->
      nasi;
    fwproxy ->
      ?TAC_PLUS_AUTHEN_SVC_FWPROXY;
    ?TAC_PLUS_AUTHEN_SVC_FWPROXY ->
      fwproxy;
    L when atom(L) ->
      ?TAC_PLUS_AUTHEN_SVC_NONE;
    _ ->
      none
  end.

authen_method(Meth) ->
  case Meth of
    none ->
      ?TAC_PLUS_AUTHEN_METH_NONE;
    ?TAC_PLUS_AUTHEN_METH_NONE ->
      none;
    krb5 ->
      ?TAC_PLUS_AUTHEN_METH_KRB5;
    ?TAC_PLUS_AUTHEN_METH_KRB5 ->
      krb5;
    line ->
      ?TAC_PLUS_AUTHEN_METH_LINE;
    ?TAC_PLUS_AUTHEN_METH_LINE ->
      line;
    enable ->
      ?TAC_PLUS_AUTHEN_METH_ENABLE;
    ?TAC_PLUS_AUTHEN_METH_ENABLE ->
      enable;
    local ->
      ?TAC_PLUS_AUTHEN_METH_LOCAL;
    ?TAC_PLUS_AUTHEN_METH_LOCAL ->
      local;
    tacacs ->
      ?TAC_PLUS_AUTHEN_METH_TACACSPLUS;
    ?TAC_PLUS_AUTHEN_METH_TACACSPLUS ->
      tacacs;
    L when atom(L) ->
      ?TAC_PLUS_AUTHEN_METH_NOT_SET;
    _ ->
      not_set
  end.

acct_type(Type) ->
  Duplicate = ?TAC_PLUS_ACCT_FLAG_WATCHDOG bor ?TAC_PLUS_ACCT_FLAG_START,
  case Type of
    start ->
      ?TAC_PLUS_ACCT_FLAG_START;
    ?TAC_PLUS_ACCT_FLAG_START ->
      start;
    stop ->
      ?TAC_PLUS_ACCT_FLAG_STOP;
    ?TAC_PLUS_ACCT_FLAG_STOP ->
      stop;
    continue ->
      ?TAC_PLUS_ACCT_FLAG_WATCHDOG;
    ?TAC_PLUS_ACCT_FLAG_WATCHDOG ->
      continue;
    duplicate ->
      Duplicate;
    Duplicate ->
      duplicate;
    _ ->
      throw_error(bad_account_type, Type)
  end.

-ifdef(TACACSV2_ACCOUNT_BUG).
-define(acct_status_success(_V), _V when _V == 0;
	                                 _V ==  ?TAC_PLUS_ACCT_STATUS_SUCCESS).
-else.
-define(acct_status_success(_V),  _V when _V == ?TAC_PLUS_ACCT_STATUS_SUCCESS).
-endif.

acct_status(Status) ->
  case Status of
    ?acct_status_success(Status) ->
      ok;
    ok ->
      ?TAC_PLUS_ACCT_STATUS_SUCCESS;  
    ?TAC_PLUS_ACCT_STATUS_ERROR ->
      error;
    error ->
      ?TAC_PLUS_ACCT_STATUS_ERROR;
    ?TAC_PLUS_ACCT_STATUS_FOLLOW ->
      follow;
    follow ->
      ?TAC_PLUS_ACCT_STATUS_FOLLOW;
    _ ->
      throw_error(bad_account_status, Status)
  end.

authen_action(Action) ->
  case Action of
    ?TAC_PLUS_AUTHEN_LOGIN ->
      login;
    login ->
      ?TAC_PLUS_AUTHEN_LOGIN; 
    ?TAC_PLUS_AUTHEN_CHPASS ->
      chpass;
    chpass ->
      ?TAC_PLUS_AUTHEN_CHPASS;
    ?TAC_PLUS_AUTHEN_SENDAUTH ->
      sendauth;
    sendauth ->
      ?TAC_PLUS_AUTHEN_SENDAUTH;
    _ ->
      throw_error(bad_authen_action, Action)
  end.

authen_type(Atype) ->
  case Atype of
    ascii ->
      ?TAC_PLUS_AUTHEN_TYPE_ASCII;
    ?TAC_PLUS_AUTHEN_TYPE_ASCII ->
      ascii;
    pap ->
      ?TAC_PLUS_AUTHEN_TYPE_PAP;
    ?TAC_PLUS_AUTHEN_TYPE_PAP ->
      pap;
    chap ->
      ?TAC_PLUS_AUTHEN_TYPE_CHAP;
    ?TAC_PLUS_AUTHEN_TYPE_CHAP ->
      chap;
    arap ->
      ?TAC_PLUS_AUTHEN_TYPE_ARAP;
    ?TAC_PLUS_AUTHEN_TYPE_ARAP ->
      arap;
    'ms-chap' ->
      ?TAC_PLUS_AUTHEN_TYPE_MSCHAP;
    ?TAC_PLUS_AUTHEN_TYPE_MSCHAP ->
      'ms-chap';
    undefined ->
      16#0;
    16#0 ->
      undefined; 
    _ ->
      throw_error(bad_authen_type, Atype)
  end.

author_status(Status) ->
  case Status of
    ?TAC_PLUS_AUTHOR_STATUS_PASS_ADD ->
      ok;
    ok ->
      ?TAC_PLUS_AUTHOR_STATUS_PASS_ADD; 
    ?TAC_PLUS_AUTHOR_STATUS_FAIL ->
      false;
    false ->
      ?TAC_PLUS_AUTHOR_STATUS_FAIL;
    ?TAC_PLUS_AUTHOR_STATUS_ERROR ->
      error;
    error ->
      ?TAC_PLUS_AUTHOR_STATUS_ERROR;
    ?TAC_PLUS_AUTHOR_STATUS_PASS_REPL ->
      replace;
    replace ->
      ?TAC_PLUS_AUTHOR_STATUS_PASS_REPL;
    ?TAC_PLUS_AUTHOR_STATUS_FOLLOW ->
      follow;
    follow ->
      ?TAC_PLUS_AUTHOR_STATUS_FOLLOW;
    _ ->
      throw_error(bad_author_status, Status)
   end.

check_priv_lvl(L) when L < ?TAC_PLUS_PRIV_LVL_MIN,
                       L > ?TAC_PLUS_PRIV_LVL_MAX ->
  throw_error(bad_privilege_level, L);
check_priv_lvl(_) ->
  true.

throw_error(Info, Desc) ->
  throw({error, {Info, Desc}}).

format_error({bad_author_status, Status}) ->
  io_lib:format("Bad authorization status ~w", [Status]);
format_error({bad_authen_type, Atype}) ->
  io_lib:format("Bad authentication type ~w", [Atype]);
format_error({bad_authen_action, Action}) ->
  io_lib:format("Bad authenication action ~w", [Action]);
format_error({bad_acct_status, Status}) ->
  io_lib:format("Bad accounting status ~w", [Status]);
format_error({bad_account_type, Type}) ->
  io_lib:format("Bad accounting flag combination ~w", [Type]);
format_error({bad_authen_status, Status}) ->
  io_lib:format("Bad authentication status ~w", [Status]);
format_error({bad_packet_type, Data}) ->
  io_lib:format("Bad packet type ~w", [Data]);
format_error({undef_type, Type}) ->
  io_lib:format("Couldn't determine packet type ~w", [Type]);
format_error(Error) ->
  io_lib:format("~p", Error).

fmt_packet(P) ->
  tac_lib:fmt_record(P).

split_on(Pos, Body) ->
  case catch listn:split_on(Pos, Body) of
    {'EXIT', _} ->
      throw({error, {bad_body_len, "Bad body length, check your key"}});
    Ret ->
      Ret
  end.

split_on_positions(Body, Lens) ->
  case catch listn:split_on_positions(Body, Lens) of
    {'EXIT', _} ->
      throw({error, {bad_body_len, "Bad body length, check your key"}});
    Ret ->
      Ret
  end.
