%%%-------------------------------------------------------------------
%%% File    : tac_conf.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : 
%%%
%%% Created : 31 May 2004 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id$
%%
%%**
%% 
%% @doc
%% @end
%%*
-module(tac_conf).

-behaviour(property_server).
-behaviour(property_file).

-export([start_link/1, stop/0, check/1, op/1, op/2, reload/0]).
-export([gv/2, gv/3, gv_all/2, match_key/2, select/2]).
-export([master_pid/0, all_pids/0]).

%% helpers
-export([verify_module/1, verify_path/3]).

%% property_server exports
-export([init/1, new_config/1, delete_config/1, config_file/1, bootstrap/2, store_entry/2, handle_op/2, config_to_list/1]).
%% property_file exports
-export([parse_config/4, store_config/2, postload_config/2, remove_config/1]).

-import(wslib.str).
-import(wslib.file).
-import(wslib.misc).
-import(wslib.props).
-import(wslib.log).

-include("tacacs.hrl").

-define(config_modules, [?MODULE, tac_event, tac_logger, tac_server]).
-define(server, ?MODULE).
-define(tacacs_group, tacacs).

-record(state, {config_file
	       }).

start_link(Config) ->
  case property_supervisor:start_link({local, ?server}, ?MODULE,
				      Config,
				      [{file_format, clike},
				       {parse_env, []},
				       {appsup_spec, {{one_for_one, 6, 300}, []}}
				      ]) of
    Ret={ok,_} ->
      case property_server:load(?server, default) of
	ok ->
	  Ret;
	{error, Error} ->
	  stop(),
	  log:write_error("~s", [Error]),
	  {error, bad_config}
      end;
    Error ->
      log:write_error("~p", [Error]),
      {error, bad_config}
  end.

stop() ->
  property_supervisor:stop(?server).

reload() ->
  property_server:remove(?server),
  property_server:load(?server, default).

check(File) ->
  property_server:check(?server, File).

master_pid() ->
  global:whereis_name(?server).

all_pids() ->
  pg2:get_members(?tacacs_group).

gv(Config, Key) ->
  gv(Config, Key, undefined).

gv(Config, Key, Default) ->
  case op(Config, {lookup, [Key]}) of
    [] ->
      Default;
    [{Key,Value}|_] ->
      Value
  end.

gv_all(Config, Key) ->
  proplists:get_all_values(Key, op(Config, {lookup, [Key]})).

match_key(Config, KeyPat) ->
  op(Config, {match_object, [{KeyPat, '_'}]}).

select(Config, MatchSpec) ->
  op(Config, {select, MatchSpec}).

op(Op) ->
  op(?server, Op).

op(Config, Op) ->
  property_server:handler_op(Config, Op).

%%
%% Helpers
%%

verify_module(Mod) ->
  case code:which(Mod) of
    non_existing ->
      exit(str:fmt("module ~s does not exist", [Mod]));
    Path ->
      Path
  end.

%%@spec verify_path(PropKey, Props, ConfigEntries) -> NewProps
%% PropKey = term()
%% Props = [{K, V}]
%% ConfigEntries = [{K, V}]
%% NewProps = [{K, V}]
%%@doc If file name, the value of <code>PropKey</code> in <code>Props</code>
%% is relative path then convert it to absolute path using
%% <code>server_root</code> from <code>ConfigEntries</code> as prefix.
%% <code>ConfigEntries</code> is a second argument of
%% <code>store_config/2</code> or third argument of
%% <code>parse_config([eof], [], ConfigEntries, Seed)</code> in
%% <code>property_file</code> behaviour.
%% @end
verify_path(PropKey, Props, ConfigEntries) ->
  FN = props:gv(PropKey, Props),
  case filename:pathtype(FN) of
    relative ->
      NewFN = filename:join(props:gv(server_root, ConfigEntries), FN),
      proplists:expand([{{PropKey, FN}, {PropKey, NewFN}}], Props);
    _ ->
      Props
  end.
%%
%% Property_server callbacks
%%
init(default) ->
  FN =
    case application:get_env(config_file) of
      {ok, V} ->
	str:to_str(V);
      undefined ->
	?CONFIG_FILE
    end,
  case find_config_file(FN) of
    {ok, N} ->
      init({file, N});
    Error={error, Reason} ->
      log:write_error("~p", [Reason]),
      Error
  end;
init({file, FN}) ->
  case file:test("fr", FN) of
    true ->
      log:write_info("using config file ~s", [FN]),
      {ok, #state{config_file=FN}};
    false ->
      {error, str:fmt("file ~p is not a regular readable file", [FN])}
  end.

config_file(#state{config_file=CF}) ->
  {ok, CF}.

bootstrap(Config, _) ->
  {ok, Modules} = find_modules(Config),
  {ok, ?config_modules++Modules, []}.

new_config(State) ->
  ets:new(?server, [named_table, protected, bag]),
  {ok, State}.

delete_config(State)  ->
  true = ets:delete(?server),
  {ok, State}.

store_entry(Entry, State) ->
  true = ets:insert(?server, Entry),
  {ok, State}.

handle_op({EtsOperation, Args}, State) ->
  {ok, apply(ets, EtsOperation, [?server|Args]), State}.

config_to_list(_State) ->
  {ok, ets:tab2list(?server)}.

%%
%% Property_file callbacks
%%
parse_config([N=modules], Strs, [], _) ->
  {N, ?config_modules++[list_to_atom(S) || S <- Strs]};
parse_config([N=server_root], [Dir], [], _) ->
  {N, filename:absname(Dir)};
parse_config([authen], [], Content, _) ->
  Content;
parse_config([N2, N1=authen], [Str], [], _)
  when (   (N2 == login_prompt)
	or (N2 == password_prompt)
	or (N2 == chpass_prompt)  
	  )
       and is_list(Str) ->
  {{N1, N2}, Str};
parse_config([_N=runmod], Strs, [], _) ->
  Mods = [list_to_atom(M) || M <- Strs],
  lists:foreach(fun verify_module/1, Mods),
  [{runmod, M} || M <- Mods];
parse_config([N=node_type], [Str], [], _) when Str == "master";
					       Str == "worker" ->
  {N, list_to_atom(Str)};
parse_config([eof], [], Content, _) ->
  case props:check_required(defaults(), Content) of
    {ok, C} ->
      C;
    {error, Error} ->
      exit(Error)
  end.

store_config(must_never_match, _Entries) ->
  [].

postload_config(Config, _Sup) ->
  Server = whereis(?server),
  case gv(Config, node_type) of
    master ->
      yes = global:register_name(?server, Server);
    worker ->
      ok
  end,
  pg2:create(?tacacs_group),
  pg2:join(?tacacs_group, Server),
  lists:foreach(fun (Mod) -> Mod:start() end, gv_all(Config, runmod)),
  ok.

remove_config(_DB) ->
  ok.
  
find_modules(Config) ->
  case lists:keysearch(modules, 1, Config) of
    {value, {_, ModStrs}} ->
      Mods = [list_to_atom(M) || M <- ModStrs],
      {ok, Mods};
    false ->
      {ok, []}
  end.

defaults() ->
  [{node_type, worker},
   {{authen, login_prompt},    "login: "},
   {{authen, password_prompt}, "Password: "},
   {{authen, chpass_prompt},   "New Password: "}
  ].

default_root() ->
  SP = case application:get_env(home) of
	 {ok, HP} ->
	   [str:to_str(HP)];
	 undefined ->
	   MP = code:which(?MODULE),
	   SMP = filename:split(MP),
	   lists:sublist(SMP, length(SMP) - 2)
       end,
  FN = filename:join(SP++["priv"]),
  case file:test(d, FN) of
    true ->
      FN;
    false ->
      "."
  end.

find_config_file(FN) ->
  TestPaths =
    case filename:pathtype(FN) of
      relative ->
	[
	 filename:join([os:getenv("HOME"), FN]),
	 filename:join([default_root(), "conf", FN]),
	 filename:join(".", FN)
	];
      _ ->
	[FN]
    end,
  case lists:dropwhile(
	 fun (N) -> file:test("rf", N) == false end,
	 TestPaths) of
    [] ->
      Reason =
	str:fmt("couldn't find configuration file ~s", [FN]),
      {error, Reason};
    [N|_] ->
      {ok, N}
  end.
