%%%-------------------------------------------------------------------
%%% File    : tac_event.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : 
%%%
%%% Created : 11 Jan 2002 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%**
%% 
%% .* tac_event
%% 
%%*
-module(tac_event).

-behaviour(gen_event).
%%--------------------------------------------------------------------
%% Include files
%%--------------------------------------------------------------------

-include("tac_event.hrl").

%%--------------------------------------------------------------------
%% External exports
-export([start_link/0, start_link/1, stop/0, stop/1]).
-export([add_handler/3, delete_handler/3, call/3]).
-export([add_handler/2, delete_handler/2, call/2]).
-export([authenticate/2, authorize/2, account/2]).
-export([error_event/1, info_event/1, add_relay_handler/1]).
-export([perform_takeover/2]).

%% property_file exports
-export([parse_config/4, store_config/2, remove_config/1, postload_config/2]).

%% gen_event callbacks
-export([init/1, handle_event/2, handle_call/2, handle_info/2, terminate/2, code_change/3]).

-define(server, ?MODULE).

-record(relay_state, {relay_fun}).

%% External functions
start_link() ->
  start_link(local).

start_link(local) ->
  gen_event:start_link({local, ?server});
start_link(global) ->
  {ok, Pid} = start_link(local),
  case global:whereis_name(?server) of
    undefined ->
      global:re_register_name(?server, Pid);
    _ ->
      ok
  end,
  {ok, Pid}.

stop() ->
  stop(local).

stop(local) ->
  gen_event:stop(?server);
stop(global) ->
  gen_event:stop({global, ?server}).

perform_takeover(_, worker) ->
  ok;
perform_takeover(FromNode, master) ->
  MgrPid = whereis(?server),
  case global:whereis_name(?server) of
    MgrPid ->
      ok; %we are seldom, nothing to do
    _ ->
      global:re_register_name(?server, MgrPid),
      ok
  end.

add_relay_handler(Type) when Type == global;
			     Type == eva ->
  RelayFun =
    case Type of
      global ->
        fun (Event) ->
            Self = self(),
            case global:whereis_name(?server) == whereis(?server) of
              false -> gen_event:notify({global, ?server}, Event);
              true -> ok %don't send it to myself
            end
        end;
      eva ->
        fun (Event) ->
            eva:asend_event(tacacs, {tacacs, node()}, Event)
        end
    end,
  add_handler({?MODULE, Type}, {relay, RelayFun}).

authenticate({UserId,Session,A=#authen{type=T,priv_lvl=PL,service=S},Req,_},
	     EventArgs) ->
  EventArgs1 = [{request,  Req}, {type, T}, {service, S},{priv_lvl, PL}]
    ++EventArgs,
  notify(authen, UserId, Session, EventArgs1, []).

authorize({UserId, Session, Authen, Args}, EventArgs) ->
  notify(author, UserId, Session, EventArgs, Args).

account({UserId, Session, Authen, AccPhase, Args}, EventArgs) ->
  notify(account, UserId, Session, [{phase, AccPhase}|EventArgs], Args).
  
error_event(Error) when is_list(Error) ->
  notify(#tac_event{reason=error, event_args=Error}).

info_event(Info) when is_list(Info) ->
  notify(#tac_event{reason=info, event_args=Info}).

%% Property_file callbacks
parse_config([N=event_handler], [Str], [], _) when Str == "global";
						   Str == "eva" ->
  Hlr = list_to_atom(Str),
  {N, Hlr}.

store_config(must_never_match, _Entries) ->
  [].

remove_config(DB) ->
  ok.

postload_config(Config, Sup) ->
  EventServer =
    {?server,
     {?server, start_link, []},
     permanent, 5000, worker, [?MODULE]
    },
  {ok, _} = supervisor:start_child(Sup, EventServer),
  case tac_conf:gv_all(Config, event_handler) of
    [] ->
      ok;
    HlrTypes ->
      ok = lists:foreach(fun (T) -> ok = add_relay_handler(T) end, HlrTypes)
  end.

%%--------------------------------------------------------------------
%% Function: add_handler/2, add_handler/3
%% Description: Adds an event handler
%%--------------------------------------------------------------------
add_handler(Name, Args) ->
  add_handler(?server, Name, Args).

add_handler(Server, Name, Args) ->
  gen_event:add_handler(Server, Name, Args).

%%--------------------------------------------------------------------
%% Function: delete_handler/2, delete_handler/3
%% Description: Delete an event handler
%%--------------------------------------------------------------------
delete_handler(Name, Args) ->
  delete_handler(?server, Name, Args).

delete_handler(Server, Name, Args) ->
  gen_event:delete_handler(Server, Name, Args).

%%--------------------------------------------------------------------
%% Function: call/3
%% Description: Request event handler
%%--------------------------------------------------------------------
call(Handler, Query) ->
  call(?server, Handler, Query).

call(Server, Handler, Query) ->
  gen_event:call(Server, Handler, Query).

%%====================================================================
%% Server functions
%%====================================================================
%%--------------------------------------------------------------------
%% Func: init/1
%% Returns: {ok, State}          |
%%          Other
%%--------------------------------------------------------------------
init({relay, RelayFun}) ->
  {ok, #relay_state{relay_fun=RelayFun}}.

%%--------------------------------------------------------------------
%% Func: handle_event/2
%% Returns: {ok, State}                                |
%%          {swap_handler, Args1, State1, Mod2, Args2} |
%%          remove_handler                              
%%--------------------------------------------------------------------
handle_event(Event, State) ->
  (State#relay_state.relay_fun)(Event),
  {ok, State}.

%%--------------------------------------------------------------------
%% Func: handle_call/2
%% Returns: {ok, Reply, State}                                |
%%          {swap_handler, Reply, Args1, State1, Mod2, Args2} |
%%          {remove_handler, Reply}                            
%%--------------------------------------------------------------------
handle_call(takeover_state, State) ->
  {ok, State, State};
  
handle_call(Request, State) ->
  Reply = ok,
  {ok, Reply, State}.

%%--------------------------------------------------------------------
%% Func: handle_info/2
%% Returns: {ok, State}                                |
%%          {swap_handler, Args1, State1, Mod2, Args2} |
%%          remove_handler                              
%%--------------------------------------------------------------------
handle_info(Info, State) ->
  {ok, State}.

%%--------------------------------------------------------------------
%% Func: terminate/2
%% Purpose: Shutdown the server
%% Returns: any
%%--------------------------------------------------------------------
terminate(Reason, State) ->
  ok.

%%--------------------------------------------------------------------
%% Func: code_change/3
%% Purpose: Convert process state when code is changed
%% Returns: {ok, NewState}
%%--------------------------------------------------------------------
code_change(OldVsn, State, Extra) ->
  {ok, State}.

%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------
notify(Event) ->
  case catch gen_event:notify(?server, Event) of
    ok ->
      ok;
    Error ->
      log:write_error("~s does not exists", [?server]),
      ok
  end.

notify(Event, UserId, #session{peer=Peer})
  when is_record( Event, tac_event) ->
  Event1 = Event#tac_event{time=calendar:universal_time(),
                           userid=UserId,
                           serverid=#serverid{peer=Peer}
                          },
  notify(Event1).

-ifdef(CISCO_ACCOUNT_BUG).
-define(message_args(_AaaReq, _As),
	(if _AaaReq == account ->
	     %% nasty Cisco bug, end of the arg is filled by '\0'
	     [MA#msg_arg{value=strip_nulls(V)} || MA=#msg_arg{value=V} <- _As];
	    true ->
	     _As
	 end
	)).
-else.
-define(message_args(_AaaReq, _As), _As).
-endif.

notify(AaaReq, UserId, S=#session{sess_no=SN}, EventArgs, MsgArgs) ->
  notify(#tac_event{reason=AaaReq,
		    event_args=[{session, SN}|EventArgs],
		    msg_args=?message_args(AaaReq, MsgArgs)
		   },
         UserId, S).

strip_nulls(V) when is_list(V) ->
  string:strip(V, right, 16#0).
