%%%----------------------------------------------------------------------
%%% File    : tac_lib.erl
%%% Author  :  <svg@surnet.ru>
%%% Purpose : 
%%% Created : 17 Dec 1999 by  <svg@surnet.ru>
%%% $Id: tac_lib.erl,v 1.8 2002/09/04 01:05:53 svg Exp $
%%%----------------------------------------------------------------------

-module(tac_lib).
-author('svg@surnet.ru').

-include("tacacs.hrl").

-export([make_userid/1, make_userid/3, make_serverid/2, make_serverid/3]).
-export([make_authen/2, make_authen/4]).
-export([ip_to_str/1, bytes_to_int/4, int_to_bytes/1]).
-export([author_check_args/4, format_error/1, fmt_record/1, dbg/3]).

-import(wslib.str).
-import(wslib.listn).
-import(wslib.cal).
-import(wslib.log).

make_userid(User) ->
  {ok, Host} = inet:gethostname(),
  make_userid(User, Host, "undefined").

make_userid(User, Host, Port) when is_list(User),
				   is_list(Host),
				   is_list(Port) ->
  #userid{user=User, peer={Host, Port}}.

make_serverid(Host, SKey) ->
  make_serverid(Host, ?DEFAULT_PORT, SKey).

make_serverid(Host, Port, SKey) when is_list(Host),
				     is_integer(Port),
				     is_list(SKey) ->
  #serverid{peer={Host,Port}, skey=SKey}.

make_authen(Type, Service) ->
  make_authen(Type, Service, 16#01, tacacs).

make_authen(Type, Service, PrivLvl, Method) when is_atom(Type),
						 is_atom(Service),
						 is_integer(PrivLvl),
						 PrivLvl >= ?TAC_PLUS_PRIV_LVL_MIN,
						 PrivLvl =< ?TAC_PLUS_PRIV_LVL_MAX,
						 is_atom(Method) ->
  %% some checks
  tac_body:authen_type(Type),
  tac_body:authen_service(Service),
  tac_body:authen_method(Method),
  #authen{priv_lvl=PrivLvl, type=Type, service=Service, method=Method}.

ip_to_str({A1, A2, A3, A4}) ->
  str:fmt("~w.~w.~w.~w", [A1, A2, A3, A4]);
ip_to_str(Addr) ->
  str:fmt("~s", [Addr]).

%%----------------------------------------------------------------------
%% bytes_to_int(I1,I2,I3,I4) -> Integer
%% convert 4 bytes to an unsigned int
%%----------------------------------------------------------------------
bytes_to_int(I1,I2,I3,I4) ->
  (I1 bsl 24) bor (I2 bsl 16) bor (I3 bsl 8) bor I4.

%%----------------------------------------------------------------------
%% int_to_bytes(I) -> [B4, B3, B2, B1]
%% convert unsigned int to an 4 bytes 
%%----------------------------------------------------------------------
int_to_bytes(I) when integer(I) ->
  [(I band 16#ff000000) bsr 24, (I band 16#ff0000) bsr 16, (I band 16#ff00) bsr 8, (I band 16#ff)].


dbg(true, {Preamble, FmtF}, Arg) ->
  Str = log:fmt(Preamble++FmtF(Arg), []),
  error_logger:info_msg(Str);
dbg(true, Fmt, Args) ->
  Str = log:fmt(Fmt, Args),
  error_logger:info_msg(Str);
dbg(_, _, _) ->
  ok.

fmt_record(P) when is_record(P, authen_start) ->
  fmt_record(record_info(fields, authen_start), P);
fmt_record(P) when is_record(P, authen_continue) ->
  fmt_record(record_info(fields, authen_continue), P);
fmt_record(P) when is_record(P, author_request) ->
  fmt_record(record_info(fields, author_request), P);
fmt_record(P) when is_record(P, authen_reply) ->
  fmt_record(record_info(fields, authen_reply), P);
fmt_record(P) when is_record(P, author_response) ->
  fmt_record(record_info(fields, author_response), P);
fmt_record(P) when is_record(P, acct_reply) ->
  fmt_record(record_info(fields, acct_reply), P);
fmt_record(P) when is_record(P, acct_request) ->
  fmt_record(record_info(fields, acct_request), P);
fmt_record(P) when is_record(P, client_request) ->
  fmt_record(record_info(fields, client_request), P);
fmt_record(P) ->
  str:to_str(P).

fmt_record(FieldNames, Rec) ->
  [RecName|FieldsValues] = tuple_to_list(Rec),
  FieldStr = lists:flatmap(fun ({N, V}) -> str:fmt("\t~s=~p\n", [N, V]) end,
			   listn:zip(FieldNames, FieldsValues)),
  str:to_str(RecName) ++ ":\n" ++ FieldStr.
  

%%----------------------------------------------------------------------
%% format_error(Error) -> string()
%%----------------------------------------------------------------------

format_error(ErrDesc={_, Module, Desc}) when is_atom(Module) ->
  case catch apply(Module, format_error, [Desc]) of
    {'EXIT', _} ->
      str:to_str(ErrDesc);
    PrintError ->
      lists:flatten(PrintError)
  end;
format_error(Msg) ->
  str:to_str(Msg).

%%@spec author_check_args(NasArgs, SrvArgs, DefaultAttr) -> {Status, RetArgs}
%% NasArgs = [msg_arg()]
%% SrvArgs = [msg_arg()]
%% RetArgs = [msg_arg()]
%% DefaultAttr = permit | deny
%% Status = ok | false | replace
%% @doc Check NAS attributes against server attributes, return status of
%% operation and changed list of attributes for returning to NAS.
%% <code>DefaultAttr</code> indicates are absent in <code>SrvArgs</code>
%% NAS attributes permitted or not.
%% @end
author_check_args(NasArgs, Attrs, DefAttr, SkipArgs) ->
  author_check_args(NasArgs, Attrs, DefAttr, SkipArgs, [], ok).

author_check_args(NasArgs=[], SrvArgs, _, _, Acc, Status) ->
  {RetNasArgs, RetStatus} =
    author_add_mandatory(SrvArgs, lists:reverse(Acc), Status),
  {RetStatus, RetNasArgs};
author_check_args(NasArgs, [], permit, _, _Acc, Status) ->
  %% all permitted, nothing to add
  {Status, []};
author_check_args([NasA=#msg_arg{name=N, value=V, mandatory=M}|NasArgs],
		  SrvArgs, Default, SkipArgs, Acc, Status) ->
  case lists:member(N, SkipArgs) of
    true ->
      author_check_args(NasArgs, SrvArgs, Default, SkipArgs, [NasA|Acc], Status);
    false ->
      Args = [X || X <- SrvArgs, X#msg_arg.name == N],
      case {Args, Default, M} of
	%%user attr was not  found, get NAS attr (d, j)
	{[], permit, _} ->
	  author_check_args(NasArgs, SrvArgs, Default, SkipArgs, [NasA|Acc], Status);
	%% (i) delete optional NAS AV pair
	{[], deny, false} ->
	  author_check_args(NasArgs, SrvArgs, Default, SkipArgs, Acc, replace);
	{_, _, true} ->
	  %% check mandatory argument
	  %% exact match in mandatory list (a, e)
	  %% or optional arg with the same name(b)
	  case (lists:member(NasA, Args)
		or lists:keymember(false, #msg_arg.mandatory, Args)) of
	    true ->
	      author_check_args(NasArgs, SrvArgs, Default, SkipArgs, [NasA|Acc], Status);
	    false ->
	      {false, []}
	  end;
	{_, _, false} ->
	  %% check optional arg
	  %% exact match in mandatory list (a,e)
	  case lists:member(NasA#msg_arg{mandatory=true}, Args) of
	    true ->
	      author_check_args(NasArgs, SrvArgs, Default, SkipArgs,
				[NasA#msg_arg{mandatory=true}|Acc], replace);
	    false ->
	      case lists:keysearch(true, #msg_arg.mandatory, Args) of
		%% found mandatory arg (f)
		{value, SrvA} ->
		  author_check_args(NasArgs, SrvArgs, Default, SkipArgs, [SrvA|Acc], replace);
		false ->
		  case lists:member(NasA, Args) of
		    %% exact founded (g)
		    true ->
		      author_check_args(NasArgs, SrvArgs, Default, SkipArgs, [NasA|Acc], Status);
		    false ->
		      %% (h) there is optional user arg with the same name
		      [Arg|_] = Args,
		      author_check_args(NasArgs, SrvArgs, Default, SkipArgs,[Arg|Acc], replace)
		  end
	      end
	  end;
	_ ->
	  {false, []}
      end
  end.

author_add_mandatory(Attrs, Acc, Status) ->
  AddAttrs = [X || X <- Attrs, X#msg_arg.mandatory == true,
		   lists:member(X, Acc) == false],
  case Status of
    replace ->
      {AddAttrs ++ Acc, replace};
    _ ->
      {AddAttrs, Status}
  end.
