%%%-------------------------------------------------------------------
%%% File    : tac_logger.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : 
%%%
%%% Created : 13 Feb 2002 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%**
%% 
%% .* tac_logger
%% 
%%*
-module(tac_logger).
-author('svg@surnet.ru').
-vsn('$Id\$').

-behaviour(gen_event).
-behaviour(property_file).

-export([add_logger_mf/2, delete_logger_mf/1, info/1, rotate/1]).
-export([dilber_format/1, rotate_file_name/1]).

%% property_file callbacks
-export([parse_config/4, store_config/2, remove_config/1, postload_config/2]).
%% gen_event callbacks
-export([init/1, handle_event/2, handle_call/2, handle_info/2, terminate/2]).

%%-export([configured_logs/1]).

-import(wslib.misc).
-import(wslib.str).
-import(wslib.cal).
-import(wslib.props).
-import(wslib.log).
-import(wslib.listn).

-include("tac_event.hrl").
-include_lib("stdlib/include/ms_transform.hrl").

-define(event_types, [authen, author, account, error, system]).
-define(disk_log_name(_EventType), {tacacs, _EventType}).

-define(is_log_type(_T), (   (_T == authen)
	                  or (_T == author)
                  	  or (_T == account)
                	  or (_T == error)
			  or (_T == system)
			  )).

-record(log_event, {timestamp=calendar:universal_time(), type, event}).
-record(state, {log, func}).
 
add_logger_mf(EventType, Props) when ?is_log_type(EventType) ->
  LogName = ?disk_log_name(EventType),
  Format = props:gv(format, Props),
  FormatF = mk_format(EventType, Format),
  LoggerProps = [{name, LogName}, {repair, true}|Props],
  EventMgr = event_manager_name(EventType),
  gen_event:add_handler(EventMgr, {?MODULE, EventType}, {FormatF, LoggerProps}).

delete_logger_mf(EventType) when ?is_log_type(EventType) ->
  EventMgr = event_manager_name(EventType),
  gen_event:delete_handler(EventMgr, {?MODULE, EventType}, stop).

rotate(EventType) when ?is_log_type(EventType) ->
  LogName = ?disk_log_name(EventType),
  case disk_log:info(LogName) of
    Error = {error, _} ->
      Error;
    Info ->
      RotName = new_rotate_name(Info),
      case disk_log:reopen(LogName, RotName) of
	ok ->
	  {ok, RotName};
	Error ->
	  Error
      end
  end.

info(logs) ->
  configured_logs(tac_conf);
info(EventType) ->
  EventMgr = event_manager_name(EventType),
  case gen_event:call(EventMgr, {?MODULE, EventType}, info) of
    {error, bad_module} ->
      {error,no_such_log};
    Result ->
      Result
  end.

%%
%% gen_event callbacks
%%
init({Func, Opts}) ->
  S = #state{func = Func},
  case disk_log:open(Opts) of
    {ok, Log} ->
      {ok, S#state{log=Log}};
    {repaired, Log, _, _} ->
      {ok, S#state{log=Log}};
    Error ->
      log:write_error(disk_log:format_error(Error), []),
      {error, open_log}
  end.

handle_event(Event, S=#state{log=L, func=F}) ->
  case F(Event) of
    false ->
      {ok, S};
    Bin ->
      case disk_log:balog(L, Bin) of
	ok ->
	  {ok, S};
	Error ->
	  Error
      end
  end.

handle_call(info, S) ->
  Reply = disk_log:info(S#state.log),
  {ok, Reply, S}.

handle_info({emulator, GL, Chars}, S) ->
  handle_event({emulator, GL, Chars}, S);
handle_info(_Info, S) ->
  {ok, S}.

terminate(_Arg, S) ->
  disk_log:close(S#state.log).

%%
%% property_file callbacks
%%
parse_config([_N=log], [EventTypeStr], Content, _) ->
  EventType = list_to_atom(EventTypeStr),
  ?is_log_type(EventType)
    orelse fatal_error("bad log type ~s", [EventTypeStr]),
  Props = check_log_props(Content),
  {{log, EventType}, check_log_props(Props)};
parse_config([N=file, log], [FileName], [], _) when is_list(FileName) ->
  {N, FileName};
parse_config([N=type, log], [TypeStr], [], _) when TypeStr == "halt";
						   TypeStr == "wrap" ->
  {N, list_to_atom(TypeStr)};
parse_config([N=format, log], [FmtStr], [], _) when FmtStr == "external";
						    FmtStr == "internal" ->
  {N, list_to_atom(FmtStr)};
parse_config([N=size, log], [Sz], [], _) when  Sz == "infinity";
					       Sz == 0 ->
  {N, infinity};
parse_config([N=size, log], [Sz], [], _) when is_integer(Sz) ->
  {N, Sz*1024};
parse_config([N=max_files, log], [Max], [], _) when is_integer(Max),
						    Max > 0, Max =< 255 ->
  {N, Max};
parse_config([log_dir], _, _, _) -> %skip it, help directive
  [].

store_config(_E={K={log, EventType}, Props}, Entries)
  when ?is_log_type(EventType) ->
  {K, tac_conf:verify_path(file, Props, Entries)}.

postload_config(Config, _Sup) ->
  lists:foreach(
    fun ({EventType, Args}) ->
	case add_logger_mf(EventType, Args) of
	  ok ->
	    ok;
	  {error, Error} ->
	    exit(Error)
	end
    end,
    configured_logs(Config)
   ),
  ok.

remove_config(Config) ->
  lists:foreach(
    fun ({EventType, _Args}) ->
	(catch delete_logger_mf(EventType))
    end,
    configured_logs(Config)
   ),
  ok.

%% Used by external log viewer
dilber_format(_Msg=#log_event{timestamp=TS, type=ET, event=E}) ->
  format_event(external, ET, E, TS);
dilber_format(Event) ->
  list_to_binary(io_lib:format("~p", [Event])).

mk_format(EventType, LogFormat) when EventType == authen;
                                     EventType == author;
                                     EventType == account ->
  fun(Event=#tac_event{reason=Reason}) when Reason == EventType ->
      format_event(LogFormat, Reason, Event);
     (_Event) ->
      false
  end;
mk_format(error, LogFormat) ->
  fun (Event=#tac_event{reason=EventType}) ->
      if EventType == authen;
         EventType == author;
         EventType == account ->
          false;
         true ->
          format_event(LogFormat, EventType, Event)
      end;
      (Event) ->
      format_event(LogFormat, error, Event)
  end;
mk_format(system, LogFormat) ->
  fun(Event={Type, _GL, _Msg}) when Type == error_report;
				    Type == info_report;
				    Type == info_msg;
				    Type == error;
				    Type == info;
				    Type == warning_report;
				    Type == warning_msg;
				    Type == emulator ->
      (catch format_event(LogFormat, system, Event));
     (_) ->
      false
  end.

format_event(internal, EventType, Event) ->
  term_to_binary(#log_event{timestamp=calendar:universal_time(),
                            type=EventType, event=Event});

format_event(external, EventType, Event) ->
  format_event(external, EventType, Event, calendar:universal_time()).

format_event(external, EventType, Event, TimeStamp) ->
  list_to_binary([fmt_log_date(TimeStamp),
		  fmt_header(EventType, Event),
		  fmt_event_ext(EventType, Event),
		  $\n
		 ]).

fmt_header(_, _E=#tac_event{userid=UserId, serverid=ServerId, node=Node}) ->
  fmt_header_args([UserId, Node, ServerId]);
fmt_header(system, _) ->
  "";
fmt_header(_, _) ->
  str:to_str(node()).

fmt_header_args(Args) ->
  [fmt_header_arg(A) || A <- Args].

fmt_header_arg(undefined) ->
  [];
fmt_header_arg(#userid{user = User, peer = {RemAddr, Port}}) ->
  [$\t, str:to_str(User),$\t, str:to_str(Port), $\t, str:to_str(RemAddr)];
fmt_header_arg(#serverid{ peer={Host,_Port} }) -> [$\t|str:to_str(Host)];
fmt_header_arg(V) -> [$\t|str:to_str(V)].

fmt_event_ext(EventType, _Event=#tac_event{event_args=EAs, msg_args=MAs}) ->
  fmt_args([EventType|EAs ++ MAs]);
%% error_logger events
fmt_event_ext(_EventType=system, {Reason, _GL, Msg}) ->
  try [$\t,system_event_name(Reason),$\n,fmt_system_event(Reason, Msg)]
  catch
    throw:false ->
      throw(false);
     exit:_ ->
      [$\t,"ERROR",$\n,str:to_str(Msg)]
  end;
%% other
fmt_event_ext(EventType, Msg) ->
  fmt_args([EventType, Msg]).

fmt_args(Args) ->
  [[$\t|fmt_arg(A)] || A <- listn:mk_list(Args)].

fmt_arg(_Arg=#msg_arg{name=N, value=V}) ->
  str:fmt("~s=~s", [N,str:to_str(V)]);
fmt_arg({format, Fmt, Args}) ->
  str:fmt(Fmt, Args);
fmt_arg({N, V}) ->
  str:fmt("~s=~s", [N,str:to_str(V)]);
fmt_arg(Arg) ->
  str:to_str(Arg).

new_rotate_name(Info) ->
  {value, {_, Name}} = lists:keysearch(file, 1, Info),
  rotate_file_name(Name).

rotate_file_name(Name) ->
  Base = filename:rootname(Name),
  Suf = filename:extension(Name),
  {{YYYY,MM,DD}, _} = calendar:local_time(),
  Date = io_lib:format(".~.4w-~.2.0w-~.2.0w", [YYYY, MM, DD]),
  RetName = Base ++ Date ++ Suf,
  RetName.

fmt_log_date(DateTime) ->
  cal:format_datetime(apache, DateTime).

configured_logs(Config) ->
  tac_conf:select(
    Config,
    [ets:fun2ms(
       fun ({{log, EventType}, LogOpts}) when ?is_log_type(EventType) ->
	   {EventType, LogOpts}
       end)
    ]).
   
check_log_props(UserProps) ->
  {ok, Props} = props:check_required(defaults(), UserProps),
  check_size(Props).

check_size(Props) ->
  NewProps = 
    case props:gv_map([type, size, max_files], Props) of
      [halt, _, _] ->
	Props;
      [_, infinity, _] ->
	Props;
      [_, Sz, MaxFs] ->
	[{size, {Sz, MaxFs}}|props:delete(size, Props)]
    end,
  props:delete(max_files, NewProps).

event_manager_name(EventType) ->
  if EventType == system ->
      error_logger;
     true ->
      tac_event
  end.

defaults() ->
  [{file},
   {type, halt},
   {size, infinity},
   {format, external},
   {max_files, 1}
  ].

fatal_error(Fmt, Args) ->
  exit(str:fmt(Fmt, Args)).

%% error_logger
system_event_name(info_report) ->
  "INFO REPORT";
system_event_name(info_msg) ->
  "INFO REPORT";
system_event_name(warning_report) ->
  "WARNING REPORT";
system_event_name(warning_msg) ->
  "WARNING REPORT";
system_event_name(_) ->
  "ERROR REPORT".

fmt_system_event(emulator, Chars) ->
  io_lib:fwrite(Chars, []);
fmt_system_event(Reason, _M={Pid, Fmt, Args}) when is_pid(Pid) ->
  [str:to_str(node(Pid)),$\t,fmt_system_event(Reason, Fmt, Args)];
fmt_system_event(_, _) ->
  throw(false).

fmt_system_event(error, Format, Args) ->
  io_lib:fwrite(Format, Args);
fmt_system_event(info, Info, _) ->
  io_lib:fwrite("~p", [Info]);
fmt_system_event(error_report, std_error, Rep) ->
  format_report(Rep);
fmt_system_event(info_report, std_info, Rep) ->
  format_report(Rep);
fmt_system_event(info_msg, Format, Args) ->
  io_lib:fwrite(Format, Args);
fmt_system_event(warning_report, std_warning, Rep) ->
  format_report(Rep);
fmt_system_event(warning_msg, Format, Args) ->
  io_lib:fwrite(Format, Args);
fmt_system_event(_, _, _) ->
  throw(false).

format_report(Rep) when list(Rep) ->
    case string_p(Rep) of
	true ->
	    io_lib:format("~s~n",[Rep]);
	_ ->
	    format_rep(Rep)
    end;
format_report(Rep) ->
    io_lib:format("~p~n",[Rep]).

format_rep([{Tag,Data}|Rep]) ->
    io_lib:format("    ~p: ~p~n",[Tag,Data]) ++ format_rep(Rep);
format_rep([Other|Rep]) ->
    io_lib:format("    ~p~n",[Other]) ++ format_rep(Rep);
format_rep(_) ->
    [].

string_p([]) ->
    false;
string_p(Term) ->
    string_p1(Term).

string_p1([H|T]) when integer(H), H >= $\s, H < 255 ->
    string_p1(T);
string_p1([$\n|T]) -> string_p1(T);
string_p1([$\r|T]) -> string_p1(T);
string_p1([$\t|T]) -> string_p1(T);
string_p1([$\v|T]) -> string_p1(T);
string_p1([$\b|T]) -> string_p1(T);
string_p1([$\f|T]) -> string_p1(T);
string_p1([$\e|T]) -> string_p1(T);
string_p1([H|T]) when list(H) ->
    case string_p1(H) of
	true -> string_p1(T);
	_    -> false
    end;
string_p1([]) -> true;
string_p1(_) ->  false.
