%%%----------------------------------------------------------------------
%%% File    : tac_packet.erl
%%% Author  :  <svg@surnet.ru>
%%% Purpose : 
%%% Created : 21 Nov 1999 by  <svg@surnet.ru>
%%%	$Id: tac_packet.erl,v 1.10 2002/09/04 01:05:53 svg Exp $	
%%%----------------------------------------------------------------------

-module(tac_packet).
-author('svg@surnet.ru').

-include("tacacs.hrl").

%% -compile(export_all).
-export([mk_packer/3]).

%% Internal export for tac_body
-import(tac_lib, [bytes_to_int/4, int_to_bytes/1]).

-import(wslib.props).
-import(wslib.cal).
-import(wslib.log).
-import(wslib.str).

-define(MD5_LEN, 16).
-define(dbg(_F, _A), tac_lib:dbg(get(dbg), _F, _A)).

-record(state, {skey,
		session,
		header,
		expect_packet,
		debug
	       }).

mk_packer(FsmType, SKey, Args) ->
  [Unenc, SingleConn, RandomF, _Debug] = check_opts(Args),
  fun () ->
      SessNo = case FsmType of
		 server -> undefined;
		 client -> RandomF()
	       end,
      SeqNo = 1,
      Session = #session{sess_no=SessNo,
			 seq_no=SeqNo,
			 unencrypt=Unenc,
			 single_conn=SingleConn
			},
      State = #state{skey=SKey,
		     session=Session
		    },
      Self = self(),
      spawn_link(fun () -> packet_loop(Self, fun start/2, State) end)
  end.

packet_loop(Parent, StateF, S=#state{session=Sess}) ->
  receive {Parent, {debug, Dbg}} -> ok end,
  put(dbg, Dbg),
  Parent ! {self(), peername},
  receive
    {Parent, {peername, {IP, Port}}} ->
      Peer = {tac_lib:ip_to_str(IP), Port},
      packet_loop({Parent, StateF,
		   S#state{session=Sess#session{peer=Peer}, debug=Dbg}})
  end.

packet_loop({Parent, StateF, State}) ->
  Self = self(),
  receive
    {Parent, {error, _}} ->
      Parent ! {Self, {done, <<>>}};
    {Parent, Msg} ->
      {Ret, NxtStateF, NxtState} = 
	state(StateF, Msg, State),
      Parent ! {Self, Ret},
      packet_loop({Parent, NxtStateF, NxtState})
  end.

state(StateF, Msg, State) ->
  case StateF(Msg, State) of
    {more, RestBin, NxtParseF, NxtState} ->
      {more, mk_continue_parse(RestBin, NxtParseF),  NxtState};
    {done, Result, RestBin, NxtState} ->
      Ret = {done, Result, RestBin},
      {Ret, fun start/2, NxtState};
    {done, Result, NxtState} ->
      Ret = {done, Result},
      {Ret, fun start/2, NxtState}
  end.

start({parse, PacketType, Bin}, State) ->
  parse(PacketType, Bin, State);
start({pack, Msg}, State) ->
  pack(Msg, State).

mk_continue_parse(Bin, ParseF) ->
  fun ({received, MoreBin}, State) ->
      ParseF(concat_binary([Bin, MoreBin]), State)
  end.

parse(PacketType, Bin, S) ->
  get_header(Bin, S#state{expect_packet=PacketType}).

pack({Version, Msg}, S=#state{session=Sess}) when is_integer(Version) ->
  pack(Msg, S#state{session=Sess#session{ver=Version}});
pack(Msg, S=#state{skey=Skey, session=Sess}) ->
  {Sess1, Body} = tac_body:pack(Sess, Msg),
  Packet = make_packet(Skey, Sess1, Body),
  {done, Packet, S#state{session=incr_session(Sess)}}.

get_header(<<HBin:12/binary, Body/binary>>, S) ->
  Header = parse_header(HBin),
  get_body(Body, S#state{header=Header});
get_header(Bin, S) ->
  {more, Bin, fun get_header/2, S}.

get_body(Bin, S=#state{header=H}) ->
  BodyLen = element(size(H), H),
  case Bin of
    <<Data:BodyLen/binary, Rest/binary>> ->
      {Ret, NxtS} = parse_body(Data, S),
      {done, Ret, Rest, NxtS};
    _ ->
      {more, Bin, fun get_body/2, S}
  end.

parse_body(Bin,
	   S=#state{skey=SKey,
		    header=H,
		    session=ExpS,
		    expect_packet=ExpP
		   }) ->
  {RecvSess, BodyData} = parse_body(SKey, H, Bin),
  CurSess = incr_session(verify_session(RecvSess, ExpS)),
  NxtS = S#state{session=CurSess},
  Ret =
    case expected_packet_type(ExpP, CurSess) of
      Error={error, _} ->
	?dbg("undefined session type ~w", [CurSess#session.type]),
	Error;
      PacketType ->
	?dbg("expecting packet type ~s", [PacketType]),
	case catch tac_body:parse(PacketType, BodyData) of
	  {Error, Reason} when Error == error;
			       Error == 'EXIT' ->
	    {error, Reason};
	  Packet ->
	    Packet
	end
    end,
  {{CurSess, Ret}, NxtS}.

parse_body(Skey, Header, Body) ->
  parse0(Skey, Header, Body).

make_packet(Skey, Header, Body) ->
  pack0(Skey, Header, Body).

%%FixMe check len, SesNo to single segment
parse_header(<<Ver, Type, SeqNo, Flags, SessNo:32, Len:32>>) ->
  {Unencrypted, SingleConnect} =
    {Flags band 16#01, (Flags band 16#04) bsr 3},
  Header = {Ver, Type, SeqNo, Unencrypted, SingleConnect, SessNo, Len},
  ?dbg({"header ", fun fmt_header/1}, Header),
  Header.

parse0(SecretKey, Header, Body) when is_binary(Body) ->
  parse0(SecretKey, Header, binary_to_list(Body));
parse0(SecretKey, Header, Body) ->
  {Ver, Type, SeqNo, Unencrypted, SingleConnect, SessNo, Len} = Header,
  if
    length(Body) =/= Len ->
      throw({error, {bad_body_len, "Bad body length, check your key"}});
    true ->
      ok
  end,
  Data = decode_body(Ver, SeqNo, Unencrypted, SessNo, Len, Body, SecretKey),
  Sess = #session{ver=Ver,
		  type=Type,
		  seq_no=SeqNo,
		  unencrypt=Unencrypted,
		  single_conn=SingleConnect,
		  sess_no=SessNo
		 },
  {Sess, Data}.

pack0(SecretKey,
      _S=#session{ver=Ver,
		  type=Type,
		  seq_no=SeqNo,
		  unencrypt=Unenc,
		  single_conn=SingleConnect,
		  sess_no=SessNo
		 },
      Body) ->
  Header = {Ver, Type, SeqNo, Unenc, SingleConnect, SessNo, length(Body)},
  Data = [pack_header(Header),
	  encode_body(Ver, SeqNo, Unenc, SessNo, Body, SecretKey)],
  Data.

pack_header(H={Ver, Type, SeqNo, Unenc, SingleConnect, SessNo, Len}) ->
  Flags =  Unenc bor (SingleConnect bsl 3),
  ?dbg({"header ", fun fmt_header/1} , H),
  << Ver, Type, SeqNo, Flags, SessNo:32, Len:32 >>.

encode_body(Ver, SeqNo, Unencrypted, SessNo, Body, SecretKey) ->
  decode_body(Ver, SeqNo, Unencrypted, SessNo, length(Body), Body, SecretKey).

decode_body(Ver, SeqNo, Unencrypted, SessNo, Len, Body, SecretKey) ->
  if
    Unencrypted == 0 ->
      Pad = md5_pad(SessNo, SecretKey, Ver, SeqNo, Len),
      xor_body(Body, Pad, []);
    true ->
      Body
  end.

xor_body([], _, Body) ->
  lists:reverse(Body);
xor_body(_, [], Body) ->
  lists:reverse(Body);
xor_body([B|BTail], [P|Ptail], Body) ->
  xor_body(BTail, Ptail, [(B bxor P)|Body]).

md5_pad(SessNo, Key, Ver, SeqNo, DataLen) ->
  Str = list_to_binary([<< SessNo:32 >>, Key, [Ver, SeqNo]]),
  MD1 = erlang:md5(Str),
  binary_to_list(md5_pad0(Str, DataLen, MD1, MD1, ?MD5_LEN)).

md5_pad0(Str, DataLen, MDP, Pad, PadLen) ->
  if
    PadLen > DataLen ->
      {RetBin, _} = erlang:split_binary(Pad, DataLen),
      RetBin;
    PadLen == DataLen ->
      Pad;
    true ->
      MDN = erlang:md5([Str,MDP]),
      md5_pad0(Str, DataLen, MDN, erlang:concat_binary([Pad, MDN]),
	       PadLen +?MD5_LEN)
  end.

incr_session(S=#session{seq_no=SeqNo}) ->
  S#session{seq_no=SeqNo+1}.

verify_session(RecvS=#session{seq_no=SeqNo},
	       _ExpS=#session{peer=Peer, seq_no=SeqNo, sess_no=undefined}
	      ) when SeqNo == 1 -> % Initial state
  RecvS#session{seq_no=SeqNo, peer=Peer};
verify_session(_RecvS=#session{seq_no=SeqNo, sess_no=SessNo},
	       ExpS=#session{seq_no=SeqNo, sess_no=SessNo}
	      ) ->
  ExpS;
verify_session(_RecvS=#session{seq_no=SqNoR, sess_no=SsNoR},
	       _ExpS=#session{seq_no=SqNoE, sess_no=SsNoE}
	      ) ->
  ?dbg(
     "bad session <seq_no:sess_no>"
     " expected <~w:~w>, received <~w:~w>",
     [SqNoE,SsNoE,SqNoR,SsNoR]),
  exit(bad_session_number).

expected_packet_type(ExpType, Sess) ->
  Type =
    case ExpType of
      aaa_start ->
	first_packet_type(Sess);
      _ ->
	ExpType
    end,
  Type.

first_packet_type(#session{type=?TAC_PLUS_AUTHEN}) ->
  authen_start;
first_packet_type(#session{type=?TAC_PLUS_AUTHOR}) ->
  author_request;
first_packet_type(#session{type=?TAC_PLUS_ACCT}) ->
  acct_request;
first_packet_type(_Sess) ->
  {error, undef_session_type}.

%% options
check_opts(Opts) ->
  Ret =
    props:gv_map(
      [{unencrypt, 16#0},
       {single_connection, 16#0},
       {random, default_random()},
       {debug, false}
      ],
      [check_opt(O) || O <- Opts]),
  Ret.

check_opt({K=unencrypt, V}) ->
  {K, check_bool(V)};
check_opt({K=single_connection, V}) ->
  {K, check_bool(V)};
check_opt({K=random, {M, F}}) ->
  {K, fun () -> M:F() end};
check_opt(KV={random, V}) when is_function(V) ->
  KV;
check_opt(KV={debug, Bool}) when Bool == true;
				 Bool == flase ->
  KV;
check_opt(O) ->
  exit({badopt, O}).

check_bool(V) when V == 16#0; V == 16#1 ->
  V;
check_bool(true) -> 16#1;
check_bool(false) -> 16#0.

default_random() ->
  fun () -> randomx:uniform(16#ffffffff) end.

fmt_header(H={_Ver,_Type,_SeqNo,_Unencrypted,_SingleConnect,_SessNo,_Len}) ->
  str:fmt("\n\t version      : ~w"
	  "\n\t type         : ~w"
	  "\n\t seqno        : ~w"
	  "\n\t unencrypted  : ~w"
	  "\n\t singleconnect: ~w"
	  "\n\t sessno       : ~w"
	  "\n\t len          : ~w",
	  tuple_to_list(H)
	 ).
