%%%----------------------------------------------------------------------
%%% File    : tac_fsm.erl
%%% Author  :  <svg@surnet.ru>
%%% Purpose : 
%%% Created : 3 Jul 2002 by  <svg@surnet.ru>
%%%	 $Id: tac_fsm.erl,v 1.1 2002/09/04 01:07:07 svg Exp $
%%%----------------------------------------------------------------------

-module(tac_protocol).
-author('svg@surnet.ru').

-export([mk_fsm/2]).

-import(wslib.props).
-import(wslib.log).

-include("tacacs.hrl").

-define(dbg(_F, _A), tac_lib:dbg(get(dbg), _F, _A)).
-define(MAX_RETRIES, 10).

-record(state, {user=#userid{},
		pass,
		data,
		authen,
		action,
		recv_timeout=?TCP_TIMEOUT,
		aaa_handler={tac_law, []},
		aaa_op,   % current aaa opration authen | author | acct
		fsm_type, % server | client
		phase,
		retries=0,
		debug=false
	       }).

%%%
%%% AaaHandler protocol
%%%
%%% AaaHandler:init(ConfigServer::pid()) -> {ok, AaaHlrState}
%%% 
%%% AaaHandler:authenticate(AuthenData, AaaHlrState) -> AaaHlrReturn
%%% 
%%% AaaHandler:authorize(AuthorData, AaaHlrState) -> AaaHlrReturn
%%% 
%%% AaaHandler:account(AccountData, AaaHlrState) -> AaaHlrReturn
%%%
%%% AuthenData = {UserId, Session, Authen, AuthenReqType, AuthenData}
%%% AuthorData = {UserId, Session, Authen, AuthorArgs}
%%% AccountData = {UserId, Session, Authen, AccPhase, AccountArgs}
%%% 
%%% AuthenReqType = enable | login | chpass
%%% AuthenData = Pass | {OldPass, NewPass}
%%% AaaHlrReturn = {Status, RetArgs, ServerMsg, UserData}
%%% 
%%% UserId = userid()
%%% Session = session()
%%% Authen = authen()
%%% AuthorArgs = [msg_arg()]
%%% AccountArgs = [msg_arg()]
%%% RetArgs = [msg_arg()]
%%% Status = ok | false | error
%%%
%%% AaaHlrState = term()
%%% 


%%%----------------------------------------------------------------------
%%% API
%%%----------------------------------------------------------------------
mk_fsm(FsmType, Args) when FsmType == server;
			   FsmType == client ->
  [RT, {AaaHlr, AaaHlrState}] = check_opts(Args),
  State = #state{recv_timeout=RT,
		 aaa_handler={AaaHlr,AaaHlrState},
		 fsm_type=FsmType
		},
  StartF = case FsmType of
	     server ->
	       fun (Parent) -> server_session(Parent, State) end;
	     client ->
	       Master = self(),
	       fun (Parent) -> client_session(Parent, Master, State) end
	   end,
  fun () ->
      Self = self(),
      spawn_link(fun () -> StartF(Self) end)
  end.

%%
%% Server
%%
server_session(Parent, State) ->
  server_session({Parent, State}).  

server_session({Parent, S=#state{recv_timeout=RT}}) ->
  receive {Parent, {debug, Dbg}} -> ok end,
  put(dbg, Dbg),
  Parent ! {self(), {recv, aaa_start, RT}},
  server_loop(Parent, fun server_start/3, S#state{debug=Dbg}).

server_loop(Parent, StateF, S=#state{recv_timeout=RT, aaa_op=AO}) ->
  Self = self(),
  receive
    {Parent, {error, Error}} ->
      server_log_error("transport/packet error, reason ~p", [Error], Parent);
    {Parent, {received, {Session, Answer}}} ->
      case
	Answer of
	{error, Reason} ->
	  server_log_error("error reason ~p", [Reason], Session),
	  SrvMsg = "packet error, check your key",
	  Parent ! done(AO, {error, [], SrvMsg, ""});
	_E=#error_reply{} ->
	  server_log_error("packet error, probably bad key",[], Session),
	  Parent ! done;
	#authen_continue{flags=?TAC_PLUS_CONTINUE_FLAG_ABORT,
			 data=Msg} ->
	  server_log_error("received abort, reason ~s",[Msg], Session),
	  Parent ! done;
	Data ->
	  case StateF(Data, Session, S) of
	    {continue, Reply, NxtPacketType, NxtStateF, NxtState} ->
	      Parent ! {Self, {continue, Reply, NxtPacketType, RT}},
	      server_loop(Parent, NxtStateF, NxtState);
	    Stop={done, _} ->
	      Parent ! {Self, Stop},
	      exit(normal)
	  end
      end
  end.

server_start(Data, Session, S) when is_record(Data, authen_start) ->
  ?dbg("start authentication ~p", [Session]),
  authen_start(Data, Session, S#state{aaa_op=authen});
server_start(Data, Session, S) when is_record(Data, author_request) ->
  ?dbg("start authorisation ~p", [Session]),
  author_start(Data, Session, S#state{aaa_op=author});
server_start(Data, Session, S) when is_record(Data, acct_request) ->
  ?dbg("start accounting ~p", [Session]),
  acct_start(Data, Session, S#state{aaa_op=acct}).

authen_start(_D=#authen_start{user=User, data=Pass, action=Action, authen=A},
	     Sess=#session{ver=Ver},
	     S) ->
  #authen{type=AT, service=Service} = A,
  NxtS = S#state{user=User, pass=Pass, authen=A, action=Action},
  case authen_check_ver(Action, AT, Ver) of
    false ->
      NxtVer = if Ver == ?TAC_PLUS_VERSION_0 -> ?TAC_PLUS_VERSION_1;
                  true -> ?TAC_PLUS_VERSION_0
	       end,
      {done, Reply} =
	authen_error({Action, AT, Ver}, "Unsupported action", Sess),
      {done, {NxtVer, Reply}};
    true ->
      case {Action, Service} of
	{login, enable} ->
	  authen_login(Sess, NxtS#state{authen=A#authen{type=ascii}});
	{login, _} ->
	  authen_login(Sess, NxtS);
	{chpass, _} ->
	  if (AT /= ascii) ->
	      authen_error(AT, "Unsupported authentication type", Sess);
	     true ->
	      authen_login(Sess, NxtS)
	  end;
	_ ->
	  authen_error({Action, Service}, "Unsupported action", Sess)
      end
  end.

authen_check_ver(login, AT, Ver) ->
  Ver == login_type_ver(AT);
authen_check_ver(_, _, ?TAC_PLUS_VERSION_0) ->
  true;
authen_check_ver(_, _, _) ->
  false.

authen_login(Sess, S=#state{user=User, authen=#authen{type=ascii}}) ->
  authen_getdata(User#userid.user, Sess, S#state{phase=getuser});
authen_login(Session, State) ->
  authen_login_done(Session, State).

authen_getdata(_P=#authen_continue{user_msg=Data}, Session, State) ->
  authen_getdata(Data, Session, State);
authen_getdata("", Sess, _S=#state{action=Ac, authen=A, retries=R})
  when R > ?MAX_RETRIES -> 
  authen_error({Ac, A, R}, "Maximum number of retries exceeded", Sess);
authen_getdata("", _Sess, S=#state{retries=R, phase=P}) ->
  PromptKey =
    case P of
      getuser ->
	{authen, login_prompt};
      getpass ->
	{authen, password_prompt};
      getdata ->
	{authen, chpass_prompt}
    end,
  Prompt = tacacs:config_gv(PromptKey),
  NxtS = S#state{retries=R+1},
  authen_continue(Prompt, fun authen_getdata/3, NxtS);
authen_getdata(Data, Sess,
	       S=#state{user=User, action=Ac, authen=A, phase=P,
			aaa_handler=AH
		       }) -> 
  ?dbg("<~p:~s> received=~p", [Sess#session.peer, P, Data]),
  case {P, Ac} of
    {getuser, _} ->
      User1 = User#userid{user=Data},
      %% is password required?
      AaaMsg = {User1, Sess, A, login, nopassword},
      Reply = handle_aaa_req(authenticate, AaaMsg, AH),
      if element(1, Reply) == ok ->
	  done(authen, Reply);
	 true ->
	  NxtS=S#state{user=User1, phase=getpass, retries=0},
	  authen_getdata("", Sess, NxtS)
      end;
    {getpass, chpass} ->
      NxtS = S#state{pass=Data, phase=getdata, retries=0},
      authen_getdata("", Sess, NxtS);
    {getpass, _} ->
      NxtS = S#state{pass=Data},
      case A#authen.service of
	enable ->
	  authen_enable_done(Sess, NxtS);
	_ ->
	  authen_login_done(Sess, NxtS)
      end;
    {getdata, chpass} ->
      authen_chpass_done(Data, Sess, S)
  end.

authen_continue(SrvMsg, StateF, S=#state{phase=Status}) ->
  EchoFlag = if Status == getuser -> 16#0;
	       true -> ?TAC_PLUS_REPLY_FLAG_NOECHO
	     end,
  Reply = #authen_reply{status=Status,
			flags=EchoFlag,
			server_msg=SrvMsg,
			data=[]
		       },
  {continue, Reply, authen_continue, StateF, S}.

authen_chpass_done(OldPass, Session,
		   _S=#state{user=User, pass=Pass, authen=A,
			     aaa_handler=AH
			    }) ->
  AaaMsg = {User, Session, A, chpass, {Pass, OldPass}},
  Reply = handle_aaa_req(authenticate, AaaMsg, AH),
  done(authen, Reply).

authen_enable_done(Session,
		   _S=#state{user=User, pass=Pass, authen=A,
			     aaa_handler=AH
			    }) ->
  AaaMsg = {User, Session, A, enable, Pass},
  Reply = handle_aaa_req(authenticate, AaaMsg, AH),
  done(authen, Reply).

authen_login_done(Session,
		  _S=#state{user=User, pass=Pass, authen=A,
			    aaa_handler=AH
			   }) ->
  AaaMsg = {User, Session, A, login, Pass},
  Reply = handle_aaa_req(authenticate, AaaMsg, AH),
  done(authen, Reply).

authen_error(Cause, SrvMsg, Session) ->
  server_log_error("authentication error <~p:~s>", [Cause, SrvMsg], Session),
  done(authen, {error, [], SrvMsg, ""}).

author_start(_P=#author_request{user=User, authen=ReqAuthen, args=ReqArgs},
	     Session,
	     S) ->
  author_done(ReqAuthen, ReqArgs, Session, S#state{user=User}).


author_done(Authen, ReqArgs,
	    Session,
	    _S=#state{user=User, aaa_handler=AH}) ->
  AaaMsg = {User, Session, Authen, ReqArgs},
  Reply = handle_aaa_req(authorize, AaaMsg, AH),
  done(author, Reply).

acct_start(_P=#acct_request{user=User, flags=Phase,
			    authen=ReqAuthen, args=ReqArgs},
	   Session,
	   S) ->
  acct_done(Phase, ReqAuthen, ReqArgs, Session, S#state{user=User}).

acct_done(Phase, Authen, ReqArgs,
	  Session,
	  _S=#state{user=User, aaa_handler=AH}) ->
  AaaMsg = {User, Session, Authen, Phase, ReqArgs},
  Reply = handle_aaa_req(account, AaaMsg, AH),
  done(acct, Reply).

done(Cause, {Status, Args, SrvMsg, Data}) ->
  Packet = pack_reply(Cause, Status, Args, SrvMsg, Data),
  {done, Packet}.

pack_reply(Cause, Status, Args, SrvMsg, Data) ->
  case Cause of
    undefined ->
      #error_reply{};
    error ->
      #error_reply{};
    authen ->
      #authen_reply{status=Status, server_msg=SrvMsg, data=Data,
                    flags=?TAC_PLUS_REPLY_FLAG_NOECHO
		   };
    author ->
      #author_response{status=Status, server_msg=SrvMsg, data=Data, args=Args};
    acct ->
      #acct_reply{status=Status, server_msg=SrvMsg, data=Data}
  end.

%%
%% Client
%%
client_session(Parent, Master, State) ->
  Master ! {Parent, {protocol_pid, self()}},
  receive {Parent, {debug, Dbg}} -> ok end,
  put(dbg, Dbg),
  client_session({Parent, Master, State#state{debug=Dbg}}).

client_session({Parent, Master, S=#state{recv_timeout=RT}}) ->
  Self = self(),
  receive
    {Master, {request, Data}} ->
      {Msg, NxtPacketType, NxtStateF, NxtS} =
	client_init_session(Data, S),
      Parent ! {Self, {continue, Msg, NxtPacketType, RT}},
      Result = client_loop(Parent, NxtStateF, NxtS),
      Master ! {Self, {answer, Result}},
      client_session({Parent, Master, S});
    {Master, stop} ->
      Parent ! done
  end.

client_loop(Parent, StateF, S=#state{recv_timeout=RT}) ->
  Self = self(),
  receive
    {Parent, Resp} ->
      case Resp of
	{received, {Session, Answer}} ->
	  case Answer of
	    #error_reply{} ->
	      {error, error_reply};
	    {error, Reason} ->
	      {error, {bad_packet, Reason}};
	    Data ->
	      case StateF(Data, Session, S) of
		{continue, Reply, NxtPacketType, NxtStateF, NxtState} ->
		  Parent ! {Self, {continue, Reply, NxtPacketType, RT}},
		  client_loop(Parent, NxtStateF, NxtState);
		_Stop={done, Return} ->
		  Return
	      end
	  end;
	Error={error, _} ->
	  Error
      end
  end.

client_init_session(Req=#client_request{
		      action = Action,
		      user   = User,
		      authen = Authen,
		      pass   = Pass,
		      data   = UserData,
		      args   = Args,
		      flags  = Flags
		     },
		    S) ->
  ?dbg({"request ", fun fmt_packet/1}, Req),
  #authen{type=AT} = Authen,
  {Msg, NxtStateF, NxtPacketType, AaaOp} =
    case Action of
      login ->
	Pass1 = if AT == ascii -> [];
		   true -> Pass
		end,
	Ver = login_type_ver(AT),
	R=#authen_start{user=User, action=Action, authen=Authen, data=Pass1}, 
	{{Ver, R},
	 fun client_authen/3,
	 authen_reply,
	 authen
	};
	chpass ->
	Pass1 = if AT == ascii -> [];
		   true -> Pass
		end,
	{#authen_start{user=User, action=Action, authen=Authen, data=Pass1},
	 fun client_authen/3,
	 authen_reply,
	 authen
	};
      enable ->
	Authen1 = Authen#authen{service=enable, type=undefined},
        {#authen_start{user=User, action=login, authen=Authen1, data=[]},
         fun client_authen/3,
	 authen_reply,
	 authen};
      authorize ->
        {#author_request{user=User, authen=Authen, args=Args},
         fun client_author_done/3,
	 author_response,
	 author};
      accounting ->
        {#acct_request{user=User, flags=Flags, authen=Authen, args=Args},
	 fun client_acct_done/3,
	 acct_reply,
	 acct}
    end,
  NxtS = S#state{user=User,
		 data=UserData,
		 pass=Pass,
		 action=Action,
		 aaa_op=AaaOp
		},
  {Msg, NxtPacketType, NxtStateF, NxtS}.

client_authen(_P=#authen_reply{status = Status,
			       server_msg = Msg,
			       flags = _Flags,
			       data =ServData},
	      _Session,
	      S=#state{user=User,
		       pass=Pass,
		       data=UserMsg}) ->
  case Status of
    _ when Status == ok;
	   Status == false;
	   Status == follow;
	   Status == restart ->
      {done, {Status, [], Msg, ServData}};
    error ->
      {done, {Status, [], Msg, ServData}};
    getuser ->
      Rep = #authen_continue{user_msg=User#userid.user},
      NxtStateF = fun client_authen/3,
      PacketType = authen_reply,
      {continue, Rep, PacketType, NxtStateF, S};
    getpass ->
      Rep = #authen_continue{user_msg=Pass},
      NxtStateF = fun client_authen/3,
      PacketType = authen_reply,
      {continue, Rep, PacketType, NxtStateF, S};
    getdata ->
      Rep = #authen_continue{user_msg=UserMsg},
      NxtStateF =  fun client_authen/3,
      PacketType = authen_reply,
      {continue, Rep, PacketType, NxtStateF, S}
  end.

client_author_done(_P=#author_response{status=Status,
				       args=Args,
				       server_msg=SrvMsg,
				       data=UserData
				      },
		   _Session,
		   _S) ->
  {done, {Status, Args, SrvMsg, UserData}}.

client_acct_done(_P=#acct_reply{status=Status,
				server_msg=SrvMsg,
				data=UserData
			       },
		 _Session,
		 _S) ->
  {done, {Status, [], SrvMsg, UserData}}.

login_type_ver(AT) when  AT == pap;
                         AT == chap;
                         AT == arap;
                         AT == 'ms-chap' ->
  ?TAC_PLUS_VERSION_1;
login_type_ver(_) ->
  ?TAC_PLUS_VERSION_0.

server_log_error(Format, Args, Parent) when is_pid(Parent) ->
  Parent ! {self(), peername},
  receive
    {Parent, {peername, {IP, Port}}} ->
      IpStr = tac_lib:ip_to_str(IP)
  end,
  server_log_error(Format, Args, IpStr, Port, undefined);


server_log_error(Format, Args,_S=#session{peer={Host, Port}, sess_no=SessNo}) ->
  server_log_error(Format, Args, Host, Port, SessNo).

server_log_error(Format, Args, Host, Port, SessNo) ->
  tacacs:log_error([{host, Host},
		    {port, Port},
		    {session, SessNo},
		    {format, Format, Args}]).

handle_aaa_req(Action, Req, {HlrMod, HlrState}) ->
  HlrMod:Action(Req, HlrState).

check_opts(Opts) ->
  Ret =
    props:gv_map(
      [{recv_timeout,?TCP_TIMEOUT}, {aaa_handler, {tac_law, []}}],
      [check_opt(O) || O <- Opts]),
  Ret.

check_opt(KV={recv_timeout,V}) when is_integer(V); V == infinity ->
  KV;
check_opt(KV={aaa_handler, {_M, _A}}) ->
  KV;
check_opt(O) ->
  exit({badopt, O}).

fmt_packet(P) ->
  tac_lib:fmt_record(P).
