%%%-------------------------------------------------------------------
%%% File    : cal.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : Some calendar functions
%%%
%%% Created :  9 Jan 2004 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id$
%%
%%**
%% 
%% @doc Some calendar functions
%% @end
%%*
-module(wslib.cal).

-export([day_name/1, day_num/1, month_name/1, month_num/1]).
-export([epoch_to_datetime/1, datetime_to_epoch/1]).
-export([format_datetime/1, format_datetime/2]).
-export([create_week/0, week_set_weekday/2, week_set_restday/2, week_is_restday/2, week_is_weekday/2]).
-export([create_time_interval/2, is_time_between/2]).
-export([valid_datetime/1, valid_time_of_day/1]).

-import(calendar).
-import(io_lib).
-import(lists).

-define(EPOCH, 62167219200). % 1970-01-01 00:00:00

%% @type datetime = {{Year, Month, Day}, {HH24, MI, SE}}
%% 

%% @spec epoch_to_datetime(T::second()) -> datetime()
%% @doc
%% Convert _epoch_ seconds to _datetime()_
%% 
epoch_to_datetime(T) when is_integer(T) ->
  calendar:gregorian_seconds_to_datetime(?EPOCH+T).

%% @spec datetime_to_epoch(T::integer()) -> second()
%% @doc
%% Convert _datetime()_ to _epoch_ seconds
%% 
datetime_to_epoch(DT={Date, Time}) ->
  case (calendar:datetime_to_gregorian_seconds(DT) - ?EPOCH) of
    V when V >= 0 ->
      V;
    _ ->
      exit(bad_datetime)
  end.

%% @spec valid_datetime(DateTime::datetime()) -> bool()
%% 
valid_datetime({Date, Time}) ->
  calendar:valid_date(Date) and valid_time_of_day(Time).

%% @spec valid_time_of_day(Time::time()) -> bool()
%% 
valid_time_of_day({HH24, MI, SE}) when HH24 =< 23,
				      HH24 >= 0,
				      MI >= 0,
				      MI =< 59,
				      SE >= 0,
				      SE =< 59 ->
  true;
valid_time_of_day(_) ->
  false.

%% @spec create_time_interval(From::datetime(), To::datetime()) -> time_interval()
%% @doc
%% Time interval between <code>From</code> and <code>To</code>.
%% If <code>To</code> is less than <code>From</code> it is treated
%% as time of next day.
%% 
create_time_interval(From, To) ->
  {From, To}.
%% @spec is_time_between(Time::datetime(),
%%                             Interval::time_interval()) -> boolean
%%                             
is_time_between({Day, Time}, Interval) ->
  is_time_between(Time, Interval);
is_time_between(Time, {From, To}) ->
  Time1 = calendar:time_to_seconds(Time),
  From1 = calendar:time_to_seconds(From),
  To1 = calendar:time_to_seconds(To),
  if (To > From) ->
      (Time >= From) and (Time =< To);
     (To < From) ->
      not ((Time >= To) and (Time =< From));
     true -> % From == To
      false
  end.

%% @doc
%% week is represented as integer, days are started from Monday
%% weekday mask = 0
%% restday mask = 1
%% @end
%%*

%% @spec create_week() -> week()
%% @doc Create default week representation <code>2#1100000</code>
%% 
create_week() ->
  2#1100000.

%% @spec week_set_weekday(DayNum::integer(), WeekMask::week()) -> week()
%% @doc Mark day in week as weekday
%% 
week_set_weekday(DayNum, WeekMask) when DayNum >=1;
					DayNum =< 7 ->
  WeekMask band (0 bsl (DayNum-1)).
%% @spec week_set_restday(DayNum::integer(), WeekMask::week()) -> week()
%% @doc Mark day in week as restday
%% 
week_set_restday(DayNum, WeekMask) when DayNum >=1;
					DayNum =< 7 ->
  WeekMask bor (1 bsl (DayNum-1)).

%% @spec week_is_restday(Day::integer(), WeekMask::week()) -> boolean()
%% @doc Return <code>true</code> if day is a restday or <code>false</code>.
%% 
week_is_restday(Day, WeekMask) ->
  (not week_is_weekday(Day, WeekMask)).
%% @spec week_is_weekday(Day::integer(), WeekMask::week()) -> boolean()
%% @doc Return <code>true</code> if day is a weekday or <code>false</code>.
%%
week_is_weekday({{Y, M, D}, {_H, _I, _S}}, WeekMask) -> 
  week_is_weekday({Y, M, D}, WeekMask);
week_is_weekday(Date={_Y, _M, _D}, WeekMask) ->
  week_is_weekday(calendar:day_of_the_week(Date), WeekMask);
week_is_weekday(DayNum, WeekMask) when DayNum >= 1,
				       DayNum =< 7 ->
  ((WeekMask band (1 bsl (DayNum-1))) == 0).

%% day of the week

day_name(1) -> "Mon";
day_name(2) -> "Tue";
day_name(3) -> "Wed";
day_name(4) -> "Thu";
day_name(5) -> "Fri";
day_name(6) -> "Sat";
day_name(7) -> "Sun".

day_num("Mon") -> 1;
day_num("Tue") -> 2;
day_num("Wed") -> 3;
day_num("Thu") -> 4;
day_num("Fri") -> 5;
day_num("Sat") -> 6;
day_num("Sun") -> 7.
%% month of the year

month_name(1) ->  "Jan";
month_name(2) ->  "Feb";
month_name(3) ->  "Mar";
month_name(4) ->  "Apr";
month_name(5) ->  "May";
month_name(6) ->  "Jun";
month_name(7) ->  "Jul";
month_name(8) ->  "Aug";
month_name(9) ->  "Sep";
month_name(10) -> "Oct";
month_name(11) -> "Nov";
month_name(12) -> "Dec".

month_num("Jan") -> 1 ;
month_num("Feb") -> 2 ;
month_num("Mar") -> 3 ;
month_num("Apr") -> 4 ;
month_num("May") -> 5 ;
month_num("Jun") -> 6 ;
month_num("Jul") -> 7 ;
month_num("Aug") -> 8 ;
month_num("Sep") -> 9 ;
month_num("Oct") -> 10;
month_num("Nov") -> 11;
month_num("Dec") -> 12.

format_datetime(FS=localtime) ->
  format_datetime(FS, calendar:local_time());
format_datetime(FormatStyle) ->
  format_datetime(FormatStyle, calendar:universal_time()).

%% borrowed from Inets module
format_datetime(apache, UniversalTime) ->
  LocalTime=calendar:universal_time_to_local_time(UniversalTime),
  Minutes=round((calendar:datetime_to_gregorian_seconds(UniversalTime) -
                 calendar:datetime_to_gregorian_seconds(LocalTime))/60),
  {{YYYY,MM,DD},{Hour,Min,Sec}}=LocalTime,
  Date=io_lib:format("~.2.0w/~.3s/~.4w:~.2.0w:~.2.0w:~.2.0w ~c~.2.0w~.2.0w",
                     [DD, month_name(MM),YYYY,Hour,Min,Sec,sign(Minutes),
                      abs(Minutes) div 60,abs(Minutes) rem 60]),  
  lists:flatten(Date);
%% borrowed from ce_calendar module of catseye@catseye.mb.ca
format_datetime(rfc1123, {{Y, M, D}, {H, I, S}}) ->
  Dow = day_name(calendar:day_of_the_week(Y, M, D)),
  Month = month_name(M),
  Date = io_lib:format(
	   "~s, ~2.2.0w ~s ~w ~2.2.0w:~2.2.0w:~2.2.0w GMT",
	   [Dow, D, Month, Y, H, I, S]),
  lists:flatten(Date);
format_datetime(localtime, {{Y, M, D}, {H, I, S}}) ->
 Date = io_lib:format(
	   "~.4w.~.2.0w.~.2.0w:~2.2.0w:~2.2.0w:~2.2.0w",
	  [Y, M, D, H, I, S]),
  lists:flatten(Date).

sign(Minutes) when Minutes > 0 ->
  $+;
sign(Minutes) ->
  $-.
