%%%-------------------------------------------------------------------
%%% File    : dbg_srv.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : 
%%%
%%% Created : 18 Jun 2002 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id: dbg_srv.erl,v 1.2 2003/10/27 16:45:59 vsekissov Exp $
%%
%%**
%% Trace server traces functions:
%% 
%%      dbg_info(Format, [Arg]) -> ok.
%%      
%% Receiving event trace server invokes
%% 
%%      error_logger:info_msg(Header++Format, [TracedModule, TracedPid| Args])
%%
%% 
%% lvl_mods -> [{DebugLevel, [Module]}]
%% 
%% start_link() - start trace server
%% 
%% stop() - stop trace server
%% 
%% info(Srv, Lvl) -> [TracedDebugLevel]
%%
%% stop/start trace for levels
%% 
%% log(Srv, Lvl, Flag) -> Result
%% Lvl - all | DebugLevel | [DebugLevel]
%% 


%%*
-module(wslib.dbg_srv).

%% External exports
-export([start_link/1, start/1, stop/0, log/2, info/1, add_level/2, del_level/1]).
%% Internal exports
-export([init/1]).

-import(.lists).
-import(.proc_lib).
-import(.error_logger).
-import(.calendar).
-import(.io).
-import(.erlang).

-define(DBG_FUN, dbg_info).
-define(DBG_FUN_ARITY, 2).

%% User functions
start(LevelDesc) ->
  case whereis(?MODULE) of
    Pid when pid(Pid) ->
      {ok, Pid};
    _ ->
      proc_lib:start(?MODULE, init, [{self(), ?MODULE, LevelDesc}])
  end.

start_link(LevelDesc) ->
  case whereis(?MODULE) of
    Pid when pid(Pid) ->
      {ok, Pid};
    _ ->
      proc_lib:start_link(?MODULE, init, [{self(), ?MODULE, LevelDesc}])
  end.

stop() ->
  case whereis(?MODULE) of
    Pid when pid(Pid) ->
      Pid ! {stop, Pid},
      ok;
    _ ->
      {error, not_started}
  end.

info(Query) when Query == level ->
  send_recv(?MODULE, Query).

log(Lvl, Flag) when Flag == true;
		    Flag == false ->
  send_recv(?MODULE, {log, Lvl, Flag}).

add_level(Level, Modules) when atom(Level);
			       list(Modules) ->
  send_recv(?MODULE, {add_level, Level, Modules}).

del_level(Level) ->
  send_recv(?MODULE, {del_level, Level}).

%% Internal

init({Parent, Name, LevelDesc}) ->
  init({Parent, Name, LevelDesc,
	fun (F, A) -> error_logger:info_msg(F, A) end});
init({Parent, Name, LevelDesc, ReportFun}) ->
  Self = self(),
  case register(Name, Self) of
    badarg ->
      exit({error, already_registered});
    true ->
      proc_lib:init_ack(Parent, {ok, Self})
  end,
  process_flag(trap_exit, true),
  erlang:trace_pattern({'_', '_', '_'}, false, [local]),
  set_trace(true, LevelDesc),
  simple_tracer_loop(ReportFun, LevelDesc).
  
simple_tracer_loop(ReportFun, LevelDesc) ->
  Self = self(),
  receive
    {Tag, Pid, Flag, {M, dbg_info, [Fmt, Args]}, TimeStamp}
    when ((Flag == call) or (Flag == return_to)) and
         ((Tag == trace) or (Tag == trace_ts)) ->
      ReportFun("*DEBUG* ~w ~p " ++ Fmt ++ "~n", [M, Pid| Args]),
      simple_tracer_loop(ReportFun, LevelDesc);
    {From, level} ->
      From ! {Self, do_info(level, LevelDesc)},
      simple_tracer_loop(ReportFun, LevelDesc);
    {From, {log, Lvl, Flag}} ->
      IsLvl = is_level(Lvl, LevelDesc),
      From ! {Self, IsLvl},
      if IsLvl ->
	  do_trace(Lvl, Flag, LevelDesc);
	 true ->
	  ok
      end,
      simple_tracer_loop(ReportFun, LevelDesc);
    {From, {add_level, Lvl, Modules}} ->
      NotLvl = (not is_level(Lvl, LevelDesc)),
      NxtLevelDesc =
	if NotLvl ->
	    [{Lvl, Modules}|LevelDesc];
	   true ->
	    LevelDesc
	end,
      From ! {Self, NotLvl},
      simple_tracer_loop(ReportFun, NxtLevelDesc);
    {From, {del_level, Lvl}} ->
      IsLvl = is_level(Lvl, LevelDesc),
      NxtLevelDesc =
	if IsLvl ->
	    lists:keydelete(Lvl, 1, LevelDesc);
	   true ->
	    LevelDesc
	end,
      From ! {Self, IsLvl},
      simple_tracer_loop(ReportFun, NxtLevelDesc);
    {stop, Self} ->
      set_trace(false, LevelDesc),
      {'EXIT', normal};
    {'EXIT', PPid, shutdown} ->
      set_trace(false, LevelDesc),
      exit(shutdown);
    {'EXIT', Self, Reson} ->
      set_trace(false,LevelDesc),
      exit(Reson);
    Other ->
      io:format("~w other message ~p~n", [?MODULE, Other]),
      simple_tracer_loop(ReportFun, LevelDesc)
  end.

set_trace(Flag, LevelDesc) when Flag == true;
				Flag == false ->
  erlang:trace(all, Flag, [call, timestamp]),
  if Flag == false ->
      do_trace(all, Flag, LevelDesc);
     true ->
      ok
  end,
  ok.

do_trace(all, Flag, LevelDesc) ->
  do_trace(all_levels(LevelDesc), Flag, LevelDesc);
do_trace(Lvl, Flag, LevelDesc) when atom(Lvl) ->
  do_trace([Lvl], Flag, LevelDesc);  
do_trace(Lvls, Flag, LevelDesc) ->
  Modules = lists:usort(
              lists:append(
                [ Ms || {L, Ms} <- LevelDesc, lists:member(L, Lvls) ])),
  Result = [{M, erlang:trace_pattern({M, ?DBG_FUN, '_'}, Flag, [local])}
            || M <- lists:usort(Modules)],
  Result.

do_info(level, LevelDesc) ->
  ModuleIsTraced = fun (M, Prev) ->
                       case is_traced(M) of
                         true ->
                           Prev;
                         _ ->
                           throw(false)
                       end
                   end,
  
  AllTraced = fun (Ms) ->
                  (catch lists:foldl(ModuleIsTraced, true, Ms))
              end,

  LevelIsTraced = fun ({L, Ms}) -> AllTraced(Ms) end,
  [L || {L, Ms} <- LevelDesc, AllTraced(Ms) == true].

is_level(all, _) -> true;
is_level(Lvl, LevelDesc) ->
  lists:keymember(Lvl, 1, LevelDesc).

is_traced(M) ->
  erlang:trace_info({M, ?DBG_FUN, ?DBG_FUN_ARITY}, traced) =/= false.

all_levels(LevelDesc) ->
  [L || {L, M} <- LevelDesc].

send_recv(Srv, Query) ->
  SrvPid = if is_pid(Srv) ->
	       Srv;
	      true ->
	       whereis(Srv)
	   end,
  Ref = erlang:monitor(process, SrvPid),
  if is_pid(SrvPid) ->
      SrvPid ! {self(), Query};
     true ->
      ok
  end,
  receive
    {SrvPid, Answer} ->
      erlang:demonitor(Ref),
      Answer;
    {'DOWN', Ref, _, _, _} ->
      {error, noproc}
  end.

