%%%-------------------------------------------------------------------
%%% File    : file.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : File functions
%%%
%%% Created : 31 May 2004 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id$
%%
%%**
%% 
%% @doc
%% @end
%%*
-module(wslib.file).

-include_lib("kernel/include/file.hrl").

-export([test/2]).

-import(file).
-import(os).
-import(string).
-import(lists).

%% @spec test(Switches, File) -> boolean()
%% Switches = [Switch] | Switch
%% Switch = integer | atom()
%% File = string() | file_info()
%% @doc Test file according to switches.
%% <br>
%% <code>wslib.file:test("xrwf", ".").</code>
%% <br>
%% <code>wslib.file:test([x,r,w,f], ".").</code>
%% 
%% <ul>
%% <li><code> e </code> - file exists</li>
%% <li><code> z </code> - file exists and has zero size</li>
%% <li><code> s </code> - file exists and has non zero size</li>
%% <li><code> r </code> - file is readable</li>
%% <li><code> w </code> - file is writable</li>
%% <li><code> x </code> - file is executable</li>
%% <li><code> o </code> - file is owned by effective uid</li>
%% <li><code> O </code> - file is owned by effective gid</li>
%% <li><code> f </code> - file is a regular file</li>
%% <li><code> d </code> - file is a directory</li>
%% <li><code> l </code> - file is a symlink</li>
%% <li><code> u </code> - file has setuid bit</li>
%% <li><code> g </code> - file has setgid bit</li>
%% </ul>
%% @end
test(Switches, File) when is_list(File) ->
  case file:read_link_info(File) of
    {ok, FI} ->
      test(Switches, FI, File);
    Error ->
      false
  end.

test(Switches, FI, FN) when is_list(Switches) ->
  (not lists:member(false, [test(Sw, FI, FN) || Sw <- Switches]));
test(Switch, FI, FN) when is_integer(Switch) ->
  test(list_to_atom([Switch]), FI, FN);
test(e, I, FN) when is_record(I, file_info) ->
  true;
test(z, I=#file_info{size=0}, FN) ->
  true;
test(s, I=#file_info{size=SZ}, FN) when SZ =/= 0 ->
  true;
test(r, I=#file_info{access=A}, FN) when A == read;
				     A == read_write ->
  true;
test(w, I=#file_info{access=A}, FN) when A == write;
				     A == read_write ->
  true;
test(x, I=#file_info{mode=M}, FN) ->
  if (8#00001 band M) =/= 0 ->
      true;
     true ->
      OwnedByUser = test(o, I, FN),
      if OwnedByUser and ((8#00100 band M) =/= 0) ->
	  true;
	 true ->
	  OwnedByGroup = test('O', I, FN), 
	  if OwnedByGroup and ((8#00010 band M) =/= 0) ->
	      true;
	     true ->
	      false
	  end
      end
  end;
test(o, I=#file_info{uid=Uid}, FN) ->
  case element(1, os:type()) of
    unix ->
      Uid == list_to_integer(string:strip(os:cmd("id -u"), right, $\n));
    _ ->
      true
  end;
test('O', I=#file_info{gid=Gid}, FN) ->
  case element(1, os:type()) of
    unix ->
      Gid == list_to_integer(string:strip(os:cmd("id -g"), right, $\n));
    _ ->
      true
  end;
test(f, I=#file_info{type=regular}, FN) ->
  true;
test(d, I=#file_info{type=directory}, FN) ->
  true;
test(T, I=#file_info{type=symlink}, L) when T == f; T == d ->
  case file:read_link(L) of
    {ok, FN} ->
      test(T, FN);
    Error ->
      false
  end;
test(l, I=#file_info{type=symlink}, FN) ->
  true;
test(u, I=#file_info{mode=M}, FN) ->
  (16#800 band M) =/= 0;
test(g, I=#file_info{mode=M}, FN) ->
  (16#400 band M) =/= 0;
test(_, _ , _) ->
  false.
