%%%-------------------------------------------------------------------
%%% File    : hof.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : Higher-order functions
%%%
%%% Created : 14 Jan 2004 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id$
%%
%%**
%% 
%% @doc
%% @end
%%*
-module(wslib.hof).

-export([compose/2, compose/1, curry2/1, curry3/1, uncurry2/1, uncurry3/1]).
-export([first/2, while/2, attempt/2, attempt/3]).

-import(lists).

%% @spec while(BodyFun, Arg) -> Ret::term()
%% BodyFun = fun (Arg) -> {true, NxtArg} | {false, Return}
%% 
while(BodyFun, Arg) ->
  while1(BodyFun(Arg), BodyFun).

while1({true, Arg}, BodyFun) ->
  while1(BodyFun(Arg), BodyFun);
while1({false, Ret}, _BodyFun) ->
  Ret.

%%**
%% @spec first(Test, Funs) -> {ok, Value} | false
%% Test = Test(V) -> true | false
%% Funs = function()
%% @doc Try <code>Funs</code> in order and return <code>{ok, Value}</code>
%% of return value first function on which Test returns <code>true</code>
%% or return <code>false</code>
%% @end
%% *
first(Test, []) ->
  false;
first(Test, [F|Funs]) ->
  V = F(),
  case Test(V) of
    true ->
      {ok, V};
    false ->
      first(Test, Funs)
  end.

%%**
%% @spec compose(X::function(), Y::function()) -> function()
%% @doc Compose functions X and Y
%% @end
%%*
compose(X, Y) ->
  fun (Z) -> apply(X, apply(Y, [Z])) end.

%%**
%% @spec compose([function()]) -> function()
%% @doc Compose list of functions
%% @end
%%*
compose([]) ->
  exit(badarg);
compose(Funs) ->
  Funs1 = lists:reverse(Funs),
  fun (Z) ->
      lists:foldl(fun (F, Prev) -> apply(F, [Prev]) end,
		  Z, Funs1)
  end.

%%**
%% @spec curry2(function()) -> function()
%% @doc Curry function of two arguments
%% @end
%%* 
curry2(F) ->
  fun (X) ->
      fun (Y) -> F(X, Y) end
  end.

%%**
%% @spec curry3(function()) -> function()
%% @doc Curry function of three arguments
%% @end
%% *

curry3(F) ->
  fun (X) ->
      fun (Y) ->
	  fun (Z) -> F(X, Y, Z) end
      end
  end.

uncurry2(F) ->
  fun (X, Y) ->
      (F(X))(Y)
  end.

uncurry3(F) ->
  fun (X, Y, Z) ->
      ((F(X))(Y))(Z)
  end.

%% Borrowed from jungerl/lib/ce/src/ce_lib.erl
%% @spec attempt(function(), [term()]) -> {ok, term()} | {error, term()}
%% @doc Catches errors from an error-throwing function.
%% Thanks to Thomas Arts for this code.

attempt(F, Args) ->
  case catch apply(F, Args) of
    {'EXIT', Why} -> {error, Why};
    Other -> {ok, Other}
  end.

%% Borrowed from jungerl/lib/ce/src/ce_lib.erl
%% @spec attempt(function(), [term()], term()) -> term()
%% @doc Catches errors and returns a default value if error thrown.
%% Also can be used with functions which return {ok, Result} | {error, Why}.

attempt(F, Args, Default) ->
  case catch apply(F, Args) of
    {'EXIT', _} -> Default;
     {error, _} -> Default;
    {ok, Value} -> Value;
          Other -> Other
  end.
