%%%-------------------------------------------------------------------
%%% File    : mnesiadb.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : Misc Mnesia helper funs
%%%
%%% Created :  5 Jan 2003 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id: mnesiadb.erl,v 1.3 2003/10/27 16:45:59 vsekissov Exp $
%%
%%
%%**
%% 
%% @doc
%% @end
%%*
-module(wslib.mnesiadb).

-export([create_schema/2, connect/2]).
-export([create_tables/2, delete_tables/2, add_tables/2, wait_for_tables/2]).
-export([add_table_replicas/2]).
-export([init_sequences/2, next_rowid/1, next_rowid/2, current_rowid/1]).
-export([update_counter/3, create_constraints/1]).
-export([delete_row/2, read_row/2, write_row/2, fetch_row/2, fetch/1, fetch/2]).
-export([delete/2, read/2, write/2, select/2]).
-export([dirty_delete/2, dirty_read/2, dirty_write/2, dirty_select/2]).
-export([make_record/1, make_record/2, mk_make_record/1]).
-export([make_object/1, make_object/2, mk_make_object/1]).
-export([dirty_foldl/3, match_spec/2]).
-export([export/2, import/2]).

%% internal exports
-export([connect/1]).

%%
-import(lists).
-import(mnesia).
-import(mnesia_schema).
-import(rdbms).
-import(rpc).
-import(io).
-import(proplists).
-import(wslib.listn).
-import(wslib.log).
-import(wslib.props).
-import(wslib.misc).
-import(wslib.hof).

-include("mnesiadb.hrl").

%% Sequences for tables
-record(table_sequence, {table,
			 val
			}).

%% @spec create_schema([node()], [table_def()]) -> ok | exit
%% table_def() = {table_name(), [attr()]}
%% @doc Create schema and tables on nodes.
%% @see create_tables/2
%% 
create_schema(Nodes, TableDefs) when is_list(Nodes) ->
  case mnesia:create_schema(Nodes) of
    ok ->
      ok;
    {error, {_,{already_exists, SchemaNode}}} ->
      log:write_error("schema already exists on ~s, "
		"trying to continue with tables install~n", [SchemaNode]),
      ok;
    Error1 ->
      exit(Error1)
  end,
  create_tables(Nodes, TableDefs).

%% @spec connect(Node::node(), MasterNode::node()) -> ok | exit
%% @doc Connect <code>Node</code> to database on <code>MasterNode</code>.
%% 
connect(Node, MasterNode) when is_atom(Node) ->
  misc:call_node(Node, ?MODULE, connect, [MasterNode]).

%% @spec connect(MasterNode::node()) -> ok | exit
%% @see connect/2
%% 
connect(MasterNode) ->
  mnesia:start(),
  Node = node(),
  DbNodes = misc:call_node(MasterNode, mnesia, system_info, [db_nodes]),

  MasterUserProps =
    misc:call_node(MasterNode, mnesia, table_info, [schema, user_properties]),
  UserProps = mnesia:table_info(schema, user_properties),

  %% Mnesia doesn't allow to merge schemas with different user_properties,
  %% but if ours are empty we can copy them from master node.
  case UserProps of
    [] ->
      WritePropsF =
	fun () ->
	    lists:foreach(
	      fun (P) -> mnesia_schema:do_write_table_property(schema, P) end,
	      MasterUserProps),
	    ok
	end,
      {atomic, ok} = mnesia_schema:schema_transaction(WritePropsF);
    _ ->
      ok
  end,
  {ok, _} = mnesia:change_config(extra_db_nodes, DbNodes),
  mnesia:change_table_copy_type(schema, Node, disc_copies),
  ok.

%% @spec create_tables(Nodes, TableDefs) -> AliveNodes | exit
%% TableDefs = [{table_name(), [attr()]}]
%% AliveNodes = [node()]
%% %% @doc Create tables on Nodes. Tables are created as disk copies
%% if storage type is not specified.

create_tables(Nodes, TableDefs) ->
  Up = mnesia_start(Nodes),
  CreateTableF =
    fun ({Name, Attrs}) ->
	Attrs1 =
	  case lists:any(fun ({K, _}) when K == disc_copies;
					   K == disc_only_copies;
					   K == ram_copies ->
			     true;
			     (_) ->
			     false
			 end, Attrs) of
	    true ->
	      Attrs;
	    false ->
	      [{disc_copies, Nodes}|Attrs]
	  end,
	case mnesia:create_table(Name, Attrs1) of
          {atomic, _} -> ok;
	  {aborted,{already_exists,_}} ->
	    log:write_error("table ~s already exists, trying to continue~n", [Name]),
	    ok;
          Error -> exit(Error)
        end
    end,
  lists:foreach(CreateTableF, TableDefs),
  Up.

%% @spec delete_tables(Nodes, TableNames) -> AliveNodes | exit
%% Nodes = [node()]
%% TableNames = [table_name()]
%%
%% @doc Delete table copies from <code>Nodes</code>. Return list of alive
%% nodes.
%% 
delete_tables(Nodes, TableNames) ->
  Up = mnesia_start(Nodes),

  DeleteTableCopyF =
    fun (Node, TableName) ->
        case mnesia:del_table_copy(TableName, Node) of
          {aborted, {badarg, _,unknown}} -> ok;
          {atomic, _} -> ok;
          Error -> exit(Error)
        end
    end,
  listn:mapn(DeleteTableCopyF, [Up, TableNames]),
  Up.


%% @spec add_replicas(Nodes, MasterNode) -> ok | exit(Error)
%% Nodes = [node()]
%% MasterNode = node()
%% @doc Create replicas of tables on <code>MasterNode</node> with the same
%% storage type on <code>Nodes</code>.
%% 
add_table_replicas(Nodes, MasterNode) ->
  Tables = lists:delete(
	     schema,
	     misc:call_node(MasterNode, mnesia, system_info, [tables])
	    ),
  TableDefs =
    [{T, misc:call_node(MasterNode, mnesia, table_info, [T, storage_type])}
     || T <- Tables],
  add_tables(Nodes, TableDefs).

%% @spec add_tables(Nodes, TableDefs) -> AliveNodes | exit
%% Nodes = [node()]
%% TableDefs = [TableDef]
%% TableDef = {table_name(), storage_type()} | table_name()
%% AliveNodes = [node()]
%% 
%% @doc Create table copies on <code>Nodes</code>. Return list of alive
%% nodes. Tables are created as disk copies if storage type is not specified.
%% 
add_tables(Nodes, TableDefs) ->
  TableDefs1 = lists:map(fun (TD={_TableName, _StorageType}) ->
			     TD;
			     (TableName) ->
			     {TableName, disc_copies}
			 end, TableDefs),
  TableNames = proplists:get_keys(TableDefs1),

  Up = mnesia_start(Nodes),

  AddTableCopyF =
    fun (Node, {TableName, StorageType}) ->
        case mnesia:add_table_copy(TableName, Node, StorageType) of
          {aborted, {already_exists, _, _}} -> ok;
          {atomic, _} -> ok;
          Error -> exit(Error)
        end
    end,
  listn:mapn(AddTableCopyF, [Up, TableDefs1]),
  Up.

wait_for_tables(Tables, Timeout) ->
  ForceLoadF = fun(Table, Bad) ->
                  case mnesia:force_load_table(Table) of
                    yes ->
                      log:write_info("force load table ~s", [Table]),
                      Bad;
                    Error ->
                      [{Table, Error}|Bad]
                  end
              end,

  case mnesia:wait_for_tables(Tables,Timeout) of
    {timeout,BadTables} ->
      case lists:foldl(ForceLoadF, [], BadTables) of
        [] ->
          ok;
        TableErrors ->
	  {error, {tables_not_loaded, TableErrors}}
      end;
    Other ->
      Other
  end.

create_constraints(Cs) ->
  create_constraints(Cs, []).

create_constraints([], Props) ->
  {atomic, ok} = rdbms:add_properties(Props);
create_constraints([{{table, TableName, attrs}, TableAttrs}|Rest],
		   Props) ->
  create_constraints(Rest, Props++create_table_attrs(TableName, TableAttrs));
create_constraints([{{record, RecName, attrs}, RecAttrNames, RecAttrs}|Rest],
		   Props) ->
  create_constraints(Rest,
		     Props++create_record_attrs(RecName, RecAttrNames, RecAttrs)).
create_table_attrs(TableName, TableAttrs) ->
  lists:flatten([[{{attr, {TableName, AttrName}, PropName}, PropValue}
		  ||{PropName, PropValue} <- AttrProps]
		 || {AttrName, AttrProps} <- TableAttrs]).

create_record_attrs(RecName, RecAttrNames, RecAttrs) ->
  [{{record, RecName, attrs}, RecAttrNames}|
   lists:flatten([[{{attr, AttrName, PropName}, PropValue}
		   || {PropName, PropValue} <- AttrProps]
		  || {AttrName, AttrProps} <- RecAttrs])
  ].


init_sequences(Nodes, Seq) ->
  {Tbl, Attrs} = ?table_def(table_sequence, [{type, set}]),
  case mnesia:create_table(Tbl, [{disc_copies, Nodes}|Attrs]) of
    {aborted, {already_exists, _, _}} -> ok;
    {atomic, _} -> ok;
    Error -> exit(Error)
  end,
  lists:foreach(fun ({T, InitVal}) ->
		    create_seq(T, InitVal);
		    (T) ->
		    create_seq(T)
		end, Seq).

create_seq(Tbl) ->
  create_seq(Tbl, 0).

create_seq(Tbl, InitVal) when integer(InitVal) ->
  Tr = fun () ->
	   mnesia:write(#table_sequence{table=Tbl, val=InitVal})
       end,
  case mnesia:transaction(Tr) of
    {atomic, _} ->
      ok;
    {aborted, Error} ->
      exit(Error)
  end.

%% Mnesia

%% @spec export(IO::io_device(), Tbl::table()) -> ok | {error, Reason}
%% @doc Exports the table to a disk file. We use the same format
%% as _ce_db:export/3 (jungerl/ce)
%% @see
export(IO, Tbl) ->
  io:fwrite(IO, "~p.~n", [Tbl]),
  MkObjectF = mk_make_object(Tbl),
  WriteF = fun (Rec, _) ->
	       io:fwrite(IO, "~p.~n", [MkObjectF(Rec)])
	   end,
  case mnesia:transaction(fun() ->
			      mnesia:foldl(WriteF, ok, Tbl),
			      io:fwrite(IO, "~p.~n", [end_of_records])
			  end) of
    {atomic, Good} -> Good;
    {aborted, Error} -> {error, Error}
  end.

%% @spec import(IO::io_device(), Tbl::table()) -> ok | {error, Reason}
%% @doc Import the table from a disk file.We use the same format
%% as _ce_db:import/3 (jungerl/ce)
import(IO, Tbl) ->
  BodyF = fun (Cnt) ->
	      case io:read(IO, '') of
		{ok, end_of_records} ->
		  {false, {ok, Cnt}};
		{ok, KVs} ->
		  case write_row(Tbl, KVs) of
		    {error, Error} ->
		      {false, {error, {{record_number, Cnt}, Error}}};
		    ok ->
		      {true, Cnt+1}
		  end;
		Error ->
		  {false, Error}
	      end
	  end,
  case io:read(IO, '') of
    {ok, Tbl} ->
      hof:while(BodyF, 0);
    _ ->
      {error, could_not_read_table_tag}
  end.
  
dirty_foldl(Fun, Acc, Tab) ->
  Key = mnesia:dirty_first(Tab),
  dirty_foldl(Fun, Acc, Tab, Key).

dirty_foldl(Fun, Ret, Tab, '$end_of_table') ->
  Ret;
dirty_foldl(Fun, Acc, Tab, Key) ->
  NxtAcc = Fun(Key, Acc),
  NxtKey = mnesia:dirty_next(Tab, Key),
  dirty_foldl(Fun, NxtAcc, Tab, NxtKey).

mk_make_record(Tbl) ->
  Attrs = mnesia:table_info(Tbl, attributes),
  RecName = mnesia:table_info(Tbl, record_name),
  fun (KVs) ->
      make_record(RecName, Attrs, KVs)
  end.

make_record(Tbl) ->
  default_record(Tbl).

make_record(Tbl, KVs) ->
  RecName = mnesia:table_info(Tbl, record_name),
  Attrs = mnesia:table_info(Tbl, attributes),
  make_record(RecName, Attrs, KVs).

make_record(RecName, Attrs, KVs) ->
  list_to_tuple([RecName|props:gv_map(Attrs, KVs)]).

default_record(Tbl) ->
  RecName = mnesia:table_info(Tbl, record_name),
  Vs = tl(tuple_to_list(rdbms:default_record(Tbl))),
  list_to_tuple([RecName|[null_to_undefined(V) || V <- Vs]]).

mk_make_object(Tbl) ->
  RecName = mnesia:table_info(Tbl, record_name),
  Attrs = mnesia:table_info(Tbl, attributes),
  fun (Record) ->
      make_object(Record, RecName, Attrs)
  end.

make_object(Tbl) ->
  make_object(Tbl, default_record(Tbl)).

make_object(Tbl, Record) when is_atom(Tbl) ->
  RecName = mnesia:table_info(Tbl, record_name),
  Attrs = mnesia:table_info(Tbl, attributes),
  make_object(Record, RecName, Attrs);
make_object(From, To) -> % replace KVs in To from From
  lists:foldl(fun (KV={K,_}, Obj) ->
		  lists:keyreplace(K, 1, Obj, KV)
	      end, To, From).

make_object(Record, RecName, Attrs) ->
  [RecName|AttrVals] = tuple_to_list(Record),
  listn:zip(Attrs, AttrVals).

null_to_undefined(V) ->
  null_to_undefined(V, rdbms:null_value()).

null_to_undefined(Null, Null) -> undefined;
null_to_undefined(T, Null) when is_tuple(T) ->
  list_to_tuple([null_to_undefined(E, Null) || E <- tuple_to_list(T)]);
null_to_undefined(V, _) ->
  V.

%% @spec update_counter(table(), key()) -> integer() | exit({aborted, Reason})
%% @doc Transaction context agnostic update counter.

update_counter(Tab, Key, Incr) ->
  case get(mnesia_activity_state) of
    TrId={_, _, _} ->
      [{T,K,Number}] = mnesia:read(Tab, Key),
      NewNumber = Number + Incr,
      mnesia:write(Tab, {T,K,NewNumber}),
      NewNumber;
    _ ->
      mnesia:dirty_update_counter(Tab, Key, Incr)
  end.

next_rowid(Tbl) ->
  next_rowid(Tbl, 1).

next_rowid(Tbl, Incr) when is_integer(Incr),
			   Incr > 0 ->
  update_counter(table_sequence, Tbl, Incr).

current_rowid(Tbl) ->
  ReadF = case get(mnesia_activity_state) of
	    TrId={_, _, _} ->
	      read;
	    _ ->
	      dirty_read
	  end,
  [#table_sequence{val=Ret}]=mnesia:ReadF({table_sequence, Tbl}),
  Ret.

fetch_row(Tbl, KVs) ->
  fetch(read_row(Tbl, KVs)).

fetch(QueryResult, Default) ->
  case fetch(QueryResult) of
    {ok, Ret} ->
      Ret;
    {error, no_data_found} ->
      Default;
    {error, Error} ->
      exit(Error)
  end.

fetch(QueryResult) ->
  case QueryResult of
    [Row] ->
      {ok, Row};
    [] ->
      {error, no_data_found};
    Rs when is_list(Rs) ->
      {error, too_many_rows}
  end.

write_row(Tbl, KVs) ->
  mnesia_do(fun () -> write(Tbl, KVs) end).

read_row(Tbl, KVs) ->
  mnesia_do(fun () -> read(Tbl, KVs) end).
  
delete_row(Tbl, KVs) ->
  mnesia_do(fun () -> delete(Tbl, KVs) end).
%%
dirty_write(Tbl, KVs) ->
  do_write(dirty_write, Tbl, KVs).

write(Tbl, KVs) ->
  do_write(write, Tbl, KVs).

dirty_read(Tbl, KVs) ->
  do_read(dirty_read, Tbl, KVs).

read(Tbl, KVs) ->
  do_read(read, Tbl, KVs).

dirty_delete(Tbl, KVs) ->
  do_delete(dirty_delete, Tbl, KVs).

delete(Tbl, KVs) ->
  do_delete(delete, Tbl, KVs).

%% we suppose that Match is written as ets:fun2ms(fun (_) -> object() end)
dirty_select(Tbl, Match) ->
  do_select(dirty_select, Tbl, Match).

select(Tbl, Match) ->
  do_select(select, Tbl, Match).

%%
do_write(F=write, Tbl, KVs) ->
  mnesia:F(Tbl, make_record(Tbl, KVs), write);
do_write(F=dirty_write, Tbl, KVs) ->
  mnesia:F(Tbl, make_record(Tbl, KVs)).

do_read(ReadF, Tbl, KVs) ->
  RecName = mnesia:table_info(Tbl, record_name),
  Attrs = [KeyName|_] = mnesia:table_info(Tbl, attributes),
  Records = mnesia:ReadF({Tbl, props:gv(KeyName, KVs)}),
  [make_object(R, RecName, Attrs) || R <- Records].

do_delete(DeleteF, Tbl, KVs) ->
  [KeyName|_] = mnesia:table_info(Tbl, attributes),
  mnesia:DeleteF({Tbl, props:gv(KeyName, KVs)}).

do_select(SelectF, Tbl, Match) ->
  MakeObjectF = mk_make_object(Tbl),
  [MakeObjectF(R) || R <- mnesia:SelectF(Tbl, Match)].

%%
match_spec(Table, Conditions) ->
  Attrs = mnesia:table_info(Table, attributes),
  {SubstConds, BindedAttrs} = match_conditions(Conditions, Attrs),
  {Bindings, _} =
    lists:mapfoldl(fun (A, Cnt) ->
		       B =
			 case lists:member(A, BindedAttrs) of
			   true ->
			     match_expr_var(Cnt);
			   false ->
			     '_'
			 end,
		       {B, Cnt+1}
		   end, 1, Attrs),
  [{list_to_tuple([Table|Bindings]), SubstConds, ['$_']}].
  
match_conditions([], _) ->
  {[], []};
match_conditions(Conds, Attrs) ->
  {SubstConds, BindedAttrs} =
    lists:mapfoldl(fun (C, Acc) ->
		       {SC, BAs} = match_condition(C, Attrs),
		       {SC, BAs++Acc}
		   end, [], Conds).

match_condition(C={const,_}, _) ->
  {C, []};
match_condition(Cond, Attrs) when is_tuple(Cond),
				  is_atom(element(1, Cond)) ->
  [GF|Exprs] = tuple_to_list(Cond),
  {SubstEexprs, BindedAttrs} =
    lists:mapfoldl(fun (Expr, Acc) ->
		       {SC, BAs} = match_cond_expr(Expr, Attrs),
		       {SC, BAs++Acc}
		   end, [], Exprs),
  {list_to_tuple([GF|SubstEexprs]), BindedAttrs};
match_condition(Term, _) ->
  {Term, []}.

match_cond_expr(Attr, Attrs) when is_atom(Attr) ->
  {match_expr_var(listn:pos(Attr, Attrs)), [Attr]};
match_cond_expr(Cond, Attrs) ->
  match_condition(Cond, Attrs).

match_expr_var(N) ->
  list_to_atom("\$"++integer_to_list(N)).

mnesia_do(Fun) ->
  case mnesia:transaction(Fun) of
    {atomic, Result} ->
      Result;
    Error ->
      Error
  end.

%% @spec mnesia_start([node()]) -> AliveNodes
%% AliveNodes = [node()]
%% @doc Start Mnesia on <code>Nodes</code>. Return list of alive nodes.
%% 
mnesia_start(Nodes) ->
  Node = node(),
  RemoteNodes = Nodes -- [node()],
  mnesia:start(),
  {_, Down} = rpc:multicall(RemoteNodes, mnesia, start, []),
  Up = Nodes -- Down.
