%%%----------------------------------------------------------------------
%%% File    : passwd.erl
%%% Author  :  <svg@disney.surnet.ru>
%%% Purpose : 
%%% Created : 21 Jan 2000 by  <svg@disney.surnet.ru>
%%% $Id: passwd.erl,v 1.2 2003/10/27 16:45:59 vsekissov Exp $
%%%----------------------------------------------------------------------

-module(wslib.passwd).
-author('svg@surnet.ru').

-export([md5_crypt/1, md5_crypt/2, des_crypt/1, des_crypt/2]).
-export([crypt/2, crypt/3, from_string/1, from_string/2, to_string/1]).
-export([authenticate/3, cipher/1, cipher/2]).

-import(erlang).
-import(lists).
-import(io_lib).
-import(crypto).
-import(string).
-import(randomx).
-import(wslib.listn).

-define(MD5_MAGIC, "\$1\$").
-define(MD5_LEN, 16).
-define(ITOA64,
	"./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz").
-define(MAX_LETTER, 122).
-define(MD5_SALTLEN, 8).
-define(DES_SALTLEN, 2).

%% @spec authenticate(Method, AuthenData, SecretData) -> true | false | Error
%% Method = login | chap
%% AuthenData = term()
%% SecretData = {cleartext, string()}
%%              | {md5, string()} 
%%              | {des, string()}
%%              | undefined
%%              | nopassword
%%              | term()
%% Error = {error, ErrorDesc}
%% ErrorDesc = atom()
%% @doc authenticate AuthenData against SecretData using Method.
%% 
authenticate(_, _, undefined) ->
  false;
authenticate(_, _, {_, undefined}) ->
  false;
authenticate(_, undefined, _) ->
  false;
authenticate(_, _, nopassword) ->
  true;
authenticate(ascii, Pass, {cleartext, Pass}) ->
  true;
authenticate(ascii, Pass, {des, Cipher}) ->
  PassCipher =
    case lists:prefix("\$1\$", Cipher) of
      true ->
	md5_crypt(Pass, string:sub_word(Cipher, 2, $$));
      _ ->
	des_crypt(Pass, lists:sublist(Cipher, 2))
    end,
  (PassCipher == Cipher);
authenticate(ascii, Pass, {md5, Cipher}) ->
  PassCipher = md5_crypt(Pass, string:sub_word(Cipher, 2, $$)),
  (PassCipher == Cipher);
authenticate(chap, Data, {cleartext, Secret}) ->
  ChalLen = length(Data) - ?MD5_LEN - 1,
  if ChalLen < 0 -> 
      {error, bad_chap_challenge};
     true ->
      [Id, Chal, Dig] = listn:split_on_positions(Data, [1, ChalLen, ?MD5_LEN]),
      (binary_to_list(erlang:md5(Id++Secret++Chal)) == Dig)
  end;
authenticate(_, _, _) ->
  false.

%% @spec cipher(Which) -> Cipher::string()
%% Which = md5 | des
%% @doc interactively generate cipher from password string
%% @see cipher/2
%% @end
cipher(Which) ->
  cipher(ask, Which).

%% @spec cipher(Pass, Which) -> Cipher::string()
%% Pass = string()
%% Which = md5 | des
%% @doc generate cipher from password string, if <code>Pass == undefined</code>
%% user is asked about password string interactively
%% @see cipher/1
%% @end
cipher(Pass, cleartext) ->
  Pass;
cipher(Pass, Which) when Which == des;
			 Which == md5 ->
  Fun = case Which of
	  md5 ->
	    fun md5_crypt/1;
	  des ->
	    fun des_crypt/1
	end,
  Pass1 = case Pass of
	    ask ->
	      {ok, V} = io:fread('Password :', "~s"),
	      V;
	    _ ->
	      Pass
	  end,
  Fun(Pass1).

%% @spec from_string(Str) -> {Type, Passwd}
%% @see from_string/2
from_string(Str) ->
  from_string(Str, cleartext).

%% @spec from_string(Str, DefaultType) -> {Type, Passwd}
%% Type = atom()
%% DefaultType = atom()
%% Passwd = undefined | string()
%% @doc Used when password is retrieved from relational database
%% @see to_string/1
from_string(Str, DefaultType) ->
  DefaultTypeStr = atom_to_list(DefaultType),
  {TypeStr, Passwd} =
    case
      lists:splitwith(fun ($\s) -> false;
			  (_) -> true
		      end, Str) of

      {"", ""} ->
	{DefaultTypeStr, undefined};
      {"*", ""} ->
	{DefaultTypeStr, undefined};
      {Str, [_|"*"]} ->
	{Str, undefined};
      {Str, ""} ->
	{DefaultTypeStr, Str};
      {Str, " "} ->
	{Str, undefined};
      {Str1, [_|Str2]} ->
	{Str1, Str2}
    end,

  Type =
    if (TypeStr == "cleartext")
       or (TypeStr == "des")
       or (TypeStr == "md5") ->
	list_to_atom(TypeStr);
     true ->
      exit({badarg, Str})
  end,

  {Type, Passwd}.

%% @spec to_string({Type, Passwd}) -> Str::string()
%% Type = atom()
%% Passwd = undefined | string()
%% @doc Convert password to string
%% @see from_string/1
to_string({Type, Passwd}) ->
  Passwd1 =
    case Passwd of
      undefined ->
	"*";
      _ ->
	Passwd
    end,
  io_lib:format("~s ~s", [Type, Passwd1]).

crypt(Pass, cleartext) ->
  Pass;
crypt(Pass, Which) ->
  Salt = case Which of
	   des ->
	     random_string(?DES_SALTLEN);
	   md5 ->
	     random_string(?MD5_SALTLEN);
	   _ ->
	     exit({error, io_lib:format("undefined method ~w", Which)})
	 end,
  crypt(Pass, Salt, Which).

crypt(Pass, Salt, cleartext) ->
  Pass;
crypt(Pass, Salt, Which) ->
  Fun = case Which of
	  md5 ->
	    fun md5_crypt/2;
	  des ->
	    fun des_crypt/2;
	  _ ->
	    exit({error, io_lib:format("undefined method ~w", Which)})
	end,
  Fun(Pass, Salt).

des_crypt(Pass) ->
  des_crypt(Pass, random_string(?DES_SALTLEN)).

des_crypt(Pass, Salt) ->
  crypto:des_fcrypt(Pass, Salt).

md5_crypt(Pass) ->
  md5_crypt(Pass, random_string(?MD5_SALTLEN)).

md5_crypt(Pass, Salt) ->
  SaltBegin =
    case lists:prefix(?MD5_MAGIC, Salt) of
      true ->
	4;
      _ ->
	1
    end,
  SaltLen =
    case string:chr(string:substr(Salt, SaltBegin), $$) of
      0 ->
	8;
      Idx ->
	Idx - 1
    end,
  RealSalt = string:substr(Salt, SaltBegin, SaltLen),
  Str1 =
    binary_to_list(
       erlang:md5_final(erlang:md5_update(erlang:md5_update(erlang:md5_update(
					     erlang:md5_init(),Pass),RealSalt),Pass))),
  Ctx1 =
    erlang:md5_update(erlang:md5_update(erlang:md5_update(
			       erlang:md5_init(),Pass),?MD5_MAGIC),RealSalt),
  Ctx2 = weird1(Ctx1, Str1, length(Pass)),
  Str2 = binary_to_list(erlang:md5_final(weird2(Ctx2, Pass))),
  Str3 = weird3(Pass, Str2, RealSalt),
  ?MD5_MAGIC++RealSalt++"\$"++weird4(Str3).

weird1(Ctx, Str1, Len) when Len =< 0 ->
  Ctx;
weird1(Ctx, Str1, Len) ->
  Len1 = 
    if
      Len > 16 ->
	16;
      true ->
	Len
    end,
  weird1(erlang:md5_update(Ctx, string:substr(Str1, 1, Len1)), Str1, Len - 16).

weird2(Ctx, Pass) ->
  [P|_] = Pass,
  weird2(Ctx, P, length(Pass)).

weird2(Ctx, _, Num) when Num =< 0 ->
  Ctx;
weird2(Ctx, P, Num) when Num band 16#01 == 1 ->
  weird2(erlang:md5_update(Ctx, [16#0]), P, Num bsr 1);
weird2(Ctx, P, Num) ->
  weird2(erlang:md5_update(Ctx, [P]), P, Num bsr 1).

weird3(Pwd, Str, Salt) ->
  weird3(Pwd, Str, Salt, 0).

weird3(_, Str, _, 1000) ->
  Str;
weird3(Pwd, CtxStr, Salt, N) ->
  Str = string:substr(CtxStr, 1, 16),
  Ctx0 = erlang:md5_init(),
  Ctx1 =
    if
      N band 16#1 == 1 ->
	erlang:md5_update(Ctx0, Pwd);
      true ->
	erlang:md5_update(Ctx0, Str)
    end,
  Ctx2 =
    if
      N rem 16#3 > 0 ->
	erlang:md5_update(Ctx1, Salt);
      true ->
	Ctx1
    end,
  Ctx3 =
    if
      N rem 16#7 > 0 ->
	erlang:md5_update(Ctx2, Pwd);
      true ->
	Ctx2
    end,
  Ctx4 =
    if
      N band 16#1 == 1 ->
	erlang:md5_update(Ctx3, Str);
      true ->
	erlang:md5_update(Ctx3, Pwd)
    end,
  weird3(Pwd, binary_to_list(erlang:md5_final(Ctx4)), Salt, N+1).

weird4(Str) ->
  L1 = (lists:nth(1, Str) bsl 16) bor (lists:nth(7, Str) bsl 8) bor
    lists:nth(13, Str),
  L2 = (lists:nth(2, Str) bsl 16) bor (lists:nth(8, Str) bsl 8) bor
    lists:nth(14, Str),
  L3 = (lists:nth(3, Str) bsl 16) bor (lists:nth(9, Str) bsl 8) bor
    lists:nth(15, Str),
  L4 = (lists:nth(4, Str) bsl 16) bor (lists:nth(10, Str) bsl 8) bor
    lists:nth(16, Str),
  L5 = (lists:nth(5, Str) bsl 16) bor (lists:nth(11, Str) bsl 8) bor
    lists:nth(6, Str),
  L6 = lists:nth(12, Str),
  to64(L1, 4) ++ to64(L2, 4) ++ to64(L3, 4) ++ to64(L4, 4) ++ to64(L5, 4) ++ 
    to64(L6, 2).

to64(V, 0) ->
  [];
to64(V, N) ->
  NV = V band 16#3f,
  [lists:nth(NV+1, ?ITOA64)|to64(V bsr 6, N - 1)].

random_string(Len) ->
  randomx:alnum_string(Len).
