%%%-------------------------------------------------------------------
%%% File    : str.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : Some string manipulation functions
%%%
%%% Created : 28 Nov 2002 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id: str.erl,v 1.4 2003/10/27 16:45:59 vsekissov Exp $
%%
%%
%%**
%% 
%% @doc
%% @end
%%*

-module(wslib.str).

-export([substitute/3, join/2, capitalize/1, to_upper/1, to_lower/1, trim/2, trim/3]).
-export([fmt/2, to_str/1, eval/1, eval/2]).

-import(lists).
-import(string).
-import(io_lib).
-import(erl_scan).
-import(erl_parse).
-import(erl_eval).

%%**
%% @spec eval(Str::string()) -> Result::term()
%% @doc Evaluate Erlang expression in <code>Str</code>.
%% @end
%% 
eval(Str) ->
  eval(Str, erl_eval:new_bindings()).

%%**
%% @spec eval(Str::string(), Bindings::bindings()) -> Result::term()
%% @doc Evaluate Erlang expression in <code>Str</code> with
%% <code>Bindings</code>.
%% @see erl_eval
%% @end
%% 

eval(Str, Bindings) ->
  {ok, Tokens, _} = erl_scan:string(Str),
  {ok, [Form]} = erl_parse:parse_exprs(Tokens),
  {value, Ret, _} = erl_eval:expr(Form, Bindings),
  Ret.

%%**
%% @spec substitute(Str::string(), From::[char()], To) -> string()
%% To = string() | char()
%%
%% @doc Substitute sequence of chars <code>From</code> to
%% <code>To</code>
%% @end
%% 
substitute(Str, From, To) ->
  From1 = listn:mk_list(From),
  To1 = listn:mk_list(To),
  lists:append(substitute(Str, From1, length(From1), To1, [])).

substitute([], From, FromLen, To, Acc) ->
  lists:reverse(Acc);
substitute(Str, From, FromLen, To, Acc) ->
  case string:str(Str, From) of
    0 ->
      lists:reverse([Str|Acc]);
    Pos ->
      Before = string:substr(Str, 1, Pos-1),
      After = string:substr(Str, Pos+FromLen),
      substitute(After, From, FromLen, To, [To,Before|Acc])
  end.

%%**
%% @spec join(Strings::[string()], Sep::string() ) -> string()
%%
%% @doc Join <code>Strings</code> with <code>Sep</code> separator
%% @end
%%*
join([], _) ->
    [];
join(Strings, Sep) when is_integer(Sep) ->
  join(Strings, [Sep]);
join([S|Rest], Sep) ->
  lists:append([S|[Sep++Str || Str <- Rest]]).

capitalize([]) ->
  [];
capitalize([C|Rest]) ->
  [char:to_upper(C)|Rest].

%% to_upper
to_upper(Str) when is_list(Str) ->
  [char:to_upper(C) || C <- Str].

%% to_lower
to_lower(Str) when is_list(Str) ->
  [char:to_lower(C) || C <- Str].

trim([], _) -> [];
trim(Str, Chrs) ->
  trim(trim(Str, Chrs, left), Chrs, right).

trim([], _, _) -> [];
trim(Str, Chrs, left) ->
  case string:span(Str, Chrs) of
    0 ->
      Str;
    N ->
      string:substr(Str, N+1)
  end;
trim(Str, Chrs, right) ->
  lists:reverse(trim(lists:reverse(Str), Chrs, left)).

fmt(Format, Args) ->
  lists:flatten(io_lib:format(Format, Args)).

to_str(I) when is_integer(I) ->
  integer_to_list(I);
to_str(F) when is_float(F) ->
  float_to_list(F);
to_str(A) when is_atom(A) ->
  atom_to_list(A);
to_str(T) ->
  case io_lib:char_list(T) of
    true ->
      T;
    false ->
      fmt("~p", [T])
  end.

max(A, B) when A > B -> A;
max(_, B) -> B.
