%%%-------------------------------------------------------------------
%%% File    : wsocket_listen.erl
%%% Author  :  <svg@disney.surnet.local>
%%% Description : TCP/IP/UDP server, accepts connections and spawns protocol
%%%               and transport modules
%%%
%%% Created : 12 Jun 2002 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% 
%%*
-module(wsocket_listen).

-behaviour(gen_server).
%%--------------------------------------------------------------------
%% Include files
%%--------------------------------------------------------------------

%%--------------------------------------------------------------------
%% External exports
-export([start_link/5, start_link/6, stop/1, info/2, debug/3]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2, code_change/3]).

%% Internal exports
-export([new_connect/2]).

-import(wslib.props).
-import(wslib.cal).
-import(wslib.log).

-record(state, {listen_pid,              % Listener pid
                acceptor,                %Current acceptor
                active=[],               %Active connections
                max_conn=infinity,       %Max connections number
                transp_mod,              %Transport module
                packet_fun,              %Packet fun
                proto_fun,               %Protocol fun
		mediator_opts=[],
		debug=false
               }).

-ifndef(SOC_LISTEN_TIMEOUT).
-define(SOC_LISTEN_TIMEOUT, 30000).
-endif.

-define(MAX_CONNECTIONS, 256).

-define(server(P),
        list_to_atom(lists:flatten(io_lib:format(?MODULE_STRING "_~w",[P])))).

-define(debug_levels, [transport, protocol, packet]).
-define(is_debug_level(_V),
	(_V == transport) or ( _V == protocol) or (_V == packet)).
%%====================================================================
%% External functions
%%====================================================================
%%--------------------------------------------------------------------
%% Function: start_link/0
%% Description: Starts the server
%%--------------------------------------------------------------------
start_link(Port, TrMod, PctFun, PrtFun, Args) ->
  start_link(?server(Port), Port, TrMod, PctFun, PrtFun, Args).

start_link(Name, Port, TrMod, PctFun, PrtFun, Args) ->
  gen_server:start_link({local, Name},
                        ?MODULE, {Port, TrMod, PctFun, PrtFun, Args},
                        []).

stop(Port) when is_integer(Port) ->
  stop(?server(Port));
stop(Srv) ->
  gen_server:cast(Srv, stop).

info(Port, Query) when is_integer(Port) ->
  info(?server(Port), Query);
info(Server, Query) ->
  gen_server:call(Server, {info, Query}).

debug(Port, Levels, Bool) when is_integer(Port) ->
  debug(?server(Port), Levels, Bool);
debug(Server, all, Bool) ->
  debug(Server, ?debug_levels, Bool);
debug(Server, Levels, Bool) when Bool == true;
				 Bool == false ->
  Cmd = if Bool -> start_debug;
	   true -> stop_debug
	end,  
  gen_server:cast(Server, {Cmd, Levels}).

%%====================================================================
%% Server functions
%%====================================================================

%%--------------------------------------------------------------------
%% Function: init/1
%% Description: Initiates the server
%% Returns: {ok, State}          |
%%          {ok, State, Timeout} |
%%          ignore               |
%%          {stop, Reason}
%%--------------------------------------------------------------------
init({Port, TrMod, PctFun, PrtFun, Args}) ->
  [Max, AclFun, TrOpts, DebugOpts] =
    props:gv_map([max, acl, transp_opts, debug], check_props(Args)),
  MediatorOpts = [{acl, AclFun}, {debug, DebugOpts}],
  Debug = props:gv(transport, DebugOpts, false),
  process_flag(trap_exit, true),
  case TrMod:listen(Port, Max, TrOpts) of
    {ok, LrPid} ->
      Pid = start_accept(LrPid, TrMod, PctFun, PrtFun, MediatorOpts, Debug),
      State = #state{listen_pid = LrPid,
		     acceptor = Pid,
		     transp_mod = TrMod,
		     proto_fun = PrtFun,
		     packet_fun = PctFun,
		     mediator_opts = MediatorOpts,
		     max_conn = Max,
		     debug=Debug
		    },
      dbg(Debug, "starting at port ~p",[TrMod:sockname(LrPid)]),
      {ok, State};
    {error, Reason} ->
      {stop, Reason};
    Other ->
      {stop, Other}
  end.

%%--------------------------------------------------------------------
%% Function: handle_call/3
%% Description: Handling call messages
%% Returns: {reply, Reply, State}          |
%%          {reply, Reply, State, Timeout} |
%%          {noreply, State}               |
%%          {noreply, State, Timeout}      |
%%          {stop, Reason, Reply, State}   | (terminate/2 is called)
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_call({info, state}, From,
            S=#state{listen_pid=LrPid,
                     acceptor=Acceptor,
                     active=Active,
                     max_conn=Max,
                     transp_mod=TrMod,
		     debug=Debug
                    }) ->
  Reply = [{listen_pid, LrPid},
           {acceptor, Acceptor},
           {transp_mod, TrMod},
           {active, length(Active)},
           {max_conn, Max},
	   {debug, Debug}
          ],
  {reply, Reply, S};
handle_call({info, active_pids}, From, S=#state{active=A}) ->
  {reply, A, S};
handle_call(stop, From, State) ->
  {stop, normal, ok, State}.

%%--------------------------------------------------------------------
%% Function: handle_cast/2
%% Description: Handling cast messages
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_cast({start_debug, Level}, State) when is_atom(Level) ->
  {noreply, set_debug(true, [Level], State)};
handle_cast({start_debug, Levels}, State) ->
  {noreply, set_debug(true, Levels, State)};
handle_cast({stop_debug, Level}, State) when is_atom(Level) ->
  {noreply, set_debug(false, [Level], State)};
handle_cast({stop_debug, Levels}, State) ->
  {noreply, set_debug(false, Levels, State)};
handle_cast(stop, State) ->
  {stop, normal, State};
handle_cast({connected, Current},
            S=#state{acceptor=Current, active=As, max_conn=Max,
                     listen_pid=LrPid,
		     transp_mod=TrMod,
		     packet_fun=PcF,
		     proto_fun=PrF,
		     mediator_opts=MOs,
		     debug=D}) ->
  AsLen = length(As),
  dbg(D, "new connection accepted by ~p,connections ~w",[Current, AsLen+1]),
  New = another_accept(LrPid, TrMod, PcF, PrF, MOs, AsLen, Max, false, D),
  {noreply, S#state{acceptor=New, active=[Current|As]}};
handle_cast(_, State) ->
  {noreply, State}.

%%--------------------------------------------------------------------
%% Function: handle_info/2
%% Description: Handling all non call/cast messages
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_info({'EXIT', LrPid, Reason}, S=#state{listen_pid=LrPid, debug=D}) ->
  dbg(D, "listener ~p exited", [LrPid]),
  {stop, Reason, S#state{listen_pid=none}};
handle_info({'EXIT', Current, Reason},
            S=#state{listen_pid=LrPid,
		     acceptor=Current,
                     active=As,
		     max_conn=Max,
                     transp_mod=TM,
		     proto_fun=PrF,
		     packet_fun=PcF,
		     mediator_opts=MOs,
		     debug=D
		    }) ->
  AsLen = length(As),
  dbg(D,"acceptor ~p exited, reason ~p, connections ~w",[Current, Reason,AsLen]),
  New = another_accept(LrPid, TM, PcF, PrF, MOs, AsLen, Max, false, D),
  {noreply, S#state{acceptor=New}};
handle_info({'EXIT', Pid, Reason},
            S=#state{listen_pid=LrPid,
		     acceptor=A,
		     active=As,
		     max_conn=Max,
                     transp_mod=TM,
		     proto_fun=PrF,
		     packet_fun=PcF,
		     mediator_opts=MOs,
		     debug=D
		    })  ->
  AsLen = length(As),
  As1 = lists:delete(Pid, As),
  AsLen1 = length(As1),

  if AsLen == AsLen1 ->
      {noreply, S};
     true ->
      dbg(D,"session ~p exited, reason ~p, connections ~w",[Pid, Reason, AsLen1]),
      New = another_accept(LrPid, TM, PcF, PrF, MOs, AsLen1, Max, A, D),
      {noreply, S#state{acceptor=New, active=As1}}
  end;
handle_info(Info, S=#state{listen_pid=LrPid, transp_mod=TM, debug=D}) ->
  Port = case TM:sockname(LrPid) of
           {ok, {_, P}} ->
             P;
           _Error ->
             unknown
         end,
  
  dbg(D, "unexpected info ~p",[Info]),
  {noreply, S};
handle_info(Info, S) ->
  {noreply, S}.

%%--------------------------------------------------------------------
%% Function: terminate/2
%% Description: Shutdown the server
%% Returns: any (ignored by gen_server)
%%--------------------------------------------------------------------
terminate(Reason, S=#state{listen_pid=LrPid, transp_mod=TrMod,
			   acceptor=Current, active=Active, debug=D}) ->
  case LrPid of
    _ when is_pid(LrPid) ->
      TrMod:close(LrPid);
    _ ->
      ok
  end,
  Children = 
    case Current of
      _ when is_pid(Current) ->
	[Current|Active];
      _ ->
	Active
    end,
  lists:foreach(fun (P) -> exit(P, stop) end, Children),
  dbg(D,"terminating", []),
  ok.

%%--------------------------------------------------------------------
%% Func: code_change/3
%% Purpose: Convert process state when code is changed
%% Returns: {ok, NewState}
%%--------------------------------------------------------------------
code_change(OldVsn, State, Extra) ->
  {ok, State}.

%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------

start_accept(LrPid, TrMod, PctFun, PrtFun, Props, Dbg) ->
  Pid = wsocket_mdr:spawn_server(LrPid, TrMod, PctFun, PrtFun, Props),
  dbg(Dbg,"starting new acceptor ~p", [Pid]),
  Pid.

%% return new acceptor
another_accept(LrPid, TrMod, PctFun, PrtFun, Props, Active, Max, false, Dbg) ->
  if Active < Max -> % lower than limit
      start_accept(LrPid, TrMod, PctFun, PrtFun, Props, Dbg);
     true ->
      dbg(Dbg,"don't start new acceptor, max connections number <~w> exided",[Max]),
      false
  end;
another_accept(_LrPid, _TM, _PcF, _PrF, _Ps, _As, _Max, Acceptor, _) ->
  Acceptor.

new_connect(Server, New) ->
  gen_server:cast(Server, {connected, New}).

set_debug(true, Levels, S=#state{mediator_opts=MOs}) ->
  case catch [check_debug_prop(P) || P <- Levels] of
    {'EXIT', _} ->
      S;
    Ls->
      DbgOpts = props:gv(debug, MOs, []),
      NxtDbgOpts = Ls ++ DbgOpts -- Ls,
      NxtDbg = lists:member(transport, NxtDbgOpts),
      S#state{mediator_opts=[{debug, NxtDbgOpts}|props:delete(debug, MOs)],
	      debug=NxtDbg
	     }
  end;
set_debug(false, Levels, S=#state{mediator_opts=MOs}) ->
  case catch [check_debug_prop(P) || P <- Levels] of
    {'EXIT', _} ->
      S;
    Ls->
      DbgOpts = props:gv(debug, MOs, []),
      NxtDbgOpts = DbgOpts -- Ls,
      NxtDbg = lists:member(transport, NxtDbgOpts),
      S#state{mediator_opts=[{debug, NxtDbgOpts}|props:delete(debug, MOs)],
	      debug=NxtDbg
	     }
  end.
    
check_props(Props) ->
  {ok, Props1} = props:check_required(
		   [{acl, fun allow_all/2},
		    {max, ?MAX_CONNECTIONS},
		    {transp_opts, []},
		    {debug, []}
		   ], Props),
  [check_prop(P) || P <- Props1].

check_prop(P={max, infinity}) -> P;
check_prop(P={max, N}) when is_integer(N), N > 0 -> P;
check_prop(P={acl, _}) -> P;
check_prop(P={transp_opts, _}) -> P;
check_prop(P={debug, Opts}) when is_list(Opts) ->
  {debug, [check_debug_prop(V) || V <- Opts]};
check_prop(P) ->
  exit({badopt, P}).

check_debug_prop(V) when ?is_debug_level(V) ->
  V;
check_debug_prop(V) ->
  exit({badopt, {debug, V}}).

allow_all(RemotePeer, LocalPeer) ->
  permit.

dbg(false, _, _) ->
  ok;
dbg(true, Fmt, Args) ->
  Str = cal:format_datetime(localtime) ++ " " ++ log:fmt(Fmt, Args),
  error_logger:info_msg(Str).
