%%%----------------------------------------------------------------------
%%% File    : wsocket_mdr.erl
%%% Author  :  <svg@disney.surnet.local>
%%% Purpose : Socket mediator between transport and protocol levels
%%% Created :  1 Jul 2002 by  <svg@disney.surnet.local>
%%%----------------------------------------------------------------------

%%
%% 
%%*
-module(wsocket_mdr).
-author('svg@surnet.ru').

-export([spawn_server/5, spawn_client/5]).

-import(wslib.props).
-import(wslib.cal).
-import(wslib.log).

-include("plain_fsm.hrl").

%%% Transport module interface
%%
%% listen(Port, Max, TrOpts) -> {ok, ListenPid} | Error
%% accept(ListenPid) -> {ok, Socket} | Error
%% recv(init, Socket, Timeout) -> {ok, InitState} | Error
%% recv(next, Socket, State) -> {ok, Bin, NextState} | Error
%% recv(cancel, Socket, State) -> ok | Error
%% send(Socket, Bin) -> ok | Error
%% close(Socket) -> ok | Error
%% sockname(Socket) -> {ok, {IP, Port}} | Error
%% peername(Socket) -> {ok, {IP, Port}} | Error
%% Error {error, Error}
%%
%%%

%% <code>Packet</code> and <code>Protocol</code> processes receive messages
%% in the form <code>{MediatorPid, Msg}</code> and return answers in the form
%% {SelfPid, Msg}.
%% 
%% <h3>Packet Interface</h3>
%% 
%% PacketFun(Opts) -> PacketPid
%% Opts = [{debug, Bool}]
%% 
%% Messages.
%% 
%% Meduator Cast
%%
%% {debug, Bool}
%%
%% Request                Meduator Answer
%%
%% sockname | peername ->
%%                     <- {Request, Peer} | exit()
%% 
%% Meduator Request             Packet Answer
%% 
%% {parse, ProtocolArg, Bin} ->
%%                           <- more % get more data
%%                           <- {done, Packet, RestBin}
%% {received, MoreBin}       ->
%%                           <- more % get more data
%%                           <- {done, Packet, RestBin}
%% {pack, Packet}            ->
%%                           <- {done, Bin}
%% {error, Reason}           ->
%%                           <- {done, RestBin}
%%
%% <h3>Protocol Interface</h3>
%%
%% ProtocolFun(Opts) -> ProtocolPid
%% Opts = [{debug, Bool}]
%% 
%% Messages.
%% 
%% Protocol Request                            Meduator Answer
%%
%% {recv, ProtocolArg, Timeout}             ->
%%                                          <- {received, Packet}
%%                                          <- Error={error, Reason}
%% {send, Packet}                           ->
%%                                          <- sended
%% {continue, Packet, ProtocolArg, Timeout} ->
%%                                          <- {received, Packet}
%% {done, Packet}                           ->
%% -                                        <-
%% done                                     ->
%% -                                        <-
%% 
%% If there is an error on packet/transport layer protocol process
%% receives exit(Error)
%% Error = {badproto, Reason} | {badsock, Reason}
%% spawn_server(LrPid, TrMod, PctFun, ProtoFun, Opts) -> Pid
%%
%%        | {deny, {RemPeer, LocalPeer}}
%%        | {badproto,Error}
%%        | {badpacket, Error}
%%        | {badsock, Error}
%% Error = term()
%% Parent = socket_server_pid()
%% LrPid listener_pid()
%% TrMod = transport_module()
%% PctFun = packet_fun()
%% ProtoFun protocol_fun()
%% Opts = [{acl, false}, {debug, false}]
%% if debug = true debug messages are logged with error_logger
%% 

spawn_server(LrPid, TrMod, PctFun, ProtoFun, Opts) ->
  [AclF, DbgOpts] = check_opts(Opts),
  Parent = self(),
  Dbg = props:gv(transport, DbgOpts, false),
  plain_fsm:spawn_link(
    ?MODULE,
    fun() ->
	Self = self(),
	case TrMod:accept(LrPid) of
	  {error, closed} ->
	    exit(normal);
	  {error, Error} ->
	    dbg(Dbg, "accept error ~p", [Error]),
	    exit(normal);
	  {ok, Socket} ->
	    ask_acl(Socket, TrMod, AclF, Dbg),
	    PctPid = PctFun(),
	    ProtoPid = ProtoFun(),
	    PctPid ! {Self, {debug, props:gv(packet, DbgOpts, false)}},
	    ProtoPid ! {Self, {debug, props:gv(protocol, DbgOpts, false)}},
	    wsocket_listen:new_connect(Parent, Self),
	    socket_loop(Parent, Socket, TrMod, PctPid, ProtoPid, Dbg)
	end
    end).

spawn_client(ConnectParams, TrMod, PctFun, ProtoFun, Opts) ->
  [_, DbgOpts] = check_opts(Opts),
  Dbg = props:gv(transport, DbgOpts, false),
  Parent = self(),
  plain_fsm:spawn_link(
    ?MODULE,
    fun () ->
	Self = self(),
	case apply(TrMod, connect, ConnectParams) of
	  {ok, Socket} ->
	    PctPid = PctFun(),
	    ProtoPid = ProtoFun(),
	    PctPid ! {Self, {debug, props:gv(packet, DbgOpts, false)}},
	    ProtoPid ! {Self, {debug, props:gv(protocol, DbgOpts, false)}},
	    socket_loop(Parent, Socket, TrMod, PctPid, ProtoPid, Dbg);
	  {error, Error} ->
	    dbg(Dbg, "connect failed", []),
	    exit(Error)
	end
    end).

ask_acl(Socket, TrMod, AclFun, Dbg) ->
  {ok, LocalPeer} = TrMod:sockname(Socket),
  {ok, RemPeer} = TrMod:peername(Socket),
  case AclFun(RemPeer, LocalPeer) of
    permit ->
      ok;
    _ ->
      dbg(Dbg,"access denied by acl from ~p to ~p", [RemPeer, LocalPeer]),
      TrMod:close(Socket),
      exit(normal)
  end.

socket_loop(Parent, Socket, TrMod, PctPid, ProtoPid, Dbg) ->
  process_flag(trap_exit, true),
  socket_loop({Parent, Socket, TrMod, PctPid, ProtoPid, <<>>, Dbg}).

socket_loop(S={Parent, Socket, TrMod, PctPid, ProtoPid, AccBin, Dbg}) ->
  Self = self(),
  plain_fsm:extended_receive(
    receive
      {Pid, TrCmd} when TrCmd == sockname;
			TrCmd == peername ->
	{ok, N} = TrMod:TrCmd(Socket),
	Pid ! {Self, {TrCmd, N}},
	socket_loop(S);
      {ProtoPid, Request} ->
	ReqResult =
	  case Request of
	    {continue, Packet, RecvParam, Timeout} ->
	      case send_packet(Socket, TrMod, PctPid, Packet, Dbg) of
		{ok, sended} ->
		  recv_packet(Socket,TrMod,PctPid,RecvParam,AccBin,Timeout,Dbg);
		PctError ->
		  PctError
	      end;
	    {recv, RecvParam, Timeout} ->
	      recv_packet(Socket,TrMod,PctPid,RecvParam,AccBin,Timeout,Dbg);
	    {send, Packet} ->
	      send_packet(Socket, TrMod, PctPid, Packet, Dbg);
	    {done, Packet} ->
	      send_packet(Socket, TrMod, PctPid, Packet, Dbg),
	      terminate(S);
	    done ->
	      terminate(S)
	  end,
	{ProtoReturn, NxtAccBin} =
	  case ReqResult of
	    {ok, Data, RestBin} ->
	      {Data, RestBin};
	    {ok, Data} ->
	      {Data, AccBin};
	    Error={error, Reason} ->
	      dbg(Dbg,"packet or transport error ~p, terminating", [Reason]),
	      exit(ProtoPid, Reason),
	      exit(PctPid, shutdown),
	      exit(normal)
	  end,
	ProtoPid ! {Self, ProtoReturn},
	socket_loop({Parent, Socket, TrMod, PctPid, ProtoPid, NxtAccBin, Dbg});
      {'EXIT', ProtoPid, Reason} ->
	dbg(Dbg,"protocol exited with reason ~p, terminating", [Reason]),
	exit(PctPid, shutdown),
	TrMod:close(Socket),
	exit(normal);
      {'EXIT', Parent, _Reason} ->
	terminate(S)
    end
   ).

terminate({_, Socket, TrMod, PctPid, ProtoPid, _, _}) ->
  TrMod:close(Socket),
  exit(PctPid, shutdown),
  exit(ProtoPid, shutdown),
  exit(normal).

recv_packet(Socket, TrMod, PctPid, RecvParam, AccBin, Timeout, Dbg) ->
  Self = self(),
  TrState = TrMod:recv(init, Socket, Timeout),
  PctPid ! {Self, {parse, RecvParam, AccBin}},
  recv_packet_loop(Socket, TrMod, PctPid, TrState, Dbg).

recv_packet_loop(Socket, TrMod, PctPid, TrState, Dbg) ->
  Self = self(),
  receive
    {Pid, TrCmd} when TrCmd == sockname;
		      TrCmd == peername ->
      Pid ! {Self, TrMod:TrCmd(Socket)},
      recv_packet_loop(Socket, TrMod, PctPid, TrState, Dbg);
    {PctPid, Answer} ->
      case Answer of
	more ->
	  case TrMod:recv(next, Socket, TrState) of
	    {ok, Bin, NxtState} ->
	      PctPid ! {Self, {received, Bin}},
	      recv_packet_loop(Socket, TrMod, PctPid, NxtState, Dbg);
	    Error={error, timeout} ->
	      PctPid ! {Self, Error},
	      receive
		{PctPid, {done, RestBin}} when is_binary(RestBin) ->
		  {ok, Error, RestBin};
		{'EXIT', PctPid, Reason} ->
		  {error, {badpacket, Reason}}
	      end;
	    {error, Reason} ->
	      {error, {badsock, Reason}}
	  end;
        {done, Data, RestBin} when is_binary(RestBin) ->
          TrMod:recv(cancel, Socket, TrState),
          {ok, {received, Data}, RestBin};
        {error, Reason} ->
          TrMod:close(Socket),
          {error, {badpacket, Reason}};
	Other ->
	  dbg(Dbg,"unexpected packet return ~p", [Other]),
	  TrMod:close(Socket),
	  exit({badpacket_answer, Other})
      end;
    {'EXIT', PctPid, Reason} ->
      {error, {badpacket, Reason}}
  end.

send_packet(Socket, TrMod, PctPid, Packet, Dbg) ->
  Self = self(),
  PctPid ! {Self, {pack, Packet}},
  receive
    {Pid, TrCmd} when TrCmd == sockname;
		      TrCmd == peername ->
      Pid ! {Self, TrMod:TrCmd(Socket)},
      send_packet(Socket, TrMod, PctPid, Packet, Dbg);
    {PctPid, Answer} ->
      case Answer of
	{done, Bin} ->
	  case TrMod:send(Socket, Bin) of
	    ok ->
	      {ok, sended};
	    {error, Reason} ->
	      {error, {badsock, Reason}}
	  end;
	{error, Reason} ->
	  {error, {badpacket, Reason}};
	Other ->
	  dbg(Dbg,"unexpected packet return ~p", [Other]),
	  TrMod:close(Socket),
	  exit({badpacket_answer, Other})
      end;
    {'EXIT', PctPid, Reason} ->
      {error, {badpacket, Reason}}
  end.

check_opts(Opts) ->
  props:gv_map([{acl, fun default_acl/2},
		{debug, []}
	       ],
	       Opts).

dbg(false, _, _) ->
  ok;
dbg(true, Fmt, Args) ->
  Str = cal:format_datetime(localtime) ++ " " ++ log:fmt(Fmt, Args),
  error_logger:info_msg(Str).

default_acl(_, _) ->
  permit.
