%%%-------------------------------------------------------------------
%%% File    : wsocket_tcp_transport.erl
%%% Author  :  <svg@disney.surnet.local>
%%% Description : 
%%%
%%% Created : 14 Jun 2002 by  <svg@disney.surnet.local>
%%%-------------------------------------------------------------------
%%
%% 
%%*
-module(wsocket_tcp_transport).

-export([listen/3, connect/3, connect/4, accept/1, close/1]).
-export([send/2, recv/3, sockname/1, peername/1]).

-record(state, {socket, master}).

listen(Port, Max, Opts) ->
  DefOpts = [binary, {active, false}],
  BackLog = if is_integer(Max) ->
		[{backlog, Max}];
	       true ->
		[]
	    end,
  Params = BackLog ++ DefOpts ++ Opts,
  case  gen_tcp:listen(Port, Params) of
    {ok, Socket} ->
      Self = self(),
      LrPid = proc_lib:spawn_link(fun () -> start_listener(Self, Socket) end),
      gen_tcp:controlling_process(Socket, LrPid),
      LrPid ! {Self, start},
      {ok, LrPid};
    Error ->
      Error
  end.

connect(Addr, Port, Opts) ->
  connect(Addr, Port, Opts, infinity).

connect(Addr, Port, Opts, Timeout) ->
  {ok, IP} = inet:getaddr(Addr, inet),
  gen_tcp:connect(IP, Port, [binary, {packet, 0}, {active, false}|Opts], Timeout).

accept(LrPid) when is_pid(LrPid) ->
  gen_tcp:accept(get_socket(LrPid)).

close(Pid) when is_pid(Pid) ->
  exit(Pid, normal);
close(S) ->
  gen_tcp:close(S).

send(S, Bin) ->
  case gen_tcp:send(S, Bin) of
    ok ->
      ok;
    Error ->
      close(S),
      Error
  end.

sockname(Pid) when is_pid(Pid)->
  sockname(get_socket(Pid));
sockname(S) ->
  inet:sockname(S).

peername(S) ->
  inet:peername(S).

get_socket(LrPid) ->
  Self = self(),
  LrPid ! {Self, get_socket},
  receive
    {LrPid, {socket, Socket}} ->
      Socket
  end.

recv(init, _S, Timeout) when is_integer(Timeout);
                            Timeout == infinity ->
  recv(init, _S, {0, Timeout});

recv(init, _S, RS={Size, Timeout}) when is_integer(Size);
                                       Size >= 0 ->
  RS;
recv(next, S, {Size, Timeout}) ->
  do_recv(S, Size, Timeout);
recv(cancel, _S, _RS) ->
  ok.

%% Internal funs

do_recv(S, Size, Timeout) ->
  NxtState = {Size, Timeout},
  case gen_tcp:recv(S, Size, Timeout) of
    {ok, Data} ->
      {ok, Data, NxtState};
    Error={error, timeout} ->
      Error;
    Error={error, Reason} ->
      gen_tcp:close(S),
      Error;
    Other ->
      gen_tcp:close(S),
      {error, Other}
  end.

start_listener(Serv, Socket) ->
  receive
    {Serv, start} ->
      process_flag(trap_exit, true),
      listener_loop(#state{socket=Socket, master=Serv})
  end.

listener_loop(S=#state{master=Mr, socket=Sock}) ->
  Self = self(),
  receive
    {Pid, get_socket} ->
      Pid ! {Self, {socket, Sock}},
      listener_loop(S);
    {'EXIT', Self, Reason} ->
      close(Sock),
      exit(Reason);
    {'EXIT', Mr, Reason} ->
      close(Sock),
      exit(Reason);
    Other ->
      {ok, Peer} = inet:sockname(Sock),
      error_logger:format("~w:~p unexpected message: ~p~n",
			  [?MODULE, Peer, Other]),
      listener_loop(S)
  end.
