%%%-------------------------------------------------------------------
%%% File    : wsocket_udp_transport.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : 
%%%
%%% Created : 14 Jun 2002 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------

%%
%% 
%%*
-module(wsocket_udp_transport).

-export([listen/3, connect/3, connect/4, info/2, accept/1, close/1]).
-export([send/2, recv/3, sockname/1, peername/1]).

-define(COMMON_TIMEOUT, 5000).

-record(state,{master,            % listen process
               acceptor=false,    % current acceptor
               active=[],         % active connections
               socket,            % socket
               inq = queue:new(), % input queue
               inq_size = 0,      % input queue size
               inq_max}).         % input queue max size
-record(wsocket_udp, {listen_pid, peer}).

listen(Port, Max, Opts) ->
  DefOpts = [binary, {active, true}],
  Params = DefOpts ++ Opts,
  case  gen_udp:open(Port, Params) of
    {ok, Socket} ->
      Self = self(),
      LrPid = proc_lib:spawn_link(fun () -> start_listener(Self, Socket, Max) end),
      gen_udp:controlling_process(Socket, LrPid),
      LrPid ! {Self, start},
      {ok, LrPid};
    Error ->
      Error
  end.

%% Start listener on any port, register ourselves as acceptor
%% and send dummy package to listener to emulate connection
%% so we can use the same interface as client
%% connect/4 is only for compatibility with common interface
%% connection timeout is never used
connect(Addr, Port, Opts) ->
  connect(Addr, Port, Opts, none).

connect(Addr, Port, Opts, _Timeout) ->
  {ok, LrPid} = listen(0, 1, Opts),
  {ok, IP} = inet:getaddr(Addr, inet),
  LrPid ! {udp, false, IP, Port, connect},
  {ok, Socket} = accept(LrPid),
  receive
    {LrPid, {recv, connect}} ->
      {ok, Socket}
  end.

accept(LrPid) ->
  Self = self(),
  LrPid ! {Self, accept},
  receive
    {LrPid, {connected, Socket}} ->
      {ok, Socket};
    {LrPid, Error={error, _}} ->
      Error
  end.

close(Pid) when is_pid(Pid) ->
  exit(Pid, normal);  
close(S=#wsocket_udp{listen_pid=LrPid, peer=Peer}) ->
  Self = self(),
  LrPid ! {Self, {close, Peer}},
  receive
    {LrPid, closed} ->
      ok;
    {LrPid,  Error={error, _}} ->
      Error
  end.

send(S=#wsocket_udp{listen_pid=LrPid, peer=Peer}, Bin) ->
  LrPid ! {self(), {send, Peer, Bin}},
  receive
    {LrPid, sended} ->
      ok;
    {LrPid, Error={error, _}} ->
      Error
  end.

recv(init, _S, Timeout) when is_integer(Timeout);
			     Timeout == infinity ->
  recv(init, _S, {0, Timeout});

recv(init, _S, RS={Size, Timeout}) when is_integer(Size),
					Size >= 0 ->
  RS;
recv(next, S, State) ->
  do_recv(S, State, <<>>);
recv(cancel, _S, _State) ->
  ok.

info(S=#wsocket_udp{listen_pid=LrPid}, Query) ->
  info(LrPid, Query);
info(LrPid, Query) when is_pid(LrPid) ->
  LrPid ! {self(), {info, Query}},
  receive
    {LrPid, {info, Answer}} ->
      {ok, Answer}
  after ?COMMON_TIMEOUT ->
      {error, timeout}
  end.

start_listener(Serv, Socket, Max) ->
  receive
    {Serv, start} ->
      process_flag(trap_exit, true),
      listener_loop(#state{socket=Socket, master=Serv, inq_max=Max})
  end.

listener_loop(S=#state{master=Mr,
                       acceptor=Current, active=Active,
                       socket=Sock}) ->
  Self = self(),
  receive
    {From, sockname} ->
      From ! {Self, {sockname, inet:sockname(Sock)}},
      listener_loop(S);
    {Pid, accept} ->
      set_acceptor(Pid, S);
    {From, {close, Peer}} ->
      close_socket(From, Peer, S);
    {From, {send, Peer, Data}} ->
      send_to_socket(From, Peer, Data, S);
    {From, {recv, Peer}} ->
      recv_from_socket(From, Peer, S);
    {From, {info, state}} ->
      From ! {Self, {info, S}},
      listener_loop(S);
    {'EXIT', Mr, Reason} ->
      exit(Reason);
    {'EXIT', Self, Reason} ->
      exit(Reason);
    {'EXIT', Current, Reason} ->
      listener_loop(S#state{acceptor=false});
    {'EXIT', Pid, Reason} ->
      listener_loop(S#state{active=active_delete(Pid, Active)});
    {udp, _S, IP, Port, Data} ->
      dispatch_packet({IP, Port}, Data, S);
    Other ->
      {ok, Peer} = inet:sockname(Sock),
      error_logger:format("~w:~p unexpected message: ~p~n",
			  [?MODULE, Peer, Other]),
      listener_loop(S)
  end.

set_acceptor(Pid, S=#state{socket=Sock, acceptor=false}) ->
%  inet:setopts(Sock, [{active, once}]),
  link(Pid),
  new_connection(S#state{acceptor=Pid});
set_acceptor(Pid, S=#state{acceptor=Current}) when is_pid(Current) ->
  Pid ! {self(), {error, already_exist}},
  listener_loop(S).

dispatch_packet(Peer, Data, S=#state{active=Active}) ->
  case active_lookup(Peer, Active) of
    undefined ->
      new_connection(Peer, Data, S);
    {Pid, _} ->
      Pid ! {self(), {recv, Data}},
      listener_loop(S)
  end.

new_connection(S=#state{acceptor=Pid, inq=Q, inq_size=QSz}) when is_pid(Pid) ->
  case queue:out(Q) of
    {empty, NQ} -> %input queue is empty
      listener_loop(S);
    {{value, {Peer, Data}}, NQ} ->
      new_connection(Peer, Data, S#state{inq=NQ, inq_size=QSz - 1})
  end.

new_connection(Peer, Data,
	       S=#state{acceptor=false, inq=Q, inq_size=QSz, inq_max=Max}) ->
  if QSz >= Max ->
      listener_loop(S); % no active acceptors, queue is full - discard
     true -> % put in queue
      listener_loop(S#state{inq=queue:in({Peer, Data}, Q), inq_size=QSz+1})
  end;
new_connection(Peer, Data,
               S=#state{master=Mr, acceptor=Current, active=Active}) ->
  Self = self(),
  NewSocket = #wsocket_udp{listen_pid=Self, peer=Peer},
  Current ! {Self, {connected, NewSocket}},
  Current ! {Self, {recv, Data}},
  listener_loop(S#state{acceptor=false,
                        active=active_add({Current, Peer}, Active)}).

send_to_socket(Pid, Peer={IP, Port}, Data,
               S=#state{socket=Sock}) ->
  Self = self(),
  case gen_udp:send(Sock, IP, Port, Data) of
    ok ->
      Pid ! {Self, sended};
    {error, Reason} ->
      Pid ! {Self, {error, Reason}}
  end,
  listener_loop(S).

recv_from_socket(From, Peer,
                 S=#state{socket=Sock}) ->
%  inet:setopts(Sock, [{active, once}]),
  listener_loop(S).

close_socket(Pid, Peer, S=#state{active=Active}) ->
  NewActive = active_delete({Pid, Peer}, Active),
  Self = self(),
  if length(NewActive) == length(Active) ->
      Pid ! {Self, {error, not_exist}};
     true ->
      Pid ! {Self, closed}
  end,
  listener_loop(S#state{active=NewActive}).

sockname(S=#wsocket_udp{listen_pid=LrPid, peer=Peer}) ->
  LrPid ! {self(), sockname},
  receive
    {LrPid, {sockname, Ret}} ->
      Ret;
    {LrPid, {error, Reason}} ->
      {error, Reason}
  end.

peername(S=#wsocket_udp{peer=Peer}) ->
  {ok, Peer}.

do_recv(S=#wsocket_udp{listen_pid=LrPid, peer=Peer},
	State = {Size, Timeout}, AccBin) ->
  receive
    {LrPid, {recv, RecvBin}} ->
      Bin = erlang:concat_binary([AccBin, RecvBin]),
      BinSize = size(Bin),
      if (Size == 0) or (BinSize == Size) ->
	  {ok, Bin, State};
	 BinSize < Size ->
	  do_recv(S, State, Bin);
	 true ->
	  {RetBin, _} = erlang:split_binary(Bin, Size),
	  {ok, RetBin, State}
      end;
    {LrPid, Error={error, _}} ->
      Error
  after Timeout ->
      {error, timeout}
  end.

active_add(New, Active) ->
  [New|Active].

active_delete(Pid, Active) when is_pid(Pid) ->
  [Conn || Conn <- Active, element(1, Conn) =/= Pid];
active_delete(Conn, Active) ->
  lists:delete(Conn, Active).

active_lookup(Pid, Active) when is_pid(Pid) ->
  case lists:keysearch(Pid, 1, Active) of
    {value, V} ->
      V;
    _ ->
      undefined
  end;
active_lookup(Peer={_IP, _Port}, Active) ->
  case lists:keysearch(Peer, 2, Active) of
    {value, V} ->
      V;
    _ ->
      undefined
  end.
