%%% File    : mdisp_tt.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : Hacked version of Ulf Wiger's mdisp with thread idle time
%%% expiring
%%% Created : 26 Dec 2002 by  <svg@surnet.ru>

%%
%% $Id: mdisp_tt.erl,v 1.1.1.1 2003/05/05 20:20:58 vsekissov Exp $
%%
%%**
%% 
%% @doc Hacked version of Ulf Wiger's mdisp with thread idle time
%% expiring
%% @end
%%*
-module(mdisp_tt).

-behaviour(gen_server).

-export([start_link/1, start_link/2]).

-export([
	 new/4, new/5,		 % start a new process instance
	 send/3,		 % send message to existing process instance
	 new_send/5, new_send/6, % start instance + send message
	 delete/2
	]).

%% Can be called _only_ from the dispatched threads.
-export([
	 sleep/3,
	 die/2
	]).


%% OTP Stuff
-export([behaviour_info/1]).
%% Gen_server callbacks
-export([init/1,
	 code_change/3,
	 terminate/2,
	 handle_call/3,
	 handle_cast/2,
	 handle_info/2]).

%%% =============================================================
%%% Description of the mdisp concept
%%%
%%% The message dispatcher takes care of spawning a worker thread
%%% for a specific task, and allows the thread to "go to sleep", preserving
%%% state. This is useful for types of task, where the number of tasks 
%%% prohibits a design based on one process per task, and/or a relatively
%%% large portion of the lifecycle of a task is idle time.
%%%
%%% NOTE: One side effect of this concept is that the pid of a task will
%%% most likely change over time, i.e. every time the task wakes up.
%%%
%%% mdisp clients must provide a callback module, exporting the
%%% following functions:

%%% - init(Key, State) -> void()
%%% - awake(Key, State) -> void()
%%% - terminate(Reason, Key, State) -> void()
%%%
%%%   Key ::= term()	-- unique identifier of the task
%%%   State ::= term()	-- internal state of the task
%%%   Reason ::= term()	-- EXIT reason signifying internal error in the task
%%%
%%% The following functions are available to the client:
%%%
%%% - new(MDisp, CBMod, Key, InitialState [, Opts]) -> ok
%%%	starts a task process using CBMod:init(Key, InitialState).
%%%     The actual task is created with spawn_opt(Fun, Opts). If no
%%%     options are given, [link] will be used. Link is mandatory and default.
%%%
%%% - send(MDisp, Key, Msg) -> ok
%%%	sends a message to the task process identified by Key, awaking it
%%%	if necessary (MDisp calls Mod:awake(Key, State) and then sends Msg.)
%%%
%%% - send_new(MDisp, CBMod, Key, Msg [, Opts]) -> ok
%%%	Combines new() and send() above. Assuming that the thread should
%%%	start by receiving a message, this function is faster and more
%%%     convenient than calling new() and then send(). One might also pass
%%%     the message as InitialState in one call to new(). This is a matter
%%%     of taste.
%%%
%%% - delete(MDisp, Key) -> ok | {error, Reason}
%%%	removing the task identified by Key; only allowed if the task process
%%%	is asleep.
%%%
%%% The following functions are available to the task process,
%%% and can *only* be called from the task process:
%%%
%%% - sleep(MDisp, Key, State) -> void()
%%%	Puts the task to sleep. The current task process is terminated and 
%%%     State is stored until the task wakes up again. If there are messages
%%%	in the queue of the task process, Mod:awake(Key, State) is called;
%%%	if not, MDisp terminates the task process using exit(Pid, kill).
%%%
%%% - die(MDisp, Key) -> void()
%%%	Kills the task, deleting its state. This is performed safely by the 
%%%     mdisp in order to ensure that the process kills itself.
%%%
%%% =============================================================



-define(call, mdisp_call).
-define(cast, mdisp_cast).
-define(spawn(M, F, A), spawn_link(M, F, A)).

-define(ttl, 30 * 60 * 1000).
-define(clean_period, 10). % percents from ttl

-record(state, {keys = new_set(keys_tt),
		active = 0,
		ttl=?ttl,
		clean_period=?clean_period,
		tm_ref}).

behaviour_info(callbacks) ->
    [{init,2}, {awake,2}, {terminate,3}].

start_link(Name) ->
  start_link(Name, []).

start_link(Name, Opts) ->
    gen_server:start_link({local, Name}, ?MODULE, {Name, Opts},
			  [{spawn_opt, [{priority, high}]}]).

new(MDisp, Mod, Key, InitialState) ->
    new(MDisp, Mod, Key, InitialState, []).

new(MDisp, Mod, Key, InitialState, Opts0) ->
    Opts1 = validate_opts(Opts0),
    gen_server:cast(MDisp, {new, Mod, Key, InitialState, Opts1}).

send(MDisp, Key, Msg) ->
    gen_server:cast(MDisp, {send, Key, Msg}).

new_send(MDisp, Mod, Key, InitialState, Msg) ->
    new_send(MDisp, Mod, Key, InitialState, Msg, []).

new_send(MDisp, Mod, Key, InitialState, Msg, Opts) ->
    Opts1 = validate_opts(Opts),
    gen_server:cast(MDisp, {new_send, Mod, Key, InitialState, Msg, Opts1}).


sleep(MDisp, Key, State) ->
    gen_server:cast(MDisp, {sleep, Key, self(), State}),
    await_kill(Key, State, MDisp).


die(MDisp, Key) ->
    gen_server:cast(MDisp, {die, Key, self()}),
    die_unconditionally().


delete(MDisp, Key) ->
    gen_server:call(MDisp, {delete, Key}).


%%% --- implementation

init({Name, Opts}) ->
  S = check_opts(Opts, #state{}),
  case (S#state.ttl div S#state.clean_period) of
    0 ->
      {stop, {badttl, S#state.ttl}};
    CleanAfter ->
      Ref = erlang:start_timer(CleanAfter, self(), expire_idle),
      process_flag(trap_exit, true),
      {ok, S#state{tm_ref=Ref}}
  end.

die_unconditionally() ->
    %% Yes, this is an endless loop, which doesn't seem to lead to death
    %% any time soon. Look in handle_cast({die, Key, Pid}, S) for full
    %% understanding of the 'die' protocol.
    receive
	_ -> die_unconditionally()
    end.

await_kill(Key, State, MDisp) ->
    receive
	{?MODULE, stay_alive, Mod} ->
	    thread_awake(Mod, Key, State, MDisp)
    end.


%%% --- gen_server callbacks

handle_cast({new, Mod, Key, InitialState, Opts}, S0) ->
    MDisp = self(),
    {Pid, S1} = start_thread(fun() ->
				     thread_init(Mod, Key, InitialState, MDisp)
			     end, Opts, S0),
    S2 = log_awake(Key, Mod, Pid, Opts, S1),
    {noreply, S2};

handle_cast({new_send, Mod, Key, InitialState, Msg, Opts}, S0) ->
    MDisp = self(),
    {Pid, S1} = start_thread(fun() ->
				     thread_init(Mod, Key, InitialState, MDisp)
			     end, Opts, S0),
    Pid ! Msg,
    S2 = log_awake(Key, Mod, Pid, Opts, S1),
    {noreply, S2};

handle_cast({send, Key, Msg}, S0) ->
    case ets:lookup(S0#state.keys, Key) of
	[] ->
	    %% treat this as a send to a nonexisting process -- ignore
	    {noreply, S0};
	[{_, Mod, awake, Opts, Pid}] ->
	    Pid ! Msg,
	    {noreply, S0};
	[{_, Mod, asleep, Opts, State, _}] ->
	    MDisp = self(),
	    {Pid,S1} = start_thread(
			 fun() ->
				 thread_awake(Mod, Key, State, MDisp)
			 end, Opts, S0),
	    Pid ! Msg,
	    S2 = log_awake(Key, Mod, Pid, Opts, S1),
	    {noreply, S2}
    end;

handle_cast({sleep, Key, Pid, State}, S) ->
    %% we perform two checks here, designed to crash mdisp if the 
    %% sleep synchronization is somehow subverted:
    %%
    %% 1) Assert that Key actually is looked as an active process, and that
    %%    it is, in fact, this process making the call;
    %%
    %% 2) Assert that Pid is in the function await_kill/3 in this module.
    %%    This function is not exported, so you shouldn't end up there by
    %%    mistake (unless of course, we compile with 'export_all'... oh well)
    %%
    %% Given these assertions, the process will be killed iff there are no
    %% more messages in its queue; otherwise, it's told to stay awake
    %%
    case ets:lookup(S#state.keys, Key) of
	[{_, Mod, awake, Opts, P}] when P == Pid ->
	    {_, {?MODULE, await_kill, 3}} = 
		process_info(Pid, current_function),
	    case process_info(Pid, message_queue_len) of
		{_, 0} ->
		    unlink(Pid),
		    exit(Pid, kill),
		    NewS = log_asleep(Key, Mod, Opts, State, S),
		    {noreply, NewS};
		{_, L} when L > 0 ->
		    Pid ! {?MODULE, stay_alive, Mod},
		    {noreply, S}
	    end
    %% no other case allowed
    end;

handle_cast({die, Key, Pid}, S) ->
    %% we perform two checks here, designed to crash mdisp if the 
    %% die protocol is somehow subverted:
    %%
    %% 1) Assert that Key actually is looked as an active process, and that
    %%    it is, in fact, this process making the call;
    %%
    %% 2) Assert that Pid is in the function ?MODULE:die_unconditionally/0.
    %%    This function is not exported, so you shouldn't end up there by
    %%    mistake (unless of course, we compile with 'export_all'... oh well)
    %%
    %% Given these assertions, the process will be killed without pardon, and
    %% all traces of it removed from mdisp.
    %%
    case ets:lookup(S#state.keys, Key) of
	[{_, Mod, awake, Opts, P}] when P == Pid ->
	    {_, {?MODULE, die_unconditionally, 0}} = 
		process_info(Pid, current_function),
	    unlink(Pid),
	    exit(Pid, kill),
	    ets:delete(S#state.keys, Key),
	    NewS = S#state{active = S#state.active-1},
	    {noreply, NewS}
    %% no other case allowed
    end.



handle_call({delete, Key}, From, S) ->
    case ets:lookup(S#state.keys, Key) of
	[{_, Mod, awake, Opts, Pid}] ->
	    {reply, {error, {awake, Key, Pid}}, S};
	[{_, Mod, asleep, Opts, State, _}] ->
	    ets:delete(S#state.keys, Key),
	    {reply, ok, S#state{active = S#state.active-1}}
    end.

handle_info({timeout, R0, expire_idle},
	    S=#state{active=A, keys=K, tm_ref=R0, ttl=TTL, clean_period=CP}) ->
  NA = A - expire_idle(K),
  R1 = erlang:start_timer((TTL div CP), self(), expire_idle),
  {noreply, S#state{active = NA, tm_ref=R1}};
handle_info(Msg, S) ->
    {noreply, S}.

terminate(Reason, S=#state{tm_ref=Ref}) ->
  stop_timer(Ref),
  ok.

code_change(FromVsn, State, Extra) ->
    {ok, State}.



start_thread(Fun, Opts, State) ->
    [{spawn_opt, SOpts}] = Opts,
    Pid = spawn_opt(Fun, SOpts),
    {Pid, State}.

%% thread_init/4
%% thread_awake/4
%%
%% The purpose of these two functions is to get a chance to perform mdisp-
%% related actions before handing control over to the dispatched process.
%% An example of such an action is load control. Another thing could be to
%% catch the Mod:init/2 or Mod:awake/2 call and perform error handling on 
%% failure.
%%
thread_init(Mod, Key, State, MDisp) ->
    case catch Mod:init(Key, State) of
	{'EXIT', Reason} ->
	    Mod:terminate(Reason, Key, State);
	_ ->
	    ok
    end.

thread_awake(Mod, Key, State, MDisp) ->
    case catch Mod:awake(Key, State) of
	{'EXIT', Reason} ->
	    Mod:terminate(Reason, Key, State);
	_ ->
	    ok
    end.


log_awake(Key, Mod, Pid, Opts, S) ->
    ets:insert(S#state.keys, {Key, Mod, awake, Opts, Pid}),
    S#state{active = S#state.active+1}.

log_asleep(Key, Mod, Opts, State, S=#state{ttl=TTL}) ->
  ets:insert(S#state.keys, {Key, Mod, asleep, Opts, State, stime()+TTL}),
  S#state{active = S#state.active-1}.


new_set(Name) ->
    ets:new(Name, [set, protected]).


validate_opts(Opts0) ->
    DefaultSpawnOpts = [link],
    {SO,Opts1} =
	lists:foldl(
	  fun({spawn_opt, SpawnOpts},{SOx,Acc}) when list(SpawnOpts) ->
		  {validate_spawn_opt(SpawnOpts,SOx), Acc};
	     (Other,_) ->
		  exit({illegal_option, Other})
	  end, {DefaultSpawnOpts, []}, Opts0),
    [{spawn_opt, SO}|Opts1].

validate_spawn_opt(Opts,DefaultOpts) ->
    lists:foldl(
      fun({priority, P}=O,Acc) when P==low;
				   P==normal;
				   P==high;
				   P==max ->
	      [O|Acc];
	 ({fullsweep_after, N}=O,Acc) when integer(N) ->
	      [O|Acc];
	 ({min_heap_size, Sz}=O,Acc) when integer(Sz) ->
	      [O|Acc];
	 (link,Acc) ->
	      %% redundant, but OK
	      Acc;
	 (Other,_) ->
	      exit({illegal_spawn_opt, Other})
      end, DefaultOpts, Opts).

expire_idle(Tbl) ->
  Now = stime(),
  P = [{{'$1', '_', '$3', '_', '$5', '$6'},
	[{'==', '$3', asleep},
	 {'<', '$6', {const, Now}}],
	['$1']}],
  expire_idle(ets:select(Tbl, P, 10000), Tbl, 0).

expire_idle('$end_of_table', _Tbl, N) ->
    N;
expire_idle({Keys, Cont}, Tbl, N0) ->
  N1 = lists:foldl(fun (K, C) ->
		       ets:delete(Tbl, K),
		       C+1
		   end, N0, Keys),
  expire_idle(ets:select(Cont), Tbl, N1).

check_opts([], S) ->
  S;
check_opts([{ttl, N}|Rest], S) when integer(N) ->
  check_opts(Rest, S#state{ttl=N});
check_opts([{clean_period, N}|Rest], S) when integer(N) ->
  check_opts(Rest, S#state{clean_period=N});
check_opts([Opt|_], _S) ->
  exit({badopt, Opt}).

stop_timer(undefined) -> ok;
stop_timer(Ref) ->
  erlang:cancel_timer(Ref),
  receive
    {timeout, Ref, _} ->
      ok
  after 0 ->
      ok
  end.

stime() ->
  {M, S, U} = now(),
  M*1000000000+S*1000+(U div 1000).
