%%%-------------------------------------------------------------------
%%% File    : wsacl.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : 
%%%
%%% Created :  4 Apr 2003 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id: wsacl.erl,v 1.3 2003/10/31 17:57:31 vsekissov Exp $
%%
%%**
%% 
%% @doc
%% @end
%%*
-module(wsacl).

-behaviour(gen_server).
%%--------------------------------------------------------------------
%% Include files
%%--------------------------------------------------------------------

%%--------------------------------------------------------------------

%% External exports
-export([start_link/2, login/4, logout/3, authenticate/4, authorize/6, check_permission/4, service_spec/3]).
-export([behaviour_info/1]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2, code_change/3]).

%% internal exports
-export([session_id/1]).

-include("wstl.hrl").

-record(state, {handler,
		hdata}).

-record(acl_arg, {server, session_id}).

behaviour_info(callbacks) ->
  [{init,1}, {terminate, 1},
   {get_login,3}, {is_login_enabled,3}, {login_successful,3}, {validate_password,4}, {logout/3},
   {authorize, 7}, {check_permission,4}].

service_spec(AppName, Proc, Props) ->
  {SessionId, AclProps} =
    extract_prop(session_id, Props, atom_to_list(AppName)),
  AclArg = #acl_arg{server=Proc, session_id=SessionId},
  InitEnv = [mk_init_env(AclArg), {wsession,init_env}],
  {ok, AclProps, InitEnv}.

start_link(Name, Opts) ->
  Module = get_prop(module, Opts),
  InitParam = proplists:get_value(init_param, Opts, []),
  start_link(Name, Module, InitParam).

start_link(Name, Module, InitParam) ->
  gen_server:start_link(Name, ?MODULE, {Module, InitParam}, []).

login(WA, LoginId, Password, Attrs) when is_record(WA, ws_arg) ->
  case find_serv(WA) of
    {ok, Srv} ->
      login(Srv, LoginId, Password, Attrs);
    undefined ->
      {ok, LoginId}
  end;
login(Srv, LoginId, Password, Attrs) ->
  gen_server:call(Srv, {login, LoginId, Password, Attrs}).

logout(WA, User, Attrs) when is_record(WA, ws_arg) ->
  case find_serv(WA) of
    {ok, Srv} ->
      logout(Srv, User, Attrs);
    undefined ->
      ok
  end;
logout(Srv, User, Attrs) ->
  gen_server:call(Srv, {logout, User, Attrs}).

authenticate(WA, LoginId, Password, Attrs) when is_record(WA, ws_arg) ->
  case find_serv(WA) of
    {ok, Srv} ->
      authenticate(Srv, LoginId, Password, Attrs);
    undefined ->
      {ok, LoginId}
  end;
authenticate(Srv, LoginId, Password, Attrs) ->
  gen_server:call(Srv, {authenticate, LoginId, Password, Attrs}).

authorize(WA, User, Event, Path, AccessId, Attrs) when is_record(WA, ws_arg) ->
  case find_serv(WA) of
    {ok, Srv} ->
      authorize(Srv, User, Event, Path, AccessId, Attrs);
    undefined ->
      permit
  end;
authorize(Srv, User, Event, Path, AccessId, Attrs) ->
  gen_server:call(Srv, {authorize, User, Event, Path, AccessId, Attrs}).

check_permission(WA, User, Permission, Attrs) when is_record(WA, ws_arg) ->
  case find_serv(WA) of
    {ok, Srv} ->
      check_permission(Srv, User, Permission, Attrs);
    undefined ->
      permit
  end;
check_permission(Srv, User, Permission, Attrs) ->
  gen_server:call(Srv, {check_permission, User, Permission, Attrs}).

%%====================================================================
%% Server functions
%%====================================================================
init({Module, InitParam}) ->
  {ok, HData} = Module:init(InitParam),
  {ok, #state{handler=Module, hdata=HData}}.

%%--------------------------------------------------------------------
%% Function: handle_call/3
%% Description: Handling call messages
%% Returns: {reply, Reply, State}          |
%%          {reply, Reply, State, Timeout} |
%%          {noreply, State}               |
%%          {noreply, State, Timeout}      |
%%          {stop, Reason, Reply, State}   | (terminate/2 is called)
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_call({logout, User, Attrs}, _From,
	    S=#state{handler=M, hdata=HD}) ->
  {Reply, NewHD} = M:logout(User, Attrs, HD),
  {reply, Reply, S#state{hdata=NewHD}};

handle_call({login, LoginId, Password, Attrs}, _From,
	    S=#state{handler=M, hdata=HD}) ->
  {Reply, NewHD} = validate_login(M, HD, LoginId, Password, Attrs),
  {reply, Reply, S#state{hdata=NewHD}};

handle_call(E={authenticate, LoginId, Password, Attrs}, _From,
	    S=#state{handler=M, hdata=HD}) ->
  ?dbg("~p", [E]),
  {Reply, NewHD} = validate_password(M, HD, LoginId, Password, Attrs),
  {reply, Reply, S#state{hdata=NewHD}};

handle_call(E={authorize, User, Event, Path, AccessId, Attrs}, _From,
	    S=#state{handler=M, hdata=HD}) ->
  ?dbg("~p", [E]),
  {Reply, NewHD} = M:authorize(User, Event, Path, AccessId, Attrs, HD),
  {reply, Reply, S#state{hdata=NewHD}};

handle_call(E={check_permission, User, Permission, Attrs}, _From,
	    S=#state{handler=M, hdata=HD}) ->
  ?dbg("~p", [E]),
  {Reply, NewHD} = M:check_permission(User, Permission, Attrs, HD),
  {reply, Reply, S#state{hdata=NewHD}};

handle_call(Request, From, State) ->
  Reply = {bad_request, Request},
  {stop, Reply, Reply, State}.
%%--------------------------------------------------------------------
%% Function: handle_cast/2
%% Description: Handling cast messages
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_cast(Msg, State) ->
  {noreply, State}.

%%--------------------------------------------------------------------
%% Function: handle_info/2
%% Description: Handling all non call/cast messages
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_info(Info, State) ->
  {noreply, State}.

%%--------------------------------------------------------------------
%% Function: terminate/2
%% Description: Shutdown the server
%% Returns: any (ignored by gen_server)
%%--------------------------------------------------------------------
terminate(Reason, S=#state{handler=M, hdata=HD}) ->
  (catch M:terminate(HD)),
  ok.

%%--------------------------------------------------------------------
%% Func: code_change/3
%% Purpose: Convert process state when code is changed
%% Returns: {ok, NewState}
%%--------------------------------------------------------------------
code_change(OldVsn, State, Extra) ->
  {ok, State}.

%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------
mk_init_env(AclArg) ->
  fun (WA=#ws_arg{ws_env=WE}) ->
      WA#ws_arg{ws_env=[{?MODULE, AclArg}|WE]}
  end.

session_id(WA) ->
  (get_env(WA))#acl_arg.session_id.

validate_password(Mod, HD, LoginId, Passwd, Attrs) ->
  {Res, HD1} = Mod:get_login(LoginId, HD, Attrs),
  case Res of
    {ok, Login} ->
      Mod:validate_password(Login, Passwd, Attrs, HD1);
    Error ->
      {Error, HD1}
  end.

validate_login(Mod, HD, LoginId, Passwd, Attrs) ->
  {Res1, HD1} = Mod:get_login(LoginId, Attrs, HD),
  case Res1 of
    {ok, Login} ->
      {Valid, HD2} = Mod:validate_password(Login, Passwd, Attrs, HD1),
      {Enabled, HD3} = Mod:is_login_enabled(Login, Attrs, HD2),
      case  {Valid, Enabled} of
	{true, true} ->
	  Mod:login_successful(Login, Attrs, HD3);
	{false, _} ->
	  {{error, bad_password}, HD3};
	{_, false} ->
	  {{error, login_disabled}, HD3}
      end;
    Error ->
      {Error, HD1}
  end.

find_serv(WA) ->
  case get_env(WA) of
    undefined ->
      undefined;
     SA=#acl_arg{server=Serv} ->
      {ok, Serv}
  end.

get_env(#ws_arg{ws_env=WE}) ->
  proplists:get_value(?MODULE, WE).

get_prop(K, Opts) ->
  case proplists:get_value(K, Opts) of
    undefined ->
      exit({required_param, K});
    V ->
      V
  end.

extract_prop(K, L, D) ->
  {proplists:get_value(K, L, D), proplists:delete(K, L)}.
