%%%-------------------------------------------------------------------
%%% File    : wsapplication_mgr.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : 
%%%
%%% Created : 28 Apr 2003 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id: wsapplication_mgr.erl,v 1.5 2003/10/31 17:57:31 vsekissov Exp $
%%
%%**
%% 
%% @doc
%%
%% Application specification
%%
%% AppSpec = [{module, AppModule},
%%            {init_run_env, AppInitRunEnv},
%%            GlobalServiceParam,
%%            ...]
%%
%% AppInitRunEnv = term()
%% AppModule = atom()   
%% GlobalServiceParam = {ServiceName, [Param]}
%% ServiceName = atom()
%% Param = term() | {use_service, UseService::boolean()}
%% 
%% AppModule module must implement:
%%
%%  init_run_env(init, AppInitRunEnv::term()) -> {ok, State}
%%  init_run_env((WA::ws_arg(), State::term()) -> {ok, AddToAppEnv, NextState}
%%  State = term()
%%  NextState = term()
%%  AddToAppEnv = [{Key, Value}]
%%  Key = atom()
%%  Value = term()
%%  
%% AddToAppEnv is added to WA#ws_arg.app_env
%%
%% Defaults
%%
%% AppModule = application_name()
%% AppInitRunEnv = undefined
%% 
%% @end
%%*
-module(wsapplication_mgr).

-behaviour(gen_server).
%%--------------------------------------------------------------------
%% Include files
%%--------------------------------------------------------------------

%%--------------------------------------------------------------------
%% External exports
%%--------------------------------------------------------------------
-export([start_link/0, start_application/2, stop_application/1, init_run_env/1]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2, code_change/3]).

-include("wstl.hrl").

-define(server, ?MODULE).
-define(services, [stl_yaws, wsacl, wslanguage]).

-record(wsapp, {name, supervisor, init_run_env}).
-record(state, {}).

start_application(Name, Opts) ->
  gen_server:call(?server, {start_application, Name, Opts}).

stop_application(Name) ->
  gen_server:call(?server, {stop_application, Name}).

init_run_env(Name) ->
  case gen_server:call(?server, {init_run_env, Name}) of
    {error, no_application} ->
      exit({no_application, Name});
    Other ->
      Other
  end.

%%====================================================================
%% External functions
%%====================================================================
%%--------------------------------------------------------------------
%% Function: start_link/0
%% Description: Starts the server
%%--------------------------------------------------------------------
start_link() ->
  gen_server:start_link({local, ?server}, ?MODULE, [], []).

%%====================================================================
%% Server functions
%%====================================================================

%%--------------------------------------------------------------------
%% Function: init/1
%% Description: Initiates the server
%% Returns: {ok, State}          |
%%          {ok, State, Timeout} |
%%          ignore               |
%%          {stop, Reason}
%%--------------------------------------------------------------------
init([]) ->
  process_flag(trap_exit, true),
  ets:new(?MODULE, [set, named_table, public, {keypos, 2}]),
  {ok, #state{}}.

%%--------------------------------------------------------------------
%% Function: handle_call/3
%% Description: Handling call messages
%% Returns: {reply, Reply, State}          |
%%          {reply, Reply, State, Timeout} |
%%          {noreply, State}               |
%%          {noreply, State, Timeout}      |
%%          {stop, Reason, Reply, State}   | (terminate/2 is called)
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_call({start_application, Name, Opts}, From, State) ->
  case ets:lookup(?MODULE, Name) of
    [WA] ->
      {reply, {error, already_started}, State};
    [] ->
      Ret = do_start_application(Name, Opts),
      {reply, Ret, State}
  end;
handle_call({stop_application, Name}, From, State) ->
  case ets:lookup(?MODULE, Name) of
    [WA] ->
      Ret = do_stop_application(WA),
      {reply, Ret, State};
    [] ->
      {reply, {error, not_started}, State}
  end;
handle_call({init_run_env, Name}, From, State) ->
  case ets:lookup(?MODULE, Name) of
    [WA] ->
      {reply, WA#wsapp.init_run_env, State};
    [] ->
      {reply, {error, no_application}, State}
  end.

%%--------------------------------------------------------------------
%% Function: handle_cast/2
%% Description: Handling cast messages
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_cast(Msg, State) ->
  {noreply, State}.

%%--------------------------------------------------------------------
%% Function: handle_info/2
%% Description: Handling all non call/cast messages
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_info({'EXIT', Pid, Reason}, State) ->
  case ets:match(?MODULE, #wsapp{supervisor=Pid, _='_'}) of
    [#wsapp{name=N}] ->
      error_logger:error_report([?MODULE, {'Application', N}, Reason]),
      ets:delete(?MODULE, N);
    [] ->
      ok
  end,
  {noreply, State}.

%%--------------------------------------------------------------------
%% Function: terminate/2
%% Description: Shutdown the server
%% Returns: any (ignored by gen_server)
%%--------------------------------------------------------------------
terminate(Reason, State) ->
  ets:foldl(fun (#wsapp{supervisor=Pid}, _) ->
		stop_application_sup(Pid, Reason)
	    end, ok, ?MODULE),
  ok.

%%--------------------------------------------------------------------
%% Func: code_change/3
%% Purpose: Convert process state when code is changed
%% Returns: {ok, NewState}
%%--------------------------------------------------------------------
code_change(OldVsn, State, Extra) ->
  {ok, State}.

%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------

do_start_application(Name, Opts) ->
  [AppInitRunArg, AppModule] =
    wslib:gv_map([init_run_env, {module, Name}], Opts),
  Opts1 = wslib:k_delete_map([init_run_env, module], Opts),
  Services = services_to_start(Opts1),
  case services_spec(Name, Services) of
    {ok, ServiceSpecs, SvcInitEnv} ->
      SupName = service_proc_name(supervisor, Name),
      case wsapplication_sup:start_link(SupName, ServiceSpecs) of
	{ok, SupPid} ->
	  InitEnv =
	    mk_init_run_env(SvcInitEnv, Name, AppModule, AppInitRunArg),
	  WsApp = #wsapp{name=Name, supervisor=SupPid, init_run_env=InitEnv},
	  ets:insert(?MODULE, WsApp),
	  ok;
	Error ->
	  error_logger:error_msg("Application ~p not started. Reason: ~p~n",
				 [Name,Error]),
	  Error
      end;
    Error ->
      Error
  end.

do_stop_application(WA=#wsapp{name=N, supervisor=Pid}) ->
  stop_application_sup(Pid, shutdown),
  ets:delete(?MODULE, N),
  ok.

stop_application_sup(Pid, Reson) ->
  exit(Pid, shutdown),
  receive
    {'EXIT', Pid, _} ->
      ok
  end.

services_spec(Name, Services) ->
  catch
    begin
      {SvcSpecs, SvcInitEnv} =
	lists:foldl(
	  fun ({Svc, SvcProps}, {Specs, InitEnv}) ->
	      ProcName = service_proc_name(Svc, Name),
	      case catch service_spec(Svc, ProcName, SvcProps, Name) of
		{ok, SvcSpec, SvcInitEnv} ->
		  {[service_sup_spec(Svc, ProcName, SvcSpec)|Specs],
		   InitEnv ++ mk_list(SvcInitEnv)};
		Error ->
		  throw({error, {service_spec, Svc, Error}})
	      end
	  end, {[], []}, Services),
      {ok, lists:reverse(SvcSpecs), SvcInitEnv}
    end.

service_sup_spec(Service, ProcName, ServiceOpts) ->
  {Service, {Service, start_link, [{local, ProcName}, ServiceOpts]},
   permanent, 5000, worker, [Service]}.

service_spec(Service, ServiceProcName, ServiceProps, Name) ->
  Service:service_spec(Name, ServiceProcName, ServiceProps).

services_to_start(Opts) ->
  services_to_start(Opts, ?services, []).

services_to_start(_, [], Ret) ->
  lists:reverse(Ret);
services_to_start(Opts, [Service|Rest], Acc) ->
  SO1 = proplists:append_values(Service, Opts),
  case extract_prop(use_service, SO1, true) of
    {true, SO2} ->
      services_to_start(Opts, Rest, [{Service, SO2}|Acc]);
    {false, _} ->
      services_to_start(Opts, Rest, Acc)
  end.
		
service_proc_name(Svc, Name) ->
  list_to_atom("ws_" ++ atom_to_list(Name) ++ "_" ++ atom_to_list(Svc)).

mk_init_run_env(SvcEnv, AppName, AppModule, AppInitRunArg) ->
  AppInitRun =
    fun (Arg, State) -> AppModule:init_run_env(Arg, State) end,
  {ok, AppState} = AppInitRun(init, AppInitRunArg),
  InitRun = 
    fun (WA, State) ->
	WA1 = WA#ws_arg{application=AppName},
	WA2 =
	  lists:foldl(fun (F, Arg) -> F(Arg) end, WA1, SvcEnv),
	{ok, AppEnv, NxtAppState} = AppInitRun(WA2, State),
	AE = WA2#ws_arg.app_env,
	WA3 = WA2#ws_arg{app_env=AE ++ mk_list(AppEnv)},
	{ok, WA3, NxtAppState}
    end,
  mk_init_run_env(InitRun, AppState).

mk_init_run_env(InitRun, State) ->
  fun (WA) ->
      {ok, WA1, NxtState} = InitRun(WA, State),
      {ok, WA1, mk_init_run_env(InitRun, NxtState)}
  end.

extract_prop(K, L, D) ->
  {proplists:get_value(K, L, D), proplists:delete(K, L)}.

mk_list(L) when list(L) ->
  L;
mk_list(V) ->
  [V].
