%%% File    : wscc.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : 
%%% Created : 13 Mar 2003 by  <svg@surnet.ru>

%%
%% $Id: wscc.erl,v 1.2 2003/05/20 11:16:34 vsekissov Exp $
%%
%%**
%% 
%% @doc
%% @end
%%*

-module(wscc).

-behaviour(mdisp_tt).

-export([init/2, awake/2, terminate/3]).

-export([start_link/0, start_link/2, stop/1, new/1, new/2, disp_call/2, disp_call/3]).
-export([enc_path/2, enc_path/3, enc_path/4, dec_path/1]).

-include("wstl.hrl").

-define(server, ?MODULE).

-define(idle_timeout, 5000).
-define(key_prefix, "kid").
-define(key_regexp, ";" ?key_prefix "[0-9]+").
-define(contnum_prefix, "kn").
-define(contnum_regexp, ";" ?contnum_prefix "[0-9]+").
-define(user_data_prefix, "ud").
-define(user_data_regexp, ";" ?user_data_prefix ".*").

-record(state, {key,
		timeout,    % idle timeout
		cont,       % continuation fun
		disp
	       }).

start_link() ->
  start_link(?server, []).

start_link(Name, Opts) ->
  {ok, Pid} = mdisp_tt:start_link(Name, Opts).

stop(Name) ->
  exit(whereis(Name), normal).

new(Fun) ->
  new(?MODULE, Fun, ?idle_timeout).

new(Name, Fun) ->
  new(Name, Fun, ?idle_timeout).

new(Name, Fun, IdleTimeout) ->
  Key = generate_key(),
  mdisp_tt:new_send(Name, ?MODULE, Key,
		    #state{timeout=IdleTimeout, disp=Name},
		    {self(), apply, {fun init_cc/1, [Fun]}}),
  ok = rec_answer(Key),
  {ok, Key}.

disp_call(Key, Args) ->
  disp_call(?server, Key, Args).

disp_call(Name, Key, Args) ->
  mdisp_tt:send(Name, Key, {self(), continue, Args}),
  rec_answer(Key).

enc_path(Path, Key) ->
  enc_path(Path, Key, 0, "").

enc_path(Path, Key, ContNum) ->
  enc_path(Path, Key, ContNum, "").

enc_path(Path, Key, ContNum, UserData) ->
  {AppPath, Query} = lists:splitwith(fun ($?) -> false;
					 (_)  -> true
				     end, Path),
  io_lib:format("~s;~s;~s~w;~s~s~s",
		[AppPath, Key,
		 ?contnum_prefix, ContNum,
		 ?user_data_prefix, yaws_api:url_encode(UserData),
		 Query]).

dec_path(Str) ->
  dec_path_key(Str).

dec_path_key(Str) ->
  case dec_path_match(Str, ?key_regexp) of
    undefined ->
      {error, nothread};
    {Path, Key, Rest} ->
      dec_path_contnum(Rest,
		       #cc_url{
			 path=Path,
			 cont_key=Key,
			 cont_num=0,
			 user_data=""})
  end.

dec_path_contnum(Str, Url) ->
  case dec_path_match(Str, ?contnum_regexp) of
    {[], ?contnum_prefix++S, Rest} ->
      dec_path_user_data(Rest, Url#cc_url{cont_num=list_to_integer(S)});
    _ ->
      {ok, Url}
  end.

dec_path_user_data(Str, Url) ->
  case dec_path_match(Str, ?user_data_regexp) of
    {[], ?user_data_prefix++S, Rest} ->
      {ok, Url#cc_url{user_data=S}};
    _ ->
      {ok, Url}
  end.

dec_path_match(Str, Reg) ->
  case regexp:first_match(Str, Reg) of
    {match, Start,  Length} ->
      %%we are excluding leading ; from the Path
      {string:substr(Str, 1, Start-1),
       string:substr(Str, Start+1, Length-1),
       string:substr(Str, Start+Length)
	};
    _ ->
      undefined
  end.
  
rec_answer(Key) ->
  receive
    {Key, iam, Pid} ->
      Ref = erlang:monitor(process, Pid),
      receive
	{Key, continue, Answer} ->
	  erlang:demonitor(Ref),
	  Answer;
	{Key, die, Answer} ->
	  Answer;
	{Key, error, Error} ->
	  exit(Error);
	{'DOWN', Ref, _, _, Reson} ->
	  exit(Reson)
      end
  after 500 ->
      {error, {nothread, Key}}
  end.

init(Key, State) ->
  loop(State#state{key=Key}).

awake(Key, State) ->
  loop(State).

terminate(Reason, Key, State) ->
  exit({Reason, {Key, State}}).

loop(S=#state{key=Key, timeout=Timeout, disp=D}) ->
  receive
    {User, Req, ReqData} when Req == apply;
			      Req == continue ->
      User ! {Key, iam, self()},
      case process(Req, ReqData, S) of
	{continue, Result, NxtCont} ->
	  User ! {Key, continue, Result},
	  loop(S#state{cont=NxtCont});
	{die, Result} ->
	  User ! {Key, die, Result},
	  mdisp_tt:die(D, Key);
	{'EXIT', Error} ->
	  User ! {Key, error, Error},
	  mdisp_tt:die(D, Key);
	Error ->
	  User ! {Key, error, Error},
	  mdisp_tt:die(D, Key)
      end;
    Other ->
      loop(S)
  after Timeout ->
      mdisp_tt:sleep(D, Key, S)
  end.

process(apply, {Fun, Args}, _) ->
  (catch apply(Fun, Args));
process(continue, _Args, S=#state{cont=undefined}) ->
  undef_continuation;
process(continue, Args, S=#state{cont=CF}) ->
  (catch apply(CF, Args)).

init_cc(Fun) ->
  {continue, ok, Fun}.

generate_key() ->
  ?key_prefix ++ integer_to_list(randomx:uniform(16#ffffffffffffffff)).
