%%%-------------------------------------------------------------------
%%% File    : wsxmlt.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : XML like EHTML transformation  
%%%
%%% Created : 10 Jan 2003 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id: wsxmlt.erl,v 1.6 2003/10/31 17:57:31 vsekissov Exp $
%%
%%**
%% 
%% @doc Toy XML like EHTML transformation
%% <p>Sources are in
%% <a href="http://okmij.org/ftp/Scheme/xml.html">XML and Scheme</a>
%% </p>
%% @end
%%*
-module(wsxmlt).

-export([transform/3, text/1, default/1, copy/0, skip/0, r/2]).
-export([m_name/1, m_attr/2, m_any/0, m_text/0]).
-export([ancestors/1, sibling_pos/1]).
-export([new_var/3, new_upvar/4, set_var/3, get_var/2, find_var/2, level/1]).
-export([test/0]).

-import(wsxmlt.var).

-record(seed, {ancestors=[{'/', []}],
	       sibling_pos=0,
	       vars,
	       post_hlr,
	       rules,
	       result=[]
	      }).

transform(Rule, Ps, Tree) when is_tuple(Rule) ->
  transform([Rule], Ps, Tree);
transform(Rs, Ps, Tree) ->
  Seed = create_seed(Ps, Rs),
  ResSeed =
    fold_tree(Tree, fun down_hlr/2, fun here_hlr/2, fun up_hlr/3, Seed),
  Result = lists:append(lists:reverse(ResSeed#seed.result)),
  erase_seed(ResSeed),
  Result.

sibling_pos(S) ->
  S#seed.sibling_pos.

ancestors(S) ->
  S#seed.ancestors.

%% Rules
default(Rs) ->
  r('#default#', Rs).

text(Rs) ->
  r(fun m_text/2, Rs).

skip() ->
  r(m_any(),
    {fun dummy_pre_hlr/2,
     fun (_, _, _) -> [] end,
     fun skip/2
    }).

skip(_, _) ->
  [skip()].

copy() ->
  default({fun dummy_pre_hlr/2,
	   fun dummy_post_hlr/3,
	   fun copy/2
	  }).

copy(_, _) ->
  [copy()].

m_name(N) when is_atom(N) ->
  fun ({Name, _}, _S) when Name == N -> true;
      (_, _) -> false
  end;
m_name(Names) ->
  fun ({Name, _}, _S) ->
      lists:member(Name, Names)
  end.

m_attr(Name, Value) ->
  fun ({_N, As}, _S) ->
      lists:member({Name, Value}, As)
  end.
  
m_any() ->
  fun (_, _) -> true end.

m_text() ->
  fun m_text/2.

r(Test, Rs) when is_atom(Test), Test =/= '#default#' ->
  r(m_name(Test), Rs);
r(Test, {PreHlr, PostHlr, Rs}) when is_list(Rs) ->
  {Test, {PreHlr, PostHlr, fun (_, _) -> Rs end}};
r(Test, {PreHlr, PostHlr, Rs}) ->
  {Test, {PreHlr, PostHlr, Rs}};
r(Test, Hlr) when is_function(Hlr) ->
  r(Test, {fun dummy_pre_hlr/2, Hlr, fun copy/2});
r(Test, Rs) when is_list(Rs) ->
  r(Test, {fun dummy_pre_hlr/2, fun dummy_post_hlr/3, Rs});
r(Test, {Hlr, Rs}) ->
  r(Test, {fun dummy_pre_hlr/2, Hlr, Rs}).

%% Vars
create_vars(Init) ->
  var:create(Init).

delete_vars(Srv) ->
  var:delete(Srv).

push_var_frame(S=#seed{vars=Srv}) ->
  var:push_frame(Srv).

pop_var_frame(S=#seed{vars=Srv}) ->
  var:pop_frame(Srv).

level(Seed) ->
  length(ancestors(Seed)).

new_var(N, V, Seed) ->
  new_upvar(N, V, 0, Seed).

new_upvar(N, V, Lvl, S=#seed{vars=Srv}) ->
  case var:new(N, V, Lvl, Srv) of
    ok ->
      ok;
    Error ->
      error({Error, N}, S)
  end.

set_var(N, V, S=#seed{vars=Srv}) ->
  case var:set(N, V, Srv) of
    undefined ->
      error({undefvar, N}, S);
    ok ->
      ok
  end.

find_var(N, S=#seed{vars=Srv}) ->
  var:find(N, Srv).

get_var(N, S=#seed{vars=Srv}) ->
  case var:find(N, Srv) of
    {ok, V} ->
      V;
    false ->
      error({undefvar, N}, S)
  end.

%% //Vars

create_seed(Ps, Rs) ->
  #seed{vars=create_vars(Ps), rules=Rs}.

erase_seed(S=#seed{vars=Vs}) ->
  delete_vars(Vs),
  ok.

down_hlr(El, Seed=#seed{ancestors=As}) ->
  {PreChildren, PostChildren, ChildrenRs} = match_rule(El, Seed),
  push_var_frame(Seed),
  PreChildren(El, Seed),
  ChildrenSeed = Seed#seed{
		   result=[],
		   sibling_pos=1,
		   ancestors=[El|As],
		   post_hlr=PostChildren,
		   rules=ChildrenRs(El, Seed)
		  },
  ChildrenSeed.

up_hlr(El, Seed=#seed{result=R, sibling_pos=SP},
       CS=#seed{result=CR, post_hlr=PostChildren}) ->
  NodeResult = mk_list(PostChildren(El, lists:append(lists:reverse(CR)), Seed)),
  pop_var_frame(Seed),
  Seed#seed{result= [NodeResult|R], sibling_pos=SP+1}.

here_hlr(Str, Seed=#seed{result=R}) ->
  {Pre, Post, _} = match_rule({Str, text}, Seed),
  Pre(Str, Seed),
  Seed#seed{result=[normalize_el_content(Post(Str, text, Seed))|R]}.

%%
fold_tree(Node={Name, Args}, DownHlr, HereHlr, UpHlr, Seed) ->
  fold_tree({Name, Args, []}, DownHlr, HereHlr, UpHlr, Seed);
fold_tree(Node={Name, Args, Content}, DownHlr, HereHlr, UpHlr, Seed) ->
  NodeEl = {Name, Args},
  ChildSeed =
    fold_content(Content,
		 DownHlr, HereHlr, UpHlr, DownHlr(NodeEl, Seed)),
  UpHlr(NodeEl, Seed, ChildSeed);
fold_tree(Nodes=[N|_], DownHlr, HereHlr, UpHlr, Seed) when is_tuple(N) ->
  fold_content(Nodes, DownHlr, HereHlr, UpHlr, Seed);
fold_tree(Str, DownHlr, HereHlr, UpHlr, Seed) when is_list(Str) ->
  HereHlr(Str, Seed).

fold_content(Nodes, DownHlr, HereHlr, UpHlr, Seed) ->
  lists:foldl(fun (N, S) ->
		  fold_tree(N, DownHlr, HereHlr, UpHlr, S)
	      end, Seed, normalize_el_content(Nodes)).

normalize_el_content(C=[I|_]) when is_integer(I) ->
  %% string() -> [string()]
  [C];
normalize_el_content(C) ->
  C.

match_rule(El, S=#seed{rules=Rs}) ->
  Rule = match_rule(El, S, Rs, nodefault).

match_rule(El, Seed, [], nodefault) ->
  error({no_rule, El}, Seed);
match_rule(El, _S, [], DefRule) ->
  DefRule;
match_rule(El, Seed, [{'#default#', Rule}|Rs], nodefault) ->
  match_rule(El, Seed, Rs, Rule);
match_rule(El, Seed, [{Test, Rule}|Rs], DefRule) ->
  Matched = Test(El, Seed),
  if Matched ->
      Rule;
     true ->
      match_rule(El, Seed, Rs, DefRule)
  end;
match_rule(El, Seed, [_|Rs], DefRule) ->
  match_rule(El, Seed, Rs, DefRule).


m_text({_, text}, _) ->
  true;
m_text(_, _) ->
  false.

mk_list(L) when list(L) ->
  L;
mk_list(V) ->
  [V].

dummy_pre_hlr(_, _) ->
  ok.

dummy_post_hlr(Str, text, Seed) ->
  Str;
dummy_post_hlr({Name, Attrs}, Content, Seed) ->
  {Name, Attrs, Content}.

error(Error, Seed) ->
  exit({wsxmlt,
	Error,
	{ancestors, lists:reverse(ancestors(Seed))},
	{sibling_pos, sibling_pos(Seed)}
       }).

test() ->
  FieldRules =
    fun (_, _) ->
	[text(fun copy/2),
	 wsxmlt:default(
	   fun ({N, As}, Content, Ctxt) ->
	       Row = wsxmlt:get_var(rowcnt, Ctxt),
	       NameStr = atom_to_list(N),
	       Out =
		 {tr, [{class, "rowset"++integer_to_list(Row band 16#1)}],
		  [{td, [], NameStr},
		   {td, [], [{input, [{name, NameStr}, {value, Content}|As]}]}
		  ]},
	       wsxmlt:set_var(rowcnt, Row+1, Ctxt),
	       Out
	   end)
	]
    end,
  StyleSheet =
    [wsxmlt:r(user,
	       {fun (_, Ctxt) ->
		    wsxmlt:new_var(rowcnt, 0, Ctxt)
		end,
		fun ({_, As}, Content, Ctxt) ->
		    Out =
		      {form,
		       [{action, wsxmlt:get_var(action, Ctxt)}|As],
		       [{table, [],
			 Content ++
			 {tr, [{colspan, "2"}],
			  [{td, [],
			    [{input, [{type, "submit"}, {name, "save"}]}]}
			  ]}
		     
			}]
		      },
		    Out
		end,
		FieldRules
	       })
    ],
  TestInput =
    {user, [],
     [
      {first_name, [], "Tom"},
      {last_name, [], [{t1, [{a, "c"}], ["Test", {t2, [], "Test2"}]}]}
     ]},
  Result =
    wsxmlt:transform(StyleSheet, [{action, "adduser"}], TestInput),
  Result.
